unit DPIAdjuster;

interface

uses Windows, Graphics, Forms, Buttons, Dialogs, Outline, ImgList, SysUtils, Messages, Classes;

{Prototypes and constants for Windows DPI-Aware API}

{This message is passed to a process when DPIs change}
{The message is only sent if the program claims to DPI-Aware}
{You can do this through the manifest or by calling "SetProcessDpiAwarenessContext"}

const WM_DPICHANGED       = $02E0;

{Windows send new scaling parameters with the message}

type TWMDpi = record
 Msg: Cardinal;
 YDpi: Word;
 XDpi: Word;
 ScaledRect: PRECT;
 Result: LRESULT;
 end;

{DPI awareness parameters}

const DPI_AWARENESS_CONTEXT_UNAWARE = 16;
const DPI_AWARENESS_CONTEXT_SYSTEM_AWARE = 17;
const DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE = 18;
const DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2 = 34;

{Routine to set the DPI Awareness state of the program}
{Must be called befor the handle is created for the Window}

function SetProcessDpiAwarenessContext(value: Integer): BOOL; stdcall; external user32 name 'SetProcessDpiAwarenessContext';

{These variables must be outside the object, because
they are used before the object is created}

var FIsDPIAware: boolean;
var FErrorStr: string;

{This is a trick to access the protected scaling routines inside a form}

type TMyForm = class(TForm);

{DPI Adjuster Object}

type TDPIAdjuster = class(TComponent)
 private
  FCurrentPPI: integer;
  FOriginalWindowProc: TWndMethod;
  FOnDPIChange: TNotifyEvent;
  function GetDPIAware: boolean;
  function GetErrorString: string;
  procedure MessageHandler(var Msg: TMessage);
  function GetNewAdjuster(New: integer): integer;
 protected
  Form: TForm;
  function GetOwnerForm: TForm;
  procedure WMDpiChanged(var Msg: TWMDpi);
 public
  WinDPIMessage: TWMDpi;
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure ScaleForm(DPIMsg: TWMDpi);
  property IsDPIAware: boolean read GetDPIAware;
  property ErrorString: string read GetErrorString;
 published
  property OnDPIChange: TNotifyEvent read FOnDPIChange write FOnDPIChange;
 end;

procedure Register;

implementation

uses  Controls;


procedure Register;
begin
RegisterComponents('Fountain', [TDPIAdjuster]);
end;


procedure TDPIAdjuster.MessageHandler(var Msg:TMessage);
{Handle messages coming to the main Form}
begin
{Look for the DPI-Changed message and pass it our handler}
case Msg.Msg of
 WM_DPICHANGED: WMDpiChanged(TWMDpi(Msg));
 end;
FOriginalWindowProc(Msg);
end;



function TDPIAdjuster.GetOwnerForm: TForm;
{Follow the chain of owners back to owning form}
var C: TComponent;
begin
Result:=nil;
C:=Owner;
while true do
	begin
	if C=nil then exit;
	if C is TForm then
		begin
		Result:=TForm(C);
		exit;
		end;
	C:=C.Owner;
	end;
end;



procedure SetDPIAwareness;
begin
{DPI Awareness must be set before a handle is generated for the Window}
FIsDPIAware:=SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2);
if not FIsDPIAware then FErrorStr:='DPI Awareness Error: '+SysErrorMessage(GetLastError)
else FErrorStr:='No Error - DPI Awareness Enabled';
end;


constructor TDPIAdjuster.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
if not (csDesigning in ComponentState) then SetDPIAwareness;
Form:=TMyForm(GetOwnerForm);
if Form=nil then raise Exception.Create('A Form MUST be the owner');
FCurrentPPI:=Form.PixelsPerInch;
if not (csDesigning in ComponentState) then
	begin
	{We need to capture the forms messages, because Windows}
	{customizes the "ScaledRectangle" parameters for that Window}
	FOriginalWindowProc := Form.WindowProc;
	Form.WindowProc := MessageHandler;
	end;
end;



destructor TDPIAdjuster.Destroy;
begin
{Restore forms message loop}
if not (csDesigning in ComponentState) then Form.WindowProc:=FOriginalWindowProc;
inherited;
end;


function TDPIAdjuster.GetDPIAware: boolean;
begin
Result:=FIsDPIAware;
end;



function TDPIAdjuster.GetErrorString: string;
begin
Result:=FErrorStr;
end;


function TDPIAdjuster.GetNewAdjuster(New: integer): integer;
begin
Result:=round((Form.PixelsPerInch/FCurrentPPI) * New);
FCurrentPPI:=New;
end;

{This is a trick to access the protected properties inside a TMyOutline}

type TMyOutline = class(TCustomOutline);


procedure TDPIAdjuster.ScaleForm(DPIMsg: TWMDpi);

	procedure ResizeForm;
	begin
	Form.Left:=DPIMsg.ScaledRect.Left;
	Form.Top:=DPIMsg.ScaledRect.Top;
	Form.Width:=DPIMsg.ScaledRect.Right-DPIMsg.ScaledRect.Left;
	Form.Height:=DPIMsg.ScaledRect.Bottom-DPIMsg.ScaledRect.Top;
	end;



	procedure ResizeGlyph(Glyph: TBitmap; M,D: integer);
	var R: TRect;
	var BM2: TBitmap;
	begin
	if (Glyph.Width=0) or (Glyph.Height=0) then exit;
	BM2:=TBitmap.Create;
	try
	Glyph.Canvas.Lock;
	try
	R:=Rect(0,0,MulDiv(Glyph.Width,M,D),MulDiv(Glyph.Height,M,D));
	BM2.Width:=Glyph.Width;
	BM2.Height:=Glyph.Height;
	BM2.Canvas.Draw(0,0,Glyph);
	Glyph.Width:=R.Right;
	Glyph.Height:=R.Bottom;
	Glyph.Canvas.StretchDraw(R,BM2);
	finally Glyph.Canvas.UnLock; end;
	finally BM2.Free; end;
	end;



	procedure ResizeImageList(IL: TImageList; M,D: integer);
	var BA: array of TBitmap;
	var I: integer;
	var BM: TBitmap;
	begin
	if IL.Count<1 then exit;
	IL.DrawingStyle:=dsNormal;
	SetLength(BA,IL.Count);
	for I:=0 to High(BA) do
		begin
		BA[I]:=TBitmap.Create;
		IL.GetBitmap(I,BA[I]);
		ResizeGlyph(BA[I],M,D);
		end;
	IL.Clear;
	IL.Masked:=False;
	IL.Width:=BA[0].Width;
	IL.Height:=BA[0].Height;
	for I:=0 to High(BA) do
		begin
		IL.Add(BA[I],nil);
		BA[I].Free;
		end;
	end;


	procedure ScaleOutline(OL: TCustomOutline; M,D: integer);
	begin
	ResizeGlyph(TMyOutline(OL).PictureOpen,M,D);
	ResizeGlyph(TMyOutline(OL).PictureClosed,M,D);
	ResizeGlyph(TMyOutline(OL).PicturePlus,M,D);
	ResizeGlyph(TMyOutline(OL).PictureMinus,M,D);
	ResizeGlyph(TMyOutline(OL).PictureLeaf,M,D);
	end;


	procedure ScaleControls(WC: TControl; M,D: integer);
	var I: Integer;
	var C: TComponent;
	begin
	Screen.MenuFont.Size:=MulDiv(Screen.MenuFont.Size,M,D);
	for I := 0 to WC.ComponentCount - 1 do
		begin
		C:=WC.Components[I];
		if C.InheritsFrom(TSpeedButton) then ResizeGlyph(TSpeedButton(C).Glyph,M,D)
		else if C.InheritsFrom(TBitBtn) then ResizeGlyph(TBitBtn(C).Glyph,M,D)
		else if C.InheritsFrom(TImageList) then ResizeImageList(TImageList(C),M,D)
		else if C.InheritsFrom(TCustomOutline) then ScaleOutline(TCustomOutline(C),M,D);
		end;
	end;


	procedure ScaleObjects;
	var A,D: integer;
	begin
	A:=GetNewAdjuster(DPIMsg.XDpi);
	D:=Form.PixelsPerInch;
	{The TMyForm trick allows the protected "ChangeScale" procedure to be accessed}
	TMyForm(Form).ChangeScale(A,D);
	ScaleControls(TMyForm(Form),A,D);
	end;

begin
if DPIMsg.XDpi>FCurrentPPI then
	begin
	ResizeForm;
	ScaleObjects;
	end
else
	begin
	ScaleObjects;
	ResizeForm;
	end;
end;



procedure TDPIAdjuster.WMDpiChanged(var Msg: TWMDpi);
begin
WinDPIMessage:=Msg;
ScaleForm(WinDPIMessage);
if assigned(FOnDPIChange) then FOnDPIChange(Self);
end;


end.
