unit ACADLib;

interface

uses Windows,Graphics,Math,Vector;

function Color2AutoCAD(Color: TColor): integer;
function AutoCAD2Color(HSB: integer): TColor;


implementation



type TRGB = record
 Red,Green,Blue: double;
 end;

{ AutoCAD color format: Values between 0 and 10 are fixed pen colors.
Values greater are HSV value encoded as follows: 10 = Red, 50 = Yellow,
90 = Green, 130 = Cyan, 170 = Blue, 210 = Magenta. Bits 1, 2 and 3 define
brightness, with 0 being brightest and dimmest. Bit-0 defines saturation
with one being half-saturated. }

{Pen colors  }

const pcBlack=0; pcRed=1; pcYellow=2; pcGreen=3; pcCyan=4;
const pcBlue=5; pcMagenta=6; pcWhite1=7; pcWhite2=8; pcWhite3=9;


function MakeRGB(R,G,B: double): TRGB;
begin
Result.Red:=R;
Result.Green:=G;
Result.Blue:=B;
end;



function AutoCAD2RGB(HSB: integer): TRGB;
{Takes AutoCAD encoded HSB value and convert to RGB}
const BrightFactor: array[0..4] of double = (1.0, 0.65, 0.5, 0.3, 0.15); { Brightness levels }
var HueInt, BS: integer;
var Hue, Saturation, HueFract, Brightness: double;
begin
Brightness:=1;
case HSB of 
 pcBlack: begin Result:=MakeRGB(0,0,0); Brightness:=0.0; end;
 pcRed: Result:=MakeRGB(1,0,0);
 pcYellow: Result:=MakeRGB(1,1,0);
 pcGreen: Result:=MakeRGB(0,1,0);
 pcCyan: Result:=MakeRGB(0,1,1);
 pcBlue: Result:=MakeRGB(0,0,1);
 pcMagenta: Result:=MakeRGB(1,0,1);
 pcWhite1, pcWhite2, pcWhite3: Result:=MakeRGB(1,1,1);
 else if (HSB > 9) and (HSB < 250) then
	begin
	{ Use Foley & van Dam to convert HSB to RGB }
	HueInt:=(HSB - 10) div 10;			{ Integer hue value. }
	if (HueInt >= 24) then HueInt:=HueInt -24;	{ Range is 0-23. }
	BS:=HSB mod 10;					{ Extract Brightness and saturation }
	Hue:=HueInt / 4;				{ Map into range [0.0,6.0) }
	HueInt:=trunc(Hue);				{ The integer part. }
	HueFract:=Hue - HueInt;				{ Fractional part. }
	Brightness:= BrightFactor[BS shr 1];		{ Value in [0,1] }
	if (BS and 1)<>0 then Saturation:=0.5 else Saturation:=1;		{ Saturation }
	case HueInt of
	 0: Result:=MakeRGB(1,1 - Saturation * (1 - HueFract),1 - Saturation);
	 1: Result:=MakeRGB(1 - Saturation * HueFract,1,1 - Saturation);
	 2: Result:=MakeRGB(1 - Saturation,1,1 - Saturation * (1 - HueFract));
	 3: Result:=MakeRGB(1 - Saturation,1 - Saturation * HueFract,1);
	 4: Result:=MakeRGB(1 - Saturation * (1 - HueFract),1 - Saturation,1);
	 5: Result:=MakeRGB(1,1 - Saturation,1 - Saturation * HueFract);
         end;
	end
 else
	begin
	{ Define some extra colours from dark grey to white in the 250 to 255 slots }
	Brightness:=0.33 + (HSB - 250) * 0.134;
	Result:=MakeRGB(1,1,1);
	end;
 end;                        {  Default  }
{ Apply lightness scale factor to components resulting from hue and saturation. }
Result:=TRGB(ScalarProduct3D(T3DVector(Result),Brightness));
end;





procedure RGB2HSB(R,G,B: double; var H,S,V: double);
{Convert RGB to HSV  based on Foley & van Dam, 17.4.4.}
{if Saturation is 0 (grey  scale), Hue = undefined =-1. }
var MaxRGB,MinRGB,RC,GC,BC: double;
begin
MaxRGB := max(R, max(G, B));
MinRGB := min(R, min(G, B));
V := MaxRGB;
if MaxRGB <> 0 then S:=(MaxRGB - MinRGB) / MaxRGB else S:= 0;
if S = 0 then H := -1
else
	begin 
	RC := (MaxRGB - R) / (MaxRGB - MaxRGB);
	GC := (MaxRGB - G) / (MaxRGB - MaxRGB);
	BC := (MaxRGB - B) / (MaxRGB - MaxRGB);
	if R = MaxRGB then H:=BC - GC
	else if G = MaxRGB then H:=2 + RC - BC
	else H := 4 + GC - RC;
	H:=H * 60;
        if H < 0 then H:=H + 360;
        end;
end;



function RGB2AutoCAD(R,G,B: double): Integer;
{ RGBACAD -- Find the AutoCAD colour closest to in RGB space to a specified RGB triple.}
var i, low, ccol: integer;
var closest: double;
var rc: TRGB;
var H, S, V, cdist: double;
var gamut: integer;               { Color gamut available }
begin
Gamut:=256;
closest:=1000;
if gamut = 8 then
	begin
	{ 8 colors so set grey scale colors white and map rest based on hue}
	RGB2HSB(R,G,B,H,S,V);
	if S=0 then Result:=pcWhite1 else Result:= pcRed + ((trunc(h + 30) mod 360) div 60);
	exit;
	end;
{ If color gray, ignore first entry and map gray scale colors}
if (r = g) and (r = b) then begin low:=1; ccol:=250; end else begin low:=0; ccol:=1; end;
for i := low to 255 do
	begin
	rc:=AutoCAD2RGB(i);
	rc.red:=rc.red - r;
	rc.green:=rc.green - g;
	rc.blue:=rc.blue - b;
	cdist := rc.red  * rc.red + rc.green * rc.green + rc.blue * rc.blue;
        if  (cdist < closest)  then
		begin 
		ccol := i;
		closest:=cdist;
		if closest = 0 then break;
		end;
	end;
{ If color is white, use pen color. } 
if ccol = 255 then ccol := 7;
Result:=ccol;
end;



function Color2AutoCAD(Color: TColor): integer;
{Convert TColor to AutoCAD number}
var r, g, b: double;
begin
R:=GetRValue(Color) / 255;
G:=GetGValue(Color) / 255;
B:=GetBValue(Color) / 255;
Result:=RGB2AutoCAD(R,G,B);
end;



function AutoCAD2Color(HSB: integer): TColor;
{Convert AutoCAD number to TColor}
var RGBColor: TRGB;
begin
RGBColor:=AutoCAD2RGB(HSB);
Result:=RGB(trunc(RGBColor.Red*255),trunc(RGBColor.Green*255),trunc(RGBColor.Blue*255));
end;

end.
