unit CursorLib;

interface

uses Windows, mmSystem, SysUtils, Classes, Graphics, IconLib,
  CursorPalette;

type JIF = DWORD;

type 
 RTag = packed record //RIFF chunk header
 ckID: FourCC;
 ckSize: DWORD;
 end;

type TAniCurHeader = packed record
 cbSizeof: DWORD;
 cFrames: DWORD;
 cSteps: DWORD;
 cx, cy: DWORD;
 cBitCount: DWORD;
 cPlanes: DWORD;
 cjifRate: JIF;
 cflags: DWORD;
 end;

type TAniCursor = class(Tobject)
 private
  fHeader: TAniCurHeader;
  fAuthor: String;
  fTitle: String;
  fJifRates: TList;
  fSequences: TList;
  fFrams: TList;
  fExcess: Boolean;  // Excess data within .ani file
  procedure FErrIn;
  procedure FErrOut;
  function GetFrameCount: Integer;
  function GetHotX: WORD;
  function GetHotY: WORD;
  function GetIsPal16: Boolean;
  function GetRateCount: Integer;
  function GetSeqCount: Integer;
  function GetPalette: TPalArray;
  function GetFrame(Index: Integer): TInternalCur;
  function GetRate(Index: Integer): DWORD;
  function GetSequence(Index: Integer): DWORD;
 public
  constructor Create; overload;
  constructor Create(const filename: String); overload;
  constructor Create(const vTitle, vAuthor: String;
  vAllJifs: DWORD; is16color: Boolean); overload;
  destructor Destroy; override;
  procedure LoadFromStream(Stream: TStream);
  procedure SaveToStream(Stream: TStream);
  procedure LoadFromFile(const fileName: String);
  procedure SaveToFile(const fileName: String);
  procedure AddImage(T: TInternalCur);
  procedure PrepareToSave(vHotX, vHotY: Word);
  property Excess: Boolean read fExcess;
  property Header: TAniCurHeader read fHeader;
  property HotX: WORD read GetHotX;
  property HotY: WORD read GetHotY;
  property IsPal16: Boolean read GetIsPal16;
  property FrameCount: Integer read GetFrameCount;
  property RateCount: Integer read GetRateCount;
  property SeqCount: Integer read GetSeqCount;
  property Palette: TPalArray read GetPalette;
  property Frames[Index: Integer]: TInternalCur read GetFrame;
  property Rates[Index: Integer]: DWORD read GetRate;
  property Sequences[Index: Integer]: DWORD read GetSequence;
  property Author: String read fAuthor write fAuthor;
  property Title: String read fTitle write fTitle;
  end;

implementation

const
  // RIFF chunk type headers
  FOURCC_ACON = $4E4F4341; // 'A','C','O','N'
  FOURCC_anih = $68696E61; // 'a','n','i','h'
  FOURCC_rate = $65746172; // 'r','a','t','e'
  FOURCC_seq  = $20716573; // 's','e','q',' '
  FOURCC_fram = $6D617266; // 'f','r','a','m'
  FOURCC_icon = $6E6F6369; // 'i','c','o','n'
  FOURCC_INFO = $4F464E49; // 'I','N','F','O'
  FOURCC_IART = $54524149; // 'I','A','R','T'
  FOURCC_INAM = $4D414E49; // 'I','N','A','M'

//"If the AF_ICON flag is specified the fields cx, cy, cBitCount, and
// cPlanes are all unused.  Each frame will be of type ICON and will
// contain its own dimensional information.

  AF_ICON   = $0001; // Windows format icon/cursor animation
  AF_SEQUENCE = $0002; // Animation is sequenced

function PadUp(D: Dword): DWord;
begin
Result:= (D+1) AND (NOT 1);
end;



function ReadTag(Stream: TStream; var tag: RTag): Boolean;
begin
try
Stream.ReadBuffer(Tag, SizeOf(Tag));
Result:= True;
except On Exception DO Result:= False; end;
end;



function ReadChunk(Stream: TStream; var Tag: RTag;  var pv): Boolean;
begin
try
Stream.ReadBuffer(pv, Tag.ckSize);
// Align to word boundary
if Odd(tag.ckSize) THEN Stream.Seek(1, soFromCurrent);
Result:= True;
except  on Exception do Result:= False;  end;
end;



function ReadChunkS(Stream: TStream; var Tag: RTag; var S: String): Boolean;
var buffer: PChar;
begin
try
GetMem(Buffer, Tag.ckSize+1);
try
FillChar(Buffer^, Tag.ckSize+1, 0);
Stream.ReadBuffer(Buffer^, Tag.ckSize);
S:= StrPas(Buffer);
// Align to word boundary
if Odd(tag.ckSize) then Stream.Seek(1, soFromCurrent);
Result:= True;
finally FreeMem(Buffer); end;
except on Exception do Result:= False;  end;
end;



function SkipChunk(Stream: TStream; var Tag: RTag): Boolean;
begin
try
Result:= True;
Stream.Seek(PadUp(Tag.ckSize), soFromCurrent);
except ON Exception DO Result:= False;  end;
end;



function WriteTag(Stream: TStream; tag: RTag): Boolean;
begin
try
Stream.WriteBuffer(Tag, SizeOf(Tag));
Result:= True;
except On Exception DO Result:= False;  end;
end;



function WriteTagData(Stream: TStream; tag: RTag; var Data): Boolean;
const zero: Byte = 0;
begin
try
Stream.WriteBuffer(Tag, SizeOf(Tag));
Stream.WriteBuffer(Data, Tag.ckSize);
// Align to word boundary
if Odd(Tag.ckSize) then Stream.WriteBuffer(Zero, 1);
Result:= True;
except  On Exception DO Result:= False;  end;
end;



function WriteTagString(Stream: TStream; tag: RTag; const S: String): Boolean;
var buffer: PChar;
begin
GetMem(Buffer, Length(S)+1);
try
FillChar(Buffer^, Length(S)+1, 0);
StrPCopy(Buffer, S);
Result:= WriteTagData(Stream, Tag, buffer^);
finally FreeMem(Buffer); end;
end;



function WriteType(Stream: TStream; ckID: DWORD): Boolean;
begin
try
Stream.WriteBuffer(ckID, SizeOf(ckID));
Result:= True;
except  On Exception DO Result:= False;  end;
end;



{ TAniCursor }

constructor TAniCursor.Create;
// Create a tabula rasa
begin
Inherited Create;
fExcess:= False;
fAuthor:= '';
fTitle:= '';
FillChar(fHeader, SizeOf(fHeader), 0);
with fHeader do
 begin
 cbSizeof:= SizeOf(fHeader);
 cPlanes:= 1;
 cjifRate:= 10;
 cflags:= 1;
 end;
fJifRates:= TList.Create;
fSequences:= TList.Create;
fFrams:= TList.Create;
end;



constructor TAniCursor.Create(const filename: String);
// Create and load from a file
begin
Create;
LoadFromFile(filename);
end;



constructor TAniCursor.Create(const vTitle, vAuthor: String;
  vAllJifs: DWORD; is16color: Boolean);
// Create, ready to have images added with AddImage
begin
Create;
fTitle:= vTitle;
fAuthor:= vAuthor;
fHeader.cjifRate:= vAllJifs;
if Is16color then fHeader.cBitCount:= 0 else fHeader.cBitCount:= 4;
end;



destructor TAniCursor.Destroy;
var I: integer;
begin
fJifRates.Free;
fSequences.Free;

for I:=0 to fFrams.Count-1 do TInternalCur(fFrams[I]).Free;
fFrams.Free;
Inherited Destroy;
end;

{ Animated cursor file has this format:
RIFF
	(
	'ACON'
	LIST
		(
		'INFO'
		INAM( <name> )
		IART( <artist> )
		)
	'anih' ( <anihdr> )
	['rate'( <rateinfo> )  ]
	['seq '( <seq_info> )]
	LIST( 'fram' icon( <icon_file> ) )
	)
}

procedure TAniCursor.LoadFromStream(Stream: TStream);
var Tag: RTag;
var N: Integer;
var P: DWORD;
var cbChunk: DWORD;
var TII: TInternalCur;

	function ReadDWORD: DWord;
	begin
	try
	Stream.ReadBuffer(Result, SizeOf(Result));
	except on Exception do FErrIn; end;
	end;


	procedure LoadFrame;
	// The actual icon-data
	begin
	while cbChunk >= SizeOf(Tag) do
		begin
		if not ReadTag(Stream, Tag) then FErrIn;
		Dec(cbChunk, SizeOf(Tag));
		case Tag.ckID of
		 FOURCC_Icon:
			begin
			P:= Stream.Position;
			TII:= TInternalCur.Create(Stream);
			if TII.Excess then fExcess:= True;
			fFrams.Add(TII);
			Stream.Seek(P+PadUp(Tag.ckSize), soFromBeginning);
			end;
		 else SkipChunk(Stream, Tag);
		 end;
		Dec(cbChunk, PadUp(Tag.ckSize));
		end;
	end;


	procedure LoadAuthorTitle;
	// INFO includes author and title
	begin
	while cbChunk >= SizeOf(Tag) do
		begin
		if NOT ReadTag(Stream, Tag) then FErrIn;
		Dec(cbChunk, SizeOf(Tag));
		case Tag.ckID of
		 FOURCC_INAM: if NOT ReadChunkS(Stream, Tag, fTitle) then FErrIn;
		 FOURCC_IART: if NOT ReadChunkS(Stream, Tag, fAuthor) then FErrIn;
		 else SkipChunk(Stream, Tag);
		 end;
		Dec(cbChunk, PadUp(Tag.ckSize));
		end;
	end;


	procedure LoadListData;
	begin
	cbChunk:= PadUp(Tag.ckSize);
	Tag.ckID:= ReadDWORD;
	Dec(cbChunk, SizeOf(Tag.ckID));
	case Tag.ckID of
	 FOURCC_fram: LoadFrame;
	 FOURCC_INFO: LoadAuthorTitle;
	  else SkipChunk(Stream, Tag);
	  end;
	end;
	

begin
// Clear out any existing data
fJifRates.Clear;
fSequences.Clear;
if fFrams.Count > 0 then
 for N:= 0 to fFrams.Count-1 do
      with TInternalCur(fFrams[N]) do Free;

fFrams.Clear;
fAuthor:= '';
fTitle:= '';
if not ReadTag(Stream, Tag) then FErrIn;
if Tag.ckID <> FOURCC_RIFF  then FErrIn; // Not valid header
if Integer(Tag.ckSize) > Stream.Size  then  FErrIn; // Not valid size
if ReadDWORD <> FourCC_ACON then FErrIn; // Not ACON chunk
while Stream.Position < Stream.Size do
	begin
	if not ReadTag(Stream, Tag) then FErrIn;
	case Tag.ckID of
	 FourCC_anih:
		begin
		if Tag.ckSize <> SizeOf(fHeader) then FErrIn;
		if not ReadChunk(Stream, Tag, fHeader) then FErrIn;
		if (fHeader.cflags AND AF_ICON = 0) or (fHeader.cFrames = 0) then FErrIn;
		end;
         FourCC_Rate:
		begin
		if Tag.ckSize DIV SizeOf(DWORD) <> fHeader.cSteps then FErrIn;
		for N:= 0 to fHeader.cSteps-1 do fJifRates.Add(Pointer(ReadDWord));
		end;
         FourCC_seq:
		begin
		if Tag.ckSize DIV SizeOf(DWORD) <> fHeader.cSteps then FErrIn;
		for N:= 0 to fHeader.cSteps-1 do fSequences.Add(Pointer(ReadDWord));
		end;
	 FourCC_LIST:
		begin
		LoadListData;
		end;
	 else SkipChunk(Stream, Tag);
      	 end;
	end;
end;



procedure TAniCursor.SaveToStream(Stream: TStream);
var i: Integer;
var cbFile: DWORD;
var cbFram: DWORD;
var cbINFO: DWORD;
var cbTitle: DWORD;
var cbAuthor: DWORD;
var fRate: Boolean; // true if rate-list used
var fSeq: Boolean; // true if sequence-list used
var Tag: RTag;
begin
fRate:= fJifRates.Count > 0;
fSeq:= fSequences.Count > 0;
cbFram:= 0;
cbINFO:= 0;
cbTitle:= Length(fTitle);
cbAuthor:= Length(fAuthor);
for i:= 0 to fFrams.Count-1 do
	begin
	Inc(cbFram, SizeOf(RTag));
	Inc(cbFram, TInternalCur(fFrams[i]).DataSize);
	end;

  // At this point, cbFram = the size required by all the frames,
  // add in the rate, seq, anih, and INFO list sizes as well as
  // all the other overhead.

Inc(cbFram, SizeOf(FOURCC));  // fram type
cbFile:= cbFram;

//ACON type //anih tag //LIST tag (for fram list)
cbFile:= cbFile + sizeof(FOURCC) + sizeof(RTAG) + PADUP(sizeof(TAniCurHeader)) + sizeof(RTAG);       
if (cbTitle > 0) OR (cbAuthor > 0) then
	begin
	if cbTitle > 0 then
		begin
		Inc(cbTitle); // add in ASCIIZ terminator
		Inc(cbINFO, SizeOf(RTAG)); //INAM Tag
		Inc(cbINFO, PadUp(cbTitle));
		end;
	if cbAuthor > 0 then
		begin
		Inc(cbAuthor); // add in ASCIIZ terminator
		Inc(cbINFO, SizeOf(RTAG)); //IART Tag
		Inc(cbINFO, PadUp(cbAuthor));
		end;
	inc(cbInfo, SizeOf(FourCC)); // INFO type
	Inc(cbFile, cbInfo);
	inc(cbFile, SizeOf(RTag)); // LIST tag
	end;

if fSeq then
	begin
	Inc(cbFile, SizeOf(RTag)); // seq tag
	Inc(cbFile, fSequences.Count*SizeOf(DWORD));
	end;
if fRate then
	begin
	Inc(cbFile, SizeOf(RTag)); // rate tag
	Inc(cbFile, fJifRates.Count*SizeOf(DWORD));
	end;

// Now we have all the structures built in memory, it's time to
// write them out in RIFF ACON format!

Tag.ckID:= FOURCC_RIFF;
Tag.ckSize:= cbFile;
if not WriteTag(Stream, Tag) then FErrOut;
if not WriteType(Stream, FOURCC_ACON) then FErrOut;

// If title or author data exists, write a LIST type block to hold them
if (cbTitle > 0) or (cbAuthor > 0) then
	begin
	Tag.ckID:= FOURCC_LIST;
	Tag.ckSize:= cbInfo;
	if not WriteTag(Stream, Tag) then FErrOut;
	if not WriteType(Stream, FOURCC_INFO) then FErrOut;
	// If title exists, write an INAM block
	if cbTitle > 0 then
		begin
		Tag.ckID:= FOURCC_INAM;
		tag.ckSize:= cbTitle;
		if not WriteTagString(Stream, Tag, fTitle) then FErrOut;
		end;
	// if author exists write an IART block
	if cbAuthor > 0 then
		begin
		Tag.ckID:= FOURCC_IART;
		tag.ckSize:= cbAuthor;
		if not WriteTagString(Stream, Tag, fAuthor) then FErrOut;
		end;
	end;

Tag.ckID:= FOURCC_anih;
Tag.ckSize:= SizeOf(fheader);
if not WriteTagData(Stream, Tag, fHeader) then FErrOut;

// If there's a list of separate rates for the steps, write it in a Rate block.
if fRate then
	begin
	Tag.ckID:= FourCC_Rate;
	Tag.ckSize:= fJifRates.Count*SizeOf(DWORD);
	if not WriteTag(Stream, Tag) then FErrOut;
	for i:= 0 to fJifRates.Count-1 do
		if not WriteType(Stream, DWORD(fJifRates[i])) then FErrOut;
	end;

// If there's a sequence list to match each step to a frame, write it in a _seq block
if fSeq then
	begin
	Tag.ckID:= FourCC_Seq;
	Tag.ckSize:= fSequences.Count*SizeOf(DWORD);
	if not WriteTag(Stream, Tag) then FErrOut;
	for i:= 0 to fSequences.Count-1 do
	        if not WriteType(Stream, DWORD(fSequences[i])) then FErrOut;
	end;
// Write a LIST block to hold the frames
Tag.ckID:= FOURCC_LIST;
Tag.ckSize:= cbFram;
if not WriteTag(Stream, Tag) then FErrOut;
if not WriteType(Stream, FOURCC_Fram) then FErrOut;
for i:= 0 to fFrams.Count-1 do
	begin
	// Write an ICON block for each frame
	Tag.ckID:= FOURCC_ICON;
	Tag.ckSize:= TInternalCur(fFrams[i]).DataSize;
	WriteTag(Stream, Tag);
	TInternalCur(fFrams[i]).SaveToStream(Stream);
	end;
end;



procedure TAniCursor.LoadFromFile(const fileName: String);
var Stream: TStream;
begin
Stream:= TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
try
LoadFromStream(Stream);
finally Stream.Free; end;
end;



procedure TAniCursor.SaveToFile(const fileName: String);
var Stream: TStream;
begin
Stream:= TFileStream.Create(FileName, fmCreate);
try
SaveToStream(Stream);
finally    Stream.Free;  end;
end;



procedure TAniCursor.AddImage(T: TInternalCur);
var TII: TInternalCur;
begin
fJifRates.Add(Pointer(T.Tag));
TII:= TInternalCur.Create(T);
TII.Tag:= 0;
fFrams.Add(TII);
fSequences.Add(Pointer(fSequences.Count));
end;



procedure TAniCursor.PrepareToSave(vHotX, vHotY: Word);
var J, N, M, P: Integer;
var Any, Match: Boolean;
begin
fHeader.cSteps:= fFrams.Count;
fHeader.cFrames:= fFrams.Count;
Any:= False;
J:= Integer(fJifRates[0]);
for N:= 1 to fJifRates.Count-1 do
 if Integer(fJifRates[N]) <> J then
	begin
	Any:= True;
	Break;
	end;
if not Any then
// if all the rates are the same, discard the list
	begin
	fJifRates.Clear;
	fHeader.cjifRate:= J;
	end;
Any:= False;
// The sequence list initially has the same number of items as
// the Frames list, and item "n" indexes the "nth" frame. If
// every frame is unique, discard the list. Otherwise, if frame
// M matches earlier frame N, discard frame M and adjust the
// sequence list

for N:= fHeader.cSteps-1 downto 1 do
	begin
	Match:=False;
	for M:= 0 to N-1 do
	 if TInternalCur(fFrams[N]).Same(TInternalCur(fFrams[M])) then
		begin
		Any:= True;
		Match:= True;
		fSequences[N]:= Pointer(M);
		// When we remove a frame, we must update all sequence
		// indices that point to *later* frames
		for P:= N+1 to fSequences.Count-1 do
		if Integer(fSequences[P]) > N then fSequences[P]:= Pointer(Integer(FSequences[P])-1);
		fFrams.Delete(N);
		Dec(fheader.cFrames);
		Break;
		end;
	if Match then Continue;
	end;
if Any then fHeader.cflags:= fHeader.cFlags OR AF_SEQUENCE
else fSequences.Clear;
for N:= 0 to fFrams.Count-1 do
 with TInternalCur(fFrams[N]) do
	begin
	HotX:= vHotX;
	HotY:= vHotY;
	end;
end;



procedure TAniCursor.FErrIn;
begin
Raise(Exception.Create('Error reading .ANI file'));
end;



procedure TAniCursor.FErrOut;
begin
Raise(Exception.Create('Error writing .ANI file'));
end;



function TAniCursor.GetFrameCount: Integer;
begin
Result:= fFrams.Count;
end;



function TAniCursor.GetHotX: WORD;
begin
if fFrams.Count > 0 then Result:= TInternalCur(fFrams[0]).HotX
else Result:= 0;
end;



function TAniCursor.GetHotY: WORD;
begin
if fFrams.Count > 0 then Result:= TInternalCur(fFrams[0]).HotY
else Result:= 0;
end;



function TAniCursor.GetIsPal16: Boolean;
begin
IF fFrams.Count > 0 then Result:= TInternalCur(fFrams[0]).Is16color
else Result:= True;
end;



function TAniCursor.GetRateCount: Integer;
begin
Result:= fJifRates.Count;
end;



function TAniCursor.GetSeqCount: Integer;
begin
Result:= fSequences.Count;
end;



function TAniCursor.GetPalette: TPalArray;
begin
IF fFrams.Count > 0 then Result:= TInternalCur(fFrams[0]).Palette
else FillChar(Result, SizeOf(Result), 0);
end;



function TAniCursor.GetFrame(Index: Integer): TInternalCur;
begin
if fSequences.Count = 0 then
	begin
	if (Index < 0) or (Index >= fFrams.Count) then Raise(Exception.Create('List index out of bounds'));
	Result:= TInternalCur(fFrams[Index]);
	end
else
	begin
	iF (Index < 0) or (Index >= fSequences.Count) then Raise(Exception.Create('List index out of bounds'));
	Result:= TInternalCur(fFrams[Sequences[Index]]);
	end;
end;



function TAniCursor.GetRate(Index: Integer): DWORD;
begin
if fJifRates.Count = 0 then Result:= fHeader.cjifRate
else
	begin
	if (Index < 0) or (Index >= fJifRates.Count) then Raise(Exception.Create('List index out of bounds'));
	Result:= DWORD(fJifRates[Index]);
	end;
end;



function TAniCursor.GetSequence(Index: Integer): DWORD;
begin
if fSequences.Count= 0 then Result:= Index
else
	begin
	if (Index < 0) or (Index >= fSequences.Count) then Raise(Exception.Create('List index out of bounds'));
	Result:= DWORD(fSequences[Index]);
	end;
end;



end.
