unit DBFLibrary;

interface

uses Classes,SysUtils,DSupport,StreamLib;

const ftChar = 'C';	{ Characters, all ascii }
const ftDate = 'D';	{Date ddmmyyyy, fix size 8}
const ftTime = 'T';	{Time hhmmss, fix size 6}
const ftFloat = 'F';	{Float point, -,.,0..9}
const ftLogical = 'L';	{Logical, ?,Y,y,N,n}
const ftMemo = 'M';	{Memo, as numeric, fix size 10}
const ftNumeric = 'N';	{Numeric, -,.,0..9}

type TDBFHeader = record
 Version: byte;		{ Should be 3 or $83 - Offset =0 }
 Year,Month,Day: byte;	{ Date of last update - Offset = 1 }
 NumRecs: longint;	{ Number of records in the file - Offset = 4 }
 HeadLen: word;		{ Length of the header - Offset = 8}
 RecLen: word;		{ Length of individual records  - Offset = 10 }
 Nets: word;		{ not used - Offset = 12 }
 Transaction: byte;	{ 0 = Not Protected, 01 = Protected - Offset = 14}
 Encrypted: byte;	{ 0 = Unencrypted, 1 = encrypted - Offset = 15}
 Network: array [1..12] of byte;	{ Offset = 16}
 Mdxfile: byte;		{ 00 = Does not exist, 01 = Exists - Offset = 28}
 LangDrv: byte;		{ language driver - Offset = 29 }
 Labeled: word;
 end;

type TChar11 = array [0..10] of char; {filename defs.}
type TFieldReq = (dbfUnique, dbfDuplicates);
type TSortByIndex = (dbfAscending, dbfDescending, dbfAlternative);


type TFieldDefinition = record
 Name: TChar11;			{ Name of the field  - Offset =  0 }
 FieldType: Char;		{ Type of data in this field - Offset =  11 }
 Data: array [0..1] of word;	{ Not used - Offset =  12 }
 FieldLength: byte;		{ Length of the field  - Offset =  14 }
 DecimalPlaces: byte;		{ Number of decimal places  - Offset =  15 }
 DupFlag: TFieldReq;		{ Duplicate   - Offset =  16 }
 SortFlag: TSortByIndex;	{ Sort method - Offset =  17 }
 DBFIdent: Byte;		{ datafield identifier ?? - Offset =  18 }
 FileIndex: TChar11;		{ here file name in index  XXXXXXXXIDX - Offset =  19 }
 end;

const  MaxFields = 128;  { maximum count for columns }


{Object for writing DBF files}
{ Steps to use this object: }

{ 1. Call InitializeDatabase. }
{ 2. Call AddField for each field. }
{ 3. For each row call Write routines for each field. }
{ 4. When a row has been written, call CommitDatabaseRow. }
{ 5. When all rows have been written, call WriteDataBase. }

type TDBFWriter = class
 private
   FDBFHeader: TDBFHeader;
   FFieldDefArray: array[0..MaxFields-1] of TFieldDefinition;
   FieldCount: integer;
   RawDataList: TStringList;
   RowFields: array [0..MaxFields] of string;
   procedure FinalizeHeader;
   procedure StoreString(S: string; FieldNumber,FieldLength: integer; Left: boolean);
   protected
 public
   constructor Create;
   destructor Destroy; override;
   procedure InitializeDatabase;
   procedure AddField(Name: string; FieldType: char; FieldLength,DecimalPlaces: byte);
   procedure WriteDouble(D: double; FieldNumber: integer);
   procedure WriteString(S: string; FieldNumber: integer);
   procedure WriteDeleted(FieldNumber: integer);
   procedure CommitDatabaseRow;
   procedure WriteDataBase(FileName: string);
 end;

implementation



function StringToChar11(S: string): TChar11;
{convert Delph string to 11 character array}
var I: integer;
begin
FillChar(Result,11,0);
for I:=1 to Length(S) do Result[I-1]:=S[I];
end;



procedure TDBFWriter.InitializeDatabase;
{Intialize database for new data}
var Y,M,D: word;
begin
FillChar(FDBFHeader,SizeOf(FDBFHeader),0);
with FDBFHeader do
 begin
 Version:=$3;
 DecodeDate(Date,Y,M,D);
 Year:=y Mod 100;
 Month:=Lo(M);
 Day:=Lo(D);
 LangDrv:=87;
 NumRecs:=0;
 HeadLen:=0;
 RecLen:=0;
 end;
FieldCount:=0;
RawDataList.Clear;
end;



constructor TDBFWriter.Create;
begin
RawDataList:=TStringList.Create;
InitializeDatabase;
end;



destructor TDBFWriter.Destroy;
begin
RawDataList.Free;
end;



procedure TDBFWriter.AddField(Name: string; FieldType: char; FieldLength,DecimalPlaces: byte);
{Specify a field for the data base}
begin
FillChar(FFieldDefArray[FieldCount],SizeOf(TFieldDefinition),0);
FFieldDefArray[FieldCount].Name:=StringToChar11(Name);
FFieldDefArray[FieldCount].FieldType:=FieldType;
FFieldDefArray[FieldCount].FieldLength:=FieldLength;
FFieldDefArray[FieldCount].DecimalPlaces:=DecimalPlaces;
Inc(FieldCount);
end;



procedure TDBFWriter.FinalizeHeader;
{Finalize all header data after all rows have been written}
begin
FDBFHeader.NumRecs:=RawDataList.Count;
FDBFHeader.HeadLen:=SizeOf(TDBFHeader)+SizeOf(TFieldDefinition) * FieldCount + 1;
if FDBFHeader.NumRecs>0 then FDBFHeader.RecLen:=Length(RawDataList.Strings[0]) + 1;
end;


{----------------- Field writing routines ------------------------}


procedure TDBFWriter.StoreString(S: string; FieldNumber,FieldLength: integer; Left: boolean);
{Store string versions of field data}
begin
if Length(S)>FieldLength then SetLength(S,FieldLength);
if Left then RowFields[FieldNumber]:=GetTabStrLeft(S,FieldLength)
else RowFields[FieldNumber]:=GetTabStrRight(S,FieldLength);
end;


procedure TDBFWriter.WriteString(S: string; FieldNumber: integer);
{Write a string field}
var FD: TFieldDefinition;
begin
FD:=FFieldDefArray[FieldNumber];
if FD.FieldType<>ftChar then raise Exception.Create('Field Mismatch');
StoreString(S,FieldNumber,FD.FieldLength,True);
end;



procedure TDBFWriter.WriteDouble(D: double; FieldNumber: integer);
{Write a double field}
var S: string;
var FD: TFieldDefinition;
begin
FD:=FFieldDefArray[FieldNumber];
if FD.FieldType<>ftNumeric then raise Exception.Create('Field Mismatch');
S:=FloatToStrF(D,ffFixed,18,FD.DecimalPlaces);
StoreString(S,FieldNumber,FD.FieldLength,False);
end;



procedure TDBFWriter.WriteDeleted(FieldNumber: integer);
{Write delete field}
var S: string;
var FD: TFieldDefinition;
var I: integer;
begin
FD:=FFieldDefArray[FieldNumber];
S:='';
for I:=1 to FD.FieldLength do S:=S + '*';
StoreString(S,FieldNumber,FD.FieldLength,False);
end;



procedure TDBFWriter.CommitDatabaseRow;
{Copy current row into database}
var I: integer;
var S: string;
begin
S:='';
for I:=0 to FieldCount-1 do S:=S+RowFields[I];
RawDataList.Add(S);
end;

{---------------------------------------------------------------}


procedure TDBFWriter.WriteDataBase(FileName: string);
{Write current database to specified file}
var FileStream: TFIleStream;
var I: integer;
var S: string;
begin
FileStream:=TFileStream.Create(Filename,fmCreate);
FinalizeHeader;
FileStream.Write(FDBFHeader,sizeof(TDBFHeader));

for I:=0 to FieldCount-1 do FileStream.Write(FFieldDefArray[I],sizeof(TDBFHeader));
WriteStreamChar(FileStream,CR_Char);

for I:=0 to RawDataList.Count-1 do
	begin
	WriteStreamChar(FileStream,SP_Char);
	S:=RawDataList.Strings[I];
	WriteStreamString(FileStream,S);
	end;

WriteStreamChar(FileStream,CtrlZ);
FileStream.Free;
end;

end.
