unit DEMLoader;

interface

uses Windows,SysUtils,Classes,Math,DEMLib,DSupport;


procedure ReadDEMFile(Stream: TStream; DEMArray: TDEMArray; Progress: TDEMProgress); overload;
procedure ReadDEMFile(Filename: string; DEMArray: TDEMArray; Progress: TDEMProgress); overload;
procedure CalcDimensions(var DEMInfo: TDEMInfo);

implementation

{------------------- Variables for Buffering DEM Data ----------------}


var Buffer: array [0..1024] of char;
var BufferPtr: integer;


var
  ProfileWidth,				{Width of all profiles }
  MaxProfLength: integer;		{ Maximum number of points a profile}


{------------------- Data For Individual Profiles ---------------------}


var ProfileInfo: TProfileInfo;



{----------------- Error Handling routines -----------------------------}

type TDEMErrors = (deCREmbedded,deLFEmbedded,deInvalidCorners,deInvalidProfile);

type TDEMErrorState = record
 EmbeddedCR: integer;
 EmbeddedLF: integer;
 InvalidCorners: integer;
 InvalidProfiles: integer;
 end;

var DEMErrorState: TDEMErrorState;


procedure ResetError;
begin
DEMErrorState.EmbeddedCR:=0;
DEMErrorState.EmbeddedLF:=0;
DEMErrorState.InvalidCorners:=0;
DEMErrorState.InvalidProfiles:=0;
end;


procedure SaveDEMError(DEMError: TDEMErrors);
begin
case DEMError of
 deCREmbedded: Inc(DEMErrorState.EmbeddedCR);
 deLFEmbedded: Inc(DEMErrorState.EmbeddedLF);
 deInvalidCorners: Inc(DEMErrorState.InvalidCorners);
 deInvalidProfile: Inc(DEMErrorState.InvalidProfiles);
 end;
end;


procedure FinalizeErrors(var DEMInfo: TDEMInfo);
var S: string;
var Cnt: integer;
begin
S:='';
Cnt:=0;
if DEMErrorState.EmbeddedCR>0 then
	begin
	S:=S+' Embedded Carriage Returns: '+IntToStr(DEMErrorState.EmbeddedCR)+CRLF_Char;
	Inc(Cnt);
	end;
if DEMErrorState.EmbeddedLF>0 then
	begin
	S:=S+' Embedded Line Feeds: '+IntToStr(DEMErrorState.EmbeddedLF)+CRLF_Char;
	Inc(Cnt);
	end;
if DEMErrorState.InvalidCorners>0 then
	begin
	S:=S+' Invalid Corners: '+IntToStr(DEMErrorState.InvalidCorners)+CRLF_Char;
	Inc(Cnt);
	end;
if DEMErrorState.InvalidProfiles>0 then
	begin
	S:=S+' Profiles Out Of Bounds: '+IntToStr(DEMErrorState.InvalidProfiles)+CRLF_Char;
	Inc(Cnt);
	end;
DEMInfo.ErrorString:=S;
DEMInfo.ErrorCount:=Cnt;
end;


{------------------- Routines for block buffered input ----------------}


procedure GetNewBuffer(Stream: TStream; var DEMInfo: TDEMInfo);
{Fill the data buffer }
var C: char;
var I: integer;
begin
for I:=0 to 1024-1 do
	begin
	Stream.ReadBuffer(C,1);
	case C of
	 CR_Char:
		begin
		Inc(DEMInfo.ErrorCount);
		SaveDEMError(deCREmbedded);
		repeat Stream.ReadBuffer(C,1) until (C<>CR_CHAR) and (C<>LF_CHAR);
		end;
	 LF_Char:
		begin
		Inc(DEMInfo.ErrorCount);
		SaveDEMError(deLFEmbedded);
		break;
		end;
	 end;
        Buffer[I]:=C;
	end;
BufferPtr:=0;
end;



function GetBufferChar(Stream: TStream; var DEMInfo: TDEMInfo): char;
{Get a character from the buffer}
begin
if BufferPtr>1023 then GetNewBuffer(Stream,DEMInfo);
Result:=Buffer[BufferPtr];
Inc(BufferPtr);
end;



procedure MODBufferFlush(Stream: TStream;  Size: integer; var DEMInfo: TDEMInfo);
{Flush buffer if item size exceed buffer size}
begin
if (BufferPtr+Size)>1023 then GetNewBuffer(Stream,DEMInfo);
end;


procedure ResetInputBuffer(Stream: TStream;  var DEMInfo: TDEMInfo);
{Rewind file and load the buffer }
begin
Stream.Position:=0;
GetNewBuffer(Stream,DEMInfo);
end;

{------------------- Routines to read Fortran style format data -------}

procedure BufferSkip(Stream: TStream;  Num,Size: integer; var DEMInfo: TDEMInfo);
{Skip items in buffer for speed}
var P: integer;
begin
for P:=1 to Num do
	begin
	if (BufferPtr+Size)>1023 then GetNewBuffer(Stream, DEMInfo);
	BufferPtr:=BufferPtr+Size;
	end;
end;


procedure GetNumberString(Stream: TStream; Size: integer; var Str: string; var DEMInfo: TDEMInfo);
{Read string eating blanks and converting D to E}
var C: integer;
var CH: char;
begin
Str:='';
MODBufferFlush(Stream,Size,DEMInfo);
for C:=1 to Size do
	begin
	CH:=GetBufferChar(Stream,DEMInfo);
	if CH='D' then Str:=Str+'E'
	else if CH<>' ' then Str:=Str+CH;
	end;
end;


function ReadInteger(Stream: TStream; Size: integer; var DEMInfo: TDEMInfo): integer;
{Read Size number of characters and convert to integer}
var S: string;
begin
GetNumberString(Stream,Size,S,DEMInfo);
Result:=StrToInt(S);
end;


function ReadReal(Stream: TStream; Size: integer; var DEMInfo: TDEMInfo): double;
{Read Size number of characters and convert to doubel}
var S: string;
begin
GetNumberString(Stream,Size,S,DEMInfo);
if S<>'' then Result:=StrToFloat(S) else Result:=0;
end;



function UnpaddedString(Stream: TStream; Size: integer; var DEMInfo: TDEMInfo): String;
{Read String, Remove padding characters}
var C: integer;
var S: string;
var Pad: boolean;
var Ch: char;
begin
Pad:=False;
S:='';
MODBufferFlush(Stream, Size, DEMInfo);
for C:=1 to Size do
	begin
	S:=S+GetBufferChar(Stream, DEMInfo);
	end;
Result:=Trim(S);
end;




procedure FixHeaderFaults(var DEMInfo: TDEMInfo);
{Fix faults in headers}
begin
if (DEMInfo.SWCorner.East = 0) and  (DEMInfo.SWCorner.North = 0) and
   (DEMInfo.NWCorner.East = 0) and  (DEMInfo.NWCorner.North = 0) and
   (DEMInfo.NECorner.East = 0) and  (DEMInfo.NECorner.North = 0) and
   (DEMInfo.SECorner.East = 0) and  (DEMInfo.SECorner.North = 0) then
	begin
	Inc(DEMInfo.ErrorCount);
	SaveDEMError(deInvalidCorners);
	if DEMInfo.MapType=mt7Minute then
		begin
		DEMInfo.SWCorner.East:=296149.78125; DEMInfo.SWCorner.North:=4388623.5;
		DEMInfo.NWCorner.East:=296517.28125; DEMInfo.NWCorner.North:=4402498.5;
		DEMInfo.NECorner.East:=307227.9375; DEMInfo.NECorner.North:=4402221.5;
		DEMInfo.SECorner.East:=306879.8125; DEMInfo.SECorner.North:=4388347;
		end
	else
		begin
		DEMInfo.SWCorner.East:=-388800; DEMInfo.SWCorner.North:=140400;
		DEMInfo.NWCorner.East:=-388800; DEMInfo.NWCorner.North:=144000;
		DEMInfo.NECorner.East:=-385200; DEMInfo.NECorner.North:=144000;
		DEMInfo.SECorner.East:=-385200; DEMInfo.SECorner.North:=140400;
		end;
	end;
end;



procedure CalcDimensions(var DEMInfo: TDEMInfo);
{Calcuatle array dimensions}
var NMax,NMin: double;
begin
DEMInfo.ActualCols:=DEMInfo.DEMRows * DEMInfo.DEMCols;
NMin:=Min(DEMInfo.SWCorner.North,DEMInfo.SECorner.North);
NMax:=Max(DEMInfo.NWCorner.North,DEMInfo.NECorner.North);
DEMInfo.SouthMax:=NMin;
DEMInfo.MapHeight:=NMax - NMin;
DEMInfo.ActualRows:=trunc(DEMInfo.MapHeight/DEMInfo.Resolution[0]);
end;




procedure ReadDEMHeader(Stream: TStream; var DEMInfo: TDEMInfo);
{Read header part of DEM file }
var P: integer;
begin
DEMInfo.ErrorCount:=0;
DEMInfo.ErrorString:='';
SetInternational(False);
DEMInfo.MapName:=UnpaddedString(Stream,144,DEMInfo);
DEMInfo.LevelCode:=ReadInteger(Stream,6,DEMInfo);
DEMInfo.Pattern:=ReadInteger(Stream,6,DEMInfo);
DEMInfo.GroundRef:=ReadInteger(Stream,6,DEMInfo);
DEMInfo.Zone:=ReadInteger(Stream,6,DEMInfo);
for P:=0 to 14 do DEMInfo.Projection[P]:=ReadReal(Stream,24,DEMInfo);
DEMInfo.GroundUnits:=TDEMUnits(ReadInteger(Stream,6,DEMInfo));
DEMInfo.ElevationUnits:=TDEMUnits(ReadInteger(Stream,6,DEMInfo));
DEMInfo.PolySides:=ReadInteger(Stream,6,DEMInfo);

DEMInfo.SWCorner.East:=ReadReal(Stream,24,DEMInfo);
DEMInfo.SWCorner.North:=ReadReal(Stream,24,DEMInfo);
DEMInfo.NWCorner.East:=ReadReal(Stream,24,DEMInfo);
DEMInfo.NWCorner.North:=ReadReal(Stream,24,DEMInfo);
DEMInfo.NECorner.East:=ReadReal(Stream,24,DEMInfo);
DEMInfo.NECorner.North:=ReadReal(Stream,24,DEMInfo);
DEMInfo.SECorner.East:=ReadReal(Stream,24,DEMInfo);
DEMInfo.SECorner.North:=ReadReal(Stream,24,DEMInfo);

DEMInfo.MinElevation:=ReadReal(Stream,24,DEMInfo);
DEMInfo.MaxElevation:=ReadReal(Stream,24,DEMInfo);
DEMInfo.Angle:=ReadReal(Stream,24,DEMInfo);
DEMInfo.Accuracy:=ReadInteger(Stream,6,DEMInfo);
for P:=0 to 2 do DEMInfo.Resolution[P]:=ReadReal(Stream,12,DEMInfo);
DEMInfo.DEMRows:=ReadInteger(Stream,6,DEMInfo);
DEMInfo.DEMCols:=ReadInteger(Stream,6,DEMInfo);


if DEMInfo.GroundRef=1 then DEMInfo.MapType:=mt7Minute
else DEMInfo.MapType:=mtOneDegree;
DEMInfo.HorzDatum:=hdNAD27;
DEMInfo.Scale:=0;
DEMInfo.DataID:='?';
DEMInfo.Date:='?';
DEMInfo.BoundBoxNW:=DEMInfo.NWCorner;
DEMInfo.DeRotated:=False;
FixHeaderFaults(DEMInfo);
CalcDimensions(DEMInfo);
end;



procedure ReadProfHead(Stream: TStream; var DEMInfo: TDEMInfo);
{Read profile header}
var Off: double;
begin
ProfileInfo.RowID:=ReadInteger(Stream,6,DEMInfo);
ProfileInfo.ColID:=ReadInteger(Stream,6,DEMInfo);
ProfileInfo.ProfRows:=ReadInteger(Stream,6,DEMInfo);
ProfileInfo.ProfCols:=ReadInteger(Stream,6,DEMInfo);
ProfileInfo.ProfileLength:=ProfileInfo.ProfRows*ProfileInfo.ProfCols;
ProfileInfo.ProfLocation.East:=ReadReal(Stream,24,DEMInfo);
ProfileInfo.ProfLocation.North:=ReadReal(Stream,24,DEMInfo);
ProfileInfo.ProfElevation:=ReadReal(Stream,24,DEMInfo);
ProfileInfo.ProfMinElevation:=ReadReal(Stream,24,DEMInfo);
ProfileInfo.ProfMaxElevation:=ReadReal(Stream,24,DEMInfo);
Off:=ProfileInfo.ProfLocation.North-DEMInfo.SouthMax;
ProfileInfo.PixelOffset:=trunc(MaxProfLength*(Off/DEMInfo.MapHeight));
if (ProfileInfo.PixelOffset>MaxProfLength) or (ProfileInfo.PixelOffset<0) then
	begin
	ProfileInfo.PixelOffset:=0;
	Inc(DEMInfo.ErrorCount);
	SaveDEMError(deInvalidProfile);
	end;
end;



procedure SetUpDEMRead(Stream: TStream; var DEMArray: TDEMArray);
{Setup for DEM read }
begin
ResetError;
DEMArray.ClearInfo;
ResetInputBuffer(Stream,DEMArray.DEMInfo);
ReadDEMHeader(Stream, DEMArray.DEMInfo);
end;



procedure ReadDEMFile(Stream: TStream; DEMArray: TDEMArray; Progress: TDEMProgress);
{Read DEM file from Stream into DEMArray}
var PRFX: integer;


	procedure HandleProgress(Percent: integer; State: TProgressState; Status: string);
	begin
	if Assigned(Progress) then Progress(Percent,State,Status);
	end;


	procedure ReadProfile(Stream: TStream);
	{Read a single DEM profile line }
       	var Elev,PRFY: integer;
	begin
	ReadProfHead(Stream,DEMArray.DEMINfo);
	SetLength(DEMArray.ProfInfoArray,Length(DEMArray.ProfInfoArray)+1);
	DEMArray.ProfInfoArray[High(DEMArray.ProfInfoArray)]:=ProfileInfo;
 	for PRFY:=0 to MaxProfLength-1 do
		begin
		if (PRFY>=ProfileInfo.PixelOffset) and
		   (PRFY<(ProfileInfo.PixelOffset+ProfileInfo.ProfileLength)) then
			begin
			Elev:=ReadInteger(Stream,6,DEMArray.DEMINfo);
			if Elev>-1000 then
				begin
				if Elev>DEMArray.DEMInfo.ActualMaxVert then
					begin
					DEMArray.DEMInfo.ActualMaxVert:=Elev;
					DEMArray.DEMInfo.MaxVertIndex:=Point(PRFX,PRFY);
					end;
				if Elev<DEMArray.DEMInfo.ActualMinVert then
					begin
					DEMArray.DEMInfo.ActualMinVert:=Elev;
					DEMArray.DEMInfo.MinVertIndex:=Point(PRFX,PRFY);
					end;
				end
			else Elev:=EmptyCell;
			end
		else Elev:=EmptyCell;
		DEMArray[PRFX,PRFY]:=Elev;
		end;
	end;


begin
HandleProgress(0,psStart,'Start');
SetLength(DEMArray.ProfInfoArray,0);
try
DEMArray.Clear(EmptyCell);
SetupDEMRead(Stream,DEMArray);
ProfileWidth:=DEMArray.DEMInfo.ActualCols;
MaxProfLength:=DEMArray.DEMInfo.ActualRows;

DEMArray.SetDimensions(ProfileWidth,MaxProfLength);
for PRFX:=0 to ProfileWidth-1 do
	begin
	GetNewBuffer(Stream,DEMArray.DEMINfo);
	ReadProfile(Stream);
	HandleProgress(MulDiv(100,PRFX,ProfileWidth),psWorking,'Loading');
	end;
FindDEMAngle(DEMArray);
DEMArray.Loaded:=True;
finally
 HandleProgress(100,psDone,'Done');
 SetInternational(True);
 FinalizeErrors(DEMArray.DEMInfo);
 end;
end;



procedure ReadDEMFile(Filename: string; DEMArray: TDEMArray; Progress: TDEMProgress);
var Stream: TMemoryStream;
begin
Stream:=TMemoryStream.Create;
try
Stream.LoadFromFile(Filename);
ReadDEMFile(Stream,DEMArray,Progress);
finally Stream.Free; end;
end;


end.
