unit DEMWriter;

interface

uses Windows,SysUtils,Classes,DSupport,DEMLib,StreamLib;

procedure WriteDEMFile(Stream: TStream; DEMArray: TDEMArray; NewLines: boolean; Progress: TDEMProgress); overload;
procedure WriteDEMFile(Filename: string; DEMArray: TDEMArray; NewLines: boolean; Progress: TDEMProgress); overload;


implementation

{----------------- Support Routines -------------------------------------------}


function RemainderFloat(const X, Y: double): double;
begin
Result := X - Int(X / Y) * Y;
end;


{-------------------- Write Routines-----------------------------------}

const BufSize=1024;

type TDEMBuffer = array [0..BufSize-1] of char;

var WriteBuf: TDEMBuffer;
var WritePtr: integer;


procedure WriteBuffer(Stream: TStream);
var P: integer;
begin
for P:=WritePtr to BufSize-1 do WriteBuf[P]:=#$20;
for P:=0 to BufSize-1 do
	begin
	WriteStreamChar(Stream,WriteBuf[P]);
	end;
WritePtr:=0;
end;



procedure WriteString(Stream: TStream; S: string);
var P,L: integer;
begin
L:=Length(S);
if (WritePtr+L)>=BufSize then WriteBuffer(Stream);
for P:=1 to L do begin WriteBuf[WritePtr]:=S[P]; Inc(WritePtr); end;
end;



procedure WriteTabRight(Stream: TStream; S: string; Size: integer);
begin
WriteString(Stream,GetTabStrRight(S,Size));
end;


procedure WriteTabLeft(Stream: TStream; S: string; Size: integer);
begin
WriteString(Stream,GetTabStrLeft(S,Size));
end;



procedure WriteInteger(Stream: TStream; V,S: integer);
begin
WriteTabRight(Stream,IntToStr(V),S);
end;



procedure WriteReal24D15X(Stream: TStream; V: double);
var S: string;
var P,L: integer;
begin
if V=0 then S:='   0.0                  '
else
	begin
	S:=FloatToStrF(V,ffExponent,16,2);
	L:=Pos('.',S); for P:=1 to 5-L do S:=' '+S;
	end;
L:=Pos('E',S); if L>0 then S[L]:='D';
WriteString(Stream,S);
end;



procedure WriteReal24D15(Stream: TStream; V: double);
var S: string;
var P,L: integer;
begin
if V=0 then S:='  0.000000000000000D+000'
else
	begin
	S:=FloatToStrF(V,ffExponent,16,3);
	L:=Pos('.',S);
	for P:=1 to 4-L do S:=' '+S;
	end;
L:=Pos('E',S); if L>0 then S[L]:='D';
WriteString(Stream,S);
end;


procedure WriteReal12E6(Stream: TStream; V: double);
var S: string;
begin
if V=0 then S:='0.0         '
else S:=FloatToStrF(V,ffExponent,7,2);
WriteString(Stream,S);
end;

{------------------------------------------------------------------}


procedure WriteDEMHeader(Stream: TStream; DEMInfo: TDEMInfo);
{Read header part of DEM file }
var P: integer;
begin
WriteTabLeft(Stream,DEMInfo.MapName,144);
WriteInteger(Stream,DEMInfo.LevelCode,6);
WriteInteger(Stream,DEMInfo.Pattern,6);
WriteInteger(Stream,DEMInfo.GroundRef,6);
WriteInteger(Stream,DEMInfo.Zone,6);
for P:=0 to 14 do WriteReal24D15(Stream,DEMInfo.Projection[P]);
WriteInteger(Stream,integer(DEMInfo.GroundUnits),6);
WriteInteger(Stream,integer(DEMInfo.ElevationUnits),6);
WriteInteger(Stream,DEMInfo.PolySides,6);

WriteReal24D15(Stream,DEMInfo.SWCorner.East);
WriteReal24D15(Stream,DEMInfo.SWCorner.North);
WriteReal24D15(Stream,DEMInfo.NWCorner.East);
WriteReal24D15(Stream,DEMInfo.NWCorner.North);
WriteReal24D15(Stream,DEMInfo.NECorner.East);
WriteReal24D15(Stream,DEMInfo.NECorner.North);
WriteReal24D15(Stream,DEMInfo.SECorner.East);
WriteReal24D15(Stream,DEMInfo.SECorner.North);

WriteReal24D15(Stream,DEMInfo.MinElevation);
WriteReal24D15(Stream,DEMInfo.MaxElevation);
WriteReal24D15(Stream,DEMInfo.Angle);
WriteInteger(Stream,DEMInfo.Accuracy,6);
for P:=0 to 2 do WriteReal12E6(Stream,DEMInfo.Resolution[P]);
WriteInteger(Stream,DEMInfo.DEMRows,6);
WriteInteger(Stream,DEMInfo.DEMCols,6);
end;



procedure WriteArcInfoExtensions(Stream: TStream; DEMInfo: TDEMInfo);
{ Write arcinfo extensions for datum info  }
begin
WriteString(Stream, '                         2');
WriteInteger(Stream, integer(DEMInfo.HorzDatum),2);
end;



procedure WriteProfHead(Stream: TStream; ProfInfo: TProfileInfo);
{Write profile header}
begin
WriteInteger(Stream,ProfInfo.RowID,6);
WriteInteger(Stream,ProfInfo.ColID,6);
WriteInteger(Stream,ProfInfo.ProfRows,6);
WriteInteger(Stream,ProfInfo.ProfCols,6);
WriteReal24D15(Stream,ProfInfo.ProfLocation.East);
WriteReal24D15(Stream,ProfInfo.ProfLocation.North);
WriteReal24D15(Stream,ProfInfo.ProfElevation);
WriteReal24D15(Stream,ProfInfo.ProfMinElevation);
WriteReal24D15(Stream,ProfInfo.ProfMaxElevation);
end;



procedure WriteDEMFile(Stream: TStream; DEMArray: TDEMArray; NewLines: boolean; Progress: TDEMProgress); overload;
var GlobalMax, GlobalMin: Integer;
var ColInx, RowInx: integer;
var FirstValid, LastValid, FirstInx, LastInx: integer;
var I, N: Integer;
var RecordCount, CellInx : Integer;
var CellValue: Integer;
var XMin, YMin : double;
var MissingCount: Integer;
var DEMWidth, DEMHeight: double;
var ProfInfo: TProfileInfo;
var DEMProfRows, DEMProfCols: Integer;
const CRecordData: array [0..11-1] of integer = (0,0,0,0,0,0,0,0,0,0,0 );



	procedure HandleProgress(Percent: integer; State: TProgressState; Status: string);
	begin
	if Assigned(Progress) then Progress(Percent,State,Status);
	end;



begin
HandleProgress(0,psStart,'Start');
MissingCount:=0;
RecordCount := 1;

DEMWidth := DEMArray.DEMInfo.Resolution[0];
DEMHeight := DEMArray.DEMInfo.Resolution[1];

XMin := DEMArray.DEMInfo.BoundBoxNW.East;
YMin := DEMArray.DEMInfo.BoundBoxNW.North - (DEMArray.DEMInfo.DEMRows - 1)  * DEMHeight;

{  ensure that profile aligns with grid (for pre-y2001 files)  }
if  (RemainderFloat(XMin, DEMWidth) > 0.0001)  or  (RemainderFloat(YMin, DEMHeight) > 0.0001)  then
	begin
	if (DEMArray.DEMInfo.SWCorner.East < DEMArray.DEMInfo.NWCorner.East) then XMin := DEMArray.DEMInfo.SWCorner.East else XMin := DEMArray.DEMInfo.NWCorner.East;
	if (DEMArray.DEMInfo.SWCorner.North < DEMArray.DEMInfo.SECorner.North) then YMin := DEMArray.DEMInfo.SWCorner.North else YMin := DEMArray.DEMInfo.SECorner.North;
	XMin := DEMWidth  * (XMin / DEMWidth + 1);
	YMin := DEMHeight * (YMin / DEMHeight + 1);
	end;

RecordCount := 1;
WriteDEMHeader(Stream,DEMArray.DEMInfo);
WriteArcInfoExtensions(Stream,DEMArray.DEMInfo);
WriteBuffer(Stream);
if newlines then Stream.Write(CRLF_Char,2);

RecordCount := RecordCount + 1;
{  Write all columns in the DEM Array starting at the bottom row  }
for ColInx:= 0 to DEMArray.DEMInfo.DEMCols-1 do
	begin
	HandleProgress(MulDiv(100,ColInx,DEMArray.DEMInfo.DEMCols),psWorking,'Writing');
 	WriteString(Stream,'');

	{  figure out the bounds of the valid (not missing) data  }
	for LastValid:= DEMArray.Rows-1 downto 0 do if DEMArray[ColInx,LastValid]<>EmptyCell then break;
	for FirstValid := 0 to LastValid-1 do if DEMArray[ColInx,FirstValid]<>EmptyCell then break;
	DEMProfRows := LastValid - FirstValid + 1;
	FirstValid:=0; LastValid:=DEMArray.Rows-1; //****
	FirstInx := FirstValid;
	LastInx := FirstValid + 146 - 1;
	if LastInx > LastValid then LastInx := LastValid;

	{ Write B records. Note that the first record for each column will
	include the header and, at most, 146 elevations. Each subsequent
	record for the column can hold 170 elevations }
	WriteProfHead(Stream,DEMArray.ProfInfoArray[ColInx]);
	for N:=FirstInx to LastInx do
		begin
		WriteInteger(Stream,round(DEMArray[ColInx,N]),6);
		if DEMArray[ColInx,N] = EmptyCell then Inc(MissingCount);
		end;
	WriteBuffer(Stream);
	if newlines then Stream.Write(CRLF_Char,2);


	RecordCount := RecordCount + 1;
	while LastInx <> LastValid  do
		begin
		{  we have to process addition output lines  }
		FirstInx := LastInx + 1;
		LastInx := LastInx + 170;
		if LastInx > LastValid then LastInx := LastValid;

		for N:= FirstInx to LastInx do
			begin
			WriteInteger(Stream,round(DEMArray[ColInx,N]),6);
			if DEMArray[ColInx,N] = EmptyCell then Inc(MissingCount);
			end;
		WriteBuffer(Stream);
		if newlines then Stream.Write(CRLF_Char,2);
		RecordCount := RecordCount + 1;
		end;
	end;

{  write record c  }
WriteString(Stream,' ');
for I:= 0 to 9 do WriteInteger(Stream,CRecordData[I],6);
WriteBuffer(Stream);
if newlines then Stream.Write(CRLF_Char,2);
HandleProgress(0,psDone,'Done');
end;



procedure WriteDEMFile(Filename: string; DEMArray: TDEMArray; NewLines: boolean; Progress: TDEMProgress); overload;
var FS: TFileStream;
begin
FS:=TFileStream.Create(Filename,fmOpenWrite or fmCreate);
try
WriteDemFile(FS,DEMArray,NewLines,Progress);
finally FS.Free; end;
end;


end.
