{
Based on Paul Bourke's "Triangulate" algorithm and C version
With clarification from Dr Steve Evans Pascal version

Concaved Hull based on the paper:

"Efficient generation of simple polygons for characterizing
the shape of a set of points in the plane"

By Duckham, Kulik, Worboys and Galton
}

unit Delaunay;

interface

uses SysUtils, Dialogs, Graphics, Math, Forms, Types, Vector;


const MinSpacing = 0.000001;	{Minimum Spacing between newly added points}


{Used to hold Bounding box for all vertices}

type TBoundBox2D = record
 Min,Max: T2DVector;
 end;

{Structure to hold a triangle and related information}

type TDelTriangle = record
 V1,V2,V3: integer;
 Adj1,Adj2,Adj3: integer;
 CircleValid: boolean;
 Center: T2DVector;
 Radius: double;
 OutOfRange: boolean;
end;

{Structure to hold polygon edges}

type TEdge = record
 P1,P2: integer;
 end;

type TEdgeArray = array of TEdge;
//type PEdgeArray = ^TEdgeArray;

{Dynamic arrays for Vertices and Triangles}

type TVertArray = T2DVectorArray;
type TIntegerArray = array of integer;

type TTriangleArray = array of TDelTriangle;



{Types used for Convex and Concaved Hulls}

type TTriSide = (tsSide1,tsSide2,tsSide3,tsNone);

type THullEdge = record
 Edge: TEdge;
 TriInx: integer;
 Side: TTriSide;
 end;

type THullArray = array of THullEdge;



type TTriangulator = class
 private
  FTrianglesDone: boolean;
  FConvexDone: boolean;
  FConcaveDone: boolean;
  EdgeMin,EdgeMax: double;
  FConcaveLength: double;
  FPercentLength: double;
  function IsInCircle(P,P1,P2,P3: T2DVector;  var Center: T2DVector; var Radius: double; J: Integer): boolean;
  procedure GetBoundsTriangle(var P1, P2, P3: T2DVector);
  procedure QuickSort;
  function GetTriangleCount: integer;
  function GetVertexCount: integer;
  procedure GetNewPoint(HE: THullEdge; var NewInx,Adj1,Adj2: integer);
  procedure OrderHull(var Hull: THullArray);
  procedure CopyHull(var Source, Dest: THullArray);
  function GetEdgeMag(Edge: TEdge): double;
  procedure SortHullByLength(var Hull: THullArray);
  procedure CopySortByLen(var Source, Dest: THullArray);
  procedure SetNotTriangulated;
  procedure SetPercentLength(const Value: double);
  procedure SetConcaveLength(const Value: double);
  function RemoveDuplicates: integer;
  procedure CopySortByArea(var Source, Dest: THullArray);
  function GetTriArea(Edge: THullEdge): double;
  procedure SortHullByArea(var Hull: THullArray);
  procedure SetVertArrayLength(Len: integer);
 protected
 public
  VertArray: TVertArray;
  OriginalOrder,UserTags: TIntegerArray;
  TriArray: TTriangleArray;
  ConvexHull: THullArray;
  ConcaveHull: THullArray;
  Name: string;
  constructor Create;
  destructor Destroy; override;
  procedure Reset;
  procedure Triangulate;
  procedure GetConvexHull;
  procedure GetConcaveHull;
  procedure GetConcaveHull2;
  procedure GetEdgeMinMax(var Min,Max: double);
  procedure AddPoint(X,Y: double); overload;
  procedure AddPoint(X,Y: double; Tag: integer); overload;
  function GetBoundingBox: TBoundBox2D;
  property TriCount: integer read GetTriangleCount;
  property VertCount: integer read GetVertexCount;
  property TrianglesDone: boolean read FTrianglesDone;
  property ConvexDone: boolean read FConvexDone;
  property ConcaveDone: boolean read FConcaveDone;
  property ConcaveLength: double read FConcaveLength write SetConcaveLength;
  property PercentLength: double read FPercentLength write SetPercentLength;
  end;

implementation




{------------------------------------------------------------------------------}

procedure TTriangulator.SetNotTriangulated;
begin
FTrianglesDone:=False;
FConvexDone:=False;
FConcaveDone:=False;
end;


procedure TTriangulator.SetVertArrayLength(Len: integer);
begin
SetLength(VertArray,Len);
SetLength(OriginalOrder,Len);
SetLength(UserTags,Len);
end;



procedure TTriangulator.Reset;
begin
SetVertArrayLength(0);
SetLength(TriArray,0);
SetLength(ConvexHull,0);
SetLength(ConcaveHull,0);
SetNotTriangulated;
end;



constructor TTriangulator.Create;
begin
FConcaveLength:=150;
Reset;
end;



destructor TTriangulator.Destroy;
begin
inherited;
end;



{------------------------------------------------------------------------------}



procedure TTriangulator.QuickSort;
{Sort all points by x and y}

	function VectorCompare(V1,V2: T2DVector): integer;
	{Compare 3D vectors. Returns 0 if equal, <0 if }
	{Returns negative if V1<V2 and positive if V1>V2}
	begin
	if Abs(V1.X-V2.X)>MinSpacing then Result:=Math.Sign(V1.X-V2.X)
	else if Abs(V1.Y-V2.Y)>MinSpacing then Result:=Math.Sign(V1.Y-V2.Y)
	else Result:=0;
	end;

	procedure DoQuickSort(L, R: Integer);
	{Sort vertices in increasing values of X}
	var I, J,IT: Integer;
	var T,C: T2DVector;
	begin
	repeat
		begin
		I := L;
		J := R;
		C := VertArray[(L + R) shr 1];
		repeat
			begin
			while VectorCompare(VertArray[I],C)<0 do Inc(I);
			while VectorCompare(VertArray[J],C)>0 do Dec(J);
			if I <= J then
				begin
				T:=VertArray[I]; VertArray[I]:=VertArray[J]; VertArray[J]:=T;
				IT:=OriginalOrder[I]; OriginalOrder[I]:=OriginalOrder[J]; OriginalOrder[J]:=IT;
				IT:=UserTags[I]; UserTags[I]:=UserTags[J]; UserTags[J]:=IT;
				Inc(I);
				Dec(J);
				end;
			end
		until I > J;
		if L < J then DoQuickSort(L, J);
		L := I;
		end
	until I >= R;
	end;


begin
if High(VertArray)>0 then DoQuickSort(0, High(VertArray));
end;



{------------------------------------------------------------------------------}

function GetCircleCenter(P1,P2,P3: T2DVector): T2DVector;
{Calculate the center of the Circle based on the three points of a triangle}
{Works by calculating the perpendicular bisector of two sides of the triangle.}
{The Center is at the intersection of the two bisectors}
var PerpSlope1,PerpSlope2: double;
var Bisect1,Bisect2: T2DVector;
var S: string;
begin
if (Abs(P1.Y - P2.Y) < MinSpacing) and
   (Abs(P2.Y - P3.Y) < MinSpacing) then Result:=P1
else if Abs(P2.Y - P1.Y) < MinSpacing then
	begin
	{Slope of P1-P2 is near zero so the slope of the perpendicular is near infinity}
	{So the intersection = X-Center = Bisect of P1.X -> P2.X}
	{And Y-Center = Slope * X-Distance + Y-Bisect}
	PerpSlope2:= -(P3.X - P2.X) / (P3.Y - P2.Y);
	Bisect2:=ScalarDivide2D(VectorAdd2D(P2,P3),2);
	Result.X:= (P2.X + P1.X) / 2;
	Result.Y:= PerpSlope2 * (Result.X - Bisect2.X) + Bisect2.Y;
	end
else if Abs(P3.Y - P2.Y) < MinSpacing then
	begin
	{Slope of P2-P3 is near zero so the perpendicular is near infinity}
	{So the intersection = X-Center = Bisect of P1.X -> P2.X}
	{And Y-Center = Slope * X-Distance + Y-Bisect}
	PerpSlope1:= -(P2.X - P1.X) / (P2.Y - P1.Y);
	Bisect1:=ScalarDivide2D(VectorAdd2D(P1,P2),2);
	Result.X:= (P3.X + P2.X) / 2;
	Result.Y:= PerpSlope1 * (Result.X - Bisect1.X) + Bisect1.Y;
	end
else
	begin
	{Both slopes are not zero so the perpendicular are not infinite}
	PerpSlope1:= -(P2.X - P1.X) / (P2.Y - P1.Y);
	PerpSlope2:= -(P3.X - P2.X) / (P3.Y - P2.Y);
	Bisect1:=ScalarDivide2D(VectorAdd2D(P1,P2),2);
	Bisect2:=ScalarDivide2D(VectorAdd2D(P2,P3),2);
	if (PerpSlope1-PerpSlope2)<>0 then
		begin
		{Calculate the insection between the two perpendicular bisectors}
		Result.X:= (PerpSlope1 * Bisect1.X - PerpSlope2 * Bisect2.X + Bisect2.Y - Bisect1.Y) / (PerpSlope1 - PerpSlope2);
		Result.Y:= PerpSlope1 * (Result.X - Bisect1.X) + Bisect1.Y;
		end
	else
		begin
		{Handle the case were both lines are co-linear}
		Result:=ScalarDivide2D(VectorAdd2D(VectorAdd2D(P1,P2),P3),3);
		end;
	end;
end;



function TTriangulator.IsInCircle(P,P1,P2,P3: T2DVector; var Center: T2DVector; var Radius: double; J: Integer): Boolean;
{Returns True if the point "P" lies inside the Triangle's Circle)
{Triangle for the Circle is defined by points P1,P2, and P3 }
{Also returns the Center and Radius of the Circle}
{Note: Any point on the edge of the circle is considered inside.}
var Delta: T2DVector;
var RadiusSqr: double;
var DistSqr: double;
begin
{Have the Radius and Center of Circle already been calculated?}
if  TriArray[J].CircleValid then
	begin
	Center:= TriArray[J].Center;
	Radius:= TriArray[J].Radius;
	RadiusSqr:= Radius*Radius;
	Delta:=VectorSubtract2D(P,Center);
	DistSqr:= Delta.X * Delta.X + Delta.Y * Delta.Y;
	end
else
	begin
	Center:=GetCircleCenter(P1,P2,P3);
	Delta:=VectorSubtract2D(P2,Center);
	RadiusSqr:= Delta.X * Delta.X + Delta.Y * Delta.Y;
	Radius:= Sqrt(RadiusSqr);
	{Attempt to deal with more than 3 co-circular points}
	RadiusSqr:=RadiusSqr * 1.0000001;
	Delta:=VectorSubtract2D(P,Center);
	DistSqr:= Delta.X * Delta.X + Delta.Y * Delta.Y;

	{store the Center and Radius for later use}
	TriArray[J].CircleValid:=True;
	TriArray[J].Center:=Center;
	TriArray[J].Radius:=Radius;
	end;
{Test if point inside the circle radius}
Result:=DistSqr <= RadiusSqr;
end;

{------------------------------------------------------------------------------}



function TTriangulator.GetBoundingBox: TBoundBox2D;
{Generate a rectangular bounding box from all vertices}
var I: integer;
begin
if Length(VertArray)<1 then exit;
Result.Min:=VertArray[0];
Result.Max:=VertArray[0];
for I:= 1 To High(VertArray) do
	begin
	if VertArray[I].x < Result.Min.X then Result.Min.X:= VertArray[I].x;
	if VertArray[I].x > Result.Max.X then Result.Max.X:= VertArray[I].x;
	if VertArray[I].y < Result.Min.Y then Result.Min.Y:= VertArray[I].y;
	if VertArray[I].y > Result.Max.Y then Result.Max.Y:= VertArray[I].y;
	end;
end;



procedure TTriangulator.GetBoundsTriangle(var P1,P2,P3: T2DVector);
{Generate Equalateral triangle that bounds all vertices}
{The height and width are 2 times "Size" times the Max Dimension of vertex bounding box}
{A large value for "Size" is required to assure that a convex hull is generated}
var Box: TBoundBox2D;
var BoxCenter: T2DVector;
var BoxSize: T2DVector;
var MaxDim: double;
const Size = 100;
begin
Box:=GetBoundingBox;
BoxSize:=VectorSubtract2D(Box.Max,Box.Min);
if BoxSize.X > BoxSize.Y then MaxDim:= BoxSize.X
else MaxDim:= BoxSize.Y;
BoxCenter:=ScalarDivide2D(VectorAdd2D(Box.Max,Box.Min),2);
P1:= MakeVector2D(BoxCenter.X - Size * MaxDim,BoxCenter.Y - MaxDim);
P2:= MakeVector2D(BoxCenter.X,BoxCenter.Y + Size * MaxDim);
P3:= MakeVector2D(BoxCenter.X + Size * MaxDim,BoxCenter.Y - MaxDim);
end;



function TTriangulator.RemoveDuplicates: integer;
{Remove duplicates - relies on the list being }
{sorted so dupes are adjacent to each other}
var I,J: integer;
begin
J:=0;
for I:=1 to High(VertArray) do
	begin
	if (Abs(VertArray[J].X-VertArray[I].X)>=MinSpacing) or
	   (Abs(VertArray[J].Y-VertArray[I].Y)>=MinSpacing) then
		begin
		Inc(J);
		VertArray[J]:=VertArray[I];
		OriginalOrder[J]:=OriginalOrder[I];
		UserTags[J]:=UserTags[I];
		end
	{Always preserve the lowest duplicate}                                        
	else if OriginalOrder[J]>OriginalOrder[I] then
		begin
		OriginalOrder[J]:=OriginalOrder[I];
		UserTags[J]:=UserTags[I];
		end;
	end;
Result:=J+1;
SetVertArrayLength(Result);
end;



procedure TTriangulator.Triangulate;
{Produces a Delanuay Triangulation of points in VertArray}
{Results are stored as triangles in TriaArray}
{The edges of the triangles are arranged in counter-clockwise order}
var Edges: TEdgeArray;
var EdgeCnt,VertCnt: integer;
var I,J,k: Integer;
var TriCnt: Integer;
var CCenter,P1,P2,P3: T2DVector;
var Radius: double;
var InCircle: Boolean;
begin
{Make sure we have enough points to for at least one triangle}
if Length(VertArray) <3 then exit;
{Sort vertices by the X-Axis}
QuickSort;
RemoveDuplicates;

{Get the current count of Vertices before we insert the super triangle}
VertCnt:=Length(VertArray);
{Allocate edge space - Max Edges = 3 * Points}
SetLength(Edges,VertCnt*3);

{Allocate Triangle space - Normally this would be 2 times the vertex count}
{But it is actually 3 times the count because extra triangles will be added and removed}
SetLength(TriArray,VertCnt * 3);

{Add vertices for bounding triangle to vertex list}
GetBoundsTriangle(P1,P2,P3);
SetVertArrayLength(Length(VertArray)+3);
VertArray[VertCnt+0]:=P1;
VertArray[VertCnt+1]:=P2;
VertArray[VertCnt+2]:=P3;

{And Make it the first triangle in the list}
TriArray[0].V1:= VertCnt + 0;
TriArray[0].V2:= VertCnt + 1;
TriArray[0].V3:= VertCnt + 2;
TriArray[0].Adj1:=-1;
TriArray[0].Adj2:=-1;
TriArray[0].Adj3:=-1;
TriArray[0].CircleValid:=False;
TriArray[0].OutOfRange:= False;
TriCnt:= 1;

{Add each point one at a time to the existing triangulation}
for I:= 0 to VertCnt-1 do
	begin
	EdgeCnt:= 0; J:= 0;
	{Iterate through all existing triangles }
	while J<TriCnt do
		begin
		{Ignore triangles whose are already circles are out of range}
		if not TriArray[J].OutOfRange then
			begin
			{Test if point is in the circle specified by }
			InCircle:=IsInCircle(VertArray[I], VertArray[TriArray[J].V1], VertArray[TriArray[J].V2], VertArray[TriArray[J].V3], CCenter, Radius,J);
			{The X-Coordinates of points are sorted, so once a point is outside the }
			{ triangle circle, all other points will be and the triangle can be ignored }
			if (CCenter.X + Radius) < VertArray[I].x then TriArray[J].OutOfRange:= True
			else if InCircle then
				begin
				{If the point is inside the triangle's circle, add the triangle edges to the list}
				Edges[EdgeCnt + 0].P1:= TriArray[J].V1;
				Edges[EdgeCnt + 0].P2:= TriArray[J].V2;
				Edges[EdgeCnt + 1].P1:= TriArray[J].V2;
				Edges[EdgeCnt + 1].P2:= TriArray[J].V3;
				Edges[EdgeCnt + 2].P1:= TriArray[J].V3;
				Edges[EdgeCnt + 2].P2:= TriArray[J].V1;
				EdgeCnt:= EdgeCnt + 3;
				{Trick to delete triangle from list: Overwrite current triangle}
				{with last Triangle and decrement count - changes list order}
				TriArray[J]:=TriArray[TriCnt-1];
				TriArray[TriCnt-1].CircleValid:=False;
				TriCnt:= TriCnt - 1;
				J:= J - 1;
				end;
			end;
		J:= J + 1;
		end;

	{ Form the polygon that is created by all the combined triangles whose circles enclose the point }
	{ Tag all duplicated edges - effectively removing edges from the interior of the polygon }
	{ Since triangle edges are specified counter clockwise, all interior edges of the polygon are clockwise}
	for J:= 0 To EdgeCnt - 2 do
	 if not (Edges[J].P1 < 0) and not (Edges[J].P2 < 0) then
		begin
		for K:= J + 1 to EdgeCnt-1 do
		 if not (Edges[K].P1 < 0) and not (Edges[K].P2 < 0) then
		 	begin
		 	if (Edges[J].P1 = Edges[K].P2) and (Edges[J].P2 = Edges[K].P1) then
		 		begin
		 		Edges[J].P1:= -1;
		 		Edges[J].P2:= -1;
		 		Edges[K].P1:= -1;
		 		Edges[K].P2:= -1;
		 		end;
		 	end;
		end;

	{ Form new triangles by connecting the vertices of the polygon to the new point}
	for J:= 0 to EdgeCnt-1 do
		begin
		{ Ignore any tagged edges with indices of -1}
		if not (Edges[J].P1 < 0) and not (Edges[J].P2 < 0) then
			begin
			TriArray[TriCnt].V1:= Edges[J].P1;
			TriArray[TriCnt].V2:= Edges[J].P2;
			TriArray[TriCnt].V3:= I;
			TriArray[TriCnt].Adj1:=-1;
			TriArray[TriCnt].Adj2:=-1;
			TriArray[TriCnt].Adj3:=-1;
			TriArray[TriCnt].CircleValid:=False;
			TriArray[TriCnt].OutOfRange:= False;
			TriCnt:= TriCnt + 1;
			end;
		end;
	end;

{Remove Triangles with supertriangle vertices}
{These are triangles which have a vertex number greater than VertCnt-1 }
I:= 0;
while I<TriCnt do
	begin
	if (TriArray[I].V1 >= VertCnt) or
	   (TriArray[I].V2 >= VertCnt) or
	   (TriArray[I].V3 >= VertCnt) then
		begin
                TriArray[I]:= TriArray[TriCnt-1];
		I:= I - 1;
		TriCnt:= TriCnt - 1;
		end;
	I:= I + 1;
	end;

{Free memory and Trim off the super triangle}
SetLength(TriArray,TriCnt);
SetVertArrayLength(Length(VertArray)-3);
FTrianglesDone:=True;
end;


{------------------------------------------------------------------------------}



function TTriangulator.GetEdgeMag(Edge: TEdge): double;
{Calculate the unsigned length of an Edge}
begin
Result:=VectorAbs2D(VectorSubtract2D(VertArray[Edge.P2],VertArray[Edge.P1]));
end;



procedure TTriangulator.OrderHull(var Hull: THullArray);
{Sort hull vertices so the lines are in order - counter clockwise}
var I,J: integer;
var ET: THullEdge;
begin
for I:=0 to High(Hull)-1 do
 if Hull[I].Edge.P2<>Hull[I+1].Edge.P1 then
  for J:=I+2 to High(Hull) do
   if Hull[I].Edge.P2=Hull[J].Edge.P1 then
	begin
	ET:=Hull[I+1];
	Hull[I+1]:=Hull[J];
	Hull[J]:=ET;
	end;
end;






procedure TTriangulator.GetConvexHull;
{Extract Convex Hull from Delaunay Triangulation}
var I: integer;

	procedure HandleMinMax(E: TEdge);
	var Mag: double;
	begin
	Mag:=GetEdgeMag(E);
	if Mag>EdgeMax then EdgeMax:=Mag;
	if Mag<EdgeMin then EdgeMin:=Mag;
	end;


	function IsDuplicate(P1,P2: integer): integer;
	{Check if edges are duplicated}
	{Since triangle edges are CCW you compare P1 to P2 and P2 to P1}
	var I: integer;
	begin
	Result:=-1;
	for I:=0 to High(ConvexHull) do
	 if (P1=ConvexHull[I].Edge.P2) and (P2=ConvexHull[I].Edge.P1) then
		begin
		Result:=I;
		exit;
		end;
	end;


	procedure AddEdge(P1,P2,Inx: integer; Side: TTriSide);
	{Add an edge if it is not a duplicate}
	{Remove any that are duplicated}
	var R,TI: integer;
	begin
	R:=IsDuplicate(P1,P2);
	if R<0 then
		begin
		{Insert non-duplicated edge}
		SetLength(ConvexHull,Length(ConvexHull)+1);
		ConvexHull[High(ConvexHull)].Edge.P1:=P1;
		ConvexHull[High(ConvexHull)].Edge.P2:=P2;
		ConvexHull[High(ConvexHull)].TriInx:=Inx;
		ConvexHull[High(ConvexHull)].Side:=Side;
		{Note: this gets all edges because it handles the}
		{first insertion which before it can be deleted}
		HandleMinMax(ConvexHull[High(ConvexHull)].Edge);
		end
	else
		begin
		{Duplicated edge - mark the triangles that share the edge}
		TI:=ConvexHull[R].TriInx;
		case Side of
		 tsSide1: TriArray[Inx].Adj1:=TI;
		 tsSide2: TriArray[Inx].Adj2:=TI;
		 tsSide3: TriArray[Inx].Adj3:=TI;
		 end;
		case ConvexHull[R].Side of
		 tsSide1: TriArray[TI].Adj1:=Inx;
		 tsSide2: TriArray[TI].Adj2:=Inx;
		 tsSide3: TriArray[TI].Adj3:=Inx;
		 end;

		{Delete duplicate edge - using the trick}
		ConvexHull[R]:=ConvexHull[High(ConvexHull)];
		SetLength(ConvexHull,Length(ConvexHull)-1);
		end;
	end;


begin
if Length(VertArray) <3 then exit;
if not FTrianglesDone then Triangulate;
EdgeMin:=1E308; EdgeMax:=-1E308;
SetLength(ConvexHull,0);
I:=0;
{Collect edges that aren't common to two triangles }
{This will produce a convex hull}
while I<Length(TriArray) do
	begin
	AddEdge(TriArray[I].V1,TriArray[I].V2,I,tsSide1);
	AddEdge(TriArray[I].V2,TriArray[I].V3,I,tsSide2);
	AddEdge(TriArray[I].V3,TriArray[I].V1,I,tsSide3);
	Inc(I);
	end;
{Now sort vertices so the lines are in order - counter clockwise}
OrderHull(ConvexHull);
FConvexDone:=True;
end;

{------------------------------------------------------------------------------}



procedure TTriangulator.SortHullByLength(var Hull: THullArray);
{Sort all Edges in a hull by Edge from Longest to shortest}


	procedure DoQuickSort(var Hull: THullArray; L, R: Integer);
	{Sort Edge Lengths in sub-hull}
	var I, J: Integer;
	var P: double;
	var T: THullEdge;
	begin
	repeat
		begin
		I := L;
		J := R;
		P := GetEdgeMag(Hull[(L + R) shr 1].Edge);
		repeat
			begin
			while GetEdgeMag(Hull[I].Edge)> P do Inc(I);
			while GetEdgeMag(Hull[J].Edge)< P do Dec(J);
			if I <= J then
				begin
				T:=Hull[I]; Hull[I]:=Hull[J]; Hull[J]:=T;
				Inc(I);
				Dec(J);
				end;
			end
		until I > J;
		if L < J then DoQuickSort(Hull, L, J);
		L := I;
		end
	until I >= R;
	end;


begin
if High(Hull)>0 then DoQuickSort(Hull, Low(Hull), High(Hull));
end;




procedure TTriangulator.CopyHull(var Source,Dest: THullArray);
{Copy all edges in a Hull}
var I: integer;
begin
SetLength(Dest,Length(Source));
for I:=0 to High(Dest) do Dest[I]:=Source[I];
end;




procedure TTriangulator.CopySortByLen(var Source,Dest: THullArray);
{Copy and sort hull by edge length}
begin
CopyHull(Source,Dest);
SortHullByLength(Dest);
end;



procedure TTriangulator.GetNewPoint(HE: THullEdge; var NewInx,Adj1,Adj2: integer);
{Get the new point oposite the specified edge side of the triangle}
{Also return the indices for two corresponding adjacent triangles}
var TI: integer;
begin
TI:=HE.TriInx;
if (TriArray[TI].V1=HE.Edge.P1) and (TriArray[TI].V2=HE.Edge.P2) then
	begin
	NewInx:=TriArray[TI].V3;
	Adj1:=TriArray[TI].Adj2;
	Adj2:=TriArray[TI].Adj3;
	end
else if (TriArray[TI].V2=HE.Edge.P1) and (TriArray[TI].V3=HE.Edge.P2) then
	begin
	NewInx:=TriArray[TI].V1;
	Adj1:=TriArray[TI].Adj3;
	Adj2:=TriArray[TI].Adj1;
	end
else
	begin
	NewInx:=TriArray[TI].V2;
	Adj1:=TriArray[TI].Adj1;
	Adj2:=TriArray[TI].Adj2;
	end;
end;



procedure TTriangulator.GetConcaveHull;
{Get convex hull, choosing those convexities with longest line length}
var BaseHull: THullArray;
var I: integer;


	function IsPointOnHull(Inx: integer): boolean;
	{Test if point is on either the newly formed concaved hull}
	{or the original  convex hull - Returns true if on either}
	var I: integer;
	begin
	Result:=True;
	for I:=0 to High(BaseHull) do
	 if (Inx=BaseHull[I].Edge.P1) or (Inx=BaseHull[I].Edge.P2) then exit;
	for I:=0 to High(ConcaveHull) do
	 if (Inx=ConcaveHull[I].Edge.P1) or (Inx=ConcaveHull[I].Edge.P2) then exit;
	Result:=False;
	end;


	procedure AnalyzeEdge(HE: THullEdge);
	{Analyze hull edge and see if we can remove it}
	var NewInx,Inx,Adj1,Adj2: integer;
	var L: double;
	begin
	L:=GetEdgeMag(HE.Edge);
	GetNewPoint(HE,NewInx,Adj1,Adj2);
	if (L<=ConcaveLength) or IsPointOnHull(NewInx) then
		begin
		{Hull Edge cannot be removed so copy it into new Concave array}
		SetLength(ConcaveHull,Length(ConcaveHull)+1);
		ConcaveHull[High(ConcaveHull)]:=HE;
		end
	else
		begin
		{Romove edge and replace with other two sides of triangle}
		SetLength(ConcaveHull,Length(ConcaveHull)+1);
		Inx:=High(ConcaveHull);
		ConcaveHull[Inx].Edge.P1:=HE.Edge.P1;
		ConcaveHull[Inx].Edge.P2:=NewInx;
		ConcaveHull[Inx].TriInx:=Adj2;

		SetLength(ConcaveHull,Length(ConcaveHull)+1);
		Inx:=High(ConcaveHull);
		ConcaveHull[Inx].Edge.P1:=NewInx;
		ConcaveHull[Inx].Edge.P2:=HE.Edge.P2;
		ConcaveHull[Inx].TriInx:=Adj1;
		end;
	end;


begin
if Length(VertArray) <3 then exit;
if not FConvexDone then GetConvexHull;
CopySortByLen(ConvexHull,BaseHull);
while true do
	begin
	SetLength(ConcaveHull,0);
	for I:=0 to High(BaseHull) do
		begin
		AnalyzeEdge(BaseHull[I]);
		end;
	if Length(BaseHull)=Length(ConcaveHull) then break;
	CopySortByLen(ConcaveHull,BaseHull);
	end;
{Now sort vertices so the lines are in order - counter clockwise}
OrderHull(ConcaveHull);
FConcaveDone:=True;
end;



{------------------------------------------------------------------------------}




procedure TTriangulator.AddPoint(X,Y: double; Tag: integer);
begin
{Duplicate points are removed in the main routine}
{Insert point in list}
SetVertArrayLength(Length(VertArray)+1);
VertArray[High(VertArray)].X:= X;
VertArray[High(VertArray)].Y:= Y;
OriginalOrder[High(OriginalOrder)]:=High(OriginalOrder);
UserTags[High(UserTags)]:=High(VertArray);
SetNotTriangulated;
end;


procedure TTriangulator.AddPoint(X,Y: double);
begin
AddPoint(X,Y,0);
end;

function TTriangulator.GetTriangleCount: integer;
begin
Result:=Length(TriArray);
end;



function TTriangulator.GetVertexCount: integer;
begin
Result:=Length(VertArray);
end;




procedure TTriangulator.GetEdgeMinMax(var Min, Max: double);
begin
GetConvexHull;
Min:=EdgeMin;
Max:=EdgeMax;
end;



procedure TTriangulator.SetPercentLength(const Value: double);
var D,Min,Max: double;
begin
FPercentLength:=Value;
GetEdgeMinMax(Min,Max);
D:=Max-Min;
FConcaveLength:=Min+(D * Value);
end;




procedure TTriangulator.SetConcaveLength(const Value: double);
var D,Min,Max: double;
begin
FConcaveLength:=Value;
GetEdgeMinMax(Min,Max);
D:=Max-Min;
FPercentLength:=(FConcaveLength-Min)/D;
end;



{------------------------------------------------------------------------------}

function TTriangulator.GetTriArea(Edge: THullEdge): double;
{Calculate the area of triangle associated with edge}
var V1,V2,V3,A,B: T2DVector;
begin
V1:=VertArray[TriArray[Edge.TriInx].V1];
V2:=VertArray[TriArray[Edge.TriInx].V2];
V3:=VertArray[TriArray[Edge.TriInx].V3];
A:=VectorSubtract2D(V2,V1);
B:=VectorSubtract2D(V3,V1);
Result:=VectorABS3D(CrossProduct2D(A,B));
end;

procedure TTriangulator.SortHullByArea(var Hull: THullArray);
{Sort all Edges in a hull by Edge from Longest to shortest}


	procedure DoQuickSort(var Hull: THullArray; L, R: Integer);
	{Sort Edge Lengths in sub-hull}
	var I, J: Integer;
	var P: double;
	var T: THullEdge;
	begin
	repeat
		begin
		I := L;
		J := R;
		P := GetTriArea(Hull[(L + R) shr 1]);
		repeat
			begin
			while GetTriArea(Hull[I])> P do Inc(I);
			while GetTriArea(Hull[J])< P do Dec(J);
			if I <= J then
				begin
				T:=Hull[I]; Hull[I]:=Hull[J]; Hull[J]:=T;
				Inc(I);
				Dec(J);
				end;
			end
		until I > J;
		if L < J then DoQuickSort(Hull, L, J);
		L := I;
		end
	until I >= R;
	end;


begin
if High(Hull)>0 then DoQuickSort(Hull, Low(Hull), High(Hull));
end;



procedure TTriangulator.CopySortByArea(var Source,Dest: THullArray);
{Copy and sort hull by edge length}
begin
CopyHull(Source,Dest);
SortHullByArea(Dest);
end;



procedure TTriangulator.GetConcaveHull2;
{Get convex hull, choosing those convexities with largest triangular area}
var BaseHull: THullArray;
var I: integer;


	function IsPointOnHull(Inx: integer): boolean;
	{Test if point is on either the newly formed concaved hull}
	{or the original  convex hull - Returns true if on either}
	var I: integer;
	begin
	Result:=True;
	for I:=0 to High(BaseHull) do
	 if (Inx=BaseHull[I].Edge.P1) or (Inx=BaseHull[I].Edge.P2) then exit;
	for I:=0 to High(ConcaveHull) do
	 if (Inx=ConcaveHull[I].Edge.P1) or (Inx=ConcaveHull[I].Edge.P2) then exit;
	Result:=False;
	end;


	procedure AnalyzeEdge(HE: THullEdge);
	{Analyze hull edge and see if we can remove it}
	var NewInx,Inx,Adj1,Adj2: integer;
	var L: double;
	begin
	L:=GetTriArea(HE);
	GetNewPoint(HE,NewInx,Adj1,Adj2);
	if (L<=Sqr(ConcaveLength)/2) or IsPointOnHull(NewInx) then
		begin
		{Hull Edge cannot be removed so copy it into new Concave array}
		SetLength(ConcaveHull,Length(ConcaveHull)+1);
		ConcaveHull[High(ConcaveHull)]:=HE;
		end
	else
		begin
		{Romove edge and replace with other two sides of triangle}
		SetLength(ConcaveHull,Length(ConcaveHull)+1);
		Inx:=High(ConcaveHull);
		ConcaveHull[Inx].Edge.P1:=HE.Edge.P1;
		ConcaveHull[Inx].Edge.P2:=NewInx;
		ConcaveHull[Inx].TriInx:=Adj2;

		SetLength(ConcaveHull,Length(ConcaveHull)+1);
		Inx:=High(ConcaveHull);
		ConcaveHull[Inx].Edge.P1:=NewInx;
		ConcaveHull[Inx].Edge.P2:=HE.Edge.P2;
		ConcaveHull[Inx].TriInx:=Adj1;
		end;
	end;


begin
if Length(VertArray) <3 then exit;
if not FConvexDone then GetConvexHull;
CopySortByArea(ConvexHull,BaseHull);
while true do
	begin
	SetLength(ConcaveHull,0);
	for I:=0 to High(BaseHull) do
		begin
		AnalyzeEdge(BaseHull[I]);
		end;
	if Length(BaseHull)=Length(ConcaveHull) then break;
	CopySortByArea(ConcaveHull,BaseHull);
	end;
{Now sort vertices so the lines are in order - counter clockwise}
OrderHull(ConcaveHull);
FConcaveDone:=True;
end;

end.
