unit Fourier2D;

interface

uses Graphics, Math, DSupport, BitmapLib;

{ Structure for Complex Data type  }


type TComplex = record
 real,imag: double;
 end;



{Dynamic array of data for FFT}

type T2DIntArray = array of array of integer;
type T2DDoubleArray = array of array of double;
type T2DComplexArray = array of array of TComplex;


type TFFT = class(TObject)
 private
   procedure ReadImage;
   procedure FFT1D(dir, m: integer;  var x,y: array of double);

 protected
 public
   Obj: TBitmap;			{ Input Object Image}
   FourierPlot: TBitmap;		{ Generated Fouruer Magnitude Plot}
   PhasePlot: TBitmap;			{ Generated Fourier Phase Plot}
   GreyImage: T2DIntArray;		{GreyScale Image Array Generated from input Image}
   FourierMagnitude: T2DDoubleArray;
   FourierPhase: T2DDoubleArray;
   FFTLog: T2DDoubleArray;		{ Log of Fourier Magnitude}
   FFTPhaseLog: T2DDoubleArray;		{ Log of Fourier Phase}
   FFTNormalized: T2DIntArray;		{ Normalized FFT Magnitude : Scale 0-1}
   FFTPhaseNormalized: T2DIntArray;	{ Normalized FFT Phase : Scale 0-1}

   nx, ny: integer;			{Number of Points in Width & height}
   Width, Height: integer;
   FourierMag: T2DComplexArray;		{Fourier Magnitude  Array Used for Inverse FFT}
   FFTShifted: T2DComplexArray;		{ Shifted FFT}
   FFTNormal: T2DComplexArray;		{ FFT Shift Removed - required for Inverse FFT}
   Output: T2DComplexArray;		{ FFT Normal}

   constructor Create; overload;
   constructor Create(Input: TBitmap); overload;
   constructor Create(Input: T2DIntArray); overload;
   constructor Create(Input: T2DComplexArray); overload;
   procedure LoadBitmap(BM: TBitmap);
   function Displayimage: TBitmap; overload;
   function Displayimage(Image: T2DIntArray): TBitmap; overload;
   procedure FFTShift;
   procedure InverseFFT(Fourier: T2DComplexArray); overload;
   procedure InverseFFT; overload;
   procedure FFTPlot; overload;
   procedure FFTPlot(Output: T2DComplexArray);  overload;
   procedure RemoveFFTShift;
   procedure ForwardFFT;
   function FFT2D(c: T2DComplexArray;  nx, ny, dir: integer): T2DComplexArray;
 end;

implementation

{
2D dynamic array layout:

1. Set arrays dimenions

SetLength(A, X, Y);

2. Reading array size:

Length(A); 	/<======== Length of X-dimension
Length(A[0]); 	/<======== Length of Y-dimension

3. Access data:

A[X, Y];

}


{Support routines }

function Complex(X,Y: double): TComplex;
{Create a complex data type}
begin
Result.real:=X;
Result.imag:=Y;
end;



function Magnitude(C: TComplex): double;
{Return the magnitude of TComplex type}
begin
Result:=Sqrt(C.real * C.real + C.imag * C.imag);
end;



function Phase(C: TComplex): double;
{Return the phase angle of TComplex type}
begin
{USe DSupport version because it handles real=0}
Result:= DSupport.Arctan2(C.imag , C.real);
end;




{ TFFT }




procedure TFFT.ReadImage;
{ Function to Read Bitmap to greyscale Array}
var i, j: integer;
var Image: TBitmap;
var Scan: pRGBQuadArray;
begin
SetLength(GreyImage,Width,Height);
Image := Obj;
Image.PixelFormat:=pf32Bit;

for i:=0 to Image.Height-1 do
	begin
	Scan:=Image.ScanLine[I];
	for j:= 0 to Image.Width-1 do
		begin
		GreyImage[j, i] := (Scan[j].rgbRed + Scan[j].rgbGreen + Scan[j].rgbBlue) div 3;
		end;
	end;
end;


procedure TFFT.LoadBitmap(BM: TBitmap);
begin
Obj := BM;
Width := BM.Width;
Height:= BM.Height;
nx := BM.Width;
ny := BM.Height;
ReadImage;
end;



constructor TFFT.Create;
{ Constructor = create FFT with an input image}
begin
Obj := nil;
Width := 0;
Height:= 0;
nx := 0;
ny := 0;
end;


constructor TFFT.Create(Input: TBitmap);
{ Parameterized Constructor for FFT Reads Input Bitmap to a Greyscale Array}
begin
LoadBitmap(Input);
end;



constructor TFFT.Create(Input: T2DIntArray);
{ Parameterized Constructor for FFT}
begin
GreyImage := Input;
nx := Length(Input);
ny := Length(Input[0]);
Width := nx;
Height := ny;
end;



constructor TFFT.Create(Input: T2DComplexArray);
{ Constructor for Inverse FFT}
begin
nx := Length(Input);
ny := Length(Input[0]);
Width := nx;
Height := ny;
FourierMag := Input;
end;




function  TFFT.Displayimage: TBitmap;
{Copy gray-scale image to bitmap}
var i, j: integer;
var Image: TBitmap;
var Scan: pRGBQuadArray;
begin
Image:=TBitmap.Create;
Image.PixelFormat:=pf32Bit;
Image.Width:=Width;
Image.Height:=Height;
for i:=0 to Image.Height-1 do
	begin
	Scan:=Image.ScanLine[I];
	for j:= 0 to Image.Width-1 do
		begin
		Scan[j].rgbRed := GreyImage[j, i];
		Scan[j].rgbGreen := GreyImage[j, i];
		Scan[j].rgbBlue := GreyImage[j, i];
		Scan[j].rgbReserved := 255;
		end;
	end;
Result:=image;
end;





function  TFFT.Displayimage(Image: T2DIntArray): TBitmap;
{Convert 2D Integer array to bitmap}
var i, j: integer;
var output: TBitmap;
var Scan: pRGBQuadArray;
var B: byte;
begin
output:=TBitmap.Create;
output.PixelFormat:=pf32Bit;
output.Width:=Length(Image);
output.Height:=Length(Image[0]);
for i:=0 to Length(Image[0])-1 do
	begin
	Scan:=output.ScanLine[I];
	for j:= 0 to Length(Image)-1 do
		begin
		B:=EnsureRange(image[j, i],0,255);
		Scan[j].rgbRed := B;
		Scan[j].rgbGreen := B;
		Scan[j].rgbBlue := B;
		Scan[j].rgbReserved := 255;
		end;
	end;
Result:=Output;
end;


{-------------------------------------------------------------------------
This computes an in-place complex-to-complex FFT
x and y are the real and imaginary arrays of 2^m points.
dir := 1 gives forward transform
dir := -1 gives reverse transform
Formula: forward
         N-1
          ---
        1 \         - j k 2 pi n / N
X(K) := --- > x(n) e                  := Forward transform
        N /                            n:=0..N-1
          ---
         n:=0

Formula: reverse
         N-1
         ---
         \          j k 2 pi n / N
X(n) :=    > x(k) e                  := Inverse transform
         /                             k:=0..N-1
         ---
         k:=0
-------------------------------------------------------------------------}


procedure TFFT.FFT1D(dir, m: integer;  var x,y: array of double);
var nn, i, i1, j, k, i2, l, l1, l2: integer;
var c1, c2, tx, ty, t1, t2, u1, u2, z: double;
begin
{ Calculate the number of points }
nn := 1;
for i := 0 to m-1 do nn:=nn * 2;
{ Do the bit reversal }
i2 := nn shr 1;
j := 0;
for i := 0 to nn - 2 do
	begin
	if i < j then
		begin
		tx := x[i];
		ty := y[i];
		x[i] := x[j];
		y[i] := y[j];
		x[j] := tx;
		y[j] := ty;
		end;

	k := i2;
	while (k <= j) do
		begin
		j:=J - k;
		k :=k shr 1;
		end;
	j:=J + k;
	end;

{ Compute the FFT }
c1 := -1.0;
c2 := 0.0;
l2 := 1;
for l := 0 to m-1 do
	begin
	l1 := l2;
	l2:=l2 shl 1;
	u1 := 1.0;
	u2 := 0.0;
	for j := 0 to l1-1 do
		begin
		i := j;
		while i < nn do
			begin
			i1 := i + l1;
			t1 := u1 * x[i1] - u2 * y[i1];
			t2 := u1 * y[i1] + u2 * x[i1];
			x[i1] := x[i] - t1;
			y[i1] := y[i] - t2;
			x[i] :=x[i] + t1;
			y[i] :=y[i] + t2;
			inc(i,l2)
			end;
		z := u1 * c1 - u2 * c2;
		u2 := u1 * c2 + u2 * c1;
		u1 := z;
		end;
	c2 := Sqrt((1.0 - c1) / 2.0);
	if (dir = 1) then c2 := -c2;
	c1 := Sqrt((1.0 + c1) / 2.0);
	end;
{ Scaling for forward transform }
if (dir = 1) then
	begin
	for i:=0 to nn-1 do
		begin
		x[i] :=x[i] / nn;
		y[i] :=y[i] / nn;
		end;
	end;
end;


{-------------------------------------------------------------------------
Perform a 2D FFT inplace given a complex 2D array
The direction dir, 1 for forward, -1 for reverse
The size of the array (nx,ny)
Return false if there are memory problems or
the dimensions are not powers of 2
 -------------------------------------------------------------------------}


function TFFT.FFT2D(c: T2DComplexArray;  nx, ny, dir: integer): T2DComplexArray;
var i,j: integer;
var m: integer;			{Power of 2 for current number of points}
var real: array of double;
var imag: array of double;
var output: T2DComplexArray;
begin
output := c;
{ Do 1D FFT for each Row}
{Setup array for number items in a row}
SetLength(real,nx);
SetLength(imag,nx);
for j:=0 to ny-1 do
	begin
	for i:=0 to nx-1 do
		begin
		real[i] := c[i,j].real;
		imag[i] := c[i,j].imag;
		end;

	{ Perform 1D FFT Function for Rows}
	{Find power of 2 for rows in array}
	m := Trunc(Log2(nx));
	FFT1D(dir,m,real,imag);

	for i:=0 to nx-1 do
		begin
		output[i, j].real := real[i];
		output[i, j].imag := imag[i];
		end;

	end;

{ Do 1D FFT for each column}
{Setup array for number items in a column}
SetLength(real,ny);
SetLength(imag,ny);
for i:=0 to nx-1 do
	begin
	for j:=0 to j-1 do
		begin
		real[j] := output[i, j].real;
		imag[j] := output[i, j].imag;
		end;
	{ Perform 1D FFT Function for Columns}
	{Find power of 2 for columns in array}
	m := Trunc(Log2(ny));
	FFT1D(dir,m, real, imag);
	for j:=0 to ny-1 do
		begin
		output[i, j].real := real[j];
		output[i, j].imag := imag[j];
		end;
	end;
Result:=output;
end;






procedure TFFT.ForwardFFT;
{ Do Forward Fast Fourier Transform}
var i,j: integer;
begin
{Initializing Fourier Transform Array}
SetLength(FourierMag,Width,Height);
SetLength(Output,Width,Height);
{Copy Image Data to the Complex Array}
for i:=0 to Width-1 do
	begin
	for j:= 0 to Height-1 do
		begin
		FourierMag[i, j].real :=GreyImage[i, j];
		FourierMag[i, j].imag := 0;
		end;
	end;
{Do Forward Fourier Transform}
Output:= FFT2D( FourierMag, nx, ny, 1);
end;



procedure TFFT.FFTShift;
{ Shift The FFT of the Image}
var i, j: integer;
begin
SetLength(FFTShifted,nx, ny);
for i:=0 to (nx div 2)-1 do
 for j:=0 to (ny div 2)-1 do
	begin
	FFTShifted[i + (nx div 2), j + (ny div 2)] := Output[i, j];
	FFTShifted[i, j] := Output[i + (nx div 2), j + (ny div 2)];
	FFTShifted[i + (nx div 2), j] := Output[i , j + (ny div 2)];
	FFTShifted[i, j + (ny div 2)] := Output[i + (nx div 2), j ];
	end;
end;



procedure TFFT.RemoveFFTShift;
{ Removes FFT Shift for FFTshift Array}
var i, j: integer;
begin
SetLength(FFTNormal, nx, ny);

for i := 0 to (nx div 2) - 1 do
 for j := 0 to (ny div 2) - 1 do
	begin
	FFTNormal[i + (nx div 2), j + (ny div 2)] := FFTShifted[i, j];
	FFTNormal[i, j] := FFTShifted[i + (nx div 2), j + (ny div 2)];
	FFTNormal[i + (nx div 2), j] := FFTShifted[i, j + (ny div 2)];
	FFTNormal[i, j + (ny div 2)] := FFTShifted[i + (nx div 2), j];
	end;
end;



procedure TFFT.FFTPlot(Output: T2DComplexArray);
{ FFT Plot Method for Shifted FFT}
var i, j: integer;
var max: double;
begin
SetLength(FFTLog,nx, ny);
SetLength(FFTPhaseLog,nx, ny);

SetLength(FourierMagnitude,nx, ny);
SetLength(FourierPhase,nx, ny);

SetLength(FFTNormalized, nx, ny);
SetLength(FFTPhaseNormalized,nx, ny);

for i := 0 to Width - 1 do
 for j := 0 to Height - 1 do
	begin
	FourierMagnitude[i, j] := Magnitude(Output[i, j]);
	FourierPhase[i, j] := Phase(Output[i, j]);
	FFTLog[i, j] := Ln(1 + FourierMagnitude[i, j]);
	FFTPhaseLog[i, j] := Ln(1 + Abs(FourierPhase[i, j]));
	end;
{Generating Magnitude Bitmap}
max := FFTLog[0, 0];
for i := 0 to Width - 1 do
 for j := 0 to Height - 1 do
	begin
	if (FFTLog[i, j] > max) then
	max := FFTLog[i, j];
	end;

for i := 0 to Width - 1 do
 for j := 0 to Height - 1 do
 	begin
 	FFTLog[i, j] := FFTLog[i, j] / max;
 	end;

for i := 0 to Width - 1 do
 for j := 0 to Height - 1 do
	begin
	FFTNormalized[i, j] := trunc(2000 * FFTLog[i, j]);
	end;
{Transferring Image to Fourier Plot}
FourierPlot := Displayimage(FFTNormalized);
{generating phase Bitmap}
FFTPhaseLog[0, 0] := 0;
max := FFTPhaseLog[1, 1];
for i := 0 to Width - 1 do
 for j := 0 to Height - 1 do
	begin
	if (FFTPhaseLog[i, j] > max) then max := FFTPhaseLog[i, j];
	end;

for i := 0 to Width - 1 do
 for j := 0 to Height - 1 do
	begin
	FFTPhaseLog[i, j] := FFTPhaseLog[i, j] / max;
	end;

for i := 0 to Width - 1 do
 for j := 0 to Height - 1 do
	begin
	FFTPhaseNormalized[i, j] := trunc(255 * FFTPhaseLog[i, j]);
	end;
{Transferring Image to Fourier Plot}
PhasePlot := Displayimage(FFTPhaseNormalized);
end;



procedure TFFT.FFTPlot;
{ generate FFT Image for Display Purpose}
var i, j: integer;
var max: double;
begin
SetLength(FFTLog,nx,ny);
SetLength(FFTPhaseLog,nx, ny);

SetLength(FourierMagnitude,nx, ny);
SetLength(FourierPhase,nx, ny);

SetLength(FFTNormalized,nx, ny);
SetLength(FFTPhaseNormalized,nx, ny);

for i:=0 to Width-1 do
 for j := 0 to Height-1 do
 	begin
 	FourierMagnitude[i, j] := Magnitude(Output[i, j]);
 	FourierPhase[i, j] := Phase(Output[i, j]);
 	FFTLog[i, j] := Ln(1 + FourierMagnitude[i, j]);
 	FFTPhaseLog[i, j] := Ln(1 + Abs(FourierPhase[i, j]));
 	end;

{Generating Magnitude Bitmap}
max := FFTLog[0, 0];
for i:=0 to Width-1 do
 for j:= 0 to Height-1 do
	begin
	if (FFTLog[i, j] > max) then max := FFTLog[i, j];
	end;

for i:=0 to Width-1 do
 for j := 0 to Height-1 do
	begin
	FFTLog[i, j] := FFTLog[i, j] / max;
	end;

for i:=0 to Width-1 do
 for j := 0 to  Height-1 do
	begin
	FFTNormalized [i,j]:=trunc(1000*FFTLog[i,j]);
	end;
{Find power of 2 for rows in array}
{Transferring Image to Fourier Plot}
FourierPlot := Displayimage(FFTNormalized);

{generating phase Bitmap}
max := FFTPhaseLog[0, 0];
for i := 0 to  Width-1 do
 for j := 0 to  Height-1 do
	begin
	if (FFTPhaseLog[i, j] > max) then max := FFTPhaseLog[i, j];
	end;

for i := 0 to  Width-1 do
 for j := 0 to  Height-1 do
	begin
	FFTPhaseLog[i, j] := FFTPhaseLog[i, j] / max;
	end;

for i := 0 to  Width-1 do
 for j := 0 to  Height-1 do
 	begin
 	FFTPhaseNormalized[i, j] := trunc(2000 * FFTLog[i, j]);
 	end;
{Transferring Image to Fourier Plot}
PhasePlot := Displayimage(FFTPhaseNormalized);
end;



procedure TFFT.InverseFFT;
{ Calculate Inverse from Complex [,]  Fourier Array}
var i, j: integer;
begin
{Initialize Fourier Transform Array}
SetLength(Output,nx,ny);
Output := FFT2D(FourierMag, nx, ny, -1);

{Copying Real Image Back to Greyscal}
Obj := nil;
for i := 0 to  Width - 1 do
 for j := 0 to  Height - 1 do
	begin
	GreyImage[i, j] := Trunc(Magnitude(Output[i, j]));
	end;

Obj := Displayimage(GreyImage);
end;


    { Generates Inverse FFT of Given Input Fourier}

procedure TFFT.InverseFFT(Fourier: T2DComplexArray);
var i, j: integer;	//Calling Forward Fourier Transform
begin
//Initializing Fourier Transform Array
SetLength(Output,nx,ny);
Output := FFT2D(Fourier, nx, ny, -1);
//Copying Real Image Back to Greyscal	//Copy Image Data to the Complex Array
for i := 0 to  Width - 1 do
 for j := 0 to  Height - 1 do
	begin
	GreyImage[i, j] := Trunc(Magnitude(Output[i, j]));
	end;

Obj := Displayimage(GreyImage);
end;




end.
 