unit GLFastArrays;

interface

uses Classes,Graphics,Math,Vector,HugeArry,Hash,Huge2D,GLLibrary;



{------------------------------------------------------------------------}
{  Declaration for huge Color array object based on huge array object    }
{  Holds vertex colors - the TFaceColor array holds Face Colors          }
{------------------------------------------------------------------------}

type TColorArray = class(THugeArray)
 private
 protected
  procedure Put(Index: integer; Item: TGLColor); virtual;
  function Get(Index: integer): TGLColor;
 public
  constructor Create(Number: integer);
  destructor Destroy; override;
  function Add(Item: TGLColor): integer; reintroduce; virtual;
  function AddPointer(Item: TGLColor; Ptr: integer): Integer; virtual;
  procedure Insert(Value: TGLColor; Index: integer); virtual;
  procedure Delete(Index: integer);
  procedure Move(CurIndex, NewIndex: Integer);
  procedure SetTransparency(Transparency: double);
  property Value[Index: integer]: TGLColor read Get write Put; default;
 end;


{------------------------------------------------------------------------}
{  Declaration for huge Face array object based on huge array object     }
{------------------------------------------------------------------------}

type TFaceList = class(THugeArray)
 private
    function GetColor(Index: integer): TColor;
    procedure SetColor(Index: integer; const Value: TColor);
 protected
  procedure Put(Index: integer; Item: TFace); virtual;
  function Get(Index: integer): TFace;
 public
  Normals: THuge3DVector;
  GLColors: TColorArray;
  constructor Create(Number: integer);
  destructor Destroy; override;
  function Add(Item: TFace; Color: TGLColor; Normal: T3DVector): integer; reintroduce; virtual;
  function AddPointer(Item: TFace; Color: TGLColor; Ptr: integer): Integer; virtual;
  procedure Insert(Value: TFace; Color: TGLColor; Normal: T3DVector; Index: integer); virtual;
  procedure Delete(Index: integer);
  procedure Move(CurIndex, NewIndex: Integer);
  property Value[Index: integer]: TFace read Get write Put; default;
  procedure Clear; override;
  property Colors[Index: integer]: TColor read GetColor write SetColor;
 end;

{------------------------------------------------------------------------}
{  Declaration for Vertex Hash table object based on custom Hash Table   }
{------------------------------------------------------------------------}

type TVertexArray = class(T3DVectorHash)
 private
  function GetColor(Index: integer): TColor;
  procedure SetColor(Index: integer; const Value: TColor);
 protected
 public
  ExtentMin, ExtentMax: T3DVector;
  AdjacentFaces: TNO2DIntArray;
  Normals: THuge3DVector;
  GLColors: TColorArray;
  constructor Create(StartAlloc: LongInt);
  destructor Destroy; override;
  procedure Assign(Hash: TVertexArray);
  procedure Append(Hash: TVertexArray);
  function InsertPointer(V: T3DVector; Ptr: integer): integer; override;
  function Insert(D: T3DVector): integer;
  function AddUnique(V: T3DVector; Color: TGLColor): integer;
  function AddNonUnique(V: T3DVector; Color: TGLColor): integer;
  procedure AddFaceLink(VertInx, FaceInx: integer);
  procedure Delete(Index: integer);
  procedure Clear; override;
  property Colors[Index: integer]: TColor read GetColor write SetColor;
 end;



type TTriInterface = class(TObject)
 private
  function GetTriangle(Index: integer): TTriangle;
  procedure PutTriangle(Index: integer; const Value: TTriangle);
  function GetCount: integer;
  function GetColors(Index: integer): TColor;
  function GetGLColors(Index: integer): TGLColor;
  procedure SetColors(Index: integer; const Value: TColor);
  procedure SetGLColors(Index: integer; const Value: TGLColor);
    function GetGLFaceColor(Index: integer): TGLFaceColor;
    procedure SetGLFaceColor(Index: integer; const Value: TGLFaceColor);
 protected
 public
  Faces: TFaceList;
  Vertices: TVertexArray;
  constructor Create(var Faces: TFaceList; var Vertices: TVertexArray);
  procedure Delete(Index: integer);
  property Triangle[Index: integer]: TTriangle read GetTriangle write PutTriangle; default;
  property Count: integer read GetCount;
  property Colors[Index: integer]: TColor read GetColors write SetColors;
  property GLColors[Index: integer]: TGLColor read GetGLColors write SetGLColors;
  property FaceColors[Index: integer]: TGLFaceColor read GetGLFaceColor write SetGLFaceColor;
 end;

implementation


{------------------------------------------------------------------------}
{ Implementation for huge TColorArray array based on huge array object   }
{------------------------------------------------------------------------}


function TColorArray.Add(Item: TGLColor): integer;
begin
Result:=inherited Add(Item);
end;



function TColorArray.AddPointer(Item: TGLColor; Ptr: integer): Integer;
begin
Result:=inherited AddPointer(Item,Ptr);
end;



constructor TColorArray.Create(Number: integer);
begin
inherited Create(Number,SizeOf(TGLColor));
end;



procedure TColorArray.Delete(Index: integer);
begin
DeleteItem(Index);
end;



destructor TColorArray.Destroy;
begin
inherited;
end;



function TColorArray.Get(Index: integer): TGLColor;
var PTR: ^TGLColor;
begin
PTR:=GetItemPtr(Index);
Result:=PTR^;
end;



procedure TColorArray.Insert(Value: TGLColor; Index: integer);
begin
InsertItem(Value,Index);
end;



procedure TColorArray.Move(CurIndex, NewIndex: Integer);
var T: TGLColor;
begin
T:=Value[CurIndex];
Delete(CurIndex);
Insert(T,NewIndex);
end;



procedure TColorArray.Put(Index: integer; Item: TGLColor);
begin
PutItem(Item,Index);
end;


procedure TColorArray.SetTransparency(Transparency: double);
var I: integer;
var PTR: ^TGLColor;
var T: byte;
begin
T:=trunc(255 * Transparency);
PTR:=GetItemPtr(0);
for I:=0 to Count-1 do
	begin
	PTR^.A:=T;
	Inc(PTR);
	end;
end;


{------------------------------------------------------------------------}
{  Implementation for huge Face array object based on huge array object}
{------------------------------------------------------------------------}



constructor TFaceList.Create(Number: Integer);
{Create a huge TFace array}
begin
inherited Create(Number,SizeOf(TFace));
Normals:=THuge3DVector.Create(Number);
GLColors:=TColorArray.Create(Number);
end;




destructor TFaceList.Destroy;
{Destroy a huge TFace array}
begin
GLColors.Free;
Normals.Free;
inherited;
end;


function TFaceList.Add(Item: TFace; Color: TGLColor; Normal: T3DVector): integer;
{Add item to TFace array}
begin
Result:=inherited Add(Item);
Normals.Add(Normal);
GLColors.Add(Color);
end;


function TFaceList.AddPointer(Item: TFace; Color: TGLColor; Ptr: integer): Integer;
begin
Result:=inherited AddPointer(Item,Ptr);
Normals.Add(MakeVector3D(0,0,0));
GLColors.Add(Color);
end;


procedure TFaceList.Put(Index: integer; Item: TFace);
{Insert Vertex at index}
begin
PutItem(Item,Index);
end;


function TFaceList.Get(Index: integer): TFace;
{Get Vertex at the index}
var PTR: ^TFace;
begin
PTR:=GetItemPtr(Index);
Result:=PTR^;
end;


procedure TFaceList.Insert(Value: TFace; Color: TGLColor; Normal: T3DVector; Index: integer);
begin
InsertItem(Value,Index);
Normals.Insert(Normal,Index);
GLColors.Insert(Color,Index);
end;


procedure TFaceList.Delete(Index: integer);
begin
DeleteItem(Index);
Normals.Delete(Index);
GLColors.Delete(Index);
end;


procedure TFaceList.Move(CurIndex, NewIndex: Integer);
var T: TFace;
var C: TGLColor;
var N: T3DVector;
begin
T:=Value[CurIndex];
C:=GLColors[CurIndex];
N:=Normals[CurIndex];
Delete(CurIndex);
Insert(T,C,N,NewIndex);
end;



procedure TFaceList.Clear;
begin
inherited;
Normals.Clear;
GLColors.Clear;
end;




function TFaceList.GetColor(Index: integer): TColor;
begin
Result:=GLColorToColor(GLColors[Index]);
end;



procedure TFaceList.SetColor(Index: integer; const Value: TColor);
begin
GLColors[Index]:=ColorToGLColor(Value);
end;




{-------------------------------------------------------------------------}
{                Implementation for TVertexArray object                   }
{-------------------------------------------------------------------------}





procedure TVertexArray.Clear;
{Clear all the hash boxes and reinitialize}
begin
inherited Clear;
AdjacentFaces.Clear;
Normals.Clear;
GLColors.Clear;
ExtentMin:=MakeVector3D(MaxDouble,MaxDouble,MaxDouble);
ExtentMax:=MakeVector3D(-MaxDouble,-MaxDouble,-MaxDouble);
end;




constructor TVertexArray.Create(StartAlloc: LongInt);
{Creation routine for hash table object }
begin
inherited Create(StartAlloc);
AdjacentFaces:=TNO2DIntArray.Create;
Normals:=THuge3DVector.Create(StartAlloc);
GLColors:=TColorArray.Create(StartAlloc);
Clear;
end;




destructor TVertexArray.Destroy;
{Destroy hash table object }
begin
GLColors.Free;
Normals.Free;
AdjacentFaces.Free;
inherited Destroy;
end;



procedure TVertexArray.Assign(Hash: TVertexArray);
begin
inherited Assign(Hash);
AdjacentFaces.Assign(Hash.AdjacentFaces);
Normals.Assign(Hash.Normals);
GLColors.Assign(Hash.GLColors);
ExtentMin:=Hash.ExtentMin;
ExtentMax:=Hash.ExtentMax;
end;




procedure TVertexArray.Append(Hash: TVertexArray);
begin
inherited Append(Hash);
AdjacentFaces.Append(Hash.AdjacentFaces);
Normals.Append(Hash.Normals);
GLColors.Append(Hash.GLColors);
ExtentMin:=GetMinimumVector(Hash.ExtentMin,ExtentMin);
ExtentMax:=GetMaximumVector(Hash.ExtentMax,ExtentMax);
end;



function TVertexArray.InsertPointer(V: T3DVector; Ptr: integer): integer;
{ Insert Vertex and pointer Into Hash Table}
{ Pointer can point to any object the user }
{wishes to associate with a specific T3DVector }
begin
Result:=inherited InsertPointer(V,Ptr);
AdjacentFaces.Add;
Normals.Add(MakeVector3D(0,0,0));
GLColors.Add(GLWhite);
ExtentMin:=GetMinimumVector(V,ExtentMin);
ExtentMax:=GetMaximumVector(V,ExtentMax);
end;




function TVertexArray.Insert(D: T3DVector): integer;
begin
Result:=InsertPointer(D,0);
end;




procedure TVertexArray.Delete(Index: integer);
{ Delete Vertex and pointer from Hash Table }
begin
Delete(Index);
AdjacentFaces.DeleteRow(Index);
Normals.Delete(Index);
GLColors.Delete(Index);
Rehash;
end;




function TVertexArray.AddUnique(V: T3DVector; Color: TGLColor): integer;
{ Insert Vertex Into Hash Table if item is unique}
{ Note, this is different from the underlying "AddUnique"}
{ in that it always returns Index where as the inherited }
{routine returns -1 if already in the list}
begin
Result:=Search(V);
if Result=-1 then
	begin
	Result:=Insert(V);
	GLColors[Result]:=Color;
	end;
end;


function TVertexArray.AddNonUnique(V: T3DVector; Color: TGLColor): integer;
begin
Result:=Insert(V);
GLColors[Result]:=Color;
end;




procedure TVertexArray.AddFaceLink(VertInx,FaceInx: integer);
begin
AdjacentFaces.Add(VertInx,FaceInx);
end;


function TVertexArray.GetColor(Index: integer): TColor;
begin
Result:=GLColorToColor(GLColors[Index]);
end;



procedure TVertexArray.SetColor(Index: integer; const Value: TColor);
begin
GLColors[Index]:=ColorToGLColor(Value);
end;


{ TTriInterface }

constructor TTriInterface.Create(var Faces: TFaceList; var Vertices: TVertexArray);
begin
Self.Faces:=Faces;
Self.Vertices:=Vertices;
end;



procedure TTriInterface.Delete(Index: integer);
var Row,Col,Link: integer;
begin
for Row:=0 to Vertices.AdjacentFaces.Rows-1 do
 for Col:=0 to Vertices.AdjacentFaces.Cols[Row]-1 do
	begin
	Link:=Vertices.AdjacentFaces[Col,Row];
	if Link=Index then Vertices.AdjacentFaces.DeleteItem(Col,Row)
	else if Link>Index then Vertices.AdjacentFaces[Col,Row]:=Link-1;
	end;
Faces.Delete(Index);
Vertices.Rehash;
end;





function TTriInterface.GetCount: integer;
begin
Result:=Faces.Count;
end;



function TTriInterface.GetTriangle(Index: integer): TTriangle;
begin
Result:=MakeTriangle(Vertices[Faces[Index].P1], Vertices[Faces[Index].P2], Vertices[Faces[Index].P3]);
end;




procedure TTriInterface.PutTriangle(Index: integer; const Value: TTriangle);
begin
Vertices[Faces[Index].P1]:=Value.P1;
Vertices[Faces[Index].P2]:=Value.P2;
Vertices[Faces[Index].P3]:=Value.P3;
end;


function TTriInterface.GetColors(Index: integer): TColor;
begin
Result:=Faces.Colors[Index];
end;


procedure TTriInterface.SetColors(Index: integer; const Value: TColor);
begin
Faces.Colors[Index]:=Value;
end;



function TTriInterface.GetGLColors(Index: integer): TGLColor;
begin
Result:=Faces.GLColors[Index];
end;


procedure TTriInterface.SetGLColors(Index: integer; const Value: TGLColor);
begin
Faces.GLColors[Index]:=Value;
end;



function TTriInterface.GetGLFaceColor(Index: integer): TGLFaceColor;
begin
Result.C1:=GLColorToColor(Vertices.GLColors[Faces[Index].P1]);
Result.C2:=GLColorToColor(Vertices.GLColors[Faces[Index].P2]);
Result.C3:=GLColorToColor(Vertices.GLColors[Faces[Index].P3]);
end;


procedure TTriInterface.SetGLFaceColor(Index: integer; const Value: TGLFaceColor);
begin
Vertices.GLColors[Faces[Index].P1]:=ColorToGLColor(Value.C1);
Vertices.GLColors[Faces[Index].P2]:=ColorToGLColor(Value.C2);
Vertices.GLColors[Faces[Index].P3]:=ColorToGLColor(Value.C3);
end;

end.
