unit CardLib;


interface

uses Forms,Windows,SysUtils,Types,Graphics,Classes,MMSystem,Math,
     Vector,IntVector,DSupport,BitmapLib,WinSupport;


{ Base locations in the bitmap resource file "Cards" }

const BackBase = 200;			{Base location of card back }
const CardMask = 400;			{Base location of card mask }
const CardPocket = 401;			{Base location of card pocket }
const CardHighlight = 402;		{Base location of card Highligh }
const CardShadow = 403;			{Base location of card Shadow }
const SmallOffset = 1000;		{Offset to start of small cards}

{Callback procedure prototype  }

type TUpdateProc = procedure(Redraw: boolean);

{ Card definitions }

type TCardSuit = (csHearts,csDiamonds,csClubs,csSpades);
type TCardFace = (cfAce,cfTwo,cfThree,cfFour,cfFive,cfSix,cfSeven,cfEight,cfNine,cfTen,cfJack,cfQueen,cfKing);

const FaceName: array [0..12] of string[2]=('A','2','3','4','5','6','7','8','9','10','J','Q','K');
const Suit: array [0..3] of string[2]=('H','D','C','S');

const LongFaceName: array [0..12] of string=
	('Ace','Two','Three','Four','Five','Six','Seven',
	'Eight','Nine','Ten','Jack','Queen','King');

const LongSuitName: array [0..3] of string=('Hearts','Diamonds','Clubs','Spades');

var IsWinNT: boolean;	{Flags Windows NT or above}

{ TCard Object to handle an individual card}


type TCard = class(TObject)
 private
  function GetHeight: integer;
  function GetWidth: integer;
  function GetRectangle: TRect;
  function CardsOverlap(Card: TCard): boolean;
  procedure CopyBackground(Canvas: TCanvas);
  procedure HighlightCard(Canvas: TCanvas);
  procedure ShadowCard(Canvas: TCanvas);
  function GetCenteredRect(BM: TBitmap): TRect;
  procedure TiltCard(Canvas: TCanvas; Angle: double);
  procedure TiltBitmap(BM: TBitmap; Canvas: TCanvas; Angle: double);
  function GetCardSize: TPoint;
 protected
 public
  Face,Back,Mask: TBitmap;
  Highlight,Shadow: TBitmap;
  BackGround: TBitmap;
  CardPos: TPoint;
  ZOrder,Index,DeckIndex: integer;
  CardSuit: TCardSuit;
  CardFace: TCardFace;
  CardName: string;
  FaceUp,DealFaceUp: boolean;
  Visible: boolean;
  Highlighted,Shadowed: boolean;
  constructor Create;
  destructor Destroy; override;
  procedure DrawCard(Canvas: TCanvas);
  function HitTest(P: TPoint): boolean;
  procedure EraseCard(Canvas: TCanvas);
  property Width: integer read GetWidth;
  property Height: integer read GetHeight;
  property ClientRect: TRect read GetRectangle;
  property CardSize: TPoint read GetCardSize;
 end;

function GetLongCardName(Face: TCardFace; Suit: TCardSuit): string; overload;
function GetLongCardName(C: TCard): string; overload;


{ TDeck object to handle a deck of cards}

const CardsPerDeck = 52;
const CardsPerSuit = 13;

type TCardArray = array of TCard;
type TDeckArray = array of integer;

type TDeckState = record
 DeckArray: TDeckArray;
 Index: integer;
 end;


type TShuffler = class(TObject)
 private
  FDeckCount: integer;
  procedure SuffleArray(var A: array of integer);
  function GetDeckEmpty: boolean;
 protected
  procedure SetDeckCount(const Value: integer);
 public
  Deck: TDeckArray;
  ShuffleIndex: integer;
  constructor Create;
  property DeckCount: integer read FDeckCount write SetDeckCount;
  procedure Shuffle;
  procedure OrderDeck;
  procedure GetDeckState(var DeckState: TDeckState);
  procedure SetDeckState(const DeckState: TDeckState);
  function GetShuffledCard(var Card: integer): boolean;
  property DeckEmpty: boolean read GetDeckEmpty;
 end;

{Deck object to hold and manipulate a deck of TCard objects}

type TDeck = class(TObject)
 private
  FBigCards: boolean;
  FCardBack: integer;
  ShuffleDeck: array of integer;
  FDeckCount: integer;
  procedure ZSortCards(List: TList);
  procedure SetBigCards(const Value: boolean);
  procedure SetCardBack(const Value: integer);
  procedure SetCards;
  function GetHeight: integer;
  function GetWidth: integer;
  function GetCardSize: TPoint;
  procedure SeDeckCount(const Value: integer);
  procedure FreeDeck;
 protected
 public
  HitOffset: TPoint;
  CardArray: TCardArray;
  DebugMode: boolean;
  PocketImage: TBitmap;
  Shuffler: TShuffler;
  function FindTopCard(P: TPoint): integer;
  constructor Create;
  destructor Destroy; override;
  procedure BuildDeck;
  procedure DrawPocket(Canvas: TCanvas; X, Y: integer);
  procedure DisplayCards(Canvas: TCanvas);
  function IndexToFace(Index: integer): TCardFace;
  function IndexToSuit(Index: integer): TCardSuit;
  function IndexToName(Index: integer): string;
  procedure ClearHighlights;
  procedure SetZOrder(Card: TCard);
  procedure SetAllVisible(Visible: boolean);
  procedure GroupMove(P: TPoint; Indices: array of integer; Moving: boolean);
  procedure Animate(Canvas: TCanvas; P1,P2: TPoint; Steps,Delay: integer; Indices: array of integer; CallBack: TUpdateProc);
  procedure FlipCard(Card: TCard; Canvas: TCanvas; Steps, Delay: integer; CallBack: TUpdateProc);
  property CardBack: integer read FCardBack write SetCardBack;
  property BigCards: boolean read FBigCards write SetBigCards;
  property Width: integer read GetWidth;
  property Height: integer read GetHeight;
  property CardSize: TPoint read GetCardSize;
  property DeckCount: integer read FDeckCount write SeDeckCount;
 end;


{Constants specifying different sounds}

type TCardSound = (csMoveTo,csDeal,csError,csMoveHome,csPlayCard,csStartMove,csUndo,csWinner,csBlast,csRocket);

var DisableSound: boolean;	{Flags all sound turned off}

procedure PlayCardSound(Sound: TCardSound; Loop: boolean);
procedure StopSound;



implementation


{$R CARDS}
{$R SOUNDS}

{ Cards are contained in the resource file "Cards.res" }
{ Resource numberin system:

1 - 53 = Standard Card faces (no 0s in resource
100 = Joker,
200's = Backs
300's = Gadgets
400 = Card  AND-Mask,
401 = Recessed Card Pocket
402 =  Card Highlight
403 = Card Drop Shadow

Add 1000 to the base number for the small cards
he order is Ace to King, Hearts, Diamonds, Clubs Spades }


{ Sounds are contained the resource file "Sounds.res"}
{ They are colled with the "TCardSound" constants above}


{ TCard }

constructor TCard.Create;
begin
Face:=TBitmap.Create;
Back:=TBitmap.Create;
BackGround:=TBitmap.Create;
Mask:=TBitmap.Create;
Highlight:=TBitmap.Create;
Shadow:=TBitmap.Create;
end;




destructor TCard.Destroy;
begin
Face.Free;
Back.Free;
BackGround.Free;
Mask.Free;
Highlight.Free;
Shadow.Free;
inherited;
end;



function TCard.GetCenteredRect(BM: TBitmap): TRect;
{Return Rectangle for specified Bitmap Centered on card position}
var X,Y: integer;
begin
X:=CardPos.X-((BM.Width-Width) div 2);
Y:=CardPos.Y-((BM.Height-Height) div 2);
Result:=Rect(X,Y,X+BM.Width,Y+BM.Height);
end;



procedure TCard.CopyBackground(Canvas: TCanvas);
{Save background under card }
{So card can be erased later}
var R1,R2: TRect;
begin
{Use shadow because it is larger than card itself}
R1:=Rect(0,0,Shadow.Width,Shadow.Height);
R2:=GetCenteredRect(Shadow);
Background.Canvas.CopyRect(R1,Canvas,R2);
end;




procedure TCard.ShadowCard(Canvas: TCanvas);
{Drop shadow under the card}
var R: TRect;
begin
{Draw bitmap as transparent highlight}
R:=GetCenteredRect(Shadow);
Canvas.CopyMode:=cmSrcCopy;
AlphaBlendBitmaps(Canvas,R,Shadow,255,True);
end;



procedure TCard.HighlightCard(Canvas: TCanvas);
var R: TRect;
var X,Y: integer;
begin
{Draw bitmap as transparent highlight}
R:=GetCenteredRect(Highlight);
AlphaBlendBitmaps(Canvas,R,Highlight,128);

{Then add non-transparent highlight}
Canvas.Pen.Color:=clYellow;
Canvas.Pen.Width:=1;
R:=ClientRect;
InflateRect(R,1,1);
Canvas.Rectangle(R);
end;



procedure TCard.DrawCard(Canvas: TCanvas);
{Draw the card or back, with shadowing and highlights}
const  ROP_DstCopy = $00AA0029;
begin
if Shadowed then ShadowCard(Canvas);
if Highlighted then HighlightCard(Canvas);
if IsWinNT then
	begin
	{Use MaskBlt to carve out the corners if Window NT or above}
	{ It is 3X faster than other techniques}
	if FaceUp then MaskBlt(Canvas.Handle,CardPos.X,CardPos.Y,
			       Face.Width,Face.Height, Face.Canvas.Handle,0,0,
			       Mask.Handle,0,0, MakeRop4(ROP_DstCopy, SrcCopy))
	else MaskBlt(Canvas.Handle,CardPos.X,CardPos.Y,
		     Back.Width,Back.Height, Back.Canvas.Handle,0,0,
		     Mask.Handle,0,0, MakeRop4(ROP_DstCopy, SrcCopy));
	end
else
	begin
	{Win98 and earlier doesn't support MaskBlt}
	{So manually use mask to cut out corners of card}
	Canvas.CopyMode:=cmSrcAnd;
	Canvas.Draw(CardPos.X,CardPos.Y,Mask);
	{Insures the card face will be white}
	Canvas.Brush.Color:=clWhite;
	{Logical OR the face into black mask area}
	Canvas.CopyMode:=cmSrcPaint;
	if FaceUp then Canvas.Draw(CardPos.X,CardPos.Y,Face)
	else Canvas.Draw(CardPos.X,CardPos.Y,Back);
	end;
end;


procedure TCard.EraseCard(Canvas: TCanvas);
{Erase Card using previously saved background}
var R: TRect;
begin
R:=GetCenteredRect(Background);
Canvas.CopyMode:=cmSrcCopy;
Canvas.Draw(R.Left,R.Top,Background);
end;



function TCard.HitTest(P: TPoint): boolean;
{Does point fall within card's rectangle?}
begin
Result:=PtInRect(ClientRect,P);
end;



function TCard.CardsOverlap(Card: TCard): boolean;
{Does the specific card overlap this card?}
begin
Result:=DoRectsOverlap(ClientRect,Card.ClientRect);
end;




function TCard.GetHeight: integer;
{Get cards Height from the underlying bitmap}
begin
Result:=Face.Height;
end;



function TCard.GetWidth: integer;
{Get cards Width from the underlying bitmap}
begin
Result:=Face.Width;
end;



function TCard.GetRectangle: TRect;
{Calculate and return the card's rectangle}
begin
Result:=Rect(CardPos.X,CardPos.Y,CardPos.X+Width,CardPos.Y+Height);
end;



function ShrinkRect(R: TRect; Angle: double): TRect;
{Shrink rectangle horizontally to simulate flipping card}
var W2,D: integer;
begin
Result:=R;
W2:=(R.Right-R.Left) div 2;
D:=W2-round(abs(Cos(DegToRad(Angle)))*W2);
Result.Left:=Result.Left+D; Result.Right:=Result.Right-D;
end;



procedure TCard.TiltBitmap(BM: TBitmap; Canvas: TCanvas; Angle: double);
{Tilt bitmap by specified angle}
var R,R2: TRect;
begin
R:=GetCenteredRect(BM);
{Simulate tilting by shrinking rectangle}
R:=ShrinkRect(R,Angle);
{Oversample to cleanly shrink bitmap}
OversampleBitmap(Canvas,R,BM);
end;


procedure TCard.TiltCard(Canvas: TCanvas; Angle: double);
{Tilt card by specified number of degrees}
{If the angle is greater than 90, start showing the opposite side}
begin
if (FaceUp and (Angle<90)) or
   (not FaceUp and (Angle>=90)) then TiltBitmap(Face,Canvas,Angle)
else TiltBitmap(Back,Canvas,Angle);
end;


function TCard.GetCardSize: TPoint;
{Property handler to retrieve the card size}
begin
Result:=Point(Width,Height);
end;


{------------------ Routine to handle TShuffler Object ----------------------------}


{ TShuffler }


procedure TShuffler.SetDeckCount(const Value: integer);
{Specify the number of decks in use}
begin
if Value<>DeckCount then
	begin
	FDeckCount:=Value;
	SetLength(Deck,CardsPerDeck* DeckCount);
	end;
end;


constructor TShuffler.Create;
begin
DeckCount:=1;
end;



procedure TShuffler.SuffleArray(var A: array of integer);
{Shuffle array in place using Fisher-Yates inside out shuffle}
{Places all values between 0 and N-1 while it is suffling}
var I,J: integer;
begin
A[0]:=0;
for I:=1 to High(A) do
	begin
	J:=Random(I+1);
	if J <> I then A[I]:=A[J];
	A[J]:=I;
	end;
end;


procedure TShuffler.Shuffle;
{Shuffle Card Deck}
begin
Randomize;
SuffleArray(Deck);
ShuffleIndex:=Length(Deck);
end;



procedure TShuffler.OrderDeck;
{Put deck in order Ace to King, Hearts, Diamonds, Clubs, Spades }
var I: integer;
begin
for I:=0 to High(Deck) do Deck[I]:=I;
ShuffleIndex:=Length(Deck);
end;



function TShuffler.GetShuffledCard(var Card: integer): boolean;
{Get one card from the shuffled deck}
begin
Result:=False;
Dec(ShuffleIndex);
if ShuffleIndex<0 then exit;
Card:=Deck[ShuffleIndex];
Result:=True;
end;



function TShuffler.GetDeckEmpty: boolean;
begin
Result:=ShuffleIndex<1;
end;



procedure TShuffler.GetDeckState(var DeckState: TDeckState);
var I: integer;
begin
SetLength(DeckState.DeckArray,Length(Deck));
for I:=0 to High(DeckState.DeckArray) do DeckState.DeckArray[I]:=Deck[I];
DeckState.Index:=ShuffleIndex;
end;



procedure TShuffler.SetDeckState(const DeckState: TDeckState);
var I: integer;
begin
for I:=0 to High(Deck) do Deck[I]:=DeckState.DeckArray[I];
ShuffleIndex:=DeckState.Index;
end;



{------------------ Routine to handle TDeck Object ----------------------------}



procedure LoadCardBitmap(BM: TBitmap; Card: integer);
{Load specified card into Bitmap Windows resource}
begin
BM.Handle:=LoadBitmap(HInstance,PChar(Card));
end;


procedure TDeck.FreeDeck;
{Free all card object and zero array}
var I: integer;
begin
for I:=0 to High(CardArray) do CardArray[I].Free;
SetLength(CardArray,0);
end;



procedure TDeck.BuildDeck;
{Build basic array of cards without loading images}
var I: integer;
begin
FreeDeck;
SetLength(CardArray,CardsPerDeck * DeckCount);
for I:=0 to High(CardArray) do
	begin
	CardArray[I]:=TCard.Create;
	CardArray[I].Index:=I;
	CardArray[I].DeckIndex:=I mod CardsPerDeck;
	CardArray[I].CardFace:=IndexToFace(I);
	CardArray[I].CardSuit:=IndexToSuit(I);
	CardArray[I].CardName:=IndexToName(I);
	CardArray[I].FaceUp:=True;
	CardArray[I].Visible:=True;
	CardArray[I].ZOrder:=I;
	end;
Shuffler.DeckCount:=DeckCount;
end;


procedure TDeck.SetCards;
{Load specific images into individual cards}
{Based on the specified card size and card backs}
var I,Inx: integer;
var CardBase,BackOff: integer;

	procedure BuildShadow;
	{Setup Alpha channels for shadow}
	begin
	SetAlphaByColor(CardArray[I].Shadow,clBlack,64,True);
	SetAlphaByColor(CardArray[I].Shadow,clWhite,0,False);
	end;

begin
if BigCards then CardBase:=0
else CardBase:=SmallOffset;
BackOff:=CardBase + BackBase;
LoadCardBitmap(PocketImage,CardBase+CardPocket);
for I:=0 to High(CardArray) do
	begin
	Inx:=I mod CardsPerDeck;
	LoadCardBitmap(CardArray[I].Face,CardBase+Inx+1);
	LoadCardBitmap(CardArray[I].Back,BackOff+CardBack);
	LoadCardBitmap(CardArray[I].Mask,CardBase+CardMask);
	LoadCardBitmap(CardArray[I].Highlight,CardBase+CardHighlight);
	LoadCardBitmap(CardArray[I].Shadow,CardBase+CardShadow);
	BuildShadow;
	CardArray[I].BackGround.Width:=CardArray[I].Shadow.Width;
	CardArray[I].BackGround.Height:=CardArray[I].Shadow.Height;
	{Mask is required to be monochrome for MaskBlt}
	CardArray[I].Mask.PixelFormat:=pf1bit;
	end;
end;




procedure TDeck.SeDeckCount(const Value: integer);
begin
if FDeckCount<>Value then
	begin
	FDeckCount:=Value;
	BuildDeck;
	SetCards;
	end;
end;


constructor TDeck.Create;
begin
PocketImage:=TBitmap.Create;
Shuffler:=TShuffler.Create;
SetLength(CardArray,0);
DebugMode:=False;
FBigCards:=True;
FCardBack:=5;
DeckCount:=1;
end;



destructor TDeck.Destroy;
begin
Shuffler.Free;
FreeDeck;
PocketImage.Free;
inherited;
end;



procedure TDeck.DrawPocket(Canvas: TCanvas; X,Y: integer);
{Draw current card pocket }
begin
Canvas.CopyMode:=cmSrcCopy;
Canvas.Draw(X,Y,PocketImage);
end;



function TDeck.FindTopCard(P: TPoint): integer;
{Fine Top visible card under specified point}
var I,Z: integer;
begin
Result:=-1;
HitOffset:=Point(0,0);
Z:=-1;
for I:=0 to High(CardArray) do
 if CardArray[I].HitTest(P) and (CardArray[I].ZOrder>Z) then
	begin
	Result:=I;
	Z:=CardArray[I].ZOrder;
	end;
{Mouse clicks can happen anywhere on the card}
{So calculate the offset between selection point and the upper-left corner }
if Result>=0 then HitOffset:=IntVectorSubtract(P,CardArray[Result].CardPos);
end;



procedure TDeck.SetZOrder(Card: TCard);
{Set the Z-Order of the specified Card}
var I,Z: integer;
begin
Z:=0;
{Find the maximum Z of all cards that are visible and overlap}
for I:=0 to High(CardArray) do
 if (Card<>CardArray[I]) and CardArray[I].Visible and
     CardArray[I].CardsOverlap(Card) and (CardArray[I].ZOrder>Z) then Z:=CardArray[I].ZOrder;
{Put this card on top by making one greater}
Card.ZOrder:=Z+1;
end;



function ZCompare(Item1, Item2: Pointer): Integer;
{Callback routine to Compare Z-Order for the Quick Sort}
begin
Result:=TCard(Item1).ZOrder-TCard(Item2).ZOrder;
end;


procedure TDeck.ZSortCards(List: TList);
{Z-Sort a list of cards}
var I: integer;
begin
{Copy the cards to a TList and use it to sort}
for I:=0 to High(CardArray) do List.Add(CardArray[I]);
List.Sort(ZCompare);
end;



procedure TDeck.DisplayCards(Canvas: TCanvas);
{Display all currently visible cards in the Deck}
var I: integer;
var List: TList;
var Card: TCard;
var S: string;
var X: integer;
begin
List:=TList.Create;
try
{Z-Sort cards }
ZSortCards(List);
{Now display them in Farest to Nearest order}
for I:=0 to List.Count-1 do
 if TCard(List[I]).Visible then
	begin
	Card:=TCard(List[I]);
	Card.DrawCard(Canvas);
	{If Debugging show info on the cards}
	if DebugMode then
		begin
		Canvas.Brush.Color:=clWhite;
		S:=IntToStr(Card.ZOrder)+' '+IntToStr(Card.Index);
		X:=(Card.CardPos.X + Card.CardSize.X) - Canvas.TextWidth(S);
		Canvas.TextOut(X,Card.CardPos.Y+5,S);
		end;
	end;
finally List.Free; end;
end;



procedure TDeck.SetAllVisible(Visible: boolean);
{Set the visibility of all cards}
var I: integer;
begin
for I:=0 to High(CardArray) do CardArray[I].Visible:=Visible;
end;


procedure TDeck.ClearHighlights;
{Clear the hightlights of all cards}
var I: integer;
begin
for I:=0 to High(CardArray) do CardArray[I].Highlighted:=False;
end;



procedure TDeck.GroupMove(P: TPoint; Indices: array of integer; Moving: boolean);
{Move the list of cards specified in "Indices" to location "P"}
{While moving "Moving" should be True and False on the last step of a move}
var I: integer;
var Off: TPoint;
var Card: TCard;
begin
{Subtract the offset to translate the }
{movement to upper left corner of the card }
P:=IntVectorSubtract(P,HitOffset);
{Find the offset from the top card}
Off:=IntVectorSubtract(P,CardArray[Indices[0]].CardPos);
{Hide these cards so they can be Z-Ordered one at a time }
for I:=0 to High(Indices) do CardArray[Indices[I]].Visible:=False;
for I:=0 to High(Indices) do
	begin
	Card:=CardArray[Indices[I]];
	{If we are moving, shadow the card}
	Card.Shadowed:=Moving;
	{Move this card}
	Card.CardPos:=IntVectorAdd(Card.CardPos,Off);
	{If the card is moving, use a very high Z-Order, so it is over all other cards}
	if Moving then Card.ZOrder:=High(Integer)-(High(Indices)-I)
	else SetZOrder(Card);
	Card.Visible:=True;
	end;
end;



procedure TDeck.Animate(Canvas: TCanvas; P1,P2: TPoint; Steps,Delay: integer; Indices: array of integer; CallBack: TUpdateProc);
{Animate moving a list of cards from one point to another}
{P1 and P2 = the starting and ending locations}
{Step = the number of steps to cover the distance}
{Delay = Time in milliseconds between each step}
{Indices = List of Indices to cards to be moved}
{Callback = Subroutine update the display after each step}
var I: integer;
var D,P: TPoint;
var DI: T2DVector;
begin
{Exit if no valid callback specified}
if not Assigned(CallBack) then exit;
{Zero out the HitOffset so we can use "GroupMove" }
HitOffset:=Point(0,0);
{Calculate the vector distance to move each step}
D:=IntVectorSubtract(P2,P1);
DI:=ScalarDivide2D(PointToVector2D(D),Steps);
for I:=0 to Steps do
	begin
	{Calculate the position for this step}
	P:=IntVectorAdd(P1,Vector2DToPoint(ScalarProduct2D(DI,I)));
	{Move the cards in the list one steop}
	GroupMove(P,Indices,I<Steps);
	{Update the screen}
	CallBack(True);
	{Delay}
	DoTimeDelay(Delay);
	end;
end;




procedure TDeck.FlipCard(Card: TCard; Canvas: TCanvas; Steps,Delay: integer; CallBack: TUpdateProc);
{Animate flipping the specified card over }
{Step = the number of steps to cover the distance}
{Delay = Time in milliseconds between each step}
{Callback = Subroutine update the display after each step}
var Angle,Increment: double;
var I: integer;
begin
{Hide the card}
Card.Visible:=False;
{Then display all the other cards}
DisplayCards(Canvas);
CallBack(True);
{Copy the background so we can erase}
Card.CopyBackground(Canvas);
{Calculate the step increment}
Increment:=180/Steps;
for I:=0 to Steps do
	begin
	{Calculate the current angle}
	Angle:=Increment * I;
	{Erase the card}
	Card.EraseCard(Canvas);
	{Draw the card titted}
	Card.TiltCard(Canvas,Angle);
	CallBack(False);
	DoTimeDelay(Delay);
	end;
{Flip the card over and show it}
Card.FaceUp:=not Card.FaceUp;
Card.Visible:=True;
CallBack(True);
end;


procedure TDeck.SetBigCards(const Value: boolean);
{Code to set the "BigCards" property}
begin
if FBigCards<>Value then
	begin
	FBigCards := Value;
	SetCards;
	end;
end;



procedure TDeck.SetCardBack(const Value: integer);
{Code to set the "CardBack" property}
begin
if FCardBack<>Value then
	begin
	FCardBack := Value;
	SetCards;
	end;
end;



function TDeck.GetHeight: integer;
{Code to Get the "Height" property's value}
begin
Result:=CardArray[0].Height;
end;



function TDeck.GetWidth: integer;
{Code to Get the "Width" property's value}
begin
Result:=CardArray[0].Width;
end;





function TDeck.GetCardSize: TPoint;
{Code to Get the "CardSize" property's value}
begin
Result:=CardArray[0].CardSize;
end;


function TDeck.IndexToFace(Index: integer): TCardFace;
{Get the face value of the card, disregarding suit}
begin
Result:=TCardFace((Index) mod CardsPerSuit);
end;



function TDeck.IndexToSuit(Index: integer): TCardSuit;
{Get the Suit value of the card, disregarding Face}
begin
Result:=TCardSuit((Index div CardsPerSuit) mod 4);
end;



function TDeck.IndexToName(Index: integer): string;
{Return string representing the card }
begin
if (Index<0) or (Index>High(CardArray)) then Result:='??'
else
	begin
	Result:=FaceName[(Index) mod CardsPerSuit];
	Result:=Result + Suit[(Index div CardsPerSuit) mod 4];
	end;
end;


{----------------- Audio Effects ---------------------------------------------}



procedure PlayCardSound(Sound: TCardSound; Loop: boolean);
{Play the specified sound, play continuously if "Loop" is true}
{Sounds are contained in the "Sounds.res" Resource file}
var S: string;
var Flags: cardinal;
begin
if DisableSound then exit;
case Sound of
 csDeal: S:='DEAL';
 csError: S:='ERROR';
 csMoveHome: S:='MOVEHOME';
 csPlayCard: S:='PLAYCARD';
 csStartMove: S:='STARTMOVE';
 csMoveTo: S:='MOVETO';
 csUndo: S:='UNDO';
 csWinner: S:='WINNER';
 csBlast: S:='BLAST';
 csRocket: S:='ROCKET';
 end;
Flags:=SND_RESOURCE or SND_ASYNC;
if Loop then Flags:=Flags or SND_LOOP;
PlaySound(PChar(S),hInstance,Flags);
end;



procedure StopSound;
{Stop all sounds from playing}
begin
PlaySound(nil,hInstance,SND_PURGE);
end;


{--------------------- Utility Subroutines ------------------------------------}



function GetLongCardName(Face: TCardFace; Suit: TCardSuit): string;
begin
Result:=LongFaceName[integer(Face)]+' of '+LongSuitName[integer(Suit)];
end;


function GetLongCardName(C: TCard): string;
begin
Result:=GetLongCardName(C.CardFace,C.CardSuit);
end;



initialization
IsWinNT:=IsWindowsNT;
finalization
end.
