unit Hash;

{
This unit is used to create a HashTable object. The object
has the following capabilities:

1. It is based on the THugeString object so it can maintain
a list of strings that automatically grows as items are added.

2. The strings added to the list are hashed so the list can be
searched with maximum efficiency.
}

interface

uses SysUtils,Classes,Math,HugeArry,Vector;

type TAutoBase = class(TObject)
 public
  constructor Create; virtual;
  end;

type TAutoClass = class of TAutoBase;


{ Variables For Hash Table Lookup }

const HashEmpty=-1;		{ Empty Symbol Pointer }


{------------------------------------------------------------}
{------ Declaration for Custom Hash Table object ------------}
{------------------------------------------------------------}


type TCustomHashTable = class(TObject)
 private
  procedure SetBoxPower(V: integer);
  procedure SetupHashBoxes;
  procedure RemoveHashKey(Index, HashKey: integer);
  procedure RehashItem(Index, HashKey: integer);
 protected
  HashValue: cardinal;			{ Current Identifier Hash Code }
  FBoxPower: integer;			{ Power of the number of boxes}
  FBoxCount: integer;			{ Number Of Boxes MUST be power of two}
  FHashMask: integer;			{ Hash Mask }
  procedure ClearHashBox;
 public
  HashBox: THugeLongint;		{ Hash Boxes (symbol List Headers) }
  HashLink: THugeLongint;		{ List Linkage Pointers }
  constructor Create(StartAlloc: integer); virtual;
  destructor Destroy; override;
  procedure Assign(Hash: TCustomHashTable);
  procedure Append(Hash: TCustomHashTable);
  procedure Rehash;  virtual;
  procedure CustomSort(Compare: TListSortCompare); virtual;
  procedure Sort; virtual;
  procedure GetDistribution(var Dist: THugeInteger);
  procedure Clear; virtual;
  property BoxPower: integer read FBoxPower write SetBoxPower;
 end;



{------------------------------------------------------------}
{------ Declaration for Short String Hash Table object ------}
{------------------------------------------------------------}



{Hashtable object Declaration}

type THashTable = class(TCustomHashTable)
 private
  FCaseSensitive: boolean;
  procedure SetCaseSensitive(V: boolean);
  function ReadString(I: integer): shortstring;
  procedure WriteString(I: integer; V: shortstring);
  function GetCount: integer;
  function GetOnChange: TNotifyEvent;
  procedure SetOnChange(const Value: TNotifyEvent);
 protected
 public
  Strings: THugeString;		{ String table }
  constructor Create(StartAlloc,StrSize: LongInt); overload;
  constructor Create(StartAlloc: LongInt); overload;
  constructor Create; overload;
  destructor Destroy; override;
  function GetHashValue(S: ShortString): cardinal;
  procedure Assign(Hash: THashTable);
  procedure Append(Hash: THashTable);
  procedure Rehash;  virtual;
  procedure CustomSort(Compare: THugeStringSortCompare); virtual;
  procedure Sort; virtual;
  function Search(S: ShortString): Longint;
  function Insert(S: ShortString): integer;
  function Add(S: ShortString): integer;
  procedure Delete(Index: integer); virtual;
  function AddUnique(S: ShortString): integer;
  function InsertPointer(S: shortString; Ptr: integer): integer; virtual;
  procedure Clear; override;
  property Value[Inx: integer]: shortstring read ReadString write WriteString; default;
  property Count: integer read GetCount;
  property CaseSensitive: boolean read FCaseSensitive write SetCaseSensitive;
  property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
 end;


{-----------------------------------------------------------------}
{----------- Declaration for Auto Hash Table object --------------}
{-----------------------------------------------------------------}

{TAutoHash is a THashTable derrivative that can be configured to automatically
{put certain object types into the pointer array. The object type is specified as}
{in the argument to the Create method. Once it is specified, the every time the}
{"Add/AddUnique" method is called, a new object is added of the type specified. It}
{automatically Frees an object when it is deleted or when the array is cleared}

{Warning - if your object has a constructor, you must base it on the TAutoBase
because the constructor for TObject and its class TClass are not "virtual" and
the polymorphism of executing the underlying objects constructor won't work.
That means Delphi objects based on TObject won't work.}


type TAutoHash = class(THashTable)
 private
  AutoClass: TAutoClass;
  ObjectClass: TClass;
  procedure CreateClassInstance(Index: integer);
 protected
 public
  constructor Create(ObjClass: TAutoClass); overload; virtual;
  constructor Create(ObjClass: TClass); overload; virtual;
  function Add(const S: string): Integer;
  function AddUnique(S: ShortString): integer;
  procedure Delete(Index: Integer);
  procedure Clear; override;
 published
 end;


{-----------------------------------------------------------------}
{----------- Declaration for Long Hash Table object --------------}
{-----------------------------------------------------------------}


{Hashtable object Declaration for long strings}

type TLongHashTable = class(TCustomHashTable)
 private
  FCaseSensitive: boolean;
  procedure SetCaseSensitive(V: boolean);
  function ReadString(I: integer): string;
  procedure WriteString(I: integer; V: string);
  function GetCount: integer;
  function GetOnChange: TNotifyEvent;
  procedure SetOnChange(const Value: TNotifyEvent);
 protected
 public
  Strings: THugeLongString;		{ String table }
  constructor Create(StartAlloc: LongInt); overload;
  constructor Create; overload;
  destructor Destroy; override;
  procedure Assign(Hash: TLongHashTable);
  function GetHashValue(S: string): cardinal;
  function Search(S: string): Longint;
  function Insert(S: string): integer;
  function Add(S: string): integer;
  function AddUnique(S: string): integer;
  procedure Delete(Index: integer);
  function InsertPointer(S: string; Ptr: integer): integer;
  procedure Rehash;
  procedure CustomSort(Compare: THugeLongStringSortCompare); virtual;
  procedure Sort; virtual;
  procedure Clear; override;
  property Value[Inx: integer]: string read ReadString write WriteString; default;
  property Count: integer read GetCount;
  property CaseSensitive: boolean read FCaseSensitive write SetCaseSensitive;
  property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
 end;

{-----------------------------------------------------------------}
{----------- Declaration for Integer Hash Table object -----------}
{-----------------------------------------------------------------}


{Hashtable object Declaration for Integer Arrays}

type TIntegerHashTable = class(TCustomHashTable)
 private
  function ReadInteger(I: integer): integer;
  procedure WriteInteger(I: integer; V: integer);
  function GetCount: integer;
    function GetOnChange: TNotifyEvent;
    procedure SetOnChange(const Value: TNotifyEvent);
 protected
  function GetHashValue(C: cardinal): cardinal;
 public
  Items: THugeInteger;		{ integer table }
  constructor Create(StartAlloc: Integer);
  destructor Destroy; override;
  procedure Assign(Hash: TIntegerHashTable);
  function Search(I: integer): Integer;
  function Insert(I: integer): integer;
  function Add(I: integer): integer;
  procedure Delete(Index: integer);
  function AddUnique(I: integer): integer;
  function InsertPointer(I,Ptr: integer): integer;
  procedure Rehash;
  procedure CustomSort(Compare: TListSortCompare); virtual;
  procedure Sort; virtual;
  procedure Clear; override;
  property Value[Inx: integer]: integer read ReadInteger write WriteInteger; default;
  property Count: integer read GetCount;
  property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
 end;


{Double Hashtable object Declaration}

type TDoubleHashTable = class(TCustomHashTable)
 private
  function ReadDouble(I: integer): double;
  procedure WriteDouble(I: integer; D: double);
  function GetCount: integer;
    function GetOnChange: TNotifyEvent;
    procedure SetOnChange(const Value: TNotifyEvent);
 protected
  function GetHashValue(D: double): cardinal;
 public
  Items: THugeDouble;
  constructor Create(StartAlloc: LongInt);
  destructor Destroy; override;
  procedure Assign(Hash: TDoubleHashTable);
  procedure Rehash;  virtual;
  procedure CustomSort(Compare: TListSortCompare); virtual;
  procedure Sort; virtual;
  function Search(D: double): Longint;
  function Insert(D: double): integer;
  function Add(D: double): integer;
  procedure Delete(Index: integer);
  function AddUnique(D: double): integer;
  function InsertPointer(D: double; Ptr: integer): integer;
  procedure Clear; override;
  property Value[Inx: integer]: double read ReadDouble write WriteDouble; default;
  property Count: integer read GetCount;
  property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
 end;


{Vector 3D Hashtable object Declaration}

type T3DVectorHash = class(TCustomHashTable)
 private
  function Read3DVector(I: integer): T3DVector;
  procedure Write3DVector(I: integer; V: T3DVector);
  function GetCount: integer;
  function GetOnChange: TNotifyEvent;
  procedure SetOnChange(const Value: TNotifyEvent);
  procedure SetCount(const Value: integer);
 protected
  function GetHashValue(V: T3DVector): cardinal;
 public
  Items: THuge3DVector;
  constructor Create(StartAlloc: LongInt);
  destructor Destroy; override;
  procedure Assign(Hash: T3DVectorHash);
  procedure Append(Hash: T3DVectorHash);
  procedure Rehash;  virtual;
  procedure CustomSort(Compare: TListSortCompare); virtual;
  procedure Sort; virtual;
  function Search(D: T3DVector): integer;
  function Insert(D: T3DVector): integer;
  function Add(D: T3DVector): integer;
  procedure Delete(Index: integer);
  function AddUnique(D: T3DVector): integer;
  function InsertPointer(D: T3DVector; Ptr: integer): integer; virtual;
  procedure Clear; override;
  property Value[Inx: integer]: T3DVector read Read3DVector write Write3DVector; default;
  property Count: integer read GetCount write SetCount;
  property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
 end;



const FNV_prime = 16777619;		{Magic prime number which depends on hash size}
const offset_basis = 2166136261;	{Magic number used as starting hash also depends on hash size}


 implementation


{-------------------------------------------------------------------------}
{           Implementation for custom hash table object                   }
{-------------------------------------------------------------------------}


procedure TCustomHashTable.RemoveHashKey(Index,HashKey: integer);
{Remove previous hash links associated with Index and HashKey}
var NewIndex,OldIndex: integer;
begin
NewIndex:=HashBox[HashKey];
{Handle case where there are no links}
if HashLink[NewIndex]=HashEmpty then
	begin
	HashBox[HashKey]:=HashEmpty;
	exit;
	end;
if NewIndex=Index then
	begin
	{Handle case where target is in HashBoxes}
	HashBox[HashKey]:=HashLink[NewIndex];
	exit;
	end
else
	begin
	{Look through links for index that corresponds to key}
	while True do
		begin
		if HashLink[NewIndex]=Index then break;
		{Degenerate case - only occurs if the old value not in list}
		if HashLink[NewIndex]=HashEmpty then exit;
		NewIndex:=HashLink[NewIndex];
		end;
	end;
{Now delete that link and move everything else up one}
while True do
	begin
	OldIndex:=NewIndex;
	NewIndex:=HashLink[NewIndex];
	HashLink[OldIndex]:=HashLink[NewIndex];
	if HashLink[NewIndex]=HashEmpty then break;
	end;
end;




procedure TCustomHashTable.RehashItem(Index,HashKey: integer);
begin
HashLink[Index]:=HashBox[HashKey];
HashBox[HashKey]:=Index;
end;


procedure TCustomHashTable.SetupHashBoxes;
{Create and intialize hash boxes}
begin
HashBox.ReAllocMem(FBoxCount);
HashBox.ItemIndex:=FBoxCount;
ClearHashBox;
Rehash;
end;



procedure TCustomHashTable.SetBoxPower(V: integer);
{Set number of hash boxes based on power of two}
begin
if V<>FBoxPower then
	begin
	if V<4 then FBoxPower:=4 else FBoxPower:=V;
	FBoxCount:=1 shl FBoxPower;
	FHashMask:=FBoxCount-1;
	SetupHashBoxes;
	end;
end;





procedure  TCustomHashTable.ClearHashBox;
{ Clear Hash Box }
var P : Integer;
begin
HashBox.FillArray(HashEmpty);
end;



procedure TCustomHashTable.Clear;
{Clear all the hash boxes and reinitialize}
begin
HashLink.Clear;
SetupHashBoxes;
end;



procedure TCustomHashTable.Rehash;
begin
{Virtual procedure filled in by descendents}
end;


procedure TCustomHashTable.GetDistribution(var Dist: THugeInteger);
{Get the distribution of hashbox values - for testing purposes}
var I,Inx,Cnt: integer;
begin
Dist.Clear;
for I:=0 to HashBox.ItemIndex-1 do
	begin
	Inx:=HashBox[I];
	Cnt:=0;
	while Inx<>HashEmpty do
		begin
		Inc(Cnt);
		Inx:=HashLink[Inx];
		end;
	Dist.Add(Cnt)
	end;
end;



procedure TCustomHashTable.CustomSort(Compare: TListSortCompare);
begin
{Virtual procedure filled in by descendents}
end;

procedure TCustomHashTable.Sort;
begin
{Virtual procedure filled in by descendents}
end;

{-------------------------------------------------------------------}


constructor TCustomHashTable.Create(StartAlloc: integer);
{Creation routine for hash table object }
begin
inherited Create;
HashLink:=THugeLongInt.Create(StartAlloc);
HashBox:=THugeLongInt.Create(5);
BoxPower:=11;	{2048}
end;


destructor TCustomHashTable.Destroy;
{Destroy hash table object }
begin
HashLink.Free;
HashBox.Free;
inherited Destroy;
end;




procedure TCustomHashTable.Assign(Hash: TCustomHashTable);
begin
HashLink.Assign(Hash.HashLink);
HashBox.Assign(Hash.HashBox);
BoxPower:=Hash.BoxPower;
end;




procedure TCustomHashTable.Append(Hash: TCustomHashTable);
begin
HashLink.Append(Hash.HashLink);
HashBox.Append(Hash.HashBox);
end;





{-------------------------------------------------------------------------}
{                  Implementation for hash table object                   }
{-------------------------------------------------------------------------}


function THashTable.GetHashValue(S: shortString): cardinal;
{ Return Hash For String S }
{ Hashing algorith based on the FNV-1as (Fowler / Noll / Vo)}
var P : Integer;
begin
if not CaseSensitive then S:=LowerCase(S);
Result:=offset_basis;
for P:=1 to Length(S) do
	begin
	Result:=Result xor byte(S[P]);
	Result:= Result * FNV_Prime;
	end;
Result:=((Result shr FBoxPower) xor Result) and  FHashMask;
end;



procedure THashTable.Clear;
{Clear all the hash boxes and reinitialize}
begin
inherited Clear;
Strings.Clear;
end;





procedure  THashTable.Rehash;
{ Re-do The Hash Links For The Table }
var P : Longint;
begin
ClearHashBox;
HashLink.ItemIndex:=Strings.ItemIndex;
for P:=0 to Strings.ItemIndex-1 do
	begin
	HashValue:=GetHashValue(Strings[P]);
	HashLink[P]:=HashBox[HashValue];
	HashBox[HashValue]:=P;
	end;
end;


function THashTable.Search(S: shortString): Longint;
{ Search For String S, return index of string }
{ Returns -1 if not found }
begin
HashValue:=GetHashValue(S);
Result:=HashBox[HashValue];
while True do
	begin
	if Result=HashEmpty then break;
	if Strings.Compare(S,Result) then break;
	Result:=HashLink[Result];
	end;
end;


procedure THashTable.SetCaseSensitive(V: boolean);
{Used to set the Case Sensitivity property}
begin
if V<>FCaseSensitive then
	begin
	FCaseSensitive:=V;
	Strings.CaseSensitive:=V;
	Rehash;
	end;
end;


{-------------------------------------------------------------------}




constructor THashTable.Create(StartAlloc,StrSize: LongInt);
{Creation routine for hash table object }
begin
inherited Create(StartAlloc);
Strings:=THugeString.Create(StartAlloc,StrSize);
CaseSensitive:=True;
end;



constructor THashTable.Create(StartAlloc: LongInt);
begin
Create(StartAlloc,255);
end;


constructor THashTable.Create;
begin
Create(5,255);
end;



destructor THashTable.Destroy;
{Destroy hash table object }
begin
Strings.Free;
inherited Destroy;
end;


procedure THashTable.Assign(Hash: THashTable);
begin
inherited Assign(Hash);
Strings.Assign(Hash.Strings);
end;



procedure THashTable.Append(Hash: THashTable);
begin
Strings.Append(Hash.Strings);
Rehash;
end;



function  THashTable.GetCount: integer;
begin
Result:=Strings.ItemIndex;
end;



function THashTable.ReadString(I: integer): shortstring;
{Property function so hash table can use array-type syntax}
begin
Result:=Strings[I];
end;



procedure THashTable.WriteString(I: integer; V: shortstring);
{Property function so hash table can use array-type syntax}
begin
if Search(V)<>-1 then exit;
RemoveHashKey(I,GetHashValue(Self[I]));
Strings[I]:=V;
RehashItem(I,GetHashValue(Self[I]));
end;



function THashTable.InsertPointer(S: shortString; Ptr: integer): integer;
{ Insert string and pointer Into Hash Table}
{ Pointer can point to any object the user }
{wishes to associate with a specific string }
begin
HashValue:=GetHashValue(S);
Result:=Strings.AddPointer(S,Ptr);
HashLink.Add(HashBox[HashValue]);
HashBox[HashValue]:=HashLink.ItemIndex-1;
end;



function THashTable.Insert(S: shortString): integer;
{ Insert string Into Hash Table }
begin
Result:=InsertPointer(S,0);
end;



function THashTable.Add(S: shortString): integer;
{ Insert string Into Hash Table }
begin
Result:=InsertPointer(S,0);
end;


procedure THashTable.Delete(Index: integer);
{ Delete string and pointer from Hash Table }
begin
Strings.Delete(Index);
Rehash;
end;



function THashTable.AddUnique(S: ShortString): integer;
{ Insert string Into Hash Table if item is unique}
{Returns -1 if already in list else returns index of new item}
begin
if Search(S)=-1 then Result:=Insert(S)
else Result:=-1;
end;



procedure THashTable.CustomSort(Compare: THugeStringSortCompare);
begin
Strings.CustomSort(Compare);
Rehash;
end;




procedure THashTable.Sort;
begin
Strings.Sort;
Rehash;
end;



function THashTable.GetOnChange: TNotifyEvent;
begin
Result:=Strings.OnChange;
end;



procedure THashTable.SetOnChange(const Value: TNotifyEvent);
begin
Strings.OnChange:=Value;
end;


{------------------------------------------------------------------------------}
{                  Implementation for auto hash table object                   }
{------------------------------------------------------------------------------}

{Using this object is necessary because TOBject's constructor is not virtual}

constructor TAutoBase.Create;
begin
inherited Create;
end;


{-----------------------------------------------------}

procedure TAutoHash.CreateClassInstance(Index: integer);
begin
if AutoClass<>nil then Strings.Pointers[Index]:=AutoClass.Create
else Strings.Pointers[Index]:=ObjectClass.Create
end;


function TAutoHash.AddUnique(S: ShortString): integer;
begin
Result:=inherited AddUnique(S);
if Result>=0 then CreateClassInstance(Result);
end;


function TAutoHash.Add(const S: string): Integer;
begin
Result:=inherited Add(S);
CreateClassInstance(Result);
end;



procedure TAutoHash.Clear;
var I: integer;
begin
for I:=0 to Count-1 do TObject(Strings.Pointers[I]).Free;
inherited;
end;



constructor TAutoHash.Create(ObjClass: TAutoClass);
begin
inherited Create;
ObjectClass:=nil;
AutoClass:=ObjClass;
end;


constructor TAutoHash.Create(ObjClass: TClass);
begin
inherited Create;
AutoClass:=nil;
ObjectClass:=ObjClass;
end;

procedure TAutoHash.Delete(Index: Integer);
begin
TObject(Strings.Pointers[Index]).Free;
inherited;
end;



{------------------------------------------------------------------------------}
{                  Implementation for long hash table object                   }
{------------------------------------------------------------------------------}



function TLongHashTable.GetHashValue(S: string): cardinal;
{ Return Hash For String S }
{ Hashing algorith based on the FNV-1as (Fowler / Noll / Vo)}
var P: Integer;
begin
if not CaseSensitive then S:=LowerCase(S);
Result:=offset_basis;
for P:=1 to Length(S) do
	begin
	Result:=Result xor byte(S[P]);
	Result:= Result * FNV_Prime;
	end;
Result:=((Result shr FBoxPower) xor Result) and  FHashMask;
end;




procedure TLongHashTable.Clear;
{Clear all the hash boxes and reinitialize}
begin
inherited Clear;
Strings.Clear;
end;



procedure  TLongHashTable.Rehash;
{ Re-do The Hash Links For The Table }
var P : Longint;
begin
ClearHashBox;
HashLink.ItemIndex:=Strings.ItemIndex;
for P:=0 to Strings.ItemIndex-1 do
	begin
	HashValue:=GetHashValue(Strings[P]);
	HashLink[P]:=HashBox[HashValue];
	HashBox[HashValue]:=P;
	end;
end;


function TLongHashTable.Search(S: string): Longint;
{ Search For String S, return index of string }
{ Returns -1 if not found }
begin
HashValue:=GetHashValue(S);
Result:=HashBox[HashValue];
while True do
	begin
	if Result=HashEmpty then break;
	if Strings.Compare(S,Result) then break;
	Result:=HashLink[Result];
	end;
end;


procedure TLongHashTable.SetCaseSensitive(V: boolean);
{Used to set the Case Sensitivity property}
begin
if V<>FCaseSensitive then
	begin
	FCaseSensitive:=V;
	Strings.CaseSensitive:=V;
	Rehash;
	end;
end;




constructor TLongHashTable.Create(StartAlloc: LongInt);
{Creation routine for hash table object }
begin
inherited Create(StartAlloc);
Strings:=THugeLongString.Create(StartAlloc);
CaseSensitive:=True;
end;



constructor TLongHashTable.Create;
{Creation routine for hash table object }
begin
Create(5);
end;


destructor TLongHashTable.Destroy;
{Destroy hash table object }
begin
Strings.Free;
inherited Destroy;
end;


procedure TLongHashTable.Assign(Hash: TLongHashTable);
begin
inherited Assign(Hash);
Strings.Assign(Hash.Strings);
end;



function  TLongHashTable.GetCount: integer;
begin
Result:=Strings.ItemIndex;
end;


function TLongHashTable.ReadString(I: integer): string;
{Property function so hash table can use array-type syntax}
begin
Result:=Strings[I];
end;


procedure TLongHashTable.WriteString(I: integer; V: string);
{Property function so hash table can use array-type syntax}
begin
if Search(V)<>-1 then exit;
RemoveHashKey(I,GetHashValue(Self[I]));
Strings[I]:=V;
RehashItem(I,GetHashValue(Self[I]));
end;




function TLongHashTable.InsertPointer(S: string; Ptr: integer): integer;
{ Insert string and pointer Into Hash Table}
{ Pointer can point to any object the user }
{wishes to associate with a specific string }
begin
HashValue:=GetHashValue(S);
Result:=Strings.AddPointer(S,Ptr);
HashLink.Add(HashBox[HashValue]);
HashBox[HashValue]:=HashLink.ItemIndex-1;
end;



function TLongHashTable.Insert(S: string): integer;
{ Insert string Into Hash Table }
begin
Result:=InsertPointer(S,0);
end;



function TLongHashTable.Add(S: string): integer;
{ Insert string Into Hash Table }
begin
Result:=InsertPointer(S,0);
end;


procedure TLongHashTable.Delete(Index: integer);
{ Delete string and pointer from Hash Table }
begin
Strings.Delete(Index);
Rehash;
end;



function TLongHashTable.AddUnique(S: string): integer;
{ Insert string Into Hash Table if item is unique}
{Returns -1 if already in list else returns index of new item}
begin
if Search(S)=-1 then Result:=Insert(S)
else Result:=-1;;
end;



procedure TLongHashTable.CustomSort(Compare: THugeLongStringSortCompare);
begin
Strings.CustomSort(Compare);
Rehash;
end;



procedure TLongHashTable.Sort;
begin
Strings.Sort;
Rehash;
end;


function TLongHashTable.GetOnChange: TNotifyEvent;
begin
Result:=Strings.OnChange;
end;



procedure TLongHashTable.SetOnChange(const Value: TNotifyEvent);
begin
Strings.OnChange:=Value;
end;


{------------------------------------------------------------------------------}
{                  Implementation for integer hash table object                }
{------------------------------------------------------------------------------}




function TIntegerHashTable.GetHashValue(C: cardinal): cardinal;
{ Return Hash For String S }
{ Hashing algorith based on the FNV-1as (Fowler / Noll / Vo)}
var P: Integer;
var Bytes: PByteArray;
begin
Bytes:=@C;
Result:=offset_basis;
for P:=0 to SizeOf(C)-1 do
	begin
	Result:=Result xor Bytes^[P];
	Result:= Result * FNV_Prime;
	end;
Result:=((Result shr FBoxPower) xor Result) and  FHashMask;
end;



procedure TIntegerHashTable.Clear;
{Clear all the hash boxes and reinitialize}
begin
Items.Clear;
HashLink.Clear;
SetupHashBoxes;
end;



procedure  TIntegerHashTable.Rehash;
{ Re-do The Hash Links For The Table }
var P : Integer;
begin
ClearHashBox;
HashLink.ItemIndex:=Items.ItemIndex;
for P:=0 to Items.ItemIndex-1 do
	begin
	HashValue:=GetHashValue(Items[P]);
	HashLink[P]:=HashBox[HashValue];
	HashBox[HashValue]:=P;
	end;
end;


function TIntegerHashTable.Search(I: integer): Integer;
{ Search For String S, return index of string }
{ Returns -1 if not found }
begin
HashValue:=GetHashValue(I);
Result:=HashBox[HashValue];
while True do
	begin
	if Result=HashEmpty then break;
	if I=Items[Result] then break;
	Result:=HashLink[Result];
	end;
end;





constructor TIntegerHashTable.Create(StartAlloc: Integer);
{Creation routine for integer hash table object }
begin
inherited Create(StartAlloc);
Items:=THugeInteger.Create(StartAlloc);
end;


destructor TIntegerHashTable.Destroy;
{Destroy hash table object }
begin
Items.Free;
inherited Destroy;
end;



procedure TIntegerHashTable.Assign(Hash: TIntegerHashTable);
begin
inherited Assign(Hash);
Items.Assign(Hash.Items);
end;


function  TIntegerHashTable.GetCount: integer;
begin
Result:=Items.ItemIndex;
end;


function TIntegerHashTable.ReadInteger(I: integer): integer;
{Property function so hash table can use array-type syntax}
begin
Result:=Items[I];
end;


procedure TIntegerHashTable.WriteInteger(I: integer; V: integer);
{Property function so hash table can use array-type syntax}
begin
if Search(V)<>-1 then exit;
RemoveHashKey(I,GetHashValue(Self[I]));
Items[I]:=V;
RehashItem(I,GetHashValue(Self[I]));
end;




function TIntegerHashTable.InsertPointer(I, Ptr: integer): integer;
{ Insert integer and pointer Into Hash Table}
{ Pointer can point to any object the user }
{wishes to associate with a specific Integer }
begin
HashValue:=GetHashValue(I);
Result:=Items.AddPointer(I,Ptr);
HashLink.Add(HashBox[HashValue]);
HashBox[HashValue]:=HashLink.ItemIndex-1;
end;



function TIntegerHashTable.Insert(I: integer): integer;
{ Insert integer Into Hash Table }
begin
Result:=InsertPointer(I,0);
end;



function TIntegerHashTable.Add(I: integer): integer;
{ Insert integer Into Hash Table }
begin
Result:=InsertPointer(I,0);
end;



procedure TIntegerHashTable.Delete(Index: integer);
{ Delete integer and pointer from Hash Table }
begin
Items.Delete(Index);
Rehash;
end;


function TIntegerHashTable.AddUnique(I: integer): integer;
{ Insert integer Into Hash Table if item is unique}
{Returns -1 if already in list else returns index of new item}
begin
if Search(I)=-1 then Result:=Insert(I)
else Result:=-1;
end;



procedure TIntegerHashTable.CustomSort(Compare: TListSortCompare);
begin
Items.Sort(Compare);
Rehash;
end;



procedure TIntegerHashTable.Sort;
begin
Items.Sort;
Rehash;
end;



function TIntegerHashTable.GetOnChange: TNotifyEvent;
begin
Result:=Items.OnChange;
end;



procedure TIntegerHashTable.SetOnChange(const Value: TNotifyEvent);
begin
Items.OnChange:=Value;
end;


{-------------------------------------------------------------------------}
{           Implementation for Double hash table object                   }
{-------------------------------------------------------------------------}


function TDoubleHashTable.GetHashValue(D: double): cardinal;
{ Return Hash For double D }
{ Hashing algorith based on the FNV-1as (Fowler / Noll / Vo)}
var P : Integer;
var PB: PByteArray;
begin
Result:=offset_basis;
PB:=@D;
for P:=0 to Sizeof(D)-1 do
	begin
	Result:=Result xor PB[P];
	Result:= Result * FNV_Prime;
	end;
Result:=((Result shr FBoxPower) xor Result) and  FHashMask;
end;



procedure TDoubleHashTable.Clear;
{Clear all the hash boxes and reinitialize}
begin
inherited Clear;
Items.Clear;
end;



procedure  TDoubleHashTable.Rehash;
{ Re-do The Hash Links For The Table }
var P : Longint;
begin
ClearHashBox;
HashLink.ItemIndex:=Items.ItemIndex;
for P:=0 to Items.ItemIndex-1 do
	begin
	HashValue:=GetHashValue(Items[P]);
	HashLink[P]:=HashBox[HashValue];
	HashBox[HashValue]:=P;
	end;
end;


function TDoubleHashTable.Search(D: double): Longint;
{ Search For double D, return index of double }
{ Returns -1 if not found }
begin
HashValue:=GetHashValue(D);
Result:=HashBox[HashValue];
while True do
	begin
	if Result=HashEmpty then break;
	if D=Items[Result] then break;
	Result:=HashLink[Result];
	end;
end;



constructor TDoubleHashTable.Create(StartAlloc: LongInt);
{Creation routine for hash table object }
begin
inherited Create(StartAlloc);
Items:=THugeDouble.Create(StartAlloc);
end;




destructor TDoubleHashTable.Destroy;
{Destroy hash table object }
begin
Items.Free;
inherited Destroy;
end;



procedure TDoubleHashTable.Assign(Hash: TDoubleHashTable);
begin
inherited Assign(Hash);
Items.Assign(Hash.Items);
end;


function  TDoubleHashTable.GetCount: integer;
begin
Result:=Items.ItemIndex;
end;


function TDoubleHashTable.ReadDouble(I: integer): double;
{Property function so hash table can use array-type syntax}
begin
Result:=Items[I];
end;


procedure TDoubleHashTable.WriteDouble(I: integer; D: double);
{Property function so hash table can use array-type syntax}
begin
if Search(D)<>-1 then exit;
RemoveHashKey(I,GetHashValue(Self[I]));
Items[I]:=D;
RehashItem(I,GetHashValue(Self[I]));
end;




function TDoubleHashTable.InsertPointer(D: double; Ptr: integer): integer;
{ Insert Double and pointer Into Hash Table}
{ Pointer can point to any object the user }
{wishes to associate with a specific double }
begin
HashValue:=GetHashValue(D);
Result:=Items.AddPointer(D,Ptr);
HashLink.Add(HashBox[HashValue]);
HashBox[HashValue]:=HashLink.ItemIndex-1;
end;



function TDoubleHashTable.Insert(D: double): integer;
{ Insert Double Into Hash Table }
begin
Result:=InsertPointer(D,0);
end;




function TDoubleHashTable.Add(D: double): integer;
{ Insert Double Into Hash Table }
begin
Result:=InsertPointer(D,0);
end;


procedure TDoubleHashTable.Delete(Index: integer);
{ Delete Double and pointer from Hash Table }
begin
Items.Delete(Index);
Rehash;
end;




function TDoubleHashTable.AddUnique(D: double): integer;
{ Insert Double Into Hash Table if item is unique}
{Returns -1 if already in list else returns index of new item}
begin
if Search(D)=-1 then Result:=Insert(D)
else Result:=-1;
end;



procedure TDoubleHashTable.CustomSort(Compare: TListSortCompare);
begin
Items.Sort(Compare);
Rehash;
end;



procedure TDoubleHashTable.Sort;
begin
Items.Sort;
Rehash;
end;


function TDoubleHashTable.GetOnChange: TNotifyEvent;
begin
Result:=Items.OnChange;
end;



procedure TDoubleHashTable.SetOnChange(const Value: TNotifyEvent);
begin
Items.OnChange:=Value;
end;


{-------------------------------------------------------------------------}
{         Implementation for T3DVectorHash                                }
{-------------------------------------------------------------------------}

function T3DVectorHash.Read3DVector(I: integer): T3DVector;
{Property function so hash table can use array-type syntax}
begin
Result:=Items[I];
end;



procedure T3DVectorHash.Write3DVector(I: integer; V: T3DVector);
{Property function so hash table can use array-type syntax}
begin
if Search(V)<>-1 then exit;
RemoveHashKey(I,GetHashValue(Self[I]));
Items[I]:=V;
RehashItem(I,GetHashValue(Self[I]));
end;


function T3DVectorHash.Add(D: T3DVector): integer;
begin
Result:=InsertPointer(D,0);
end;



function T3DVectorHash.AddUnique(D: T3DVector): integer;
{ Insert T3DVector Into Hash Table if item is unique}
{Returns -1 if already in list else returns index of new item}
begin
if Search(D)=-1 then Result:=Insert(D)
else Result:=-1;
end;



procedure T3DVectorHash.Assign(Hash: T3DVectorHash);
begin
inherited Assign(Hash);
Items.Assign(Hash.Items);
end;



procedure T3DVectorHash.Append(Hash: T3DVectorHash);
begin
inherited Append(Hash);
Items.Append(Hash.Items);
end;



procedure T3DVectorHash.Clear;
begin
inherited Clear;
Items.Clear;
end;



constructor T3DVectorHash.Create(StartAlloc: Integer);
begin
inherited Create(StartAlloc);
Items:=THuge3DVector.Create(StartAlloc);
end;



procedure T3DVectorHash.CustomSort(Compare: TListSortCompare);
begin
Items.Sort(Compare);
Rehash;
end;



procedure T3DVectorHash.Delete(Index: integer);
begin
Items.Delete(Index);
Rehash;
end;



destructor T3DVectorHash.Destroy;
begin
Items.Free;
inherited Destroy;
end;



function T3DVectorHash.GetCount: integer;
begin
Result:=Items.ItemIndex;
end;




procedure T3DVectorHash.SetCount(const Value: integer);
begin
Items.ItemIndex:=Value;
Items.ExpandArray;
Rehash;
end;


function T3DVectorHash.GetHashValue(V: T3DVector): cardinal;
{ Return Hash For T3DVector V }
{ Hashing algorith based on the FNV-1as (Fowler / Noll / Vo)}
var P : Integer;
var PB: PByteArray;
begin
Result:=offset_basis;
PB:=@V;
for P:=0 to Sizeof(V)-1 do
	begin
	Result:=Result xor PB[P];
	Result:= Result * FNV_Prime;
	end;
Result:=((Result shr FBoxPower) xor Result) and  FHashMask;
end;



function T3DVectorHash.Insert(D: T3DVector): integer;
begin
Result:=InsertPointer(D,0);
end;



function T3DVectorHash.InsertPointer(D: T3DVector; Ptr: integer): integer;
{ Insert Vector and pointer Into Hash Table}
{ Pointer can point to any object the user }
{wishes to associate with a specific double }
begin
HashValue:=GetHashValue(D);
Result:=Items.AddPointer(D,Ptr);
HashLink.Add(HashBox[HashValue]);
HashBox[HashValue]:=HashLink.ItemIndex-1;
end;




procedure T3DVectorHash.Rehash;
{ Re-do The Hash Links For The Table }
var P : Longint;
begin
ClearHashBox;
HashLink.ItemIndex:=Items.ItemIndex;
for P:=0 to Items.ItemIndex-1 do
	begin
	HashValue:=GetHashValue(Items[P]);
	HashLink[P]:=HashBox[HashValue];
	HashBox[HashValue]:=P;
	end;
end;




function AreSinglesEqual(const V1,V2: T3DVector): boolean;
{Return True if vectors are equal}
var X1,Y1,Z1,X2,Y2,Z2: single;
begin
X1:=V1.X; Y1:=V1.Y; Z1:=V1.Z;
X2:=V2.X; Y2:=V2.Y; Z2:=V2.Z;
Result:=(X1=X2) and (Y1=Y2) and (Z1=Z2);
end;



function T3DVectorHash.Search(D: T3DVector): integer;
{ Search For String S, return index of TVertex }
{ Returns -1 if not found }
begin
HashValue:=GetHashValue(D);
Result:=HashBox[HashValue];
while True do
	begin
	if Result=HashEmpty then break;
	if AreSinglesEqual(D,Items[Result]) then break;
	Result:=HashLink[Result];
	end;
end;


procedure T3DVectorHash.Sort;
begin
Items.Sort;
Rehash;
end;


function T3DVectorHash.GetOnChange: TNotifyEvent;
begin
Result:=Items.OnChange;

end;

procedure T3DVectorHash.SetOnChange(const Value: TNotifyEvent);
begin
Items.OnChange:=Value;
end;


end.
