unit IconLib;

{ This object matches the format of the internal icon-file data
 blocks within an .ANI file. It's also a fair match for the .ICO
 file format, as long as the icon has just a single 32x32 image
 in 16 or 256 colors. }

interface

uses Windows, SysUtils, Classes, Graphics, CursorPalette;

type TICONDIR = record
 idReserved: Word;
 idType: Word;
 idCount: Word;
 end;

type TIconDirectoryEntry = packed record
 bWidth: Byte;
 bHeight: Byte;
 bColorCount: Byte;
 bReserved: Byte;
 xHotSpot: Word;
 yHotSpot: Word;
 dwBytesInRes: LongInt;
 dwImageOffset: LongInt;
 end;

type TColBitType = array[0..1023] of Byte;
type  TMaskBitType = array[0..127] of Byte;

type TInternalCur = class(TObject)
 private
  fValid: Boolean;
  fExcess: Boolean;
  fID: TIconDir;
  fIDE: TIconDirectoryEntry;
  fBIH: TBitmapInfoheader;
  fPal: TPalArray;
  fcolbits: TColBitType;
  fmaskbits: TMaskBitType;
  fIcon: TIcon;
  fDataSize: Integer;
  fTag: Integer;
  fBlackIdx: Integer; { used for transparent}
  fWhiteIdx: Integer; { used for inverted}
  function GetAsIcon: TIcon;
  function GetIs16Color: Boolean;
  function GetColBits(aCol,aRow: Integer): Integer;
  procedure SetColBits(aCol,aRow: Integer; const Value: Integer);
  function GetMaskBits(aCol,aRow: Integer): Integer;
  procedure SetMaskBits(aCol,aRow: Integer; const Value: Integer);
  function GetBlackIdx: Integer;
  function GetWhiteIdx: Integer;
 public
  constructor Create; overload;
  constructor Create(Stream: TStream); overload;
  constructor Create(orig: TInternalCur); overload;
  constructor Create(orig: TInternalCur; vtag: Integer); overload;
  constructor Create(is16color: Boolean; xHot, yHot: Word; const vpal: TPalArray); overload;
  destructor Destroy; override;
  procedure LoadFromStream(Stream: TStream);
  procedure SaveToStream(Stream: TStream);
  procedure LoadFromFile(const fileName: String);
  procedure SaveToFile(const fileName: String);
  function Same(Oth: TInternalCur): Boolean;
  property DataSize: Integer read fDataSize;
  property AsIcon: TIcon read GetAsIcon;
  property Valid: Boolean Read fValid;
  property Palette: TPalArray read fPal;
  property Is16color: Boolean read GetIs16Color;
  property Excess: Boolean read fExcess;
  property BlackIdx: Integer read GetBlackIdx;
  property WhiteIdx: Integer read GetWhiteIdx;
  property Tag: Integer read fTag write fTag;
  property HotX: Word read fIDE.xHotspot write fIDE.xHotspot;
  property HotY: Word read fIDE.yHotspot write fIDE.yHotspot;
  property ColBits[aCol, aRow: Integer]: Integer read GetColBits write SetColBits;
  property MaskBits[aCol, aRow: Integer]: Integer read GetMaskBits write SetMaskBits;
  end;

implementation

{ TInternalCur }

constructor TInternalCur.Create;
{ Just initialize; called by other constructors}
begin
Inherited Create;
fValid:= False;
fExcess:= False;
fIcon:= nil;
fBlackIdx:= -1;
fWhiteIdx:= -1;
Tag:= 0;
FillChar(fID, SizeOf(fID), 0);
FillChar(fIDE, SizeOf(fIDE), 0);
FillChar(fBIH, SizeOf(fBIH), 0);
FillChar(fPal, SizeOf(fPal), 0);
FillChar(fcolbits, SizeOf(fcolbits), 0);
FillChar(fmaskbits, SizeOf(fmaskbits), 0);
end;


constructor TInternalCur.Create(Stream: TStream);
{ Create and load from stream in one operation}
begin
Create;
LoadFromStream(Stream);
fBlackIdx:= -1;
fWhiteIdx:= -1;
end;


constructor TInternalCur.Create(orig: TInternalCur);
{ Create clone of existing object}
begin
Create;
fValid:= orig.fValid;
fExcess:= orig.fExcess;
fID:= orig.fID;
fIDE:= orig.fIDE;
fBIH:= orig.fBIH;
fPal:= orig.fPal;
fcolbits:= orig.fcolbits;
fmaskbits:= orig.fmaskbits;
fIcon:= nil;;
fDataSize:= orig.fDataSize;
fTag:= orig.fTag;
end;


constructor TInternalCur.Create(orig: TInternalCur; vtag: Integer);
{ Clone and set tag}
begin
Create(orig);
Tag:= vTag;
end;


constructor TInternalCur.Create(is16color: Boolean; xHot, yHot: Word; const vpal: TPalArray);
{ Create a fully functional blank frame}
begin
Create;
fID.idType:= 2;
fID.idCount:= 1;
fIDE.bWidth:= 32;
fIDE.bHeight:= 32;
fIDE.xHotSpot:= xHot;
fIDE.yHotSpot:= yHot;
fDataSize:= SizeOf(fID) + SizeOf(fIDE) + SizeOf(fBIH) + SizeOf(fMaskBits);
if is16Color then
	begin
	fIDE.dwBytesInRes:= 744;
	fBIH.biClrUsed:= 0;
	fBIH.biBitCount:= 4;
	Inc(fDataSize, 16*SizeOf(TColorRef));
	Inc(fDataSize, 512);
	end
else
	begin
	fIDE.dwBytesInRes:= 2216;
	fBIH.biClrUsed:= 256;
	fBIH.biBitCount:= 8;
	Inc(fDataSize, 256*SizeOf(TColorRef));
	Inc(fDataSize, 1024);
	end;
fIDE.dwImageOffset:= 22;
fBIH.biSize:= SizeOf(fBIH);
fBIH.biWidth:= 32;
fBIH.biHeight:= 64;
fBIH.biPlanes:= 1;
fBIH.biSizeImage:= 128+(1024*fBIH.biBitCount div 8);
MOVE(vPal, fPal, SizeOf(fPal));
fValid:= True;
end;



destructor TInternalCur.Destroy;
begin
if Assigned(fIcon) then fIcon.Free;
Inherited Destroy;
end;


procedure TInternalCur.LoadFromStream(Stream: TStream);
var N: Integer;
var RQ: TRGBQuad;
var P: Integer;
begin
P:= Stream.Position;
try
Stream.ReadBuffer(fID, SizeOf(fID));
if fID.idCount > 1 then
	begin
	fID.idCount:= 1;
	fExcess:= True;
	end;
Stream.ReadBuffer(fIDE, SizeOf(fIDE));
if fIDE.dwImageOffset <> 22 then
	begin
	Stream.Seek(P+fIDE.dwImageOffset, soFromBeginning);
	fIDE.dwImageOffset:= 22;
	end;
Stream.ReadBuffer(fBIH, SizeOf(fBIH));
if (fBIH.biBitCount <> 4) and (fBIH.biBitCount <> 8) then raise(Exception.Create('Can only load 16-color or '+'256-color cursors.'));
for N:= 0 to (1 shl fBIH.biBitCount)-1 do
	begin
	Stream.ReadBuffer(RQ, SizeOf(RQ));
	with RQ do fPal[N]:= RGB(RGBRed, RGBGreen, RGBBlue);
	end;
Stream.ReadBuffer(fcolbits, 32*32*fBIH.biBitCount div 8);
Stream.ReadBuffer(fmaskbits, 32*32 div 8);
fDataSize:= Stream.Position - P;
if fID.idReserved <> 0 then Exit;
if fIDE.bWidth <> 32 then Exit;
if fIDE.bHeight <> 32 then Exit;
fValid:= True;
except on EReadError do; end;
end;



procedure TInternalCur.SaveToStream(Stream: TStream);
var N: Integer;
var RQ: TRGBQuad;
begin
if not fValid then Raise(Exception.Create('No icon data loaded'));
try
Stream.WriteBuffer(fID, SizeOf(fID));
Stream.WriteBuffer(fIDE, SizeOf(fIDE));
Stream.WriteBuffer(fBIH, SizeOf(fBIH));
for N:= 0 to (1 shl fBIH.biBitCount)-1 do
	begin
	RQ.rgbRed:= GetRValue(fPal[N]);
	RQ.rgbGreen:= GetGValue(fPal[N]);
	RQ.rgbBlue:= GetBValue(fPal[N]);
	RQ.rgbReserved:= 0;
	Stream.WriteBuffer(RQ, SizeOf(RQ));
	end;
Stream.WriteBuffer(fcolbits, 32*32*fBIH.biBitCount div 8);
Stream.WriteBuffer(fmaskbits, 32*32 div 8);
except on EWriteError do; end;
end;


procedure TInternalCur.LoadFromFile(const fileName: String);
var Stream: TStream;
begin
Stream:= TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
try
LoadFromStream(Stream);
finally Stream.Free; end;
end;


procedure TInternalCur.SaveToFile(const fileName: String);
var Stream: TStream;
begin
Stream:= TFileStream.Create(FileName, fmCreate);
try
SaveToStream(Stream);
finally Stream.Free; end;
end;


function TInternalCur.Same(Oth: TInternalCur): Boolean;
var TM1, TM2: TMemoryStream;
begin
TM1:= TMemoryStream.Create;
TM2:= TMemoryStream.Create;
try
SaveToStream(TM1);
Oth.SaveToStream(TM2);
Result:= CompareMem(TM1.Memory, TM2.Memory, TM1.Size);
finally
 TM1.Free;
 TM2.Free;
 end;
end;


function TInternalCur.GetAsIcon: TIcon;
var TMS: TMemoryStream;
var Was: Integer;
begin
if not Assigned(fIcon) then
	begin
	fIcon:= TIcon.Create;
	TMS:= TMemoryStream.Create;
	try
	Was:= fID.idType;
	fID.idType:= 1; { set it to icon type}
	SaveToStream(TMS);
	fID.idType:= Was;
	TMS.Seek(0, soFromBeginning);
	fIcon.LoadFromStream(TMS);
	finally TMS.Free; end;
	end;
Result:= fIcon;
end;



function TInternalCur.GetIs16Color: Boolean;
begin
if not fValid then Raise(Exception.Create('No icon data loaded'));
Result:= fBIH.biBitCount = 4;
end;


function TInternalCur.GetColBits(aCol, aRow: Integer): Integer;
begin
if not fValid then Raise(Exception.Create('No icon data loaded'));
if (aCol < 0) or (aRow < 0) or (aCol > 31) or (aRow > 31) then Raise(Exception.Create('Pixel index out of range'));
if fBIH.biBitCount = 4 then
	begin
	if Odd(aCol) then Result:= fColBits[aRow*16+(aCol div 2)] and $F
	else Result:= fColBits[aRow*16+(aCol div 2)] shr 4;
	end
else if fBIH.biBitCount = 8 then Result:= fColBits[aRow*32+aCol]
else Result:= 0;
end;



procedure TInternalCur.SetColBits(aCol, aRow: Integer; const Value: Integer);
begin
if (aCol < 0) or (aRow < 0) or (aCol > 31) or (aRow > 31) then Raise(Exception.Create('Pixel index out of range'));
if fBIH.biBitCount = 4 then
	begin
	if Odd(aCol) then fColBits[aRow*16+(aCol div 2)]:= (fColBits[aRow*16+(aCol div 2)] and $F0) or (Value and $F)
	else fColBits[aRow*16+(aCol div 2)]:= (fColBits[aRow*16+(aCol div 2)] and $F) or ((Value and $F) SHL 4);
	end
else if fBIH.biBitCount = 8 then fColBits[aRow*32+aCol]:= Value;
end;



function TInternalCur.GetMaskBits(aCol, aRow: Integer): Integer;
var idx, bit: Integer;
begin
if not fValid then Raise(Exception.Create('No icon data loaded'));
if (aCol < 0) or (aRow < 0) or (aCol > 31) or (aRow > 31) then Raise(Exception.Create('Pixel index out of range'));
idx:= (aRow*4) + (aCol div 8);
bit:= 1 SHL (7-(aCol mod 8));
if fmaskBits[idx] and bit > 0 then Result:= 1
else Result:= 0;
end;



procedure TInternalCur.SetMaskBits(aCol, aRow: Integer; const Value: Integer);
var idx, bit: Integer;
begin
if (aCol < 0) or (aRow < 0) or (aCol > 31) or (aRow > 31) then Raise(Exception.Create('Pixel index out of range'));
idx:= (aRow*4) + (aCol div 8);
bit:= 1 SHL (7-(aCol mod 8));
if value = 1 then fMaskBits[idx]:= fMaskBits[idx] or bit
else fMaskBits[idx]:= fMaskBits[idx] and (not bit);
end;



function TInternalCur.GetBlackIdx: Integer;
var N: Integer;
begin
if fBlackIdx = -1 then
	begin
	for N:= 0 to (1 SHL fBIH.biBitCount) - 1 do
	 if fPal[N] and $FFFFFF = $000000 then
	 	begin
	 	fBlackIdx:= N;
	 	Break;
	 	end;
	end;
if fBlackIdx = -1 then fBlackIdx:= 0;
Result:= fBlackIdx;
end;



function TInternalCur.GetWhiteIdx: Integer;
var N: Integer;
begin
if fWhiteIdx = -1 then
	begin
	for N:= 0 to (1 SHL fBIH.biBitCount) - 1 do
	 if fPal[N] and $FFFFFF = $FFFFFF then
	 	begin
	 	fWhiteIdx:= N;
	 	Break;
	 	end;
	end;
if fWhiteIdx = -1 then fWhiteIdx:= (1 SHL fBIH.biBitCount) - 1;
Result:= fWhiteIdx;
end;



end.
