unit FFMPEGObject;

interface

uses
 Forms, Windows, ComCtrls, MMSystem, Messages, Classes, SysUtils, Controls, Graphics, ExtCtrls,
 Math, ThreadLib, ShlObj, ffutils,libavformat, libavutil, libavcodec, libswscale,
 libswresample,  libavdevice, sdl2, sdl2_ttf, NetLib, DSupport;

{ How it works:

1. Thread-one. "ReadThread" requests packet, decodes packets into frames and
stores them in the Audio or Video queue.

2. Thread-two. SDL calls "SDLAudioCallback" and requests an audio packet, which
it gets from the Audio Queue.

3. Thread-three. A 66 hz SDL timer event invalidates the window and triggers
a Paint event. Paint takes a frame out of the video queue and draws it.

4. Synchronization between audio and video is handled by comparing the audio
and video presentation times. If video is behind audio, let video run at 66 hz
to catch up. If audio is behind video, hold video until audio catches up.


Frame Information:

I-Frame = Key-Frame = complete image.
P-Frame = Delta-Frame = only changes from previous frame
B-Frame = uses both previous and next frame

DTS = Decoding Time Stamp
PTS = Presentation Time Stamp.

Data can be transmitted out of sync so PTS is used
to sort it out and put it in the right order.

}



var LogList: TThreadStringList;

const LM_USER = $400;

const
 PLM_PAINTRESIZE    = LM_USER;
 PLM_UPDATEPROGRESS = LM_USER+2;
 PLM_EOF            = LM_USER+1;
 PLM_TIMER          = LM_USER+3;


type
 TCustomFFMPegPlayer=class;

 TPlayerQueueRec = record
  FValid: boolean;
  FFrame: PAVFrame;
  end;

 { TPlayerQueue }

 TPlayerQueue=class
  private
   FPlayer: TCustomFFMPegPlayer;
   FPutEvent, FGetEvent: THandle;
   FPutIndex,
   FGetIndex: integer;
   FQueue: array of TPlayerQueueRec;
   FUnblockGet: boolean;
   FUnblockPut: boolean;
   procedure IncIndex(var Inx: integer);
  protected
   function GetCount: integer;
  public
   Name: string;
   constructor Create(player: TCustomFFMPegPlayer; Maxframes: integer); virtual;
   destructor Destroy; override;
   function Put(fr: PAVFrame; wait: boolean=true): boolean;
   function Get(var fr: PAVFrame; wait: boolean=true): boolean;
   function Peek(var fr:PAVFrame): boolean;
   procedure Del;
   procedure WakeUpPut;
   procedure WakeUpGet;
   procedure Clear;
   property UnblockPut: boolean read FUnblockPut write FUnblockPut;
   property UnblockGet: boolean read FUnblockGet write FUnblockGet;
   property Count: integer read GetCount;
 end;


{ TCustomFFMPegPlayer }

TCustomFFMPegPlayer=class(TCustomControl)
 private
  FTimeDisplayHeight: double;
  FTimerId: TSDL_TimerID;
  FBadFrameTolerance: double;
  FSeeking: boolean;
  FDisableAudio: boolean;
  FDisableVideo: boolean;
  FLastErrorString: string;
  FPausing,
  FPlaying: boolean;
  FMute: boolean;
  FShowTimes: boolean;
  FTimerInterval: integer;
  FUrl: string;
  FWindow: PSDL_Window;
  FRenderer: PSDL_Renderer;
  FTexture: PSDL_Texture;
  FTextureValid: boolean;
  FSDLAudioBufferSize: word;
  FSDLPixelFormat: Cardinal;
  FAVPixelFormat: TAVPixelFormat;
  FFont: PTTF_Font;
  FReadThreadId: TThreadId;
  FFormatCtx: PAVFormatContext;
  FVideoStreamIndex: integer;
  FVideoTimeBase: double;
  FVideoCodec: PAVCodec;
  FVideoCtx: PAVCodecContext;
  FVideoSwsCtx: PSwsContext;
  FVideoDecodeFrame: PAVFrame;
  FVideoDecodeFrameBuffer: PByte;
  FVideoQueue: TPlayerQueue;
  FVideoSynchPts,              { Synchronize-frame-pts}
  FVideoSynchTime,             { Synchronize-real-time}
  FVideoPts: int64;             { pts of the last videoframe shown}
  FVideoWidth,
  FVideoHeight: integer;
  FVideoShowNextFrame: boolean;		{ will be set to show the first frame if the state is paused}
  FFVideoAspect: TAVRational;		{ aspect-ratio of the last frame shown }

  { Audio stream variables}
  FAudioStreamIndex: integer;
  FAudioTimeBase: double;
  FAudioCodec: PAVCodec;
  FAudioCtx: PAVCodecContext;
  FAudioSwrCtx: PSwrContext;
  FAudioBuffer: PByte;
  FAudioBufSizeMax,
  FAudioBufSize,
  FAudioBufIndex: integer;
  FAudioQueue: TPlayerQueue;
  FAudioPts: int64;
  MemBuffer: PByte;
  FMPEGStream: TStream;  {User provided Stream}

  PutThreadToSleep: boolean;	{Command to Put thread to sleep}
  FReadThreadSleeping: boolean; {Flags that thread has received sleep command}
  FEof: boolean;               { Flags playback has reach the End Of File}
  FLastTimeString: string;
  FDelayAudioQueueing: boolean;
  FMaxAudioFrames,
  FMaxVideoFrames: integer;
  FOnPlay: TNotifyEvent;
  function GetAudioAvailable: boolean;
  function GetDuration: int64;
  function GetPosition: int64;
  function GetVideoAvailable: boolean;
  function GetVideoTimerRate: integer;
  procedure SetMaxAudioFrames(AValue: integer);
  procedure SetMaxVideoFrames(AValue: integer);
  procedure SetVideoTimerRate(AValue: integer);
  procedure WMTimer(var m: TMessage); message PLM_TIMER;
  procedure WMPaintResize(var m: TMessage); message PLM_PAINTRESIZE;
  procedure WMEof(var m: TMessage); message PLM_EOF;
  procedure WMUpdateProgress(var m: TMessage); message PLM_UPDATEPROGRESS;
  procedure InitalizePlayer;
  procedure FinalizePlayer;
  procedure WaitQueuesEmpty;
  function GetAudioQueDepth: integer;
  function GetVideoQueDepth: integer;
  function GetFrame: int64;
  procedure SetErrorStr(S: string; ErrNum: integer);
  function HandleError(IsError: boolean; Msg: string; ErrNum: integer): boolean;
  procedure SaveFileFormat;
  function GetFileFormatStr: string;
  function SetBackgroundColor(Renderer: PSDL_Renderer; Color: TColor): integer;
  procedure GetMemoryContext;
  procedure SetURL(const Value: string);
    procedure SetMPegStream(const Value: TStream);
 protected
  procedure CreateWindowAndRenderer; virtual;
  procedure DestroyWindowAndRenderer; virtual;

  function LoadText(s: string; co: TSDL_Color; var w,h: integer): PSDL_Texture;virtual;
  procedure Resize; override;
  procedure DoOnEof; virtual;
  procedure DoUpdateProgress; virtual;
  procedure StartTimer; virtual;
  procedure StopTimer; virtual;
  function GetRendererFlags: longword; virtual;
 public
  MaxDepth: integer;
  LastContext: TAVFormatContext;
  constructor Create(AOwner: TComponent); override;
  procedure DestroyWnd; override;
  destructor Destroy; override;
  procedure SetPixelFormat(SDLPixelFormat: Cardinal; AVPixelFormat: TAVPixelFormat); virtual;
  property AudioQueDepth: integer read GetAudioQueDepth;
  property VideoQueDepth: integer read GetVideoQueDepth;
  property OnPlay: TNotifyEvent read FOnPlay write FOnPlay;
  procedure PaintOverlays(renderer: PSDL_Renderer; var present: boolean); virtual;
  procedure PaintTimes(renderer: PSDL_Renderer; var present: boolean); virtual;

  procedure PaintImage; virtual;
  procedure Paint; override;

  function Play: boolean; overload; virtual;
  procedure Pause; virtual;
  procedure Resume; virtual;
  procedure Stop(WaitQueues: boolean); virtual;
  procedure Seek(ms: int64); virtual;

  property VideoTimerRate: integer read GetVideoTimerRate write SetVideoTimerRate;
  property LastErrorString: string read FLastErrorString write FLastErrorString;
  property URL: string read FURL write SetURL;
  property Stream: TStream read FMPEGStream write SetMPegStream;
  property Mute: boolean read FMute write FMute;

   {Allows user to disable audio or video before playing}
   property DisableAudio: boolean read FDisableAudio write FDisableAudio;
   property DisableVideo: boolean read FDisableVideo write FDisableVideo;

   {Allows user to change the capcity of the audio and video queues}
   property MaxAudioFrames: integer read FMaxAudioFrames write SetMaxAudioFrames;
   property MaxVideoFrames: integer read FMaxVideoFrames write SetMaxVideoFrames;
   property BadFrameTolerance: double read FBadFrameTolerance write FBadFrameTolerance;

   property Playing: boolean read FPlaying;
   property Pausing: boolean read FPausing;
   property ShowTimes: boolean read FShowTimes write FShowTimes;

   property Duration: int64 read GetDuration;

   property Position: int64 read GetPosition;
   property Frame: int64 read GetFrame;
   property TimeDisplayHeight: double read FTimeDisplayHeight write FTimeDisplayHeight;

   property Renderer: PSDL_Renderer read FRenderer;
   property SDLFont: PTTF_Font read FFont;

   property AudioAvailable: boolean read GetAudioAvailable;
   property VideoAvailable: boolean read GetVideoAvailable;
   property FileFormat: string read GetFileFormatStr;
end;

 { TFFMPegPlayer }

 TFFMPegPlayerEof=procedure (sender: TObject; var handled: boolean) of Object;

 TFFMPegPlayer=class(TCustomFFMPegPlayer)
  private
    FOnEof: TFFMPegPlayerEof;
    FOnProgress: TNotifyEvent;
  protected
    procedure DoOnEof; override;
    procedure DoUpdateProgress; override;
  public
    procedure DisplayFileFormat(Tree: TTreeView);
  published
    property Action;
    property Align;
    property Anchors;
    property BadFrameTolerance;
    property Constraints;
    property DisableAudio;
    property DisableVideo;
    property Duration;
    property Enabled;
    property Font;
    property Hint;
    property MaxAudioFrames;
    property MaxVideoFrames;
    property Mute;
    property ParentBidiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
   property Position;
    property ShowHint;
    property ShowTimes;
    property Url;
    property VideoTimerRate;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnEof: TFFMPegPlayerEof read FOnEof write FOnEof;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnProgress: TNotifyEvent read FOnProgress write FOnProgress;
    property OnResize;
 end;

function GetTimeString(t: int64; ShowMS: boolean=false): string;



implementation

uses MPEGPlayer, FFContextDisplay;

resourcestring
 rsNoStreamFound='There was no video- and audiostream found.';
 rsPlayerActive='The player already started.';
 rsVideoCtxNotAlloc='Video-context could not be allocated.';
 rsVideoSwsCtxNotAlloc='Video-SWS-context could not be allocated.';
 rsVideoDecodeFrameNotAlloc='Video-decode-frame could not be allocated.';
 rsVideoFrameBufferNotAlloc='Video-frame-buffer could not be allocated.';
 rsAudioCtxNotAlloc='Audio-context could not be allocated.';
 rsAudioSwrCtxNotAlloc='Audio-SWR-context could not be allocated.';


{================= Support Routines ===========================================}

var FileFormatStr: TThreadString;	{String containing file format info - it has to be outside the object}

var InitializeError: boolean;	{Flags initialization error}
var InitSFCount: integer;
var InitSFErrorString: string; {Initialization error string}


procedure WriteToLog(S: string);
begin
LogList.Add(S);
end;



function GetTimeString(t: int64; ShowMS: boolean=false): string;
{Convert miliseconds to hours, minutes, seconds, miliseconds}
var ms,dd,hh,mm,ss:integer;
begin
ms:=t mod 1000;
t:=t div 1000;
ss:=t mod 60;
t:=t div 60;
mm:=t mod 60;
t:=t div 60;
hh:=t mod 60;
dd:=t div 24;
if dd<>0 then Result:=format('%d:%.2d:%.2d:%.2d',[dd,hh,mm,ss])
else Result:=format('%.2d:%.2d:%.2d',[hh,mm,ss]);
if ShowMS then Result:=Result+'.'+format('%.3d',[ms]);
end;


{================= Non-Object Routines ========================================}


function ReadThread(p: Pointer): Integer;
{ Read and decode all packets - put them }
{ in the Queues if there is space}
var rc: integer;
var Player: TCustomFFMPegPlayer;
var FirstVideoFrame: boolean;
var Packet: PAVPacket;



procedure DecodePacket(Context: PAVCodecContext; Queue: TPlayerQueue; Packet: PAVPacket; video: boolean);
{Decode one Packet and put all frames in the queue }
var rc: integer;
var TimeStamp: int64;
var PutInQueue: boolean;
var Frame: PAVFrame;
begin
{Send packet to decoder}
rc := avcodec_send_packet(Context, Packet);
if rc < 0 then exit;

while true do
	begin
	if Player.PutThreadToSleep then break;
	{Allocate new frame}
	Frame:=av_frame_alloc;
	if Frame=nil then break;
	{Read a frame from the codec}
	rc:=avcodec_receive_frame(Context, Frame);
	{Zero means complete packet received}
	if rc=0 then
		begin
		TimeStamp:=av_frame_get_best_effort_timestamp(Frame);
		PutInQueue:=false;
		{ Queue only the frames with timestamp>=0 }
		if (not Player.PutThreadToSleep) and (TimeStamp>=0) then
			begin
			if video then
				begin
				{We must have at least one video frame to enable audio}
				FirstVideoFrame:=True;
				PutInQueue:=true;
				end
			else
				begin
				if FirstVideoFrame then PutInQueue:=True;
				end;
			end;
		if PutInQueue then
			begin
			if not Queue.Put(Frame) then
				begin
				{Exit if queue is full}
				av_frame_free(@Frame);
				break;
				end;
			end
		else
			begin
			av_frame_free(@Frame);
			end;
		end
	else
		begin
		{Here because there wasn't enough data - usually rc = -11 }
		av_frame_free(@Frame);
		break;
		end;
	end;
end;



begin
Result:=0;
Packet:=av_packet_alloc;
Player:=TCustomFFMPegPlayer(p);
FirstVideoFrame:=Player.FVideoStreamIndex<0;

{Note: we Delay audio queuing until we have at least one video frame }
while Player.FPlaying do
	begin
	if Player.PutThreadToSleep or Player.FEof then
		begin
		{Flag that the thread has noticed the Sleep command}
		{it is flagged here because we know the thread isn't blocked.}
		Player.FReadThreadSleeping:=true;
		Sleep(20);
		Continue;
		end;
	if Player.FDelayAudioQueueing then
		begin
		{We delay audio queuing until after first videoframe }
		if Player.FVideoStreamIndex>=0 then FirstVideoFrame:=false;
		Player.FDelayAudioQueueing:=false;
		end;
	rc := av_read_frame(Player.FFormatCtx, Packet);
	if (rc < 0) then
		begin
		av_packet_unref(Packet);
		if rc=AVERROR_EOF then
			begin
			Player.FEof:=true;
			continue;
			end
		else
			begin
			break;
			end;
		end;
	{Is valid audio packet?}

	if (Player.FAudioStreamIndex>=0) and (Packet.stream_index = Player.FAudioStreamIndex) then
		begin
		DecodePacket(Player.FAudioCtx,Player.FAudioQueue,Packet,false);
		end
	else
	if (Player.FVideoStreamIndex>=0) and (Packet.stream_index = Player.FVideoStreamIndex) then
		begin
		DecodePacket(Player.FVideoCtx,Player.FVideoQueue,Packet,true);
		if Player.FVideoQueue.Count>Player.MaxDepth then
			begin
			Player.MaxDepth:=Player.FVideoQueue.Count;
			end;
		end;
	av_packet_unref(Packet);
	end;
av_packet_free(@Packet);
end;




procedure SDLAudioCallback(UserData: Pointer; stream: PByte; len: integer); cdecl;
{Used by SDL to grab next block of Audio data}
var BufSize,len1 :integer;
var BuffPtr: PByte;
var Frame: PAVFrame;
var DecodedAudio: PByte;
var ts: int64;
var Player: TCustomFFMPegPlayer;
begin
Player:=TCustomFFMPegPlayer(UserData);
DecodedAudio:=nil;
while (len>0) and Player.FPlaying do
	begin
	{If Index past end of Buffer then all data has sent to SDL }
	if Player.FAudioBufIndex >= Player.FAudioBufSize then
		begin
		{Get the next Audio Data frame queue  - don't wait because it blocks the thread}
		if not Player.FAudioQueue.Get(Frame,false) then
			begin
			{Exit if player stopped playing}
			if not Player.FPlaying then break;
			{ If no data, clear buffer so no sound is produced }
			fillchar(stream^,len,0);
			{If end of file, trigger user event}
			if Player.FEof then PostMessage(Player.Handle,PLM_EOF,0,0);
			{And exit}
			break;
			end;
		{Get necessary Audio buffer size for specified format}
		BufSize:=av_samples_get_buffer_size(nil,Player.FAudioCtx.channels, Frame.nb_samples, AV_SAMPLE_FMT_FLT, 1);
		{Allocate necessary memory}
		ReAllocMem(DecodedAudio,BufSize);
		{Muted or Showing first frame}
		if Player.FMute or Player.FVideoShowNextFrame then
			begin
			{Zero memory so no sound produced}
			fillchar(Player.FAudioBuffer^,BufSize,0);
			end
		else
			begin
			{Decode Audio}
			swr_convert(Player.FAudioSwrCtx, @DecodedAudio, Frame.nb_samples, @Frame.data[0], Frame.nb_samples);
			{Move audio data to player buffer}
			Move(DecodedAudio^,Player.FAudioBuffer^,BufSize);
			end;
		Player.FAudioBufIndex:=0;
		Player.FAudioBufSize:=BufSize;
		ts:=av_frame_get_best_effort_timestamp(Frame);
		if ts>=0 then
			begin
			Player.FAudioPts:=ts;
			PostMessage(Player.Handle,PLM_UPDATEPROGRESS,0,0);
			end;

		av_frame_free(@Frame);
		end;
	{ Find the amount of data left to be played }
	len1:=Player.FAudioBufSize-Player.FAudioBufIndex;
	{Is Size greater than the buffer size?}
	if (len1 > len) then len1 := len;
	{ Get pointer for pointer-math incrementing}
	BuffPtr:=Player.FAudioBuffer;
	{Point to beginning of data}
	inc(BuffPtr,Player.FAudioBufIndex);
	{Move it into stream}
	Move(BuffPtr^,stream^,len1);
	{Subtract data sent}
	dec(len,len1);
	{Point to next block in stream}
	inc(stream,len1);
	inc(Player.FAudioBufIndex,len1);
	end;
ReAllocMem(DecodedAudio,0);
end;



function TimerThread(Interval: UInt32; param: Pointer): UInt32; cdecl;
{Handle Time Event}
begin
{Post message to main thread}
PostMessage(THandle(param),PLM_TIMER,1,0);
Result:=Interval;
end;




{ TPlayerQueue }

procedure TPlayerQueue.IncIndex(var Inx: integer);
begin
inc(Inx);
if Inx>=Length(FQueue) then Inx:=0;
end;



function TPlayerQueue.GetCount: integer;
begin
if FPutIndex>=FGetIndex then Result:=FPutIndex-FGetIndex
else Result:=(Length(FQueue)-FGetIndex) + FPutIndex;
if (Result=0) and FQueue[FPutIndex].FValid then Result:=Length(FQueue);
end;


constructor TPlayerQueue.Create(player: TCustomFFMPegPlayer; maxframes: integer);
var i:integer;
begin
FPlayer:=player;
FGetEvent:=CreateEvent(nil,false,false,nil);
FPutEvent:=CreateEvent(nil,false,false,nil);

SetLength(FQueue,maxframes);
for i:=0 to Length(FQueue)-1 do
 with FQueue[i] do
 	begin
 	FValid:=false;
 	FFrame:=nil;
 	end;
end;


destructor TPlayerQueue.Destroy;
begin
Clear;
SetLength(FQueue,0);
CloseHandle(FGetEvent);
CloseHandle(FPutEvent);
inherited Destroy;
end;



function TPlayerQueue.Put(fr: PAVFrame; wait: boolean): boolean;
begin
Result:=false;
if fr=nil then exit;
if (not FPlayer.FPlaying) or FUnblockPut then exit;
{Wait for item to become available}
while FQueue[FPutIndex].FValid do
	begin
	{ If item is "Valid" the current slot is in use }
	if (not FPlayer.FPlaying) or FUnblockPut then exit;
	if not wait then exit;
	{Wait for the slot to be free}
	WaitForSingleObject(FGetEvent,INFINITE);	{ Wait until frame is picked up}
	end;
{ Slot is available, so store the frame and mark it valid}
FQueue[FPutIndex].FFrame:=fr;
FQueue[FPutIndex].FValid:=true;

{Point to next item in list}
IncIndex(FPutIndex);
{Signal that data has been inserted}
SetEvent(FPutEvent);
Result:=true;
end;



function TPlayerQueue.Get(var fr: PAVFrame; wait: boolean): boolean;
begin
Result:=false;
if (not FPlayer.FPlaying) or FUnblockGet then exit;
while not FQueue[FGetIndex].FValid do
	begin
	if (not FPlayer.FPlaying) or FUnblockGet then exit;
	if not wait then exit;
	{Wait for an item to be put in the queue}
	WaitForSingleObject(FPutEvent,INFINITE);
	end;
{Get Frame}
fr:=FQueue[FGetIndex].FFrame;
{Flag that slot for frame is empty}
FQueue[FGetIndex].FFrame:=nil;
FQueue[FGetIndex].FValid:=false;

IncIndex(FGetIndex);
SetEvent(FGetEvent);
Result:=true;
end;



function TPlayerQueue.Peek(var fr: PAVFrame): boolean;
begin
with FQueue[FGetIndex] do
 begin
 fr:=FFrame;
 Result:=FValid;
 end;
end;



procedure TPlayerQueue.Del;
{ Invalidate at the current getindex, but don't free frame }
begin
with FQueue[FGetIndex] do
 begin
 if not FValid then exit;
 FFrame:=nil;
 FValid:=false;
 end;
IncIndex(FGetIndex);
SetEvent(FGetEvent);
end;


procedure TPlayerQueue.WakeUpPut;
{Signal get-event in case it is waiting}
begin
SetEvent(FGetEvent);
end;


procedure TPlayerQueue.WakeUpGet;
{Signal put-event in case it is waiting}
begin
SetEvent(FPutEvent);
end;


procedure TPlayerQueue.Clear;
var I: integer;
begin
for I:=0 to Length(FQueue)-1 do
	begin
	FQueue[I].FValid:=false;
	if FQueue[I].FFrame<>nil then av_frame_free(@FQueue[I].FFrame);
	end;
FPutIndex:=0;
FGetIndex:=0;
end;




{ TCustomFFMPegPlayer }


procedure TCustomFFMPegPlayer.SetErrorStr(S: string; ErrNum: integer);
var S1: string;
begin
{Don't lookup errors numbers of zero}
if ErrNum = 0 then FLastErrorString:=S
else
	begin
	{Look up error code}
	SetLength(S1,64);
	av_strerror(ErrNum,PChar(S1),Length(S1));
	FLastErrorString:=S+': ('+IntToStr(ErrNum)+') - '+S1;
	end;
end;



function TCustomFFMPegPlayer.HandleError(IsError: boolean; Msg: string; ErrNum: integer): boolean;
begin
{A trick to simplify calling code}
Result:=IsError;
if IsError then SetErrorStr(Msg,ErrNum);
end;



function TCustomFFMPegPlayer.GetVideoTimerRate: integer;
{Get video timer rate}
begin
Result:=FTimerInterval;
end;

function TCustomFFMPegPlayer.GetDuration: int64;
{Duration in ms, returns -1 for Livestreams}
begin
Result:=-1;
if FFormatCtx=nil then exit;
Result:=FFormatCtx.duration div 1000;
end;

function TCustomFFMPegPlayer.GetAudioAvailable: boolean;
begin
 Result:=FAudioStreamIndex>=0;
end;



function TCustomFFMPegPlayer.GetPosition: int64;
{Get the Current Position in ms, Livestreams return -1}
begin
Result:=-1;
if (FAudioStreamIndex>=0) then
	begin
	if (FAudioPts>=0) then
		begin
		Result:=round(FAudioPts*FAudioTimebase*1000);
		if FFormatCtx.start_time <> AV_NOPTS_VALUE then Result:=Result-round(FFormatCtx.start_time/1000);
		end;
	end;
if Result>=0 then exit;
if (FVideoStreamIndex>=0) then
	begin
	if (FVideoPts>=0) then
		begin
		Result:=round(FVideoPts*FVideoTimebase*1000);
		if FFormatCtx.start_time <> AV_NOPTS_VALUE then Result:=Result-round(FFormatCtx.start_time/1000);
		end;
	end;
end;


function TCustomFFMPegPlayer.GetFrame: int64;
var FR: TAVRational;
begin
if (FVideoStreamIndex>=0) then
	begin
	if (FVideoPts>=0) then
		begin
		FR:=PPtrIdx(FFormatCtx.streams,FVideoStreamIndex).codec.framerate;
		Result:=round(FVideoPts*FVideoTimebase*1000);
		if FFormatCtx.start_time <> AV_NOPTS_VALUE then Result:=Result-round(FFormatCtx.start_time/1000);
		end;
	end;

end;


function TCustomFFMPegPlayer.GetVideoAvailable: boolean;
begin
Result:=FVideoStreamIndex>=0;
end;


procedure TCustomFFMPegPlayer.SetMaxAudioFrames(AValue: integer);
begin
if (AValue<=0) then exit;
FMaxAudioFrames:=AValue;
end;


procedure TCustomFFMPegPlayer.SetMaxVideoFrames(AValue: integer);
begin
if (AValue<=0) then exit;
 FMaxVideoFrames:=AValue;
end;

procedure TCustomFFMPegPlayer.SetVideoTimerRate(AValue: integer);
{Set video timer rate}
begin
FTimerInterval:=AValue;
end;

procedure TCustomFFMPegPlayer.WMTimer(var m: TMessage);
{Message to Handle event in main thread}
begin
if (not FPlaying) or
   (FPausing and not FVideoShowNextFrame) or
   (m.wParam<>1) then exit;
{Don't use Repaint, it will block message queue}
Invalidate;
end;

procedure TCustomFFMPegPlayer.WMPaintResize(var m: TMessage);
var pr: boolean;
begin
 if FRenderer=nil then exit;
 PaintImage;
 pr:=true;
 PaintOverLays(FRenderer,pr);
 SDL_RenderPresent(FRenderer)
end;

procedure TCustomFFMPegPlayer.WMEOF(var m: TMessage);
begin
DoOnEOF;
end;


procedure TCustomFFMPegPlayer.WMUpdateProgress(var m: TMessage);
begin
DoUpdateProgress;
end;


function TCustomFFMPegPlayer.SetBackgroundColor(Renderer: PSDL_Renderer; Color: TColor): integer;
var R,G,B: byte;
begin
R:=GetRValue(Color);
G:=GetGValue(Color);
B:=GetBValue(Color);
Result:=SDL_SetRenderDrawColor(Renderer, R, G, B, 255 );
end;


procedure TCustomFFMPegPlayer.CreateWindowAndRenderer;
{Create SDL Window and Renderer}
var OK: boolean;
begin
{Exit if already created}
if FWindow<>nil then exit;
OK:=false;
try
{Create SDL Window from Windows handle}
FWindow:= SDL_CreateWindowFrom(Pointer(Handle));
if FWindow=nil then
	begin
	FlastErrorString:=format('SDL_CreateWindowFrom-Error: %s',[string(SDL_GetError)]);
	exit;
	end;
{Create SDL Renderer from the SDL Window}
FRenderer := SDL_CreateRenderer(FWindow, -1,GetRendererFlags);
if FRenderer=nil then
	begin
	FLastErrorString:=format('SDL_CreateRenderer: %s',[string(SDL_GetError)]);
	exit;
	end;
{Set Renderer color to black opaque}
if SetBackgroundColor(FRenderer, Color)<0 then
	begin
	FLastErrorString:=format('SDL_SetRenderDrawColor: %s',[string(SDL_GetError)]);
	exit;
	end;
OK:=true;

finally
 if not OK then DestroyWindowAndRenderer;
 end;
end;



procedure TCustomFFMPegPlayer.DestroyWindowAndRenderer;
begin
if (FRenderer<>nil) then
	begin
	SDL_DestroyRenderer(FRenderer);
	FRenderer:=nil;
	end;
if (FWindow<>nil) then
	begin
	SDL_DestroyWindow(FWindow);
	FWindow:=nil;
	end;
end;



function TCustomFFMPegPlayer.LoadText(s: string; co: TSDL_Color; var w,
 h: integer): PSDL_Texture;
var su:PSDL_Surface;
begin
Result:=nil;
su:=TTF_RenderUTF8_Blended(FFont,PAnsiChar(s),co);
if su=nil then exit;
try
Result := SDL_CreateTextureFromSurface( FRenderer, su );
if Result=nil then exit;
SDL_SetTextureBlendMode(Result, SDL_BLENDMODE_BLEND);
SDL_SetTextureAlphaMod(Result,co.a);

if SDL_QueryTexture(Result,nil,nil,@w,@h)<>0 then
	begin
	SDL_DestroyTexture(Result);
	Result:=nil;
	end;
finally
 if su<>nil then SDL_FreeSurface( su );
 end;
end;

procedure TCustomFFMPegPlayer.Resize;
begin
inherited Resize;
PostMessage(handle,PLM_PAINTRESIZE,0,0);
{Don't call paint here. }
end;


procedure TCustomFFMPegPlayer.DoOnEof;
begin
Stop(True);
end;



procedure TCustomFFMPegPlayer.DoUpdateProgress;
{Virtual method filled in by child objects}
begin
end;


procedure TCustomFFMPegPlayer.StartTimer;
{Create a SDL timer to Audio/Video rendering in main thread}
begin
if FTimerId<>0 then exit;
FTimerId:=SDL_AddTimer(FTimerInterval,TimerThread,Pointer(handle));
end;



procedure TCustomFFMPegPlayer.StopTimer;
begin
if FTimerId<>0 then
	begin
	SDL_RemoveTimer(FTimerId);
	FTimerId:=0;
	end;
end;



function TCustomFFMPegPlayer.GetRendererFlags: longword;
begin
Result:=SDL_RENDERER_ACCELERATED or SDL_RENDERER_PRESENTVSYNC;
end;



procedure TCustomFFMPegPlayer.Paint;
var NextFrame,fr: PAVFrame;
var its,ts: int64;
var AudioTime,VideoTime: double;
var RenderImage: boolean;
begin
{}
if InitializeError then
	begin
	Canvas.Brush.Color:=self.Color;
	Canvas.fillrect(ClientRect);
	exit;
	end;
{Create a Window and Renderer if necessary }
CreateWindowAndRenderer;
RenderImage:=false;
NextFrame:=nil;
its:=-1;
repeat
	begin
	{if we've stopped playing, are seeking or there is no image}
	if (not FPlaying) or (FSeeking) or (FTexture=nil) then
		begin
		{Clear the render target to black}
		SDL_RenderClear(FRenderer);
		{Flag to render blank screen}
		RenderImage:=true;
		break;
		end;
	if FPausing and not FVideoShowNextFrame then break;
	{Check for frame waiting in queue}
	if not FVideoQueue.Peek(fr) then
		begin
		{No frame found - if end of file, trigger EOF event}
		if FEof then PostMessage(handle,PLM_EOF,0,0);
	 	{Exit - if there are more packets, Paint will be called again}
		break;
		end;
	{It is a valid video frame}
	ts:=av_frame_get_best_effort_timestamp(fr);
	if ts<0 then
		begin
		{ could be a B-Frame }
		FVideoQueue.Del;
		av_frame_free(@fr);
		continue;
		end;
	if FVideoSynchPts<0 then
		begin
		{Synchronize when there is no audio}
		FVideoSynchPts:=ts;
		FVideoSynchTime:=av_gettime_relative;
		end;
	VideoTime:=ts*FVideoTimebase;
	if (FAudioStreamIndex>=0) then
		begin
		AudioTime:=FAudioPts*FAudioTimeBase;
		end
	else
		begin
		AudioTime:=(av_gettime_relative-FVideoSynchTime)/1000000.0 + FVideoSynchPts*FVideoTimebase;
		end;

	{Check if the video time is way ahead of the audio }
	if VideoTime > (AudioTime + FBadFrameTolerance) then
		begin
		{Delete the video frame}
		FVideoQueue.Del;
		av_frame_free(@fr);
		end
	else if (AudioTime>=0) and (VideoTime<=AudioTime) then
		begin
		{Keep rendering video until has caught up with Audio}
		if NextFrame<>nil then av_frame_free(@NextFrame);
		FVideoQueue.Del;
		{Setup the next frame to Render}
		NextFrame:=fr;
		{Set the working frame to empty}
		fr:=nil;
		its:=ts;
		end
	else
		begin
		{ Hold frame until audio catches up. }
		break;
		end;
	end
until false;

if NextFrame<>nil then
	begin
	if FVideoShowNextFrame then
		begin
		if FPausing then
			begin
			{Stop the video Timer}
			StopTimer;
			if FAudioStreamIndex>=0 then
				begin
				{Stop audio}
				SDL_PauseAudio(1);
				end;
			end;
		FVideoShowNextFrame:=false;
		end;

	{Convert image to the right size and pixel format}
	sws_scale(FVideoSwsCtx, @NextFrame.data, @NextFrame.linesize, 0, FVideoCtx.height, @FVideoDecodeFrame.data, @FVideoDecodeFrame.linesize);
	SDL_UpdateTexture(FTexture,nil, FVideoDecodeFrame.data[0], FVideoDecodeFrame.linesize[0]);
	FTextureValid:=true;
	FFVideoAspect:=av_guess_sample_aspect_ratio(FFormatCtx,PPtrIdx(FFormatCtx.streams,FVideoStreamIndex), NextFrame);
	{Save presentation time of last image}
	FVideoPts:=its;
	{Display current image}
	PaintImage;
	RenderImage:=true;
	{Update progress}
	if (FAudioStreamIndex<0) then PostMessage(handle,PLM_UPDATEPROGRESS,0,0);
	av_frame_free(@NextFrame);
	end;
{Show any overlays - in this case, overaly times}
PaintOverLays(FRenderer,RenderImage);
if RenderImage then SDL_RenderPresent(FRenderer);
end;



procedure TCustomFFMPegPlayer.SetPixelFormat(SDLPixelFormat: Cardinal; AVPixelFormat: TAVPixelFormat);
{ set the SDL-renderformat and the compatible FFMPEG-decoderformat}
begin
FSDLPixelFormat:=SDLPixelFormat;
FAVPixelFormat:=AVPixelFormat;
end;



procedure TCustomFFMPegPlayer.PaintOverlays(renderer:PSDL_Renderer; var present: boolean);
{Do any overlays here}
begin
PaintTimes(renderer,present);
end;

procedure TCustomFFMPegPlayer.PaintTimes(renderer: PSDL_Renderer; var Present: boolean);
{Display current time of stream in lower right corner}
{ Show current position and the duration if media has them
{ If the media has no duration then dispaly current time }
var tx: PSDL_Texture;
var co: TSDL_Color;
var tis: string;
var ww,hh: integer;
var rd: TSDL_Rect;
var d,p: int64;
begin
if not FShowTimes then exit;
if not FPlaying then exit;
co.r:=255; co.g:=0; co.b:=0; co.a:=255;

d:=Duration;
p:=Position;

if (d>=0) and (p>=0) then tis:=GetTimeString(p)+'   '+ GetTimeString(d)
else tis:=FormatDateTime('hh:nn:ss',now);

if tis<>FLastTimeString then
	begin
	if not Present then PaintImage;
	Present:=true;
	end;

FLastTimeString:=tis;
tx:=LoadText(tis,co,ww,hh);

rd.h:=round(self.Height*FTimeDisplayHeight/100);
rd.w:=round(ww*rd.h/hh);
rd.x:=Self.Width-rd.w-10;
rd.y:=Self.Height-rd.h-10;

SDL_RenderCopy(FRenderer, tx, nil, @rd );
SDL_DestroyTexture(tx);
end;



procedure TCustomFFMPegPlayer.PaintImage;
{Show last SDL texture image}
var rd: TSDL_Rect;

	function GetDestRect: TSDL_Rect;
	var Scale: double;
	begin
	Scale:=min(Self.Width/FVideowidth, Self.Height/FVideoheight);
	Result.w:=Round(FVideoWidth * Scale);
	Result.h:=Round(FVideoheight * Scale);
	Result.x:=(self.Width-Result.w) div 2;
	Result.y:=(self.Height-Result.h) div 2;
	end;

begin
if (FWindow=nil) or (FRenderer=nil) or (FTexture=nil) or
   (FVideoWidth=0) or (FVideoHeight=0) then exit;
RD:=GetDestRect;
SetBackgroundColor(FRenderer, Color);
SDL_RenderClear(FRenderer);
SDL_SaveBMP(
if FTextureValid then SDL_RenderCopy(FRenderer, FTexture, nil, @rd);
end;



function BeginThread2(ThreadFunc: TThreadFunc; Param: Pointer): integer;
var ThreadId: Cardinal;
begin
Result := BeginThread(nil, 0, ThreadFunc, Param, 0, ThreadID);
end;



{================ Pause/Resume code ===========================================}



procedure TCustomFFMPegPlayer.Resume;
begin
if (not FPlaying) or (not FPausing) then exit;
{Reset sync time values}
FVideoSynchPts:=-1;
FVideoSynchTime:=-1;

FPausing:=false;
StartTimer;
if FAudioStreamIndex>=0 then
	begin
	SDL_PauseAudio(0);
	end;
end;


procedure TCustomFFMPegPlayer.Pause;
begin
if not FPlaying then exit;
FPausing:=true;
StopTimer;
if FAudioStreamIndex>=0 then
	begin
	SDL_PauseAudio(1);
	end;
end;

{================ Play/Stop/Seek code ===========================================}


procedure TCustomFFMPegPlayer.Seek(ms: int64);
{Seek specified position in miliseconds}
var rc,si,flags: integer;
var ts: int64;
begin
if not FPlaying then exit;
if (ms<0) or (Duration<=0) or (ms>Duration) then exit;
if FSeeking then exit;
FSeeking:=true;
try
StopTimer;
{Stop SDL from sending audio requests through SDLAudioCallback}
if (FAudioStreamIndex>=0) and not FPausing then
	begin
	SDL_PauseAudio(1);
	end;
FReadThreadSleeping:=false;
PutThreadToSleep:=true;

if FAudioQueue<>nil then FAudioQueue.FUnblockPut:=true;
if FAudioQueue<>nil then FAudioQueue.WakeUpPut;
if FAudioQueue<>nil then FAudioQueue.Clear;

if FVideoQueue<>nil then FVideoQueue.FUnblockPut:=true;
if FVideoQueue<>nil then FVideoQueue.WakeUpPut;
if FVideoQueue<>nil then FVideoQueue.Clear;

{Wait until thread has handled sleep command}
while not FReadThreadSleeping do sleep(10);

if FAudioQueue<>nil then FAudioQueue.FUnblockPut:=false;
if FVideoQueue<>nil then FVideoQueue.FUnblockPut:=false;

si:=-1;
Flags:=0;

{We have a choice of audio or video stream
audio is the first choice because it is higher resolution}
if FAudioStreamIndex>=0 then
	begin
	{Specify the audio stream}
	si:=FAudioStreamIndex;
	{Calculate the seek target in audio timebase Units}
	ts:=round(ms /(1000 * FAudioTimeBase));
	end
else if FVideoStreamIndex>=0 then
	begin
	{Specify the video stream}
	si:=FVideoStreamIndex;
	{Calculate the seek target in video timebase Units}
	ts:=round(ms /(1000 * FVideoTimeBase));
	{Use this for a video only file}
	flags:=AVSEEK_FLAG_ANY;
	if FFormatCtx.start_time <> AV_NOPTS_VALUE then ts:=ts+FFormatCtx.start_time;
	end
{The mpeg default unit is microseconds}
else ts:=ms * int64(1000);

rc:=avformat_seek_file(FFormatCtx,si,Low(int64),ts,High(int64),flags);
if rc>=0 then
	begin
	{clear all queue and buffers}
	if FAudioQueue<>nil then FAudioQueue.Clear;
	if FVideoQueue<>nil then FVideoQueue.Clear;
	FAudioPts:=-1;
	FVideoPts:=-1;
	FVideoSynchPts:=-1;
	FVideoSynchTime:=-1;
	FAudioBufSize:=0;
	FAudioBufIndex:=0;
	if FVideoCtx<>nil then avcodec_flush_buffers(FVideoCtx);
	if FAudioCtx<>nil then avcodec_flush_buffers(FAudioCtx);
	avformat_flush(FFormatCtx);
	end;
FEof:=false;
{Tell program to wait for first video frame}
FDelayAudioQueueing:=true;
{Enable Read thread}
PutThreadToSleep:=false;
{ Set up delay for next frame }
if (FVideoStreamIndex>=0) and Pausing then FVideoShowNextFrame:=true;
StartTimer;
if (FAudioStreamIndex>=0) then
	begin
	SDL_PauseAudio(0);
	end;
Invalidate;
finally
 FSeeking:=false;
 end;
end;


procedure TCustomFFMPegPlayer.WaitQueuesEmpty;
{Wait for queues to empty and all frames to be displayed}
const Limit = 1000;
var Stop: DWord;
begin
Stop:=timeGetTime+Limit;
while (timeGetTime<Stop) and (AudioQueDepth>0) do
 Application.ProcessMessages;
Stop:=timeGetTime+Limit;
while (timeGetTime<Stop) and (VideoQueDepth>0) do
 Application.ProcessMessages;
end;

var Stopping: boolean;



procedure TCustomFFMPegPlayer.Stop(WaitQueues: boolean);
{Stop playing - if WaitQueues is true then}
{wait to allow the last bit of the video/audio data to be displayed}
begin
if Stopping then exit;
Stopping:=True;
try
if WaitQueues then WaitQueuesEmpty;
if FPlaying then
	begin
	StopTimer;
	FPlaying:=false;
	end;
if FVideoQueue<>nil then
 with FVideoQueue do
  begin
  WakeUpGet;
  WakeUpPut;
  end;
if FAudioQueue<>nil then
 with FAudioQueue do
	begin
	WakeUpGet;
	WakeUpPut;
	end;

if FAudioCodec<>nil then
	begin
	SDL_CloseAudio;
	end;
if FReadThreadId<>0 then
	begin
	{Wait for thread to terminate}
	WaitForSingleObject(FReadThreadId,INFINITE);
	FReadThreadId:=0;
	end;
FreeAndNil(FVideoQueue);
FreeAndNil(FAudioQueue);

if (FTexture<>nil) then
	begin
	SDL_DestroyTexture(FTexture);
	FTexture:=nil;
	end;
FTextureValid:=false;
ReAllocMem(FAudioBuffer,0);
if (FAudioSwrCtx<>nil) then
	begin
	swr_free(@FAudioSwrCtx);
	end;
if (FAudioCtx<>nil) then
	begin
	avcodec_free_context(@FAudioCtx);
	end;

if (FVideoDecodeFrameBuffer<>nil) then
	begin
	av_free(FVideoDecodeFrameBuffer);
	FVideoDecodeFrameBuffer:=nil;
	end;

if (FVideoDecodeFrame<>nil) then
	begin
	av_frame_free(@FVideoDecodeFrame);
	end;
if (FVideoSwsCtx<>nil) then
	begin
	sws_freeContext(FVideoSwsCtx);
	FVideoSwsCtx:=nil;
	end;

if (FVideoCtx<>nil) then
	begin
	avcodec_free_context(@FVideoCtx);
	end;

FAudioStreamIndex:=-1;
FAudioTimeBase:=0;
FVideoTimeBase:=0;

FVideoStreamIndex:=-1;
FAudioCodec:=nil;
FVideoCodec:=nil;
if (FFormatCtx<>nil) then
	begin
	avformat_close_input(@FFormatCtx);
	end;
FVideoWidth:=0;
FVideoHeight:=0;
FFVideoAspect.den:=0;
FFVideoAspect.num:=0;
PostMessage(handle,PLM_UPDATEPROGRESS,0,0);

Invalidate;
finally Stopping:=False; end;
end;


{================ Get File format information ===================================}


procedure LogCallback(ptr: Pointer; level: Integer; fmt: PAnsiChar; vl: pva_list); cdecl;
{Routine captures AV_Log information }
var  line: array [0 .. 1023] of AnsiChar;
var print_prefix: Integer;
var S: string;
begin
{This routine must be thread-safe}
print_prefix := 1;
av_log_format_line(ptr, level, fmt, vl, @line, sizeof(line), print_prefix);
S:=line;
{Data contains Linux-style LF for EOL - Replace with CRLF}
if (S<>'') and (S[Length(S)]=LF_Char) then
	begin
	S:=Trim(S)+CRLF_Char;
	end;

{This might not be thread safe}
FileFormatStr.Value:=FileFormatStr.Value+S;
end;


procedure TCustomFFMPegPlayer.SaveFileFormat;
begin
FileFormatStr.Value:='';
av_log_set_callback(LogCallback);
av_dump_format(FFormatCtx, 0,PChar(Url), 0);
av_dump_format(FFormatCtx, 1,PChar(Url), 0);
{Set it to nil to prevent extraneous messages being logged.}
av_log_set_callback(nil);
end;

{==================== Routines to handle playing from memory =============}


function ReadBuffer(opaque: Pointer; buf: puint8_t; buf_size: int): int; cdecl;
{Callback to Read next block of memory}
var Stream: TStream;
begin
Stream:=TStream(opaque);
Result:=Stream.read(Buf^,buf_size);
end;


function SeekBuffer(opaque: Pointer; Offset: int64; Whence: integer): int64; cdecl;
{Callback to Seek to a specific position in the memory buffer}
{Can also return the memory size if requested}
var Origin: Word;
const AVSEEK_SIZE=$10000;
const AVSEEK_FORCE=$20000;
var Stream: TStream;
begin
Stream:=TStream(opaque);
Result:=0;
case Whence of
 0: Result:=Stream.Seek(Offset,soFromBeginning);
 AVSEEK_SIZE: Result:=Stream.Size;
 end;
end;


procedure TCustomFFMPegPlayer.GetMemoryContext;
{Setup the memory context to read from memory}
var ct: PAVIOContext;
const BufferSize = 1024 * 32;
const AV_INPUT_BUFFER_PADDING_SIZE = 64;
const AVFMT_FLAG_CUSTOM_IO = $0080;

	function GetFormat: pAVInputFormat;
	{We have to read the format so FFMPeg doesn't do it}
	{If FFMPeg does it, it will overflow its buffer}
	var BytesRead: integer;
	var ProbeData: TAVProbeData;
	begin
	BytesRead:=FMPEGStream.read(MemBuffer^,BufferSize);
	FMPEGStream.Seek(0,soFromBeginning);
	ProbeData.buf:=MemBuffer;
	probeData.buf_size:=BytesRead;
	probeData.filename:='';
	Result:=av_probe_input_format(@ProbeData, 1);
	end;


begin
FMPEGStream.Seek(0,soFromBeginning);
FFormatCtx:=avformat_alloc_context;
MemBuffer:=av_malloc(BufferSize + AV_INPUT_BUFFER_PADDING_SIZE );
ct:=avio_alloc_context(MemBuffer, BufferSize, 0, FMPEGStream, ReadBuffer, nil, SeekBuffer);
FFormatCtx.pb:=ct;
FFormatCtx.flags:=AVFMT_FLAG_CUSTOM_IO;
FFormatCtx.iformat:=GetFormat;
avformat_open_input(@FFormatCtx,PAnsiChar('memory stream'), nil, nil);
end;




function TCustomFFMPegPlayer.Play: boolean;
var rc,NumBytes: integer;
var WantedSpec: TSDL_AudioSpec;
begin
MaxDepth:=0;
Result:=false;
FLastErrorString:='';
{Already playing?}
if HandleError(FPlaying,rsPlayerActive,0) then exit;
try
{Initialization Failure? }
if HandleError(InitializeError,InitSFErrorString,0) then exit;

fillchar(wantedSpec,sizeof(wantedSpec),0);

{Test if we are reading from a file or a stream (memory)}
if FMPEGStream<>nil then GetMemoryContext
else if HandleError(not FileExists(FURL),'No valid file or stream',0) then exit
else
	begin
	{Open mpeg file - Note: this automatically calls "avformat_alloc_context"}
	rc := avformat_open_input(@FFormatCtx, PAnsiChar(FURL), nil, nil);
	if HandleError(rc<0,'avformat_open_input',rc) then exit;
	end;

{Load stream info}
rc := avformat_find_stream_info(FFormatCtx, nil);
if HandleError(rc<0,'avformat_find_stream_info',rc) then exit;

{Look for video stream}
FVideoStreamIndex:=-1;
if not FDisableVideo then
	begin
	rc:=av_find_best_stream(FFormatCtx, AVMEDIA_TYPE_VIDEO,-1,-1,@FVideoCodec,0);
	if (rc>=0) then FVideoStreamIndex:=rc;
	end;

{Look for Audio stream}
FAudioStreamIndex:=-1;
if not FDisableAudio then
	begin
	rc:=av_find_best_stream(FFormatCtx, AVMEDIA_TYPE_AUDIO,-1,FVideoStreamIndex,@FAudioCodec,0);
	if (rc>=0) then FAudioStreamIndex:=rc;
	end;

{Exit if there isn't at least one Audio or Video stream}
if HandleError((FVideoStreamIndex<0) and (FAudioStreamIndex<0),rsNoStreamFound,0) then exit;

FVideoWidth:=0;
FVideoHeight:=0;
{Handle Video stream, if any}
if (FVideoStreamIndex >= 0) then
	begin
	{Create a video queue}
	FVideoQueue:=TPlayerQueue.Create(self,FMaxVideoFrames);
	FVideoQueue.Name:='Video';
	{Get video time-base}
	FVideoTimeBase:=av_q2d(PPtrIdx(FFormatCtx.streams,FVideoStreamIndex).time_base);
	{Create Video context}
	FVideoCtx := avcodec_alloc_context3(FVideoCodec);
	if HandleError(FVideoCtx=nil,rsVideoCtxNotAlloc,0) then exit;

	{Set context parameters}
	rc := avcodec_parameters_to_context(FVideoCtx, PPtrIdx(FFormatCtx.streams,FVideoStreamIndex).codecpar);
	if HandleError(rc<0,'avcodec_parameters_to_context (video): ',rc) then exit;

	{Set codec timebase}
	av_codec_set_pkt_timebase(FVideoCtx,PPtrIdx(FFormatCtx.streams,FVideoStreamIndex).time_base);
	{Open codec }
	rc := avcodec_open2(FVideoCtx, FVideoCodec, nil);
	if HandleError(rc<0,'avcodec_open2 (video): ',rc) then exit;

	{Get the image scaler}
	FVideoSwsCtx := sws_getContext(FVideoCtx.width, FVideoCtx.height, FVideoCtx.pix_fmt, FVideoCtx.width,
				FVideoCtx.height, FAVpixelformat, SWS_BILINEAR, nil, nil, nil);
	{Handle scaler creation error}
	if HandleError(FVideoSwsCtx=nil,rsVideoSwsCtxNotAlloc,0) then exit;

	{Get video dimensions}
	FVideoWidth := PPtrIdx(FFormatCtx.streams,FVideoStreamIndex).codecpar.width;
	FVideoHeight := PPtrIdx(FFormatCtx.streams,FVideoStreamIndex).codecpar.height;

	{Setup image buffer}
	NumBytes := av_image_get_buffer_size(FAVPixelFormat, FVideoWidth, FVideoHeight, 8);
	if HandleError(numbytes < 1,'av_image_get_buffer_size (video): ',numbytes) then exit;

	{Create a frame for raw video}
	FVideoDecodeFrame := av_frame_alloc;
	if HandleError(FVideoDecodeFrame=nil,rsVideoDecodeFrameNotAlloc,0) then exit;

	{Create a buffer for decoded video}
	FVideoDecodeFrameBuffer := av_malloc(numBytes);
	if HandleError(FVideoDecodeFrameBuffer=nil,rsVideoFrameBufferNotAlloc,0) then exit;

	{Configure raw video frame}
	rc := av_image_fill_arrays(@FVideoDecodeFrame.data[0], @FVideoDecodeFrame.linesize[0], FVideoDecodeFrameBuffer, FAVPixelFormat, FVideoWidth, FVideoHeight, 1);
	if HandleError(rc < 0,'av_image_fill_arrays: ',rc) then exit;

	{Create viewing window and Renderer}
	CreateWindowAndRenderer;
	if HandleError(FRenderer=nil,'No SDL Window/Renderer',0) then exit;

	{Create SLD texture (bitmap) for drawing}
	FTexture := SDL_CreateTexture( FRenderer, FSDLPixelFormat, SDL_TEXTUREACCESS_STREAMING, FVideoWidth, FVideoHeight);
	if HandleError(FTexture=nil,'SDL_CreateTexture: '+string(SDL_GetError),0) then exit;

	{Clear SDL image buffers}
	SDL_RenderClear(FRenderer);
	SDL_RenderPresent(FRenderer);
	end;

{Handle Audio stream, if any}
if (FAudioStreamIndex >= 0) then
	begin
	{Create Audio queue}
	FAudioQueue:=TPlayerQueue.Create(self,FmaxAudioFrames);
	FAudioQueue.Name:='Audio';
	{Get audio time-base}
	FAudioTimeBase:=av_q2d(PPtrIdx(FFormatCtx.streams,FAudioStreamIndex).time_base);
	{Create Audio Context}
	FAudioCtx := avcodec_alloc_context3(FAudioCodec);
	if HandleError(FAudioCtx=nil,rsAudioCtxNotAlloc,0) then exit;

	{Set context parameters}
	rc := avcodec_parameters_to_context(FAudioCtx, PPtrIdx(FFormatCtx.streams,FAudioStreamIndex).codecpar);
	if HandleError(rc<0,'avcodec_parameters_to_context (audio): ',rc) then exit;

	{Set codec timebase}
	av_codec_set_pkt_timebase(FAudioCtx,PPtrIdx(FFormatCtx.streams,FAudioStreamIndex).time_base);
	{Open codec}
	rc := avcodec_open2(FAudioCtx, FAudioCodec, nil);
	if HandleError(rc<>0,'avcodec_open2 (audio): ',rc) then exit;

	{Get audio resampler}
	FAudioSwrCtx := swr_alloc;
	if HandleError(FAudioSwrCtx=nil,rsAudioSwrCtxNotAlloc,0) then exit;
	{Set audio channel layout}
	if (FAudioCtx.channel_layout<>0) then
		begin
                {Use supplied layout}
		av_opt_set_channel_layout(FAudioSwrCtx, 'in_channel_layout', FAudioCtx.channel_layout, 0);
		av_opt_set_channel_layout(FAudioSwrCtx, 'out_channel_layout', FAudioCtx.channel_layout, 0);
		end
	else if (FAudioCtx.channels=2) then
		begin
		{Use stereo audio layout}
		av_opt_set_channel_layout(FAudioSwrCtx, 'in_channel_layout',AV_CH_LAYOUT_STEREO, 0);
		av_opt_set_channel_layout(FAudioSwrCtx, 'out_channel_layout', AV_CH_LAYOUT_STEREO, 0);
		end
	else if (FAudioCtx.channels=1) then
		begin
		{Use mono audio layout}
		av_opt_set_channel_layout(FAudioSwrCtx, 'in_channel_layout',AV_CH_LAYOUT_MONO, 0);
		av_opt_set_channel_layout(FAudioSwrCtx, 'out_channel_layout', AV_CH_LAYOUT_MONO, 0);
		end
	else
		begin
		{Use default format}
		av_opt_set_channel_layout(FAudioSwrCtx, 'in_channel_layout',AV_CH_LAYOUT_NATIVE, 0);
		av_opt_set_channel_layout(FAudioSwrCtx, 'out_channel_layout', AV_CH_LAYOUT_NATIVE, 0);
		end;

	{Set audio sample rate}
	av_opt_set_int(FAudioSwrCtx, 'in_sample_rate', FAudioCtx.sample_rate, 0);
	av_opt_set_int(FAudioSwrCtx, 'out_sample_rate', FAudioCtx.sample_rate, 0);
	{Set audio sample format }
	av_opt_set_sample_fmt(FAudioSwrCtx, 'in_sample_fmt', FAudioCtx.sample_fmt, 0);
	av_opt_set_sample_fmt(FAudioSwrCtx, 'out_sample_fmt', AV_SAMPLE_FMT_FLT, 0);
	{Initialize audio resampler}
	rc := swr_init(FAudioSwrCtx);
	if HandleError(rc<0,'swr_init: ',rc) then exit;

	{Allocate memory for audio buffer}
	ReAllocMem(FAudioBuffer,FAudioBufSizeMax);

	wantedSpec.channels := FAudioCtx.channels;
	wantedSpec.freq := FAudioCtx.sample_rate;
	wantedSpec.format := AUDIO_F32;
	wantedSpec.silence := 0;
	wantedSpec.samples := FSDLAudioBufferSize;
	wantedSpec.userdata := Self;
	wantedSpec.callback := @SDLAudioCallback;

	{Ignore returned spec}
	if (SDL_OpenAudio(@wantedSpec, nil) < 0) then
		begin
		FLastErrorString:=format('SDL_OpenAudio: %s',[string(SDL_GetError)]);
		exit;
		end;
	end;
{Reset everything}
FEof:=false;
FAudioBufSize:=0;
FAudioBufIndex:=0;
FAudioPts:=-1;
FVideoPts:=-1;
FVideoSynchPts:=-1;
FVideoSynchTime:=-1;

{Create the Read thread}
FReadThreadId:=BeginThread2(ReadThread,self);

{Unpause audio if there is an audio stream}
if FAudioStreamIndex>=0 then SDL_PauseAudio(0);
{Start Video Stimer}
StartTimer;

FPlaying:=true;
FPausing:=false;
FLastTimeString:='';
SaveFileFormat;

if Assigned(FOnPlay) then FOnPlay(Self);
Result:=true;
finally if not Result then Stop(False); end;
end;


{================== Create/Destroy Code ========================================}


procedure TCustomFFMPegPlayer.InitalizePlayer;
{ Initialize SDL and FFMpeg }
begin
inc(InitSFCount);
if InitSFCount<>1 then exit;

InitializeError:=True;
av_register_all;
avcodec_register_all;
avdevice_register_all;
avformat_network_init;

if SDL_Init(SDL_INIT_AUDIO or SDL_INIT_VIDEO or SDL_INIT_TIMER)<0 then
	begin
	InitSFErrorString:=format('SDL_Init-Error: %s',[string(SDL_GetError)]);
	exit;
	end;

if not SDL_SetHint( SDL_HINT_RENDER_SCALE_QUALITY, '1') then
	begin
	InitSFErrorString:='SDL_SetHint LinearTexturRendering not set';
	exit;
	end;

{ We switch off then SDL-Cursor because there are some problems }
{ if the SDL-Cursor is on and the control-cursoris off}
SDL_ShowCursor(0);

if TTF_Init<0 then
	begin
	InitSFErrorString:=format('TTF_Init-Error: %s',[string(TTF_GetError)]);
	exit;
	end;

InitializeError:=False;
end;



procedure TCustomFFMPegPlayer.FinalizePlayer;
begin
if InitSFCount>0 then
	begin
	dec(InitSFCount);
	end
else exit;
if InitSFCount>0 then exit;

avformat_network_deinit;
TTF_Quit;
SDL_Quit;
end;

constructor TCustomFFMPegPlayer.Create(AOwner: TComponent);
var a:array[0..MAX_PATH] of AnsiChar;
var fp:AnsiString;
begin
inherited Create(AOwner);
{Thread-safe string for file-format data}
FileFormatStr:=TThreadString.Create;
if not (csDesigning in ComponentState) then InitalizePlayer;

Color:=RGB(32,32,32);
FTimerInterval:=15;
FTimeDisplayHeight:=5;
if not InitializeError then
	begin
	fillchar(a,sizeof(a),0);
	SHGetSpecialFolderPath(0,@a,CSIDL_FONTS,false);
	fp:=string(a)+'\Tahoma.ttf';
	FFont:=TTF_OpenFont(PAnsiChar(fp),28);
	TTF_SetFontStyle(FFont,TTF_STYLE_BOLD);
	end;
FSDLAudioBufferSize:=1024;
FAudioBufSizeMax:=(192000*3) div 2;
{This must be large or seek will stall}
FMaxVideoFrames:=2000;
FMaxAudioFrames:=250;
FAudioStreamIndex:=-1;
FVideoStreamIndex:=-1;
FBadFrameTolerance:=20.0;
SetPixelFormat(SDL_PIXELFORMAT_IYUV,AV_PIX_FMT_YUV420P);
end;


procedure TCustomFFMPegPlayer.DestroyWnd;
begin
Stop(False);
DestroyWindowAndRenderer;
FileFormatStr.Free;
inherited DestroyWnd;
end;


destructor TCustomFFMPegPlayer.Destroy;
begin
Stop(False);
if FFont<>nil then TTF_CloseFont(FFont);
DestroyWindowAndRenderer;

if not (csDesigning in ComponentState) then FinalizePlayer;
inherited Destroy;
end;


function TCustomFFMPegPlayer.GetAudioQueDepth: integer;
begin
Result:=0;
if FAudioQueue<>nil then Result:=FAudioQueue.Count;
end;

function TCustomFFMPegPlayer.GetVideoQueDepth: integer;
begin
Result:=0;
if FVideoQueue<>nil then Result:=FVideoQueue.Count;
end;


function TCustomFFMPegPlayer.GetFileFormatStr: string;
begin
Result:=FileFormatStr.Value;
end;


procedure TCustomFFMPegPlayer.SetURL(const Value: string);
begin
if FURL<>Value then
	begin
	FURL := Value;
	FMPEGStream:=nil;
	end;
end;

procedure TCustomFFMPegPlayer.SetMPegStream(const Value: TStream);
begin
if FMPEGStream<>Value then
	begin
	FMPEGStream := Value;
	FURL:='';
	end;
end;

{ TFFMPegPlayer }

procedure TFFMPegPlayer.DoOnEof;
{Handle End of File Event}
var Handled: boolean;
begin
Handled:=false;
if Assigned(FOnEof) then FOnEof(self,handled);
if not Handled then inherited DoOnEof;
end;


procedure TFFMPegPlayer.DoUpdateProgress;
begin
inherited DoUpdateProgress;
if Assigned(FOnProgress) then FOnProgress(self);
end;


procedure TFFMPegPlayer.DisplayFileFormat(Tree: TTreeView);
var Context: PAVFormatContext;
var rc: integer;
begin
Tree.Items.Clear;
if FURL='' then ShowNode(Tree,nil,'No file open',0)
else if FileExists(FURL) or CheckUrl(FURL) then
	begin
	Context:=nil;
	rc := avformat_open_input(@Context, PAnsiChar(FURL), nil, nil);
	try
	DisplayContext(Tree,Context^);
	finally avformat_close_input(@Context); end;
	end
else ShowNode(Tree,nil,'File doesn''''t Exist.',0);
end;



initialization
LogList:=TThreadStringList.Create;
finalization
LogList.Free;
end.

