
unit MagField;

{ Subroutine to calculate magnetic declination and other geomagnetic }
{ information based on longitude, latitude, elevation, and year. }

interface


uses DSupport,MagIGRFData;

type
 TModelType = (AnnualChange, YearModel);

{ Geodetic = Elevation based on mean sea level}
{ Geocentric = Elevation based radial distance from earth's center }

type
 TCoordSystem = (Geodetic,Geocentric);


{Input Parameter for conversion }

type
 TMagneticParams = record
 CoordSystem: TCoordSystem;
 Year,Longitude,Latitude,Elevation: double;
 end;

{Result parameters}

type
 TMagneticData = record
 Decl,Incl: double;
 XField,YField,ZField: double;
 TotalField,HorzField: double;
 DecChange,IncChange: double;
 XChange,YChange,ZChange: double;
 TotalChange,HorzChange: double;
 end;

function CalculateMagnetics(Params: TMagneticParams; var MagneticData: TMagneticData): boolean;

implementation

const IEXT=0;

const EXT_COEFF1=0;
const EXT_COEFF2=0;
const EXT_COEFF3=0;

const MaxOrder = 13;					{Maximum order found in the models}
const ArrayMax = ((MaxOrder + 1) * (MaxOrder + 1)) - 1;	{Caculate the maxium array size based on order}

procedure GetSphericalHarmonics(Model: TModelType; EpochIndex,MaxOrder: Integer;var GH: array of double);
{Reads spherical harmonic coefficients from the }
{model specified in EpochIndes into an array GH}
var I,II,M,N,MM,NN : Integer;
var G,HH : double;
begin 
II := 0; I:=0;
for NN := 1 to  MaxOrder do
 for MM := 0 to NN do
	begin
	M:=IGRFData[EpochIndex].Coeff[I].M;
	N:=IGRFData[EpochIndex].Coeff[I].N;
	if Model = YearModel then
		begin
		g:=IGRFData[EpochIndex].Coeff[I].EpochG;
		hh:=IGRFData[EpochIndex].Coeff[I].EpochH;
		end
	else
		begin 
		g:=IGRFData[EpochIndex].Coeff[I].ChangeG;
		hh:=IGRFData[EpochIndex].Coeff[I].ChangeH;
		end;
	Inc(I); Inc(II);
	GH[II] := G;
	if M <> 0 then begin Inc(II); gh[II] := HH; end;
	end;
end;



function Interpolate(Year, Epoch1: double;  NMax1: integer; Epoch2: double; NMax2: Integer; Model1, Model2: array of double; var NewModel: array of double) : Integer;
{Interpolates linearly, in time, between two spherical harmonic models.}
{ Year = Year of new model }
{ Model1, Model2 = Contains first and second model. 
{ Epoch1, Epoch2 = Year of first and second model}
{ NMax1, NMax2 = maximum degree and order of base and change models}
var K, L: Integer;
var II: Integer;
var Factor: double;
begin 
Factor := (Year - Epoch1) / (Epoch2 - Epoch1);
if NMax1 = NMax2 then
	begin 
	K :=  NMax1  * (NMax1 + 2);
	Result := NMax1;
	end
else
	begin 
	if NMax1 > NMax2 then
		begin 
		K := NMax2  * (NMax2 + 2);
		L := NMax1  * (NMax1 + 2);
		for II := K + 1 to L do NewModel[II] := Model1[II] + Factor * (-Model1[II]);
		Result := NMax1;
		end
	else
		begin 
		K := NMax1  * (NMax1 + 2);
		L := NMax2  * (NMax2 + 2);
		for II := K + 1 to L do NewModel[II] := Factor  * Model2[II];
		Result := NMax2;
		end;
	end;

for II := 1 to K do NewModel[II] := Model1[II] + Factor * (Model2[II] - Model1[II]);
end;



function Extrapolate(Year, Epoch: double; NMax1, NMax2: Integer; BaseModel, ChangeModel: array of double; var NewModel: array of double) : Integer;
{Extrapolates linearly from a spherical harmonic model with a rate-of-change model.}
{Year = Year of new model, Epoch = Year of base model}
{NMax1, NMax2 = maximum degree and order of base and change models}
var K, L : Integer;
var II : Integer;
var Factor : double;
begin
Factor := Year - Epoch;
if NMax1 = NMax2 then
	begin 
	K :=  NMax1  * (NMax1 + 2);
	Result := NMax1;
	end
else
	begin 
	if NMax1 > NMax2 then
		begin 
		K := NMax2  * (NMax2 + 2);
		L := NMax1  * (NMax1 + 2);
		for II:= K + 1 to L do NewModel[II] := BaseModel[II];
		Result := NMax1;
		end
	else
	      begin 
	      K := NMax1  * (NMax1 + 2);
	      L := NMax2  * (NMax2 + 2);
	      for II := K + 1 to L do NewModel[II] := Factor  * ChangeModel[II];
	      Result := NMax2;
	      end;
	end;
for II := 1 to K do NewModel[II] := BaseModel[II] + Factor * ChangeModel[II];
end;



function ModelToField(Params: TMagneticParams; nmax: integer; gh: array of double; var X,Y,Z: double): Integer;
{Calculates field components from spherical harmonic models.}
{Based on models in gh. nmax = maximum degree and order of coefficients}
const EarthsRadius = 6371.2;
const MajorAxis = 40680925;	{square of semi-major axes of the reference spheroid}
const MinorAxis = 40408588;	{square of semi-minor axes of the reference spheroid}
var SinLat, CosLat: double;
var ratio: double;
var aa, bb, cc, dd: double;
var sd,cd: double;
var r, rr: double;
var fm, fn: double;
var sl,cl: array [0..14-1] of double;
var p,q: array [0..119-1] of double;
var ii,j,k,l,m,n: Integer;
var npq: Integer;
const iext=0; ext1=0; ext2=0; ext3=0;
begin
r := Params.Elevation;
SinLat := sin(Params.Latitude  * Degree2Radian);
if (90 - Params.Latitude) < 0.001 then aa := 89.999		{   300 ft. from North pole   }
else
	begin
	if (90 + Params.Latitude) < 0.001 then aa := -89.999	{   300 ft. from South pole   }
	else aa := Params.Latitude;
	end;

CosLat := cos(aa  * Degree2Radian);
sl[1] := sin(Params.Longitude  * Degree2Radian);
cl[1] := cos(Params.Longitude  * Degree2Radian);
X := 0; Y := 0; Z := 0;
sd := 0; cd := 1; l := 1;
n := 0; m := 1;
npq := (nmax  * (nmax + 3)) div 2;
if Params.CoordSystem = Geodetic then
	begin
	aa := MajorAxis  * CosLat  * CosLat;
	bb := MinorAxis  * SinLat  * SinLat;
	cc := aa + bb;
	dd := sqrt(cc);
	r := sqrt(Params.Elevation  * (Params.Elevation + 2  * dd) + (MajorAxis  * aa + MinorAxis  * bb) / cc);
	cd := (Params.Elevation + dd) / r;
	sd := (MajorAxis - MinorAxis) / dd * SinLat * CosLat / r;
	aa := SinLat;
	SinLat := SinLat  * cd - CosLat * sd;
	CosLat := CosLat  * cd + aa * sd;
	end;
ratio := EarthsRadius / r;
aa := sqrt(3);
p[1] := 2  * SinLat;
p[2] := 2  * CosLat;
p[3] := 4.5  * SinLat  * SinLat - 1.5;
p[4] := 3  * aa  * CosLat  * SinLat;
q[1] := -CosLat;
q[2] := SinLat;
q[3] := -3  * CosLat  * SinLat;
q[4] := aa  * (SinLat  * SinLat - CosLat  * CosLat);
for  k := 1 to npq do
	begin 
	if n < m then
		begin 
		m := 0;
		n := n + 1;
		rr := power(ratio,n + 2);
		fn := n;
		end;
	fm := m;
	if k >= 5 then
		begin 
		if m = n then
			begin 
			aa := sqrt(1 - 0.5/fm);
			j := k - n - 1;
			p[k] := (1 + 1/fm)  * aa  * CosLat  * p[j];
			q[k] := aa  * (CosLat  * q[j] + SinLat/fm  * p[j]);
			sl[m] := sl[m-1] * cl[1] + cl[m-1] * sl[1];
			cl[m] := cl[m-1] * cl[1] - sl[m-1] * sl[1];
			end
		else
			begin 
			aa := sqrt(fn *fn - fm*fm);
			bb := sqrt(((fn - 1) *(fn-1)) - (fm  * fm))/aa;
			cc := (2  * fn - 1)/aa;
			ii := k - n;
			j := k - 2 * n + 1;
			p[k] := (fn + 1)  * (cc  * SinLat/fn  * p[ii] - bb/(fn - 1)  * p[j]);
			q[k] := cc  * (SinLat  * q[ii] - CosLat/fn  * p[ii]) - bb * q[j];
			end;
		end;

	aa := rr  * gh[l];
	if m = 0 then
		begin 
		X := X + aa * q[k]; Z := Z - aa * p[k];
		l := l + 1;
		end
	else
		begin 
		bb := rr  * gh[l+1];
		cc := aa  * cl[m] + bb * sl[m];
		X := X + cc * q[k];
		Z := Z - cc * p[k];
		if CosLat > 0 then Y := Y + (aa  * sl[m] - bb  * cl[m])  * fm  * p[k]/((fn + 1) * CosLat)
		else Y := Y + (aa  * sl[m] - bb  * cl[m])  * q[k] * SinLat;
		l := l + 2;
		end;
	m := m + 1;
	end;
if iext <> 0 then
	begin 
	aa := ext2  * cl[1] + ext3 * sl[1];
	X := X - ext1 * CosLat + aa * SinLat;
	Y := Y + ext2 * sl[1] - ext3 * cl[1];
	Z := Z + ext1 * SinLat + aa * CosLat;
	end;

aa := X;
X := X  * cd + Z * sd;
Z := Z  * cd - aa * sd;
Result:=0;
end;



function CartesianToPolar(X,Y,Z: double; var Declin, Inclin, HorizonInt ,TotalInt: double): boolean;
{Converts X, Y, and Z geomagnetic components to Declination,}
{ Inclination,Horizontal Intensity, and Total Intensity.}
{ if field is too weak return false and Dec/Inc values of 999 degress}
var HPX: double;
const ErrorValue = 999 * Degree2Radian;
const LowerLimit = 0.0001;
begin
Result:=False;
Declin := ErrorValue;        {  Preset Dec and Inc in case there is an error  }
Inclin := ErrorValue; 


HorizonInt := sqrt(X*X + Y*Y);		{  calculate horizontal intensity  }
TotalInt := sqrt(X*X + Y*Y + Z*Z);	{  calculate total intensity  }
if TotalInt < LowerLimit then exit;		{ Exit if field is too weak to do calculation }

Inclin := ArcTan2(Z,HorizonInt);
if HorizonInt < LowerLimit then exit;		{ Exit if field is too weak to do calculation }

HPX := HorizonInt + X;
if HPX < LowerLimit then Declin := Pi else Declin := 2 * ArcTan2(Y,HPX);
Result:=True;
end;




function FindNearestModel(Year: double; var Inx: integer): boolean;
{Return the index of the model closest to the Year}
var I: integer;
begin
Result:=False;
{Test for out of range too low}
Inx:=0;
if Year<IGRFData[0].YearMin then exit;

{Test for out of range too high}
Inx:=MaxModels-1;
if Year>IGRFData[Inx].YearMax then exit;
for I:=0 to MaxModels-1 do
 if (Year>=IGRFData[I].YearMin) and (Year<=IGRFData[I].YearMax) then  break;
Inx:=I;
Result:=True;
end;


function CalculateMagnetics(Params: TMagneticParams; var MagneticData: TMagneticData): boolean;
{Calculate magnetics basic on values in Params. }
{The result is returned in MagneticData }
{Returns False if date is out of range }
var ModelInx: integer;
var NMax1,NMax2: integer;
var DecNext,IncNext: double;
var TotalNext,HorzNext : double;
var XNext,YNext,ZNext : double;
var gh1,gh2: array [0..ArrayMax] of double;
var gha,ghb: array [0..ArrayMax] of double;
begin
Result:=FindNearestModel(Params.Year,ModelInx);

{Check for annual change model (Secular variation)}
if (IGRFData[ModelInx].Max2 = 0)  then
	begin
	{Get model for starting and ending years}
	GetSphericalHarmonics(YearModel,  ModelInx, IGRFData[ModelInx].Max1, gh1);
	GetSphericalHarmonics(YearModel,  ModelInx+1, IGRFData[ModelInx+1].Max1, gh2);
	NMax1 := Interpolate(Params.Year, IGRFData[ModelInx].YearMin, IGRFData[ModelInx].Max1, IGRFData[ModelInx+1].YearMin, IGRFData[ModelInx+1].Max1, gh1,gh2, gha);
	NMax2 := Interpolate(Params.Year+1, IGRFData[ModelInx].YearMin , IGRFData[ModelInx].Max1, IGRFData[ModelInx+1].YearMin, IGRFData[ModelInx+1].Max1,gh1,gh2, ghb);
	end
 else
 	begin
	{Get model for start year and anual change}
 	GetSphericalHarmonics(YearModel,  ModelInx, IGRFData[ModelInx].Max1, gh1);
 	GetSphericalHarmonics(AnnualChange,  ModelInx, IGRFData[ModelInx].Max2, gh2);
 	NMax1 := Extrapolate(Params.Year, IGRFData[ModelInx].Epoch, IGRFData[ModelInx].Max1, IGRFData[ModelInx].Max2,gh1,gh2, gha);
	NMax2 := Extrapolate(Params.Year+1, IGRFData[ModelInx].Epoch, IGRFData[ModelInx].Max1, IGRFData[ModelInx].Max2,gh1,gh2, ghb);
 	end;

{ Do the calculations for the first model}
with Params,MagneticData do
 begin
 ModelToField(Params,NMax1,gha,XField,YField,ZField);
 CartesianToPolar(XField,YField,ZField,Decl,Incl,HorzField,TotalField);

{ Do the calculations for the second model}
 ModelToField(Params,NMax2,ghb,XNext,YNext,ZNext);
 CartesianToPolar(XNext,YNext,ZNext,DecNext,IncNext,HorzNext,TotalNext);
 end;

with MagneticData do
 begin
 DecChange := ((DecNext - Decl) * Radian2Degree) *60;
 IncChange := ((IncNext - Incl) * Radian2Degree) *60;
 Decl := Decl * Radian2Degree;
 Incl := Incl * Radian2Degree;
 HorZChange := HorzNext - HorzField;
 XChange := XNext - XField;
 YChange := YNext - YField;
 ZChange := ZNext - ZField;
 TotalChange := TotalNext - TotalField;
 end;

end;




end.
 