unit MatrixLibrary;

interface

uses Windows,SysUtils,Math,DSupport,Vector,GDIPObj;

{ 3x3 Matrix library for transforming 2D vectors}


type TX3Matrix = array [0..8] of double;
type TX3StandMatrix = array [1..3,1..3] of double;
type TX3ProgMatrix = array [0..2,0..2] of double;

const X3IdentityMatrix: TX3Matrix = (1,0,0,0,1,0,0,0,1);

function X3MatrixMultiply(const M1, M2: TX3Matrix): TX3Matrix;
function X3MatrixVectorMult(M: TX3Matrix; V: T2DVector): T2DVector;
function GetX3Determinate(M: TX3Matrix): double;
function InvertX3Matrix(M: TX3Matrix): TX3Matrix;

procedure SetX3Matrix(var M: TX3Matrix; A,B,C,D,E,F: double);
procedure AddX3Matrix(var M: TX3Matrix; M2: TX3Matrix);
procedure SetX3Translation(var M: TX3Matrix; Off: T2DVector);
procedure AddX3Translation(var M: TX3Matrix; Off: T2DVector);
procedure SetX3Scale(var M: TX3Matrix; Scale: T2DVector);
procedure AddX3Scale(var M: TX3Matrix; Scale: T2DVector);
procedure SetX3Rotation(var M: TX3Matrix; A: double); overload;
procedure SetX3Rotation(var M: TX3Matrix; A: double; Center: T2DVector); overload;
procedure AddX3Rotation(var M: TX3Matrix; A: double); overload;
procedure AddX3Rotation(var M: TX3Matrix; A: double; Center: T2DVector); overload;
procedure SetX3SkewX(var M: TX3Matrix; A: double);
procedure SetX3SkewY(var M: TX3Matrix; A: double);
procedure AddX3SkewX(var M: TX3Matrix; A: double);
procedure AddX3SkewY(var M: TX3Matrix; A: double);
function GetX3MatrixStr(M: TX3Matrix; Digits: integer): string;

procedure DecomposeMatrix3X(M: TX3Matrix; var Angle,ScaleX,ScaleY,SkewX,SkewY: double);

function Convert2X3ToX3(MA: TMatrixArray): TX3Matrix;
function XFormToX3(XF: TXForm): TX3Matrix;
function X3ToXForm(MX3: TX3Matrix): TXForm;


implementation


function Convert2X3ToX3(MA: TMatrixArray): TX3Matrix;
begin
SetX3Matrix(Result, MA[0],MA[1],MA[2],MA[3],MA[4],MA[5]);
end;

{
 [1,1] [2,1] [3,1]
 [1,2] [2,2] [3,2]
 [1,3] [2,3] [3,3]
}


function XFormToX3(XF: TXForm): TX3Matrix;
begin
TX3StandMatrix(Result)[1,1]:=XF.eM11;
TX3StandMatrix(Result)[1,2]:=XF.eM12;
TX3StandMatrix(Result)[2,1]:=XF.eM21;
TX3StandMatrix(Result)[2,2]:=XF.eM22;
TX3StandMatrix(Result)[3,1]:=XF.eDx;
TX3StandMatrix(Result)[3,2]:=XF.eDy;
TX3StandMatrix(Result)[1,3]:=0;
TX3StandMatrix(Result)[2,3]:=0;
TX3StandMatrix(Result)[3,3]:=1;
end;


function X3ToXForm(MX3: TX3Matrix): TXForm;
begin
Result.eM11:=TX3StandMatrix(MX3)[1,1];
Result.eM12:=TX3StandMatrix(MX3)[2,1];
Result.eM21:=TX3StandMatrix(MX3)[1,2];
Result.eM22:=TX3StandMatrix(MX3)[2,2];
Result.eDx:=TX3StandMatrix(MX3)[1,3];
Result.eDy:=TX3StandMatrix(MX3)[2,3];
end;



function X3MatrixMultiply(const M1, M2: TX3Matrix): TX3Matrix;
var I, J: Integer;
begin
for I := 0 to 2 do
 for J := 0 to 2 do
	begin
	TX3ProgMatrix(Result)[I, J] :=
		TX3ProgMatrix(M1)[I, 0] * TX3ProgMatrix(M2)[0, J] +
		TX3ProgMatrix(M1)[I, 1] * TX3ProgMatrix(M2)[1, J] +
		TX3ProgMatrix(M1)[I, 2] * TX3ProgMatrix(M2)[2, J];
	end;
end;



function X3MatrixVectorMult(M: TX3Matrix; V: T2DVector): T2DVector;
{Matrix-Vector multiply for column vector memory storage }
{The Format by DirectX and OpenGL  }
begin
Result.X:= V.X * M[0] + V.Y * M[1] + M[2];
Result.Y:= V.X * M[3] + V.Y * M[4] + M[5];
end;



function GetX3Determinate(M: TX3Matrix): double;
{Find matrix determinate}
begin
Result:=M[0]*M[4]*M[8] - M[0]*M[5]*M[7] - M[1]*M[3]*M[8] +
	M[1]*M[5]*M[6] + M[2]*M[3]*M[7] - M[2]*M[4]*M[6];
end;



function InvertX3Matrix(M: TX3Matrix): TX3Matrix;
var Det: double;
begin
Det := M[0]*M[4]*M[8] + M[3]*M[7]*M[2] + M[6]*M[1]*M[5] - M[0]*M[7]*M[5] - M[3]*M[1]*M[8] - M[6]*M[4]*M[2];
Result[0] := (M[4]*M[8] - M[7]*M[5])/Det;
Result[3] := (M[6]*M[5] - M[3]*M[8])/Det;
Result[6] := (M[3]*M[7] - M[6]*M[4])/Det;
Result[1] := (M[7]*M[2] - M[1]*M[8])/Det;
Result[4] := (M[0]*M[8] - M[6]*M[2])/Det;
Result[7] := (M[6]*M[1] - M[0]*M[7])/Det;
Result[2] := (M[1]*M[5] - M[4]*M[2])/Det;
Result[5] := (M[3]*M[2] - M[0]*M[5])/Det;
Result[8] := (M[0]*M[4] - M[3]*M[1])/Det;
end;



procedure SetX3Matrix(var M: TX3Matrix; A,B,C,D,E,F: double);
begin
M[0]:=A; M[3]:=B; M[1]:=C;
M[4]:=D; M[2]:=E; M[5]:=F;
M[6]:=0; M[7]:=0; M[8]:=1;
end;



procedure AddX3Matrix(var M: TX3Matrix; M2: TX3Matrix);
begin
M:=X3MatrixMultiply(M,M2);
end;



procedure SetX3Translation(var M: TX3Matrix; Off: T2DVector);
begin
M[2]:=Off.X; M[5]:=Off.Y;
end;


procedure AddX3Translation(var M: TX3Matrix; Off: T2DVector);
var M2: TX3Matrix;
begin
M2:=X3IdentityMatrix;
SetX3Translation(M2,Off);
M:=X3MatrixMultiply(M,M2);
end;

procedure SetX3Scale(var M: TX3Matrix; Scale: T2DVector);
begin
M[0]:=Scale.X; M[4]:=Scale.Y;
end;


procedure AddX3Scale(var M: TX3Matrix; Scale: T2DVector);
var M2: TX3Matrix;
begin
M2:=X3IdentityMatrix;
SetX3Scale(M2,Scale);
M:=X3MatrixMultiply(M,M2);
end;


procedure SetX3Rotation(var M: TX3Matrix; A: double); overload;
begin
A:=Degree2Radian * A;
SetX3Matrix(M, cos(A),sin(A),-sin(A),cos(A),0,0);
end;



procedure SetX3Rotation(var M: TX3Matrix; A: double; Center: T2DVector); overload;
begin
SetX3Translation(M,Center);
SetX3Rotation(M,A);
SetX3Translation(M,MakeVector2D(-Center.X,-Center.Y));
end;



procedure AddX3Rotation(var M: TX3Matrix; A: double); overload;
var M2: TX3Matrix;
begin
M2:=X3IdentityMatrix;
SetX3Rotation(M2,A);
M:=X3MatrixMultiply(M,M2);
end;




procedure AddX3Rotation(var M: TX3Matrix; A: double; Center: T2DVector); overload;
var M2: TX3Matrix;
begin
M2:=X3IdentityMatrix;
SetX3Rotation(M2,A,Center);
M:=X3MatrixMultiply(M,M2);
end;


procedure SetX3SkewX(var M: TX3Matrix; A: double);
begin
A:=Degree2Radian * A;
SetX3Matrix(M,1,0,tan(A),1,0,0);
end;



procedure SetX3SkewY(var M: TX3Matrix; A: double);
begin
A:=Degree2Radian * A;
SetX3Matrix(M,1,tan(a),0,1,0,0);
end;




procedure AddX3SkewX(var M: TX3Matrix; A: double);
var M2: TX3Matrix;
begin
M2:=X3IdentityMatrix;
SetX3SkewX(M2,A);
M:=X3MatrixMultiply(M,M2);
end;



procedure AddX3SkewY(var M: TX3Matrix; A: double);
var M2: TX3Matrix;
begin
M2:=X3IdentityMatrix;
SetX3SkewY(M2,A);
M:=X3MatrixMultiply(M,M2);
end;



function GetX3MatrixStr(M: TX3Matrix; Digits: integer): string;
{Displays matrix as column vectors which is the X3 text format}
{Note: this display is oposite of the way it appears in memory. See notes at the top}
var X,Y: integer;
var S: string;
begin
Result:='';
for Y:=0 to 2 do
	begin
	S:='';
	for X:=0 to 2 do
		begin
		if X<>0 then S:=S+' ';
		S:=S+FloatToStrF(M[Y*3+X],ffFixed,18,Digits);
		end;
	Result:=Result+S;
	if Y<2 then Result:=Result+CRLF_Char;
	end;
end;




procedure DecomposeMatrix3X(M: TX3Matrix; var Angle,ScaleX,ScaleY,SkewX,SkewY: double);
var Denom: double;
begin
{Length of collumn vectors = Scale}
ScaleX := sqrt(sqr(M[0]) + sqr(M[3]));
ScaleY := sqrt(sqr(M[1]) + sqr(M[4]));

{Remove scaling by dividing each column by its length}
M[0]:=M[0]/ScaleX; M[3]:=M[3]/ScaleX;
M[1]:=M[1]/ScaleY; M[4]:=M[4]/ScaleY;

Angle := ArcTan2(M[1], M[0]);


SkewX := ArcTan2(M[4], M[3]) - Pi;
SkewY := ArcTan2(M[1], M[0]);
end;


end.
