unit GLCustomMesh;

interface

uses Windows,SysUtils,Classes,Graphics,Contnrs, Vector, Hash, GLLibrary, GLOctTree, GLRenderer;

{Information for the adjacent edge routine}

type TEdgeInfo = record
 Inx: integer;
 Edge: integer;
 end;

{Defines call back progress routine for mesh building process}

type TGLProgress= procedure(Sender: TObject; Action: string; PercentDone: integer);


type TGLCustomMesh = class(TGLRenderer)
 private
  FOnChange: TNotifyEvent;
  FOnNormalProgress: TGLProgress;
  FShowNormals: boolean;
  FNormalColor: TColor;
  FShowFaceNormals: boolean;
  FFaceNormalColor: TColor;
    function GetWholeFaceColor(Index: integer): TColor;
    procedure SetWholeFaceColor(Index: integer; const Value: TColor);
 protected
  procedure Change; virtual;
  procedure DisplayProgress(Status: string; Progress, Total: integer);
  function GetVertex(Index: integer): T3DVector; virtual;
  procedure SetVertex(Index: integer; const Value: T3DVector); virtual;
  function GetVertexColor(Index: integer): TColor; virtual;
  procedure SetVertexColor(Index: integer; const Value: TColor); virtual;
  function GetVertexCount: integer; virtual;
  procedure SetVertexCount(const Value: integer); virtual;
  function GetVertexNormal(Index: integer): T3DVector; virtual;
  procedure SetVertexNormal(Index: integer; const Value: T3DVector); virtual;
  function GetFaceCount: integer; virtual;
  function GetFaceTriangles(Index: integer): TTriangle; virtual;
  procedure SetFaceTriangles(Index: integer; const Value: TTriangle); virtual;
  function GetGLColor(Index: integer): TGLFaceColor; virtual;
  procedure SetGLColor(Index: integer; const Value: TGLFaceColor); virtual;
  function GetFaceNormal(Index: integer): T3DVector; virtual;
  procedure SetFaceNormal(Index: integer; const Value: T3DVector); virtual;
 public
  MeshBuilt: boolean;
  TextureBuilt: boolean;
  OctalTree: TOctTree;
  constructor Create; override;
  destructor Destroy; override;
  procedure Assign(GLObject: TGLObject); override;
  procedure Append(GLObject: TGLObject); override;
  procedure PrepareTexture; virtual;
  procedure CalculateMinMax; virtual;

  procedure SetAllNormals(DoVertex: boolean); overload; virtual;
  procedure SetAllNormals; overload; virtual;
  procedure SetAllNormals(DoVertex: boolean; CreaseAngle: double); overload; virtual;
  procedure SetAllNormals(CreaseAngle: double); overload; virtual;

  procedure SetAdjacentNormals(Target: T3DVector; CreaseAngle: double); virtual;
  procedure RemoveDegenerateTriangles; virtual;
  procedure RemoveAdjacentDegenerates(Target: T3DVector); virtual;
  function RemoveTriangle(I: integer): T3DVector; virtual;
  procedure FindAdjacentEdge(L: T3DLine; Inx: integer; var Edge: TEdgeInfo); virtual;

  function AddTriangle(T: TTriangle; GLC: TGLColor): integer; overload; virtual;
  function AddTriangle(T: TTriangle; C: TColor): integer; overload; virtual;
  function AddTriangle(P1,P2,P3: T3DVector; C: TColor): integer; overload; virtual;
  procedure AddRectangle(X1,Y1,Z1,X2,Y2,Z2,X3,Y3,Z3,X4,Y4,Z4: double; C:TColor); overload; virtual;
  procedure AddRectangle(P1,P2: T3DVector; C: TColor); overload; virtual;
  procedure AddRectangle(P1,P2,P3,P4: T3DVector; C: TColor); overload; virtual;
  procedure AddRectangle(R: T3DRectangle; C: TColor); overload; virtual;
  procedure AddPrism(P1,P2,P3,P4,P5,P6,P7,P8: T3DVector; C: TColor); overload; virtual;
  procedure AddPrism(Min,Max: T3DVector; C: TColor); overload; virtual;
  procedure AddPrism(Min,Max: T3DVector; XCount,YCount,ZCount: integer; C: TColor); overload; virtual;
  procedure AddHexagon(P1, P2, P3, P4, P5, P6: T3DVector; C: TColor); virtual;
  procedure AddPentagon(P1, P2, P3, P4, P5: T3DVector; C: TColor); virtual;
  procedure DeleteTriangle(Index: integer); virtual;
  procedure NormalizeScale; virtual;
  procedure CenterMesh; virtual;
  procedure MeshClear; virtual;

  procedure Clear; override;
  procedure ClearData; virtual;
  procedure Render; override;
  procedure SetColor(const Color: TColor); override;
  function GetColor: TColor; override;
  procedure SetFaceColors(Inx: integer; C1,C2,C3: TColor); virtual;
  function FaceToTriangle(FaceInx: integer): TTriangle; virtual;
  procedure FindAdjacentVertices(Old: T3DVector; var VA: TIntegerHashTable); virtual;
  function GetFaceIndices(Index: integer): TFace; virtual;
  procedure SetFaceIndices(Index: integer; const Value: TFace); virtual;
  function FindFirstDuplicate(Target: T3DVector): integer; virtual;
  procedure FindDuplicateVertices(Target: T3DVector; var VA: TIntegerHashTable); virtual;
  procedure ChangeAllDuplicateVertices(Old,New: T3DVector); virtual;

  property RenderMode;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property OnNormalProgress: TGLProgress read FOnNormalProgress write FOnNormalProgress;
  property ShowNormals: boolean read FShowNormals write FShowNormals;
  property NormalColor: TColor read FNormalColor write FNormalColor;
  property ShowFaceNormals: boolean read FShowFaceNormals write FShowFaceNormals;
  property FaceNormalColor: TColor read FFaceNormalColor write FFaceNormalColor;
  property Vertex[Index: integer]: T3DVector read GetVertex write SetVertex;
  property VertexColor[Index: integer]: TColor read GetVertexColor write SetVertexColor;
  property VertexCount: integer read GetVertexCount write SetVertexCount;
  property VertexNormal[Index: integer]: T3DVector read GetVertexNormal write SetVertexNormal;
  property FaceCount: integer read GetFaceCount;
  property FaceTriangles[Index: integer]: TTriangle read GetFaceTriangles write SetFaceTriangles;
  property FaceColors[Index: integer]: TGLFaceColor read GetGLColor write SetGLColor;
  property WholeFaceColor[Index: integer]: TColor read GetWholeFaceColor write SetWholeFaceColor;
  property FaceNormal[Index: integer]: T3DVector read GetFaceNormal write SetFaceNormal;
  property FaceIndices[Index: integer]: TFace read GetFaceIndices write SetFaceIndices;
 end;

{ TMeshList }


type  TGLCustomMeshList = class(TGLObjectList)
  private
    function GetTriangles: integer;
  protected
    function GetItem(Index: Integer): TGLCustomMesh;
    procedure SetItem(Index: Integer; AObject: TGLCustomMesh);
  public
    function Add(var Mesh: TGLCustomMesh): integer; overload;
    function Add: integer; overload;
    function AddFastMesh: integer;
    function Insert(Inx: integer; var Mesh: TGLCustomMesh): integer; overload;
    function Insert(Inx: integer): integer; overload;
    function InsertFastMesh(Inx: integer): integer;
    property Items[Index: Integer]: TGLCustomMesh read GetItem write SetItem; default;
    procedure Assign(ObjectList: TObjectList);
    property Triangles: integer read GetTriangles;
  end;

implementation

uses GLMesh,GLFastMesh;

{ TGLCustomMesh }


constructor TGLCustomMesh.Create;
begin
inherited;
OctalTree:=TOctTree.Create;
MeshBuilt:=False;
TextureBuilt:=False;
FShowNormals:=False;
FNormalColor:=clLime;
FShowFaceNormals:=False;
FFaceNormalColor:=clYellow;
end;




destructor TGLCustomMesh.Destroy;
begin
OctalTree.Free;
inherited;
end;



procedure TGLCustomMesh.SetAllNormals;
begin
end;


procedure TGLCustomMesh.SetAllNormals(DoVertex: boolean);
begin
end;


procedure TGLCustomMesh.FindAdjacentEdge(L: T3DLine; Inx: integer;
  var Edge: TEdgeInfo);
begin
end;


procedure TGLCustomMesh.RemoveAdjacentDegenerates(Target: T3DVector);
begin
end;


procedure TGLCustomMesh.RemoveDegenerateTriangles;
begin
end;


function TGLCustomMesh.RemoveTriangle(I: integer): T3DVector;
begin
end;


procedure TGLCustomMesh.SetAdjacentNormals(Target: T3DVector;
  CreaseAngle: double);
begin
end;


procedure TGLCustomMesh.SetAllNormals(CreaseAngle: double);
begin
end;


procedure TGLCustomMesh.SetAllNormals(DoVertex: boolean;
  CreaseAngle: double);
begin
end;


procedure TGLCustomMesh.AddHexagon(P1, P2, P3, P4, P5, P6: T3DVector;
  C: TColor);
begin
end;


procedure TGLCustomMesh.AddPentagon(P1, P2, P3, P4, P5: T3DVector;
  C: TColor);
begin
end;


procedure TGLCustomMesh.AddPrism(P1, P2, P3, P4, P5, P6, P7, P8: T3DVector;
  C: TColor);
begin
end;


procedure TGLCustomMesh.AddPrism(Min, Max: T3DVector; XCount, YCount,
  ZCount: integer; C: TColor);
begin
end;


procedure TGLCustomMesh.AddPrism(Min, Max: T3DVector; C: TColor);
begin
end;


procedure TGLCustomMesh.AddRectangle(P1, P2: T3DVector; C: TColor);
begin
end;


procedure TGLCustomMesh.AddRectangle(P1, P2, P3, P4: T3DVector; C: TColor);
begin
end;


procedure TGLCustomMesh.AddRectangle(R: T3DRectangle; C: TColor);
begin
end;


procedure TGLCustomMesh.AddRectangle(X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3,
  X4, Y4, Z4: double; C: TColor);
begin
end;


function TGLCustomMesh.AddTriangle(T: TTriangle; C: TColor): integer;
begin
end;


function TGLCustomMesh.AddTriangle(P1, P2, P3: T3DVector; C: TColor): integer;
begin
end;




procedure TGLCustomMesh.DeleteTriangle(Index: integer);
begin

end;



procedure TGLCustomMesh.CenterMesh;
begin
end;


procedure TGLCustomMesh.MeshClear;
begin
end;


procedure TGLCustomMesh.NormalizeScale;
begin
end;



procedure TGLCustomMesh.CalculateMinMax;
begin
end;


function TGLCustomMesh.AddTriangle(T: TTriangle; GLC: TGLColor): integer;
begin
end;


procedure TGLCustomMesh.Append(GLObject: TGLObject);
begin
inherited;
end;


procedure TGLCustomMesh.Assign(GLObject: TGLObject);
begin
inherited;
end;


procedure TGLCustomMesh.Clear;
begin
inherited;
end;

procedure TGLCustomMesh.ClearData;
begin
end;

procedure TGLCustomMesh.Render;
begin
inherited Render;
end;

procedure TGLCustomMesh.Change;
begin
end;


function TGLCustomMesh.GetColor: TColor;
begin
Result:=inherited GetColor;
end;


procedure TGLCustomMesh.SetColor(const Color: TColor);
begin
inherited;
end;


procedure TGLCustomMesh.SetFaceColors(Inx: integer; C1, C2, C3: TColor);
begin
end;


procedure TGLCustomMesh.DisplayProgress(Status: string; Progress,
  Total: integer);
{This routine calls 'progress" callback routine if the user has set one up}
begin
if not Assigned(FOnNormalProgress) then exit;
if (Progress mod (Total div 100 + 1))<>0 then exit;
FOnNormalProgress(Self,Status,MulDiv(100,Progress,Total));
end;



function TGLCustomMesh.FaceToTriangle(FaceInx: integer): TTriangle;
begin
end;


function TGLCustomMesh.GetVertex(Index: integer): T3DVector;
begin
end;


procedure TGLCustomMesh.SetVertex(Index: integer; const Value: T3DVector);
begin
end;


function TGLCustomMesh.GetVertexColor(Index: integer): TColor;
begin
end;


procedure TGLCustomMesh.SetVertexColor(Index: integer;
  const Value: TColor);
begin
end;


function TGLCustomMesh.GetVertexCount: integer;
begin
end;


procedure TGLCustomMesh.SetVertexCount(const Value: integer);
begin
end;


function TGLCustomMesh.GetVertexNormal(Index: integer): T3DVector;
begin
end;


procedure TGLCustomMesh.SetVertexNormal(Index: integer;
  const Value: T3DVector);
begin
end;


function TGLCustomMesh.GetFaceCount: integer;
begin
end;


function TGLCustomMesh.GetFaceTriangles(Index: integer): TTriangle;
begin
end;


procedure TGLCustomMesh.SetFaceTriangles(Index: integer;
  const Value: TTriangle);
begin
end;


function TGLCustomMesh.GetGLColor(Index: integer): TGLFaceColor;
begin
end;


procedure TGLCustomMesh.SetGLColor(Index: integer;
  const Value: TGLFaceColor);
begin
end;


function TGLCustomMesh.GetFaceNormal(Index: integer): T3DVector;
begin
end;


procedure TGLCustomMesh.SetFaceNormal(Index: integer;
  const Value: T3DVector);
begin
end;


procedure TGLCustomMesh.FindAdjacentVertices(Old: T3DVector;
  var VA: TIntegerHashTable);
begin
end;



function TGLCustomMesh.GetFaceIndices(Index: integer): TFace;
begin
end;



procedure TGLCustomMesh.SetFaceIndices(Index: integer; const Value: TFace);
begin
end;



function TGLCustomMesh.GetWholeFaceColor(Index: integer): TColor;
begin
Result:=FaceColors[Index].C1;
end;


procedure TGLCustomMesh.SetWholeFaceColor(Index: integer; const Value: TColor);
var GLC: TGLFaceColor;
begin
GLC.C1:=Value;
GLC.C2:=Value;
GLC.C3:=Value;
FaceColors[Index]:=GLC;
end;



function TGLCustomMesh.FindFirstDuplicate(Target: T3DVector): integer;
begin
end;



procedure TGLCustomMesh.ChangeAllDuplicateVertices(Old, New: T3DVector);
begin
end;



procedure TGLCustomMesh.FindDuplicateVertices(Target: T3DVector;
  var VA: TIntegerHashTable);
begin
end;



procedure TGLCustomMesh.PrepareTexture;
begin
end;



{ TMeshList }



function TGLCustomMeshList.Add(var Mesh: TGLCustomMesh): integer;
begin
if Mesh is TGLMesh then Mesh:=TGLMesh.Create
else Mesh:=TGLFastMesh.Create;
Mesh.Name:='Mesh-'+IntToStr(Count);
Result:=inherited Add(Mesh);
end;




function TGLCustomMeshList.Add: integer;
var Mesh: TGLMesh;
begin
Result:=Add(TGLCustomMesh(Mesh));
end;




function TGLCustomMeshList.AddFastMesh: integer;
var Mesh: TGLFastMesh;
begin
Result:=Add(Mesh);
end;



function TGLCustomMeshList.Insert(Inx: integer; var Mesh: TGLCustomMesh): integer;
begin
if Mesh is TGLMesh then Mesh:=TGLMesh.Create
else Mesh:=TGLFastMesh.Create;
Mesh.Name:='Mesh-'+IntToStr(Count);
inherited Insert(Inx,Mesh);
Result:=Inx;
end;



function TGLCustomMeshList.Insert(Inx: integer): integer;
var Mesh: TGLCustomMesh;
begin
Result:=Insert(Inx,Mesh);
end;


function TGLCustomMeshList.InsertFastMesh(Inx: integer): integer;
var Mesh: TGLFastMesh;
begin
Result:=Insert(Inx,TGLCustomMesh(Mesh));
end;



procedure TGLCustomMeshList.Assign(ObjectList: TObjectList);
var I: integer;
var O: TGLCustomMesh;
begin
if ObjectList is TGLCustomMeshList then
	begin
	Clear;
	for I:=0 to ObjectList.Count-1 do
		begin
		Add(O);
		O.Assign(TGLCustomMesh(ObjectList[I]));
		end;
	end
else inherited Assign(ObjectList);
end;




function TGLCustomMeshList.GetItem(Index: Integer): TGLCustomMesh;
begin
Result := TGLCustomMesh(inherited GetItem(Index));
end;



function TGLCustomMeshList.GetTriangles: integer;
var I: integer;
begin
Result:=0;
for I:=0 to Count-1 do Result:=Result + Self[I].FaceCount;
end;



procedure TGLCustomMeshList.SetItem(Index: Integer; AObject: TGLCustomMesh);
begin
{This is done so user can replace}
inherited SetItem(Index, AObject);
end;


end.
