unit GLFont2D;

interface

uses Forms,Windows,Classes,SysUtils,Graphics,Contnrs,Vector,
     OpenGL15,GLFonts,GLRenderer,GLLibrary,GLMatrix,VCLLibrary;


type TGLStringMode = (smPixel,smWorld,smWorldDepth);




type TGLStringInfo = class(TObject)
 private
 protected
 public
  Pos: T3DVector;
  Mode: TGLStringMode;
 end;



type TGL2DString = class(TGLRenderer)
 private
  FStrings: TAutoString;
  FFontBuilt: boolean;
  FFontDisplayList: GLuint;
  FDelphiFont: TFont;
  FOnChange: TNotifyEvent;
  FViewPort: TVector4i;
  FModelView: TGLMatrixd4;
  FProjection: TGLMatrixd4;

  procedure SetFont(F: TFont);
  procedure Change;
  procedure HandleFontChange(Sender: TObject);
    function GetString(Index: integer): string;
    function GetStringInfo(Index: integer): TGLStringInfo;
    procedure SetString(Index: integer; const Value: string);
    procedure SetStringInfo(Index: integer; const Value: TGLStringInfo);
    function GetMode(Index: integer): TGLStringMode;
    function GetPosition(Index: integer): T3DVector;
    procedure SetMode(Index: integer; const Value: TGLStringMode);
    procedure SetPosition(Index: integer; const Value: T3DVector);
    procedure DrawWorld(Pos: T3DVector; text: string);
    procedure WorldStringDepth(Pos: T3DVector; S: string);
    procedure GetCurrentMatrices;
    function ScreenToWorld(V: T3DVector): T3DVector;
    function WorldToScreen(V: T3DVector): T3DVector;
 public
  function Add(Text: string; Pos: T3DVector; Mode: TGLStringMode): integer;
  procedure Delete(Index: integer);
  procedure Clear;
  property Text[Index: integer]: string read GetString write SetString; default;
  property Position[Index: integer]: T3DVector read GetPosition write SetPosition;
  property Mode[Index: integer]: TGLStringMode read GetMode write SetMode;
  constructor Create;
  constructor CreateFrom(Font: TGL2DString);
  destructor Destroy; override;
  procedure DrawFixedPixel(X, Y, Z: double; Text: string);
  procedure PixelStringOut(Pos: T3DVector; S: string);
  procedure WorldStringOut(Pos: T3DVector; S: string);
  procedure DrawAllStrings;
  procedure Render; override;
  procedure BuildFont(DC: HDC; Font: TFont); overload;
  procedure BuildFont; overload;
  procedure Assign(AObject: TGLObject);
  property Font: TFont read FDelphiFont write SetFont;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
 end;


type  TGLStringList2D = class(TGLObjectList)
  private
    procedure HandleStringChange(Sender: TObject);
  protected
    function GetItem(Index: Integer): TGL2DString;
    procedure SetItem(Index: Integer; AObject: TGL2DString);
  public
    function Add(var Str: TGL2DString): integer; overload;
    function Add: integer; overload;
    function AddFrom(Str: TGL2DString): integer;
    function Insert(Inx: integer; var Str: TGL2DString): integer; overload;
    function Insert(Inx: integer): integer; overload;
    property Items[Index: Integer]: TGL2DString read GetItem write SetItem; default;
    procedure Assign(ObjectList: TObjectList);
  end;


implementation


{ TGL2DString }


procedure TGL2DString.Change;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;




procedure TGL2DString.HandleFontChange(Sender: TObject);
begin
FFontBuilt:=False;
Change;
end;


procedure TGL2DString.BuildFont(DC: HDC; Font: TFont);
var FontHandle,Old: HFONT;
begin
FontHandle:=Font.Handle;
Old:=SelectObject(DC,FontHandle);
SetTextAlign(DC, TA_BASELINE);
FFontBuilt:=wglUseFontBitmaps(DC,0,256,FFontDisplayList);
SelectObject(DC,Old);
{Don't use DeleteObject to delete FontHandle here - it belongs to Delphi}
Change;
FFontBuilt:=True;
end;



procedure TGL2DString.BuildFont;
begin
BuildFont(wglGetCurrentDC,FDelphiFont);
end;



constructor TGL2DString.Create;
begin
inherited;
FStrings:=TAutoString.Create(TGLStringInfo);
FFontBuilt:=False;
FDelphiFont:=TFont.Create;
FDelphiFont.Height:=16;
FDelphiFont.Style:=[fsBold];
FDelphiFont.Charset:=ANSI_CHARSET;
FDelphiFont.Name:='Arial';
FDelphiFont.Color:=clLime;
FFontDisplayList:=glGenLists(256);
FDelphiFont.OnChange:=HandleFontChange;
BuildFont;
end;



constructor TGL2DString.CreateFrom(Font: TGL2DString);
begin
inherited;
Create;
Self.Assign(Font);
end;



destructor TGL2DString.Destroy;
begin
FStrings.Free;
FDelphiFont.Free;
inherited;
end;


{------------------------------------------------------------------------------}





procedure TGL2DString.GetCurrentMatrices;
{Note: loading these matrices are slow - so do it only onece }
begin
glGetDoublev(GL_MODELVIEW_MATRIX, @FModelview);
glGetDoublev(GL_PROJECTION_MATRIX, @FProjection);
glGetIntegerv(GL_VIEWPORT, @FViewport);
end;




function TGL2DString.ScreenToWorld(V: T3DVector): T3DVector;
{Fast version of the subroutine in GLLibrary}
begin
gluUnProject(V.X, FViewport[3] - V.Y, V.Z, FModelView, FProjection, FViewPort, @Result.X, @Result.Y, @Result.Z);
end;




function TGL2DString.WorldToScreen(V: T3DVector): T3DVector;
{Fast version of the subroutine in GLLibrary}
begin
gluProject(V.X, V.Y, V.Z, FModelView, FProjection, FViewPort, @Result.X, @Result.Y, @Result.Z);
Result.Y:=FViewPort[3]-Result.Y;
end;



procedure TGL2DString.DrawWorld(Pos: T3DVector; Text: string);
{Draw string at specified world coordinates}
var C3F: TGLArrayf3;
begin
glDisable(GL_LIGHTING);
{Color must come before setting raster position}
C3F:=ColorToGLRGB(Font.Color);
glColor3f(C3F[0],C3F[1],C3F[2]);
glRasterPos3f(Pos.X,Pos.Y,Pos.Z);
glListBase(FFontDisplayList);
glCallLists(length(text),GL_UNSIGNED_BYTE,PChar(text));
glEnable(GL_LIGHTING);
end;



procedure TGL2DString.DrawFixedPixel(X,Y,Z: double; Text: string);
{Draw text in fixed pixel position relative to the upper left corner}
var R: TRect;
var C3F: TGLArrayf3;
begin
glGetIntegerv(GL_VIEWPORT, @R);
glMatrixMode(GL_PROJECTION);
glPushMatrix;

glLoadIdentity;
gluOrtho2D(R.Left, R.Right, R.Bottom, R.Top);

glMatrixMode(GL_MODELVIEW);
glPushMatrix;
glLoadIdentity;

glDisable(GL_LIGHTING);
{Color must come before setting raster position}
C3F:=ColorToGLRGB(Font.Color);
glColor3f(C3F[0],C3F[1],C3F[2]);
glRasterPos3d(X,Y+Abs(FDelphiFont.Height),Z);
glListBase(FFontDisplayList);
glCallLists(length(Text),GL_UNSIGNED_BYTE,PChar(Text));

glMatrixMode(GL_PROJECTION);
glPopMatrix;

glEnable(GL_LIGHTING);
glMatrixMode(GL_MODELVIEW);
glPopMatrix;
end;



procedure TGL2DString.PixelStringOut(Pos: T3DVector; S: string);
{Output strings set at starting at specified pixel postion}
{Handles returns chars by moving to the next line}
var SL: TStringList;
var I: integer;
var Y2: double;
var S1: string;
begin
SL:=TStringList.Create;
try
SL.Clear;
{Caption separate lines in string list}
SL.Text:=S;
for I:=0 to SL.Count-1 do
	begin
	{Add the pixel height to each new line}
	Y2:=Pos.Y + I * Abs(FDelphiFont.Height);
	DrawFixedPixel(Pos.X,Y2,Pos.Z,SL[I]);
	end;
finally SL.Free; end;
end;





procedure TGL2DString.WorldStringDepth(Pos: T3DVector; S: string);
{Output strings set at starting at specified world postion}
{Handles returns chars by moving to the next line}
var V: T3DVector;
var SL: TStringList;
var I: integer;
begin
SL:=TStringList.Create;
try
SL.Clear;
{Caption separate lines in string list}
SL.Text:=S;
for I:=0 to SL.Count-1 do
	begin
	V:=Pos;
	V:=WorldToScreen(Pos);
	V.Y:=V.Y + (I+1) * Abs(FDelphiFont.Height);
	V:=ScreenToWorld(V);
	DrawWorld(V,SL[I]);
	end;
finally SL.Free; end;
end;




procedure TGL2DString.WorldStringOut(Pos: T3DVector; S: string);
var V: T3DVector;
begin
V:=WorldToScreen(Pos);
PixelStringOut(V,S);
end;




procedure TGL2DString.DrawAllStrings;
var I: integer;
begin
BuildFont;
GetCurrentMatrices;
for I:=0 to FStrings.Count-1 do
	begin
	case Mode[I] of
	 smWorld: WorldStringOut(Position[I],FStrings[I]);
	 smWorldDepth: WorldStringDepth(Position[I],FStrings[I]);
	 smPixel: PixelStringOut(Position[I],FStrings[I]);
	 end;
	end;
end;




procedure TGL2DString.SetFont(F: TFont);
begin
FDelphiFont.Assign(F);
FFontBuilt:=False;
end;




procedure TGL2DString.Assign(AObject: TGLObject);
begin
inherited Assign(AObject);
if AObject is TGL2DString then
	begin
	FDelphiFont.Assign(TGL2DString(AObject).FDelphiFont);
	FFontBuilt:=False;
	end;
end;



function TGL2DString.Add(Text: string; Pos: T3DVector; Mode: TGLStringMode): integer;
begin
Result:=FStrings.Add(Text);
TGLStringInfo(FStrings.Objects[Result]).Pos:=Pos;
TGLStringInfo(FStrings.Objects[Result]).Mode:=Mode;
end;



procedure TGL2DString.Clear;
begin
FStrings.Clear;
end;



procedure TGL2DString.Delete(Index: integer);
begin
FStrings.Delete(Index);
end;



function TGL2DString.GetStringInfo(Index: integer): TGLStringInfo;
begin
Result:=TGLStringInfo(FStrings.Objects[Index]);
end;



procedure TGL2DString.SetStringInfo(Index: integer;
  const Value: TGLStringInfo);
begin
TGLStringInfo(FStrings.Objects[Index]).Pos:=Value.Pos;
end;


function TGL2DString.GetString(Index: integer): string;
begin
Result:=FStrings[Index];
end;



procedure TGL2DString.SetString(Index: integer; const Value: string);
begin
FStrings[Index]:=Value;
end;



function TGL2DString.GetMode(Index: integer): TGLStringMode;
begin
Result:=TGLStringInfo(FStrings.Objects[Index]).Mode;
end;



function TGL2DString.GetPosition(Index: integer): T3DVector;
begin
Result:=TGLStringInfo(FStrings.Objects[Index]).Pos;
end;



procedure TGL2DString.SetMode(Index: integer; const Value: TGLStringMode);
begin
TGLStringInfo(FStrings.Objects[Index]).Mode:=Value;
end;



procedure TGL2DString.SetPosition(Index: integer; const Value: T3DVector);
begin
TGLStringInfo(FStrings.Objects[Index]).Pos:=Value;
end;

procedure TGL2DString.Render;
begin
glPushAttrib(GL_ALL_ATTRIB_BITS);
try
inherited Render;
DrawAllStrings;
finally glPopAttrib; end;
end;


{ TGLStringList }


procedure TGLStringList2D.HandleStringChange(Sender: TObject);
begin
Change;
end;



function TGLStringList2D.Add(var Str: TGL2DString): integer;
begin
Str:=TGL2DString.Create;
Str.OnChange:=HandleStringChange;
Str.Name:='String3D-'+IntToStr(Self.Count+1);
Result:=inherited Add(Str);
Self[Result].ResetTransforms;
Self[Result].ResetMaterials;
end;



function TGLStringList2D.AddFrom(Str: TGL2DString): integer;
var NewStr: TGL2DString;
begin
NewStr:=TGL2DString.CreateFrom(Str);
Result:=inherited Add(NewStr);
NewStr.Assign(Str);
end;


function TGLStringList2D.Add: integer;
var Str: TGL2DString;
begin
Result:=Add(Str);
end;



function TGLStringList2D.Insert(Inx: integer; var Str: TGL2DString): integer;
begin
Str:=TGL2DString.Create;
Str.Name:='String3D-'+IntToStr(Count+1);
inherited Insert(Inx,Str);
Result:=Inx;
end;



function TGLStringList2D.Insert(Inx: integer): integer;
var Str: TGL2DString;
begin
Result:=Insert(Inx,Str);
end;



function TGLStringList2D.GetItem(Index: Integer): TGL2DString;
begin
Result:=TGL2DString(inherited GetItem(Index));
end;


procedure TGLStringList2D.SetItem(Index: Integer; AObject: TGL2DString);
begin
inherited SetItem(Index, AObject);
end;



procedure TGLStringList2D.Assign(ObjectList: TObjectList);
var I,Inx: integer;
begin
if ObjectList is TGLStringList2D then
	begin
	Clear;
	for I:=0 to ObjectList.Count-1 do
		begin
		Inx:=AddFrom(TGL2DString(ObjectList[I]));
		Self[Inx].Assign(TGL2DString(ObjectList[I]));
		end;
	end
else inherited Assign(ObjectList);
end;




end.
