unit GLOctTree;

{

This code creates an "Oct-Tree" object which is used to speed
up the rendering of a large mesh in OpenGL. Here is how it
works:

1. The enclosed volume of a mesh is divided into 512 (8*8*8)
boxes.

2. Each face in the mesh is tested to find which box it
belongs in. This done only when the mesh is first created or
if it changes.

3. When the mesh is rendered, each box is tested to see if it
is in the Frustum and if it is, all faces associated with the
box are rendered. This is done everytime the mesh is rendered,
but since there are only 512 boxes, the run time is fixed no
matter how big the mesh is.

4. Some triangles in the mesh may straddle boxes. Those that do
are put in a separate list and always rendered, even if they are
not in the frustum.

}

interface

uses Graphics,SysUtils,Vector,HugeArry,GLFrustumLib,GLLibrary;


{Define box in the tree. The indexes define which faces are in the box}

type TTreeBox = record
 Box: TBox;
 Indexes: THugeInteger;
 InFrustum: boolean;
 end;


type T3DIndex = record
 X,Y,Z: integer;
 end;


{Dimensions of the tree}

const DimX =8; DimY = 8; DimZ = 8;

{Dimensions converted to 3D Vector}

const BoxDims: T3DVector = (X:DimX; Y:DimY; Z:DimZ);

{Iterator procedure declaration}

type TIterateProc = procedure (Inx,CellNumber: integer) of object;


{Oct-Tree object}

type TOctTree = class(TObject)
 private
  FGlobalBox: TBox;
  procedure CreateBoxes;
  procedure DestroyBoxes;
  procedure SetBoxRanges(MeshDim: TBox);
  procedure SetMeshDimensions(V: TBox);
 protected
 public
  FBoxArray: array [0..DimX-1,0..DimY-1,0..DimZ-1] of TTreeBox;
  Straddlers: THugeInteger;
  constructor Create;
  destructor Destroy; override;
  function WhichBox(P: T3DVector; var Inx: T3DIndex): boolean;
  procedure ResetBoxes;
  property MeshDimensions: TBox read FGlobalBox write SetMeshDimensions;
  function PutTriangleInBox(T: TTriangle; Index: integer): boolean;
  function PutLineInBox(T: T3DLine; Index: integer): boolean;
  function PutPointInBox(T: T3DVector; Index: integer): boolean;
  function IsBoxInFrustum(Frustum: TFrustum; Box: TBox): boolean;
  procedure FlagBoxesInFrustum;
  procedure IterateFaces(DrawFaceProc: TIterateProc);
  procedure IterateLines(DrawLineProc: TIterateProc);
 end;

var OutsideBox: Integer;


implementation


{--------------- Box creation routines ----------------------------------------}

procedure TOctTree.ResetBoxes;
{Reset the list of triangles associcate with each box}
var X,Y,Z: integer;
begin
for Z:=0 to DimZ-1 do
 for Y:=0 to DimY-1 do
  for X:=0 to DimX-1 do
	begin
        FBoxArray[X,Y,Z].Indexes.ItemIndex:=0;
        FBoxArray[X,Y,Z].Indexes.ReAllocMem(5);
	FBoxArray[X,Y,Z].InFrustum:=False;
	end;
Straddlers.ItemIndex:=0;
Straddlers.ReAllocMem(5);
end;



procedure TOctTree.CreateBoxes;
{Create all boxes in tree, creating the Index lists}
var X,Y,Z: integer;
begin
for Z:=0 to DimZ-1 do
 for Y:=0 to DimY-1 do
  for X:=0 to DimX-1 do
	begin
        FBoxArray[X,Y,Z].Indexes:=THugeInteger.Create(5);
	end;
end;


procedure TOctTree.DestroyBoxes;
{Destroy all boxes, freeing memory for Index Lists}
var X,Y,Z: integer;
begin
for Z:=0 to DimZ-1 do
 for Y:=0 to DimY-1 do
  for X:=0 to DimX-1 do
	begin
        FBoxArray[X,Y,Z].Indexes.Free;
	end;
end;



procedure TOctTree.SetBoxRanges(MeshDim: TBox);
{Set the ranges of all boxes by subdividing the Dimensions of the mesh}
var X,Y,Z: integer;
var Step,Index1,Index2: T3DVector;
begin
{Calculate the X,Y and Z dimension of each box}
Step:=VectorSubtract3D(MeshDim.Max,MeshDim.Min);
Step:=VectorDivide3D(Step,BoxDims);
for Z:=0 to DimZ-1 do
 for Y:=0 to DimY-1 do
  for X:=0 to DimX-1 do
	begin
	{Convert 3D indexes to a vector}
	Index1:=MakeVector3D(X,Y,Z);
	Index2:=MakeVector3D(X+1,Y+1,Z+1);
	{Then use vector multiply and to set box dimensions}
	FBoxArray[X,Y,Z].Box.Min:=VectorAdd3D(MeshDim.Min,VectorMultiply3D(Step,Index1));
	FBoxArray[X,Y,Z].Box.Max:=VectorAdd3D(MeshDim.Min,VectorMultiply3D(Step,Index2));
	end;
end;


procedure TOctTree.SetMeshDimensions(V: TBox);
{Set oct-tree dimensions based on mesh dimensions}
begin
FGlobalBox:=V;
SetBoxRanges(FGlobalBox);
end;




constructor TOctTree.Create;
{Create Oct-Tree Object}
begin
CreateBoxes;
Straddlers:=THugeInteger.Create(5);
FGlobalBox.Min:=MakeVector3D(0,0,0);
FGlobalBox.Max:=MakeVector3D(8,8,8);
SetBoxRanges(FGlobalBox);

ResetBoxes;
end;


destructor TOctTree.Destroy;
{Destroy boxes - freeing memory}
begin
DestroyBoxes;
Straddlers.Free;
end;


{------------- Routines to test the boxes against a frustum -------------------}



function TOctTree.IsBoxInFrustum(Frustum: TFrustum; Box: TBox): boolean;
{Returns true if the specified box is in the specified
frustum. This is done by checking each corner to see which
side of the frustum planes the point is on. If any frustum
plane doesn't have a corner in front of it, the box is out.
May generate a false positive if all the corners of the
bounding box are not behind any one plane.}
var FSide: integer;
begin
Result:=True;
for FSide:=0 to 5 do
	begin
	if (Frustum[FSide,PlaneA] * (Box.Min.X) + Frustum[FSide,PlaneB] * (Box.Min.Y) + Frustum[FSide,PlaneC] * (Box.Min.Z) + Frustum[FSide,PlaneD] > 0) then continue;
	if (Frustum[FSide,PlaneA] * (Box.Max.X) + Frustum[FSide,PlaneB] * (Box.Min.Y) + Frustum[FSide,PlaneC] * (Box.Min.Z) + Frustum[FSide,PlaneD] > 0) then continue;
	if (Frustum[FSide,PlaneA] * (Box.Min.X) + Frustum[FSide,PlaneB] * (Box.Max.Y) + Frustum[FSide,PlaneC] * (Box.Min.Z) + Frustum[FSide,PlaneD] > 0) then continue;
	if (Frustum[FSide,PlaneA] * (Box.Max.X) + Frustum[FSide,PlaneB] * (Box.Max.Y) + Frustum[FSide,PlaneC] * (Box.Min.Z) + Frustum[FSide,PlaneD] > 0) then continue;

	if (Frustum[FSide,PlaneA] * (Box.Min.X) + Frustum[FSide,PlaneB] * (Box.Min.Y) + Frustum[FSide,PlaneC] * (Box.Max.Z) + Frustum[FSide,PlaneD] > 0) then continue;
	if (Frustum[FSide,PlaneA] * (Box.Max.X) + Frustum[FSide,PlaneB] * (Box.Min.Y) + Frustum[FSide,PlaneC] * (Box.Max.Z) + Frustum[FSide,PlaneD] > 0) then continue;
	if (Frustum[FSide,PlaneA] * (Box.Min.X) + Frustum[FSide,PlaneB] * (Box.Max.Y) + Frustum[FSide,PlaneC] * (Box.Max.Z) + Frustum[FSide,PlaneD] > 0) then continue;
	if (Frustum[FSide,PlaneA] * (Box.Max.X) + Frustum[FSide,PlaneB] * (Box.Max.Y) + Frustum[FSide,PlaneC] * (Box.Max.Z) + Frustum[FSide,PlaneD] > 0) then continue;
	Result:=false;
	end;
end;



procedure TOctTree.FlagBoxesInFrustum;
{Flag all boxes that fall within the frustum}
var X,Y,Z: integer;
var F: TFrustum;
begin
F:=CalculateFrustum;
for Z:=0 to DimZ-1 do
 for Y:=0 to DimY-1 do
  for X:=0 to DimX-1 do
	begin
        FBoxArray[X,Y,Z].InFrustum:=IsBoxInFrustum(F,FBoxArray[X,Y,Z].Box);
	end;
end;


{----------- Routines to test points and triangles against the boxex ---------}



function TOctTree.WhichBox(P: T3DVector; var Inx: T3DIndex): boolean;
{Test 3D Point and return the indexes for the proper box}
{Returns false if outside the range of all boxes}

	function PointOutOfRange(P: T3DVector): boolean;
	{Do trival test to see if point is out of range}
	begin
	Result:=True;
	if P.X>FGlobalBox.Max.X then exit;
	if P.X<FGlobalBox.Min.X then exit;
	if P.Y>FGlobalBox.Max.Y then exit;
	if P.Y<FGlobalBox.Min.Y then exit;
	if P.Z>FGlobalBox.Max.Z then exit;
	if P.Z<FGlobalBox.Min.Z then exit;
	Result:=False;
	end;


	function GetXBox(V: double): integer;
	{Do binary search of X range box index}
	var  L, H: Integer;
	begin
	Result:=0;
	L := 0; H:=DimX-1;
	while True do
		begin
		Result:=(L + H) shr 1;
		if L>=H then break
		else if V<FBoxArray[Result,0,0].Box.Min.X then H:=Result
		else if V>FBoxArray[Result,0,0].Box.Max.X then L:=Result +1
		else break;
		end;
	end;


	function GetYBox(V: double): integer;
	{Do binary search of Y range box index}
	var  L, H: Integer;
	begin
	Result:=0;
	L := 0; H:=DimY-1;
	while True do
		begin
		Result:=(L + H) shr 1;
		if L>=H then break
		else if V<FBoxArray[0,Result,0].Box.Min.Y then H:=Result
		else if V>FBoxArray[0,Result,0].Box.Max.Y then L:=Result +1
		else break;
		end;
	end;


	function GetZBox(V: double): integer;
	{Do binary search of Z range box index}
	var  L, H: Integer;
	begin
	Result:=0;
	L:=0; H:=DimZ-1;
	while True do
		begin
		Result:=(L + H) shr 1;
		if L>=H then break
		else if V<FBoxArray[0,0,Result].Box.Min.Z then H:=Result
		else if V>FBoxArray[0,0,Result].Box.Max.Z then L:=Result +1
		else break;
		end;
	end;

begin
Result:=False;
Inx.X:=0; Inx.Y:=0; Inx.Z:=0;
if PointOutOfRange(P) then exit;
Inx.X:=GetXBox(P.X);
Inx.Y:=GetYBox(P.Y);
Inx.Z:=GetZBox(P.Z);
Result:=True;
end;



function Equal3DIndex(I1,I2: T3DIndex): boolean;
{Return if the two 3D indexes are equal}
begin
Result:=(I1.X = I2.X) and (I1.Y = I2.Y) and (I1.Z = I2.Z);
end;



function TOctTree.PutTriangleInBox(T: TTriangle; Index: integer): boolean;
{Put specifed vertex in corresponding box}
var I1,I2,I3: T3DIndex;
begin
Result:=False;
if not WhichBox(T.P1,I1) then exit;
if not WhichBox(T.P2,I2) then exit;
if not WhichBox(T.P3,I3) then exit;
{If any of the indexes are equal, the triangle straddles boxes}
if Equal3DIndex(I1,I2) and Equal3DIndex(I2,I3) then FBoxArray[I1.X,I1.Y,I1.Z].Indexes.Add(Index)
else Straddlers.Add(Index);
Result:=True;
end;



function TOctTree.PutLineInBox(T: T3DLine; Index: integer): boolean;
{Put specifed vertex in corresponding box}
var I1,I2: T3DIndex;
begin
Result:=False;
if not WhichBox(T.P1,I1) then exit;
if not WhichBox(T.P2,I2) then exit;
{If any of the indexes are not equal, the triangle straddles boxes}
if Equal3DIndex(I1,I2) then FBoxArray[I1.X,I1.Y,I1.Z].Indexes.Add(Index)
else Straddlers.Add(Index);
Result:=True;
end;





function TOctTree.PutPointInBox(T: T3DVector; Index: integer): boolean;
{Put specifed vertex in corresponding box}
var I: T3DIndex;
begin
Result:=False;
if not WhichBox(T,I) then exit;
FBoxArray[I.X,I.Y,I.Z].Indexes.Add(Index);
Result:=True;
end;



{------------ Support for rendering the mesh -----------------------------------}

procedure TOctTree.IterateFaces(DrawFaceProc: TIterateProc);
{Iterrate through all visible faces and call the user supplied subroutine.}
var TreeBox: TTreeBox;
var X,Y,Z,I: integer;
begin
OutsideBox:=0;
for Z:=0 to DimZ-1 do
 for Y:=0 to DimY-1 do
  for X:=0 to DimX-1 do
	begin
	TreeBox:=FBoxArray[X,Y,Z];
	{Ignore all boxes outside the frustum}
	if not TreeBox.InFrustum then begin Inc(OutsideBox); continue; end;
	for I:=0 to TreeBox.Indexes.ItemIndex-1 do DrawFaceProc(TreeBox.Indexes.Value[I],X+(Y*8)+(Z*8*8));
	end;
{Iterrate through straddlers}
for I:=0 to Straddlers.ItemIndex-1 do DrawFaceProc(Straddlers.Value[I],-1);
end;




procedure TOctTree.IterateLines(DrawLineProc: TIterateProc);
{Iterrate through all visible Lines and call the user supplied subroutine.}
begin
IterateFaces(DrawLineProc);
end;



end.
