unit GLQuatLib;

{Library for using and manipulation Quaternions with OpenGL}

interface

uses SysUtils,Math, Vector, GLMatrix,DSupport;

{structure to represent quaternion}

type TQuaternion = record
 case integer of
  0: (X,Y,Z,W: double);
  1: (V: T3DVector);
 end;

{Structure to represent Euler angles}

type TEulerAngles = record
 case integer of
 0: (Pitch,Roll,Yaw: double);
 1: (X,Y,Z: double);
 2: (V: T3DVector);
 end;

{Public declarations}

function MakeQuat(X,Y,Z,W: double): TQuaternion; overload;
function MakeQuat(V: T3DVector; W: double): TQuaternion; overload;

function GetIdentityQuaternion: TQuaternion;
function QuatMultiply(Q1,Q2: TQuaternion): TQuaternion;
function QuatConjugate(Q: TQuaternion): TQuaternion;
function QuatMagnitude(Q: TQuaternion): double;
function QuatNormalize(Q: TQuaternion): TQuaternion;
function QuatDotProduct(Q1,Q2: TQuaternion): double;

function EulerToQuat(Pitch,Yaw,Roll: double): TQuaternion;
procedure QuatToAxisAngle(Q: TQuaternion; var V: T3DVector; var Angle: double);
function AxisAngleToQuat(V: T3DVector; Angle: double): TQuaternion;
function QuatBetweenVectors(FromVect,ToVect: T3DVector): TQuaternion;
procedure QuatToEulerAngles(Q: TQuaternion; var Pitch,Yaw,Roll: double);
function QuaternionSlerp(QA, QB: TQuaternion; T: double): TQuaternion;
function InterpolateAngles(A1,A2: TEulerAngles; T: double): TEulerAngles;
function GetQuatString(Q: TQuaternion; Digits: integer): string; overload;
function GetQuatString(Q: TQuaternion): string; overload;
function GLMatrixToQuaternion(M: TGLMatrix): TQuaternion;
function QuatToMatrix(Q: TQuaternion): TGLMatrix;
function EulerToMatrix(Pitch,Yaw,Roll: double): TGLMatrix;
function QuaternionTransform(Q: TQuaternion; P: T3DVector): T3DVector;
function QuatAdd(Q1,Q2: TQuaternion): TQuaternion;
function QuatScalarProduct(Q: TQuaternion; S: double): TQuaternion;
function QuatBetweenPlanes(P1A,P1B,P2A,P2B: T3DVector): TQuaternion;
procedure EulerBetweenPlanes(P1A,P1B,P2A,P2B: T3DVector; var Pitch,Yaw,Roll: double);

const QuatIndentity: TQuaternion = (X:0; Y:0; Z:0; W:1);

{Dual Quaternion code}

type TDualQuat = record
 Real,Dual: TQuaternion;
 end;

function MakeDualQuat(Real,Dual: TQuaternion): TDualQuat; overload;
function MakeDualQuat: TDualQuat; overload;
function MakeDualQuat(R: TQuaternion; T: T3DVector): TDualQuat; overload;

function DualQuatDotProduct(A,B: TDualQuat): double;
function DualQuatScalarProduct(Q: TDualQuat; Scale: double): TDualQuat;
function DualQuatNormalize(Q: TDualQuat): TDualQuat;
function DualQuatAdd(LHS,RHS: TDualQuat): TDualQuat;
function DualQuatProduct(LHS,RHS: TDualQuat): TDualQuat;
function DualQuatConjugate(Q: TDualQuat): TDualQuat;
function GetDualQuatRotation(Q: TDualQuat): TQuaternion;
function GetDualQuatTranslation(Q: TDualQuat): T3DVector;
function DualQuaternionToMatrix(Q: TDualQuat): TGLMatrix;

implementation


function MakeQuat(X,Y,Z,W: double): TQuaternion;
{Create quaternion from the individual elements}
begin
Result.X:=X; Result.Y:=Y; Result.Z:=Z; Result.W:=W;
end;



function MakeQuat(V: T3DVector; W: double): TQuaternion;
{Create quaternion from the individual elements}
begin
Result.X:=V.X; Result.Y:=V.Y; Result.Z:=V.Z; Result.W:=W;
end;


function EulerToMatrix(Pitch,Yaw,Roll: double): TGLMatrix;
{Convert Euler angles to rotation matrix}
var Q: TQuaternion;
begin
Q:=EulerToQuat(Pitch,Yaw,Roll);
Result:=QuatToMatrix(Q);
end;


function GetIdentityQuaternion: TQuaternion;
{Returns Quaternion Identiy}
begin
Result:=QuatIndentity
end;



function QuatConjugate(Q: TQuaternion): TQuaternion;
{Return quaternion conjugate (inverse)}
begin
Result.W:=Q.W;
Result.V:=ScalarProduct3D(Q.V,-1);
end;


function QuatScalarProduct(Q: TQuaternion; S: double): TQuaternion;
{Return product of scalar and Quaternion}
begin
Result:=Q;
Result.V:=ScalarProduct3D(Q.V,S);
Result.W:=Result.W * S;
end;


function QuatAdd(Q1,Q2: TQuaternion): TQuaternion;
{Return sum of two Quaternion}
begin
Result.X:=Q1.X + Q2.X;
Result.Y:=Q1.Y + Q2.Y;
Result.Z:=Q1.Z + Q2.Z;
Result.W:=Q1.W + Q2.W;
end;



function QuatMultiply(Q1,Q2: TQuaternion): TQuaternion;
{Multiply Quaternion Q1 and Q2 }
begin
Result.W:= Q1.W*Q2.W - Q1.X*Q2.X - Q1.Y*Q2.Y - Q1.Z*Q2.Z;
Result.X:= Q1.W*Q2.X + Q1.X*Q2.W + Q1.Y*Q2.Z - Q1.Z*Q2.Y;
Result.Y:= Q1.W*Q2.Y + Q1.Y*Q2.W + Q1.Z*Q2.X - Q1.X*Q2.Z;
Result.Z:= Q1.W*Q2.Z + Q1.Z*Q2.W + Q1.X*Q2.Y - Q1.Y*Q2.X;
end;



function QuaternionTransform(Q: TQuaternion; P: T3DVector): T3DVector;
{Transform point based on specified Quaternion}
begin
Result.X := Q.W*Q.W*P.X + 2*Q.Y*Q.W*P.Z - 2*Q.Z*Q.W*P.Y + Q.X*Q.X*P.X + 2*Q.Y*Q.X*P.Y + 2*Q.Z*Q.X*P.Z - Q.Z*Q.Z*P.X - Q.Y*Q.Y*P.X;
Result.Y := 2*Q.X*Q.Y*P.X + Q.Y*Q.Y*P.Y + 2*Q.Z*Q.Y*P.Z + 2*Q.W*Q.Z*P.X - Q.Z*Q.Z*P.Y + Q.W*Q.W*P.Y - 2*Q.X*Q.W*P.Z - Q.X*Q.X*P.Y;
Result.Z := 2*Q.X*Q.Z*P.X + 2*Q.Y*Q.Z*P.Y + Q.Z*Q.Z*P.Z - 2*Q.W*Q.Y*P.X - Q.Y*Q.Y*P.Z + 2*Q.W*Q.X*P.Y - Q.X*Q.X*P.Z + Q.W*Q.W*P.Z;
end;



function QuatDotProduct(Q1,Q2: TQuaternion): double;
begin
Result:=Q1.X * Q2.X + Q1.Y * Q2.Y + Q1.Z * Q2.Z + Q1.W * Q2.W;
end;



function QuatMagnitude(Q: TQuaternion): double;
{Get magnitude of 4D Quaternion vector}
begin
Result:=Sqrt(Sqr(Q.W)+Sqr(Q.X)+Sqr(Q.Y)+Sqr(Q.Z));
end;



function QuatNormalize(Q: TQuaternion): TQuaternion;
{Get Quaternion unit vector - i.e. vector with magnitude = 1}
var Mag: double;
begin
Mag:=QuatMagnitude(Q);
if Mag=0 then Mag:=1E-99;
Result.X:= Q.X / Mag;
Result.Y:= Q.Y / Mag;
Result.Z:= Q.Z / Mag;
Result.W:= Q.W / Mag;
end;



procedure QuatToAxisAngle(Q: TQuaternion; var V: T3DVector; var Angle: double);
{Convert quaternion to direction vector and angle}
var AngleTemp,Scale: double;
begin
if Q.W>1 then Q.W:=1
else if Q.W<-1 then Q.W:=-1;
AngleTemp := ArcCos(Q.W);
Scale:=sqrt(SQR(Q.X) + SQR(Q.Y) + SQR(Q.Z));
{If scale is zero, angle is zero and any vector will do}
if Scale=0 then
	begin
	Angle:= 0;
	V:=MakeVector3D(0,0,1);
	end
else
	begin
	angle:= AngleTemp * 2;
	V:=MakeVector3D(Q.X / Scale, Q.Y / Scale, Q.Z / Scale);
	V:=UnitVector3D(V);
	end;
end;




function AxisAngleToQuat(V: T3DVector; Angle: double): TQuaternion;
{Convert direction vector and angle to Quaternion}
var S: double;
begin
S:=sin(Angle/2);
Result.X:= V.X * s;
Result.Y:= V.Y * s;
Result.Z:= V.Z * s;
Result.W:= cos(Angle/2);
end;



{---------------------------------------------------------------------}


function QuatToMatrix(Q: TQuaternion): TGLMatrix;
{Convert quaternion to OpenGL matrix}
var X2, Y2, Z2, XY, XZ, YZ, WX, WY, WZ: double;
begin
Result:=IdentityMatrix;
X2:= sqr(Q.X); Y2:= sqr(Q.Y); Z2:= sqr(Q.Z);

XY:= Q.X * Q.Y; XZ:= Q.X * Q.Z; YZ:= Q.Y * Q.Z;
WX:= Q.W * Q.X; WY:= Q.W * Q.Y; WZ:= Q.W * Q.Z;

Result[0]:= 1 - 2*(Y2 + Z2);
Result[1]:= 2 * (XY + WZ);
Result[2]:= 2 * (XZ - WY);

Result[4]:= 2 * (XY - WZ);
Result[5]:= 1 - 2*(X2 + Z2);
Result[6]:= 2 * (YZ + WX);

Result[8]:= 2 * (XZ + WY);
Result[9]:= 2 * (YZ - WX);
Result[10]:= 1 - 2 * (X2 + Y2);
end;



function MatrixToQuaternion(M: TGLMatrix): TQuaternion;
{Extract rotation from matrix to quaternion}
var PM: TProgMatrix;
begin
M:=InvertGLMatrix(M);
PM:=TProgMatrix(RemoveGLMatrixScaling(M));
Result.w := sqrt( Max( 0, 1 + PM[0,0] + PM[1,1] + PM[2,2] ) ) / 2;
Result.x := sqrt( Max( 0, 1 + PM[0,0] - PM[1,1] - PM[2,2] ) ) / 2;
Result.y := sqrt( Max( 0, 1 - PM[0,0] + PM[1,1] - PM[2,2] ) ) / 2;
Result.z := sqrt( Max( 0, 1 - PM[0,0] - PM[1,1] + PM[2,2] ) ) / 2;
Result.x := CopySign( Result.x, PM[2,1] - PM[1,2] );
Result.y := CopySign( Result.y, PM[0,2] - PM[2,0] );
Result.z := CopySign( Result.z, PM[1,0] - PM[0,1] );
end;



function GLMatrixToQuaternion(M: TGLMatrix): TQuaternion;
{Convert Matrix to Quaternion, removing scaling in advance}
var Trace,S: double;
var Mat: TProgMatrix;
begin
{Convert to my standard format}
M:=InvertGLMatrix(M);
{Remove scaling}
Mat:=TProgMatrix(RemoveGLMatrixScaling(M));
Trace:=Mat[0,0] + Mat[1,1] + Mat[2,2];
if Trace > 0 then
	begin
	S:=0.5 / sqrt(Trace+ 1.0);
	Result.w:=0.25 / s;
	Result.x:=( Mat[2,1] - Mat[1,2] ) * s;
	Result.y:=( Mat[0,2] - Mat[2,0] ) * s;
	Result.z:=( Mat[1,0] - Mat[0,1] ) * s;
	end
else
	begin
	if (Mat[0,0] > Mat[1,1]) and (Mat[0,0] > Mat[2,2]) then
		begin
		s:=2.0 * sqrt( 1.0 + Mat[0,0] - Mat[1,1] - Mat[2,2]);
		Result.w:=(Mat[2,1] - Mat[1,2] ) / s;
		Result.x:=0.25 * s;
		Result.y:=(Mat[0,1] + Mat[1,0] ) / s;
		Result.z:=(Mat[0,2] + Mat[2,0] ) / s;
		end
	else if Mat[1,1] > Mat[2,2] then
		begin
		s:=2.0 * sqrt( 1.0 + Mat[1,1] - Mat[0,0] - Mat[2,2]);
		Result.w:=(Mat[0,2] - Mat[2,0] ) / s;
		Result.x:=(Mat[0,1] + Mat[1,0] ) / s;
		Result.y:=0.25 * s;
		Result.z:=(Mat[1,2] + Mat[2,1] ) / s;
		end
	else
		begin
		S:=2.0 * sqrt( 1.0 + Mat[2,2] - Mat[0,0] - Mat[1,1] );
		Result.w:=(Mat[1,0] - Mat[0,1] ) / s;
		Result.x:=(Mat[0,2] + Mat[2,0] ) / s;
		Result.y:=(Mat[1,2] + Mat[2,1] ) / s;
		Result.z:=0.25 * s;
		end;
	end;
end;



function EulerToQuat(Pitch,Yaw,Roll: double): TQuaternion;
{Convert Euler Angles to proper quaternion}
var c1,s1,c2,s2,c3,s3: double;
var c1c2,s1s2: double;
begin
Pitch:= DegToRad(Pitch);
Yaw:= DegToRad(Yaw);
Roll:= DegToRad(Roll);
c1:= cos(Yaw/2);
s1:= sin(Yaw/2);
c2:= cos(Roll/2);
s2:= sin(Roll/2);
c3:= cos(Pitch/2);
s3:= sin(Pitch/2);

c1c2:= c1*c2;
s1s2:= s1*s2;

Result.w:=c1c2*c3 - s1s2*s3;
Result.x:=c1c2*s3 + s1s2*c3;
Result.y:=s1*c2*c3 + c1*s2*s3;
Result.z:=c1*s2*c3 - s1*c2*s3;
end;



function QuatBetweenVectors(FromVect,ToVect: T3DVector): TQuaternion;
{Find the quaternion that rotates FromVect to ToVect}
var Dir: T3DVector;
var Angle: double;
begin
Dir:=CrossProduct3D(FromVect,ToVect);
Angle:=VectorAngle3D(FromVect,ToVect);
Result:=AxisAngleToQuat(Dir,Angle);
end;




procedure QuatToEulerAngles(Q: TQuaternion; var Pitch,Yaw,Roll: double);
{Convert Quaternion to Euler angles - Quaternion can be non-normalized}
var sqw,sqx,sqy,sqz: double;
var UnitQ: double;
var test: double;
begin
sqw:= Q.w * Q.w;
sqx:= Q.x * Q.x;
sqy:= Q.y * Q.y;
sqz:= Q.z * Q.z;
{ if normalised is one, otherwise is correction factor }
UnitQ:= sqx + sqy + sqz + sqw;
if UnitQ = 0 then UnitQ:=1E-99;
test:= Q.x*Q.y + Q.z*Q.w;

if test > 0.499*UnitQ then
	begin
	{ singularity at north pole }
	Yaw:= RadToDeg(2 * ArcTan2(Q.x,Q.w));
	Roll:= RadToDeg(PI/2);
	Pitch:= RadToDeg(0);
	exit;
	end;
if test < -0.499*UnitQ then
	begin
	{ singularity at south pole }
	Yaw:= RadToDeg(-2 * ArcTan2(Q.x,Q.w));
	Roll:= RadToDeg(-PI/2);
	Pitch:= RadToDeg(0);
	exit;
	end;
Yaw:= RadToDeg(ArcTan2(2*Q.y*Q.w-2*Q.x*Q.z , sqx - sqy - sqz + sqw));
Roll:= RadToDeg(ArcSin(2*test/UnitQ));
Pitch:= RadToDeg(ArcTan2(2*Q.x*Q.w-2*Q.y*Q.z , -sqx + sqy - sqz + sqw));
end;




function QuaternionSlerp(QA, QB: TQuaternion; T: double): TQuaternion;
{Interpolate between two Quaternions by the amount specified in T}
{if T=0 QA will be returned, T=1 QB will be returned }
var cosHalfTheta: double;
var halfTheta: double;
var sinHalfTheta: double;
var ratioA: double;
var ratioB: double;
begin
// Calculate angle between them.
cosHalfTheta:= qa.w * qb.w + qa.x * qb.x + qa.y * qb.y + qa.z * qb.z;
// if qa=qb or qa=-qb then theta = 0 and we can return qa
if abs(cosHalfTheta) >= 1.0 then
	begin
	Result.w:= qa.w; Result.x:= qa.x;
	Result.y:= qa.y; Result.z:= qa.z;
	exit;
	end;

// Calculate temporary values.
halfTheta:= arccos(cosHalfTheta);
sinHalfTheta:= sqrt(1.0 - cosHalfTheta*cosHalfTheta);
// if theta = 180 degrees then result is not fully defined
// we could rotate around any axis normal to qa or qb
if abs(sinHalfTheta) < 0.001 then
	begin // fabs is floating point absolute
	Result.w:= (qa.w * 0.5 + qb.w * 0.5);
	Result.x:= (qa.x * 0.5 + qb.x * 0.5);
	Result.y:= (qa.y * 0.5 + qb.y * 0.5);
	Result.z:= (qa.z * 0.5 + qb.z * 0.5);
	exit;
	end;
ratioA:= sin((1 - t) * halfTheta) / sinHalfTheta;
ratioB:= sin(t * halfTheta) / sinHalfTheta; 
//calculate Quaternion.
Result.w:= (qa.w * ratioA + qb.w * ratioB);
Result.x:= (qa.x * ratioA + qb.x * ratioB);
Result.y:= (qa.y * ratioA + qb.y * ratioB);
Result.z:= (qa.z * ratioA + qb.z * ratioB);
end;



function InterpolateAngles(A1,A2: TEulerAngles; T: double): TEulerAngles;
{Interpolate between two Euler angle sets}
var Q1,Q2,Q3: TQuaternion;
begin
Q1:=EulerToQuat(A1.Pitch,A1.Yaw,A1.Roll);
Q2:=EulerToQuat(A2.Pitch,A2.Yaw,A2.Roll);
Q3:=QuaternionSlerp(Q1,Q2,T);
QuatToEulerAngles(Q3,Result.Pitch,Result.Yaw,Result.Roll);
end;



function GetQuatString(Q: TQuaternion; Digits: integer): string;
{Return a string representing the specified quaternion}
{Digits specifies the number of digits past the decimal}
begin
Result:=FloatToStrF(Q.X,ffFixed,18,Digits)+', '+ FloatToStrF(Q.Y,ffFixed,18,Digits)+', '+ FloatToStrF(Q.Z,ffFixed,18,Digits)+', '+FloatToStrF(Q.W,ffFixed,18,Digits);
end;



function GetQuatString(Q: TQuaternion): string;
{Return a string representing the specified quaternion}
begin
Result:=FloatToStrF(Q.X,ffGeneral,18,4)+', '+ FloatToStrF(Q.Y,ffGeneral,18,4)+', '+ FloatToStrF(Q.Z,ffGeneral,18,4)+', '+FloatToStrF(Q.W,ffGeneral,18,4);
end;




function QuatBetweenPlanes(P1A,P1B,P2A,P2B: T3DVector): TQuaternion;
{Get quaternion of rotation between planes P1 and P2}
var N1,N2,M,Axis: T3DVector;
var Angle: double;
begin
N1:=CrossProduct3D(P1A,P1B);
N2:=CrossProduct3D(P2A,P2B);
N1:=UnitVector3D(N1); N2:=UnitVector3D(N2);
M:=VectorAdd3D(N1,N2);
M:=UnitVector3D(M);
Axis:= CrossProduct3D(M,N2);
Angle:= DotProduct3D(M,N2);
Result:=MakeQuat(Axis.X, Axis.Y, Axis.Z,Angle);
Result:=QuatNormalize(Result);
end;



procedure EulerBetweenPlanes(P1A,P1B,P2A,P2B: T3DVector; var Pitch,Yaw,Roll: double);
{Find the Euler angle between two plane specified by Vector pairs P1 and P2}
var Q: TQuaternion;
begin
Q:=QuatBetweenPlanes(P1A,P1B,P2A,P2B);
{Zero quaternion = 180 rotation - but could be around any axis}
if (Q.X=0) and (Q.Y=0) and (Q.Y=0) and (Q.W=0) then
	begin
	{Arbitrarily assume Yaw-rotation}
	Pitch:=0; Yaw:=180; Roll:=0;
	end
else QuatToEulerAngles(Q,Pitch,Yaw,Roll);
end;


{------------- Dual Quaternion -----------------------------------------------}




function MakeDualQuat(Real,Dual: TQuaternion): TDualQuat;
begin
Result.Real:=QuatNormalize(Real);
Result.Dual:=Dual;
end;



function MakeDualQuat: TDualQuat;
begin
Result:=MakeDualQuat(MakeQuat(0,0,0,1),MakeQuat(0,0,0,0));
end;



function MakeDualQuat(R: TQuaternion; T: T3DVector): TDualQuat;
begin
Result.Real:= QuatNormalize(R);
Result.Dual:= QuatScalarProduct(QuatMultiply(MakeQuat(T,0), Result.Real),0.5);
end;




function DualQuatDotProduct(A,B: TDualQuat): double;
begin
Result:=QuatDotProduct( A.Real, B.Real );
end;




function DualQuatScalarProduct(Q: TDualQuat; Scale: double): TDualQuat;
begin
Result:=Q;
Result.Real:=QuatScalarProduct(Result.Real,Scale);
Result.Dual:=QuatScalarProduct(Result.Dual,Scale);
end;




function DualQuatNormalize(Q: TDualQuat): TDualQuat;
var Mag: double;
begin
Mag:=QuatDotProduct(Q.Real,Q.Real);
if Mag=0 then Mag:=1E-99;
Result:=Q;
Result.Real:=QuatScalarProduct(Result.Real, 1/mag);
Result.Dual:=QuatScalarProduct(Result.Dual, 1/mag);
end;



function  DualQuatAdd(LHS,RHS: TDualQuat): TDualQuat;
begin
Result:=MakeDualQuat(QuatAdd(LHS.Real, RHS.Real), QuatAdd(LHS.Dual, RHS.Dual));
end;


function DualQuatProduct(LHS,RHS: TDualQuat): TDualQuat;
{Multiplication order - left to right}
var Real,Dual,Temp: TQuaternion;
begin
Real:=QuatMultiply(rhs.Real,lhs.Real);
Dual:=QuatAdd(QuatMultiply(rhs.Dual,lhs.Real), QuatMultiply(rhs.Real,lhs.Dual));
Result:=MakeDualQuat(Real,Dual);
end;




function DualQuatConjugate(Q: TDualQuat): TDualQuat;
begin
Result:=MakeDualQuat(QuatConjugate(Q.Real ),QuatConjugate(Q.Dual));
end;


function GetDualQuatRotation(Q: TDualQuat): TQuaternion;
begin
Result:=Q.Real;
end;



function GetDualQuatTranslation(Q: TDualQuat): T3DVector;
var T: TQuaternion;
begin
T := QuatMultiply(QuatScalarProduct(q.Dual,2),QuatConjugate(q.Real));
Result:=MakeVector3D(T.X, T.y, T.z );
end;




function DualQuaternionToMatrix(Q: TDualQuat): TGLMatrix;
var M: TProgMatrix;
var X,Y,Z,W: double;
var T: TQuaternion;
begin
Q:=DualQuatNormalize(Q);
M := TProgMatrix(IdentityMatrix);
W := Q.Real.W;
X := Q.Real.X;
Y := Q.Real.Y;
Z := Q.Real.Z;
 // Extract rotational information
 M[0,0] := w*w + x*x - y*y - z*z;
 M[0,1] := 2*x*y + 2*w*z;
 M[0,2] := 2*x*z - 2*w*y;

 M[1,0] := 2*x*y - 2*w*z;
 M[1,1] := w*w + y*y - x*x - z*z;
 M[1,2] := 2*y*z + 2*w*x;
 M[2,0] := 2*x*z + 2*w*y;
 M[2,1] := 2*y*z - 2*w*x;
 M[2,2] := w*w + z*z - x*x - y*y;
// Extract translation information
T:= QuatMultiply(QuatScalarProduct(Q.Dual,2), QuatConjugate(Q.Real));
 M[3,0] := t.X;
 M[3,1] := t.Y;
 M[3,2] := t.Z;
Result:=TGLMatrix(M);
end;


end.
