unit GLXFileTextReader;

interface

uses Windows,Graphics,Classes,SysUtils,StreamLib,OpenGL15,Math,GLMatrix,
     GLViewer,GLMesh,GLLibrary,GLColorArray,HugeArry,Vector,GLXFileGlobal, DSupport;

procedure ParseTextXFile(Stream: TStream; Viewer: TGLViewer; ProgInfo: TProgressInfo; NewMeshes: boolean);


implementation


var MeshIndex: integer;		{Count/Index of meshes being used}
var UseNewMeshes: boolean;	{Flags create new meshes and not reuse old ones}

var GlobalMatrix: TGLMatrix;



{---------------- Objects used to parse file ----------------------------------}

type TMaterial = record
 Name: AnsiString;
 FaceColor: TGLArrayf4;
 Power: double;
 SpecularColor: TGLArrayf4;
 EmissiveColor: TGLArrayf4;
 TextureFile: AnsiString;
 end;



type TMaterialList = class(TObject)
 public
  MaterialCount,FaceCount: integer;
  FaceIndices: array of Integer;
  Materials: array of TMaterial;
  constructor Create;
  procedure Clear;
 end;


{ TMaterialList }

procedure TMaterialList.Clear;
begin
MaterialCount:=0;
FaceCount:=0;
SetLength(FaceIndices,0);
SetLength(Materials,0);
end;


constructor TMaterialList.Create;
begin
inherited;
Clear;
end;

{------------------ Local Variables ------------------------------------------}

var GlobalMaterials: array of TMaterial;


{------------------ Progress handling routines -------------------------------}


var StreamProgress: TStreamProgress;


{------------------ Low Level Support Routines -------------------------------}



function GetXNumberStr(Stream: TStream): AnsiString;
{Returns string up to comma, semicolon or EOF}
var C: AnsiChar;
begin
Result:='';
while StreamProgress.CheckStreamPosition(Stream) do
	begin
	C:=GetStreamChar(Stream);
	case C of
	 ';',',': break;
	 '/': if PeekStreamChar(Stream)='/' then StreamSkipToChar(Stream,LF_Char);
	 else Result:=Result+C;
	 end;
	end;
Result:=Trim(Result);
end;



function GetXFloat(Stream: TStream): double;
var S: AnsiString;
begin
S:=GetXNumberStr(Stream);
Result:=StrToFloat(S);
end;


function GetXInteger(Stream: TStream): integer;
var S: AnsiString;
begin
S:=GetXNumberStr(Stream);
Result:=StrToInt(S);
end;



function ReadVector(Stream: TStream): T3DVector;
begin
Result.X:=GetXFloat(Stream);
Result.Y:=GetXFloat(Stream);
Result.Z:=GetXFloat(Stream);
{Convert left hand to right hand}
Result.Z:=-Result.Z;
GetStreamChar(Stream);
end;



function ReadVectorTransformed(Stream: TStream): T3DVector;
begin
Result.X:=GetXFloat(Stream);
Result.Y:=GetXFloat(Stream);
Result.Z:=GetXFloat(Stream);
{Do matrix conversion before converting right to left}
Result:=MatrixVectorMult(GlobalMatrix,Result);
{Convert left hand to right hand}
Result.Z:=-Result.Z;
GetStreamChar(Stream);
end;



function ReadXRGBA(Stream: TStream): TGLArrayf4;
begin
Result[0]:=GetXFloat(Stream);
Result[1]:=GetXFloat(Stream);
Result[2]:=GetXFloat(Stream);
Result[3]:=GetXFloat(Stream);
GetStreamChar(Stream);
end;


function ReadXRGB(Stream: TStream): TGLArrayf3; overload;
begin
Result[0]:=GetXFloat(Stream);
Result[1]:=GetXFloat(Stream);
Result[2]:=GetXFloat(Stream);
GetStreamChar(Stream);
end;



function ReadXRGB(Stream: TStream; A: Single): TGLArrayf4; overload;
begin
Result[0]:=GetXFloat(Stream);
Result[1]:=GetXFloat(Stream);
Result[2]:=GetXFloat(Stream);
Result[3]:=A;
GetStreamChar(Stream);
end;


function HandleComment(Stream: TStream): TXObjectToken;
{Handle ordinary comment by eating to the end of line}
begin
StreamSkipToChars(Stream,[CR_Char,LF_Char]);
Result:=xfComment;
end;



procedure SkipBlock(Stream: TStream);
var Com: AnsiChar;
begin
StreamSkipToChar(Stream,'{');
while StreamProgress.CheckStreamPosition(Stream) do
	begin
	Com:=PeekStreamChar(Stream);
	if Com = '{' then SkipBlock(Stream);
	GetStreamChar(Stream);
	if Com = '}' then exit;
   	end;
end;


function HandleSlashComment(Stream: TStream): TXObjectToken;
{Handle double slash comment}
begin
if PeekStreamChar(Stream)<>'/' then Result:=xfError
else
	begin
	Result:=xfComment;
	HandleComment(Stream);
	end;
end;




function IsIndentifierChar(C: AnsiChar): boolean;
begin
Result:=(C>#$21) and (C<#$80);
end;



function ReadIdentifier(Stream: TStream): AnsiString;
var C: AnsiChar;
begin
Result:='';
StreamSkipWhiteSpace(Stream);
while StreamProgress.CheckStreamPosition(Stream) do
	begin
	C:=GetStreamChar(Stream);
	if IsIndentifierChar(C) then Result:=Result + C
	else break;
	end;

end;



function CheckForIdentifier(Stream: TStream): AnsiString;
var Com: AnsiChar;
begin
StreamSkipWhiteSpace(Stream);
Com:=PeekStreamChar(Stream);
if Com <> '{' then Result:=ReadIdentifier(Stream)
else Result:='';
end;




function GetBlockID(Stream: TStream): TXObjectToken;
begin
Result:=GetObjectID(ReadIdentifier(Stream));
end;



function ProcessBlock(Stream: TStream): TXObjectToken;
var Com: AnsiChar;

	function HandlePeekedChar(OT: TXObjectToken): TXObjectToken;
	{We don't need the char we peeked at so read it
	and pass through decoded value}
	begin
	Result:=OT;
	GetStreamChar(Stream);
	end;

begin
Com:=PeekStreamChar(Stream);
if Com<#$21 then Result:=HandlePeekedChar(xfWhiteSpace)
else Case Com of
 '{': Result:=HandlePeekedChar(xfOpenBrace);
 '}': Result:=HandlePeekedChar(xfCloseBrace);
 '/': Result:=HandleSlashComment(Stream);
 '#': Result:=HandleComment(Stream);
 else Result:=GetBlockID(Stream);
 end;
end;



procedure ProcessAnimationSets(Stream: TStream);
begin
SkipBlock(Stream);
end;


procedure ProcessMeshTextureCoords(Stream: TStream);
{Read but ignore Texture coordinates}
var CoordCnt: integer;
var I: integer;
var CList: THuge2DVector;
var V: T2DVector;
begin
CList:=THuge2DVector.Create(5);
try
StreamSkipToChar(Stream,'{');
CoordCnt:=GetXInteger(Stream);
for I:=0 to CoordCnt-1 do
	begin
	V.X:=GetXFloat(Stream);
	V.Y:=GetXFloat(Stream);
	{Eat last semicolon or comma}
	GetStreamChar(Stream);
	end;
StreamSkipToChar(Stream,'}');
finally CList.Free; end;
end;




function GetXFilename(Stream: TStream): AnsiString;
begin
StreamSkipToChar(Stream,'{');
StreamSkipToChar(Stream,'"');
Result:=StreamGetToChar(Stream,'"');
StreamSkipToChar(Stream,'}');
end;



function ProcessMaterial(Stream: TStream): TMaterial;
var Token: TXObjectToken;
begin
Result.Name:=CheckForIdentifier(Stream);
StreamSkipToChar(Stream,'{');
Result.FaceColor:=ReadXRGBA(Stream);
Result.Power:=GetXFloat(Stream);
Result.SpecularColor:=ReadXRGB(Stream,1);
Result.EmissiveColor:=ReadXRGB(Stream,1);
while StreamProgress.CheckStreamPosition(Stream) do
	begin
	Token:=ProcessBlock(Stream);
	case Token of
	 xfCOMMENT,xfWhiteSpace: ;
         xfCloseBrace: break;
	 xfTextureFilename: Result.TextureFile:=GetXFilename(Stream);
         else SkipBlock(Stream);
	 end;
	end;
end;



function GetDefaultMaterial: TMaterial;
begin
Result.FaceColor:=MakeGLArrayf4(1,1,1,1);
Result.Power:=10;
Result.SpecularColor:=MakeGLArrayf4(0.5,0.5,0.5,1.0);
Result.EmissiveColor:=MakeGLArrayf4(0,0,0,0);
end;




function LookGlobalMaterial(Name: AnsiString): TMaterial;
var I: integer;
begin
Result:=GetDefaultMaterial;
for I:=0 to High(GlobalMaterials) do
 if Name=GlobalMaterials[I].Name then
	begin
	Result:=GlobalMaterials[I];
	break;
	end;
end;



procedure ProcessMeshMaterials(Stream: TStream; var MatList: TMaterialList);
var I: integer;
var Token: TXObjectToken;
var MaterialName: AnsiString;
begin
StreamSkipToChar(Stream,'{');
MatList.MaterialCount:=GetXInteger(Stream);
MatList.FaceCount:=GetXInteger(Stream);

SetLength(MatList.FaceIndices,MatList.FaceCount);
SetLength(MatList.Materials,MatList.MaterialCount);
for I:=0 to MatList.FaceCount-1 do MatList.FaceIndices[I]:=GetXInteger(Stream);
{Eat last semicolon or comma}
GetStreamChar(Stream);
I:=0;
while StreamProgress.CheckStreamPosition(Stream) do
	begin
	Token:=ProcessBlock(Stream);
	case Token of
	 xfCOMMENT,xfWhiteSpace: ;
         xfCloseBrace: break;
	 xfOpenBrace:
		begin
		MaterialName:=ReadIdentifier(Stream);
		if GetStreamPreviousChar(Stream)<>'}' then StreamSkipToChar(Stream,'}');
		MatList.Materials[I]:=LookGlobalMaterial(MaterialName);
		Inc(I);
		end;
	 xfMaterial:
		begin
		MatList.Materials[I]:=ProcessMaterial(Stream);
		Inc(I);
		end;
         else SkipBlock(Stream);
	 end;
	end;
end;





procedure ProcessSkinWeights(Stream: TStream);
begin
SkipBlock(Stream);
end;





procedure ReadVertexList(Stream: TStream; VertexList: TVertexList; Transform: boolean);
{Read generic vertex list - i.e. face vertices or normal vertices}
var I,J,Vertices,FaceCount,PntCount: integer;
begin
Vertices:=GetXInteger(Stream);

if Transform then for I:=0 to Vertices-1 do VertexList.Vertices.Add(ReadVectorTransformed(Stream))
else for I:=0 to Vertices-1 do VertexList.Vertices.Add(ReadVector(Stream));

{Eat second semicolon at the end of the vertices}
GetStreamChar(Stream);

FaceCount:=GetXInteger(Stream);
VertexList.FaceCount:=FaceCount;
VertexList.Indices.Add(FaceCount);
for I:=0 to FaceCount-1 do
	begin
	PntCount:=GetXInteger(Stream);
	VertexList.Indices.Add(PntCount);
	for J:=0 to PntCount-1 do VertexList.Indices.Add(GetXInteger(Stream));
	{Eat last semicolon or comma}
	GetStreamChar(Stream);
	end;
end;



procedure BuildMesh(var Mesh: TGLMesh; Vertices: THuge3DVector; Faces: THugeInteger);
{Build mesh from list of vertices and Faces}
var FaceCnt,VertexCnt,Inx,I,F: integer;
var Pnts: array [0..120] of integer;
var C: TColor;
begin
Mesh.Clear;
C:=clWhite;
FaceCnt:=Faces[0];
Inx:=1;
for F:=1 to FaceCnt do
	begin
	VertexCnt:=Faces[Inx];
	Inc(Inx);
	for I:=0 to VertexCnt-1 do
		begin
		Pnts[I]:=Faces[Inx];
		Inc(Inx);
		end;
	{Order reversed to convert left to right handed coordinates}
	if VertexCnt=3 then Mesh.AddTriangle(Vertices[Pnts[2]],Vertices[Pnts[1]],Vertices[Pnts[0]],C)
	else if VertexCnt= 4 then Mesh.AddRectangle(Vertices[Pnts[3]],Vertices[Pnts[2]],Vertices[Pnts[1]],Vertices[Pnts[0]],C)
	else for I:=0 to VertexCnt-3 do Mesh.AddTriangle(Vertices[Pnts[0]],Vertices[Pnts[I+2]],Vertices[Pnts[I+1]],C);
	end;
Mesh.ResetTransforms;
Mesh.SetAllNormals(False);
end;




procedure ProcessMeshNormals(Stream: TStream; Mesh: TGLMesh);
var FaceCnt,VertexCnt,Inx,I,F: integer;
var Pnts: array [0..120] of integer;
var VertList: TVertexList;

	procedure AddNormal3(P1,P2,P3: T3DVector);
	begin
	Mesh.VertexNormals.Add(P1);
	Mesh.VertexNormals.Add(P2);
	Mesh.VertexNormals.Add(P3);
	end;


	procedure AddNormal4(P1,P2,P3,P4: T3DVector);
	begin
	{Must add points in exactly the same way as Mesh.AddRectangle}
	AddNormal3(P1,P2,P4);
	AddNormal3(P2,P3,P4);
	end;

begin
VertList:=TVertexList.Create;
try
if not StreamSkipToChar(Stream,'{') then raise Exception.Create('Syntax Error');
ReadVertexList(Stream, VertList,False);
Mesh.VertexNormals.Clear;
FaceCnt:=VertList.Indices[0];
Inx:=1;
for F:=1 to FaceCnt do
	begin
	VertexCnt:=VertList.Indices[Inx];
	Inc(Inx);
	for I:=0 to VertexCnt-1 do
		begin
		Pnts[I]:=VertList.Indices[Inx];
		Inc(Inx);
		end;
	{Order reversed to convert left to right handed coordinates}
	if VertexCnt=3 then AddNormal3(VertList.Vertices[Pnts[2]],VertList.Vertices[Pnts[1]],VertList.Vertices[Pnts[0]])
	else if VertexCnt= 4 then AddNormal4(VertList.Vertices[Pnts[3]],VertList.Vertices[Pnts[2]],VertList.Vertices[Pnts[1]],VertList.Vertices[Pnts[0]])
	else for I:=0 to VertexCnt-3 do AddNormal3(VertList.Vertices[Pnts[0]],VertList.Vertices[Pnts[I+2]],VertList.Vertices[Pnts[I+1]]);
	end;
finally VertList.Free; end;
StreamSkipToChar(Stream,'}');
end;




function ProcessVertexColors(Stream: TStream; var Mesh: TGLMesh; Faces: THugeInteger): integer;
var FaceCnt,VertexCnt,Inx,I,F: integer;
type TColorArray = array of TColor;
var CA: TColorArray;
var Pnts: array [0..120] of integer;
begin
Result:=0;
if not StreamSkipToChar(Stream,'{') then raise Exception.Create('Syntax Error');
Result:=GetXInteger(Stream);
SetLength(CA,Result);
for I:=0 to Result-1 do
	begin
	{Read index}
        GetXInteger(Stream);
	CA[I]:=GLRGBToColor(ReadXRGBA(Stream));
	{Eat last semicolon or comma}
	GetStreamChar(Stream);
	end;
StreamSkipToChar(Stream,'}');

FaceCnt:=Faces[0];
Inx:=1;
for F:=0 to FaceCnt-1 do
	begin
	VertexCnt:=Faces[Inx];
	Inc(Inx);
	for I:=0 to VertexCnt-1 do
		begin
		Pnts[I]:=Faces[Inx];
		Inc(Inx);
		end;

	{Order reversed to convert left to right handed coordinates}
	if VertexCnt=3 then Mesh.FaceColor.SetFaceColors(F,CA[Pnts[2]],CA[Pnts[1]],CA[Pnts[0]])
	else if VertexCnt= 4 then
		begin
		Mesh.FaceColor.SetFaceColors(F,CA[Pnts[3]],CA[Pnts[2]],CA[Pnts[0]]);
		Mesh.FaceColor.SetFaceColors(F+1,CA[Pnts[2]],CA[Pnts[1]],CA[Pnts[0]]);
		end
	else for I:=0 to VertexCnt-3 do Mesh.FaceColor.SetFaceColors(F+I,CA[Pnts[0]],CA[Pnts[I+2]],CA[Pnts[I+1]]);
	end;
end;

procedure ProcessMesh(Stream: TStream; Viewer: TGLViewer);
var Mesh: TGLMesh;
var I,MInx,ColorCnt: integer;
var VertList: TVertexList;
var Token: TXObjectToken;
var MatList: TMaterialList;
var C: TColor;
var GLC: TGLFaceColor;
begin
ColorCnt:=0;
VertList:=TVertexList.Create;
try
MatList:=TMaterialList.Create;
try
if UseNewMeshes or (MeshIndex>Viewer.BaseFrame.MeshList.Count-1) then
	begin
	Viewer.BaseFrame.MeshList.Add;
	Minx:=Viewer.BaseFrame.MeshList.Count-1;
	end
else MInx:=MeshIndex;
Mesh:=TGLMesh(Viewer.BaseFrame.MeshList[MInx]);
Mesh.Clear;
Mesh.Reset;
Mesh.Name:=CheckForIdentifier(Stream);
if Mesh.Name='' then Mesh.Name:='Mesh'+IntToStr(MInx);
if not StreamSkipToChar(Stream,'{') then raise Exception.Create('Syntax Error');

ReadVertexList(Stream, VertList,True);
BuildMesh(Mesh,VertList.Vertices,VertList.Indices);

MatList.Clear;
while StreamProgress.CheckStreamPosition(Stream) do
	begin
	Token:=ProcessBlock(stream);
	case Token of
	 xfComment,xfWhiteSpace: ;
	 xfCloseBrace: break;
	 xfMeshNormals: ProcessMeshNormals(Stream,Mesh);
	 xfMeshVertexColors: ColorCnt:=ProcessVertexColors(Stream,Mesh,VertList.Indices);
	 xfMeshTextureCoords: ProcessMeshTextureCoords(Stream);
	 xfMeshMaterialList: ProcessMeshMaterials(Stream, MatList);
	 xfSkinMeshHeader: SkipBlock(Stream);
	 xfSkinWeights: ProcessSkinWeights(Stream);
	 else SkipBlock(Stream);
	 end;
	end;

if MatList.MaterialCount>0 then
	begin
	Mesh.LightEmission:=MatList.Materials[0].EmissiveColor;
	Mesh.LightSpecular:=MatList.Materials[0].SpecularColor;
	Mesh.LightShininess:=EnsureRange(MatList.Materials[0].Power,0,128);
	if ColorCnt<1 then
		begin
		if MatList.FaceCount<Mesh.FaceColor.Count then
			begin
			Mesh.SetColor(GLRGBToColor(MatList.Materials[0].FaceColor));
			end
		else
			begin
			for I:=0 to Mesh.FaceColor.Count-1 do
				begin
				C:=GLRGBToColor(MatList.Materials[MatList.FaceIndices[I]].FaceColor);
				GLC.C1:=C;  GLC.C2:=C; GLC.C3:=C;
				Mesh.FaceColor[I]:=GLC;
				end;
			end;
		end;
	end;

Mesh.SetAllNormals(Mesh.VertexNormals.Count=0);
finally MatList.Free; end;
finally VertList.Free; end;
Inc(MeshIndex);
Viewer.BaseFrame.MeshList.Change;
end;




procedure ProcessFrameTransformMatrix(Stream: TStream);
var I: integer;
var S: AnsiString;
begin
StreamSkipToChar(Stream,'{');
for I:=0 to 15 do GlobalMatrix[I]:=GetXFloat(Stream);
S:=GLMatrix.GetMatrixStr(GlobalMatrix,3);
StreamSkipToChar(Stream,'}');
end;




procedure ProcessFrame(Stream: TStream; Viewer: TGLViewer);
var FrameName: AnsiString;
var Com: AnsiChar;
var Token: TXObjectToken;
begin
Com:=PeekStreamChar(Stream);
if Com <> '{' then FrameName:=ReadIdentifier(Stream)
else FrameName:='Frame';
if not StreamSkipToChar(Stream,'{') then raise Exception.Create('Syntax Error');

while StreamProgress.CheckStreamPosition(Stream) do
	begin
	Token:=ProcessBlock(Stream);
	case Token of
	 xfComment,xfWhiteSpace: ;
	 xfOpenBrace: SkipBlock(Stream);
     	 xfFrame: ProcessFrame(Stream, Viewer);
         xfFrametransformmatrix: ProcessFrameTransformMatrix(Stream);
         xfMesh: ProcessMesh(Stream, Viewer);
	 xfCloseBrace: break;
         else SkipBlock(Stream);
	 end;
	end;
GlobalMatrix:=IdentityMatrix;
end;



procedure ParseTextXFile(Stream: TStream; Viewer: TGLViewer; ProgInfo: TProgressInfo; NewMeshes: boolean);
var Token: TXObjectToken;
begin
SetLength(GlobalMaterials,0);
GlobalMatrix:=IdentityMatrix;
StreamProgress.ProgInfo:=ProgInfo;
Stream.Seek(0,soFromBeginning);
if not ReadXFileHeader(Stream,FileInfo) then raise Exception.Create('Invalid File');
if FileInfo.FileType<>ffText then raise Exception.Create('Expected Text File');
MeshIndex:=0;
UseNewMeshes:=NewMeshes;
while StreamProgress.CheckStreamPosition(Stream) do
	begin
	Token:=ProcessBlock(Stream);
	case Token of
	 xfMaterial:
		begin
		SetLength(GlobalMaterials,Length(GlobalMaterials)+1);
		GlobalMaterials[High(GlobalMaterials)]:=ProcessMaterial(Stream);
		end;
	 xfWhiteSpace,xfComment,xfCloseBrace: ;
	 xfError: raise Exception.Create('Syntax Error');
         xfFrame: ProcessFrame(Stream,Viewer);
         xfMesh: ProcessMesh(Stream,Viewer);
         xfAnimationSet: ProcessAnimationSets(Stream);
	 xfOpenBrace: SkipBlock(Stream);
	 else SkipBlock(Stream);
	 end;
	end;
end;




initialization
StreamProgress:=TStreamProgress.Create;
finalization
StreamProgress.Free;
end.
