unit Resizebm;

{
Based on the Dr.Dobbs article, "Augural Image Zooming" by Wm. Douglas Withers,
the book Graphics Gems and "General Filtered Image Rescaling" by Dale Schumacher
appearing in the book Graphics Gems III, published by Academic Press, Inc.
}

interface

uses  Windows,SysUtils, Classes, Graphics, Math;

type TFilterProc = function(Value: Single): Single;
type TProgressProc = procedure(Value: integer);
var ProgressInc: double;
var ProgressValue: double;

type TImageQuality = (iqSoft,iqMedium,iqSharp);

procedure StretchBitmap(SourceBitmap: TBitmap; NewWidth,NewHeight: integer; FilterInx: integer; ProgressProc:TProgressProc); overload;
procedure StretchBitmap(SourceBitmap: TBitmap; NewWidth,NewHeight: integer; Quality: TImageQuality; ProgressProc:TProgressProc); overload;
function GetFunctionValue(Value: single; Inx: integer): single;
procedure GetFilterNames(List: TStrings);

implementation

uses DSupport;



{---------------- Pseudo Bitmap so we are not using GDI memory -------}


type
 TPseudoBitmap = class(TObject)
 private
 MemPointer: Pointer;
 BytesPerScanline,MemSize: integer;
 FWidth, FHeight, FBytesPerPixel: integer;
 function GetScanLine(Row: Integer): Pointer;
 public
 constructor Create(Width,Height,BytesPerPixel: integer);
 destructor Destroy; override;
 property ScanLine[Row: Integer]: Pointer read GetScanLine;
 property Height: integer read FHeight;
 property Width: integer read FWidth;
 end;


constructor TPseudoBitmap.Create(Width,Height,BytesPerPixel: integer);
begin
inherited Create;
FWidth:=Width;
FHeight:=Height;
FBytesPerPixel:=BytesPerPixel;
BytesPerScanline:=Width*BytesPerPixel;
MemSize:=Width*Height*BytesPerPixel;
GetMem(MemPointer,MemSize);
end;


function TPseudoBitmap.GetScanLine(Row: Integer): Pointer;
begin
if (Row < 0) or (Row > FHeight) then raise Exception.Create('Scanline out of range');
Integer(Result) := Integer(MemPointer) + Row * BytesPerScanline;
end;


destructor TPseudoBitmap.Destroy;
begin
FreeMem(MemPointer);
inherited Destroy;
end;

{------------------ Slice bitmap into pieces -------------------------}


procedure GetHorzontalSlice(DBM,SBM: TBitmap; Top,Bottom: integer);
{Get a horizontal slice of the source bitmap and put into destination}
begin
if Bottom>SBM.Height then Bottom:=SBM.Height;
DBM.Width:=SBM.Width;
DBM.Height:=Bottom-Top;
DBM.Canvas.CopyRect(Rect(0,0,DBM.Width,DBM.Height),SBM.Canvas,Rect(0,Top,DBM.Width,Bottom));
end;



procedure PutVerticalSlice(DBM,SBM: TBitmap; Left,Right: integer);
{Put a source as a vertical slice into destination}
begin
if Right>DBM.Width then Right:=DBM.Width;
DBM.Canvas.CopyRect(Rect(Left,0,Right,DBM.Height),SBM.Canvas,Rect(0,0,Right-Left,DBM.Height));
end;

{------------------ Handle Progress Functions -------------------------}


procedure AdvanceProgress(ProgressProc: TProgressProc);
{Increment progress value and call Progress event }
begin
ProgressValue:=ProgressValue+ProgressInc;
if Assigned(ProgressProc) then ProgressProc(Trunc(ProgressValue));
end;


procedure SetProgress(Steps: integer);
{Initialize Progress }
begin
ProgressValue:=0;
ProgressInc:=100/Steps;
end;



{------------------ Filter Functions ----------------------------------}

function HermiteFilter(Value: Single): Single;
begin
{ f(t) = 2|t|^3 - 3|t|^2 + 1, -1 <= t <= 1}
if Value < 0 then Value := -Value;
if Value < 1 then Result := (2 * Value - 3) * Sqr(Value) + 1
else Result := 0;
end;



function BoxFilter(Value: Single): Single;
{ a.k.a. "Nearest Neighbour" filter}
begin
if (Value > -0.5) and (Value <= 0.5) then Result := 1
else Result := 0;
end;



function TriangleFilter(Value: Single): Single;
{ a.k.a. "Linear" or "Bilinear" filter}
begin
if Value < 0 then Value := -Value;
if Value < 1 then Result := 1 - Value
else Result := 0;
end;


function BellFilter(Value: Single): Single;
begin
if Value < 0 then Value := -Value;
if Value < 0.5 then Result := 0.75 - Sqr(Value)
else if Value < 1.5 then
	begin
	Value := Value - 1.5;
	Result := 0.5 * Sqr(Value);
	end
else Result := 0;
end;


function SplineFilter(Value: Single): Single;
{ B-spline filter}
var tt: single;
begin
if Value < 0 then Value := -Value;
if Value < 1 then
	begin
	tt := Sqr(Value);
	Result := 0.5*tt*Value - tt + 2 / 3;
	end
else if Value < 2 then
	begin
	Value := 2 - Value;
	Result := 1/6 * Sqr(Value) * Value;
	end
else Result := 0;
end;


function Lanczos3Filter(Value: Single): Single;

	function SinC(Value: Single): Single;
	begin
	if Value <> 0 then
		begin
		Value := Value * Pi;
		Result := sin(Value) / Value;
		end
	else Result := 1;
	end;

begin
if Value < 0 then Value := -Value;
if Value < 3 then Result := SinC(Value) * SinC(Value / 3)
else Result := 0;
end;


function MitchellFilter(Value: Single): Single;
const B = (1 / 3);
const C = (1 / 3);
var tt: single;
begin
if Value < 0 then Value := -Value;
tt := Sqr(Value);
if Value < 1 then
	begin
	Value := (((12 - 9 * B - 6 * C) * (Value * tt))
	+ ((-18 + 12 * B + 6 * C) * tt)
	+ (6 - 2 * B));
	Result := Value / 6;
	end
else if Value < 2 then
	begin
	Value := (((-1 * B - 6 * C) * (Value * tt))
	+ ((6 * B + 30 * C) * tt)
	+ ((-12 * B - 48 * C) * Value)
	+ (8 * B + 24 * C));
	Result := Value / 6;
	end
else Result := 0;
end;

const MaxFilters=7;

const
  ResampleFilters: array[0..MaxFilters-1] of record
    Name: string;
    Filter: TFilterProc;
    Width: Single;
  end = (
    (Name: 'Box';	Filter: BoxFilter;	Width: 0.5),
    (Name: 'Triangle';	Filter: TriangleFilter;	Width: 1),
    (Name: 'Hermite';	Filter: HermiteFilter;	Width: 1),
    (Name: 'Bell';	Filter: BellFilter;	Width: 1.5),
    (Name: 'B-Spline';	Filter: SplineFilter;	Width: 2),
    (Name: 'Mitchell';	Filter: MitchellFilter;	Width: 2),
    (Name: 'Lanczos3';	Filter: Lanczos3Filter;	Width: 3)
    );


type
  { Contributor for a pixel}

TContributor = record
  PixelInx: integer;		{ Source pixel}
  PixelWeight: single;		{ Pixel weight}
  end;

TContributorList = array[0..0] of TContributor;
PContributorList = ^TContributorList;

{ List of source pixels contributing to a destination pixel}

TContribItem = record
  PixelInx: integer;
  ListPointer: PContributorList;
  end;

TCListList = array[0..0] of TContribItem;
PCListList = ^TCListList;

TRGB = packed record
  Red, Green, Blue: single;
  end;

{ Physical bitmap pixel}
TColorRGB = packed record
  Red, Green, Blue: BYTE;
  end;
PColorRGB = ^TColorRGB;

{ Physical bitmap scanline (row)}
TRGBList = packed array[0..0] of TColorRGB;
PRGBList = ^TRGBList;



procedure StretchBitmap(SourceBitmap: TBitmap; NewWidth,NewHeight: integer; FilterInx: integer; ProgressProc:TProgressProc);
var XScale, YScale: single;
var I: integer;
var WorkBitmap: TPseudoBitmap;
var ContribList: PCListList;
var SourceLine,DestLine: PRGBList;
var DestPixel: PColorRGB;
var Delta,DestDelta: integer;
var SourceWidth,SourceHeight: integer;
var FilterProc: TFilterProc;
var SampleRadius: single;



procedure BuildWeightArray(Size, Limit: integer; Radius,Scale: single);
{ Pre-calculate Filter contributions for rows or columns}
var J,K,I: integer;
var ScaleFactor: single;
var CenterPixel: single;		
var PixelWeight: single;
var LeftPixel, RightPixel: integer;
var PixelInx: integer;
begin
if Scale < 1 then
	begin
	Radius:=Radius / Scale;
	ScaleFactor:=1 / Scale;
	end
else ScaleFactor:=1;
for I:=0 to Size-1 do
	begin
	AdvanceProgress(ProgressProc);
	ContribList^[I].PixelInx:=0;
	GetMem(ContribList^[I].ListPointer, trunc(Radius * 2 + 1) * sizeof(TContributor));
	CenterPixel:=I / Scale;
	LeftPixel:=floor(CenterPixel - Radius);
	RightPixel:=ceil(CenterPixel + Radius);
	for J:=LeftPixel to RightPixel do
		begin
		PixelWeight:=FilterProc((CenterPixel - J) / ScaleFactor) / ScaleFactor;
		if PixelWeight = 0 then continue;
		if J < 0 then PixelInx:=-J
		else if J >= Limit then PixelInx:=Limit - J + Limit - 1
		else PixelInx:=J;
		K:=ContribList^[I].PixelInx;
		ContribList^[I].ListPointer^[K].PixelInx:=PixelInx;
		ContribList^[I].ListPointer^[K].PixelWeight:=PixelWeight;
		Inc(ContribList^[I].PixelInx);
		end;
	end;
end;

procedure ApplyFilter(Size, ItemDelta, BlockDelta: integer);
{Apply filter using acquired weights }
{Data can be rows or columns depending on ItemDelta and BlockDelta}
var I,J: integer;
var PixelWeight: single;
var RGBValue: TRGB;
var CurColor: TColorRGB;
begin
for I:=0 to Size-1 do
	begin
	RGBValue.Red:=0;
	RGBValue.Green:=0;
	RGBValue.Blue:=0;
	for J:=0 to ContribList^[I].PixelInx-1 do
		begin
		CurColor:=PColorRGB(integer(SourceLine)+ContribList^[I].ListPointer^[J].PixelInx*ItemDelta)^;
		PixelWeight:=ContribList^[I].ListPointer^[J].PixelWeight;
		if PixelWeight = 0 then continue;
		RGBValue.Red:=RGBValue.Red + CurColor.Red * PixelWeight;
		RGBValue.Green:=RGBValue.Green + CurColor.Green * PixelWeight;
		RGBValue.Blue:=RGBValue.Blue + CurColor.Blue * PixelWeight;
		end;
	if RGBValue.Red > 255 then CurColor.Red:=255
	else if RGBValue.Red < 0 then CurColor.Red:=0
	else CurColor.Red:=round(RGBValue.Red);
	if RGBValue.Green > 255 then CurColor.Green:=255
	else if RGBValue.Green < 0 then CurColor.Green:=0
	else CurColor.Green:=round(RGBValue.Green);
	if RGBValue.Blue > 255 then CurColor.Blue:=255
	else if RGBValue.Blue < 0 then CurColor.Blue:=0
	else CurColor.Blue:=round(RGBValue.Blue);
	DestPixel^:=CurColor;
	inc(Integer(DestPixel), BlockDelta);
	end;
end;

procedure DoHorizontalSlices;
{ Apply Filter to sample horizontally from Source to WorkBitmap}
{ Get slices of the source so you don't have to convert the whole}
{thing to 24-bit. This helps avoid overwhelming the GDI under Win95/98/Me}
var S,I,Slices,SliceTop,SliceHeight: integer;
var TBM: TBitmap;
const MaxSize=1024*1024;
begin
TBM:=TBitmap.Create;
TBM.PixelFormat:=pf24bit;
Slices:=Round(SourceBitmap.Height/(MaxSize/(3*SourceBitmap.Width)))+1;
SliceHeight:=SourceBitmap.Height div Slices;
S:=0;
while true do
	begin
	SliceTop:=S*SliceHeight;
	if SliceTop>=SourceBitmap.Height then break;
	GetHorzontalSlice(TBM,SourceBitmap, SliceTop,SliceTop+SliceHeight);
	for I:=0 to TBM.Height-1 do
		begin
		AdvanceProgress(ProgressProc);
		SourceLine:=TBM.ScanLine[I];
		DestPixel:=WorkBitmap.ScanLine[SliceTop+I];
		ApplyFilter(NewWidth,SizeOf(TColorRGB),SizeOf(TColorRGB));
		end;
	Inc(S);
	end;
TBM.Free;
end;



procedure DoVerticalSlice;
{ Apply Filter to sample vertically from WorkBitmap to Dest}
var I: integer;
var TBM: TBitmap;
var S,Slices,SliceWidth,SliceLeft: integer;
const MaxSize=1024*1024;
begin
Slices:=Round(WorkBitmap.Height/(MaxSize/(3*WorkBitmap.Width)))+1;

SourceBitmap.Width:=NewWidth;
SourceBitmap.Height:=NewHeight;
SliceWidth:=NewWidth div Slices;

TBM:=TBitmap.Create;
TBM.Width:=SliceWidth;
TBM.Height:=NewHeight;
TBM.PixelFormat:=pf24bit;

SourceLine:=WorkBitmap.ScanLine[0];
Delta:=integer(WorkBitmap.ScanLine[1]) - integer(SourceLine);

S:=0;
while true do
	begin
	SliceLeft:=S*SliceWidth;
	if SliceLeft>=SourceBitmap.Width then break;
	DestLine:=TBM.ScanLine[0];
	DestDelta:=integer(TBM.ScanLine[1]) - integer(DestLine);
	for I:=0 to SliceWidth-1 do
		begin
		AdvanceProgress(ProgressProc);
		DestPixel:=pointer(DestLine);
		ApplyFilter(NewHeight,Delta,DestDelta);
		Inc(SourceLine, 1);
		Inc(DestLine, 1);
		end;
	PutVerticalSlice(SourceBitmap,TBM,SliceLeft,SliceLeft+SliceWidth);
	Inc(S);
	end;
TBM.Free;
end;



begin
FilterProc:=ResampleFilters[FilterInx].Filter;
SampleRadius:=ResampleFilters[FilterInx].Width;

SourceWidth:=SourceBitmap.Width;
SourceHeight:=SourceBitmap.Height;
if (SourceWidth < 1) or (SourceHeight < 1) then
	raise Exception.Create('Source bitmap too small');
SetProgress(NewWidth+SourceHeight+NewHeight+NewWidth);

{ Create intermediate image to hold horizontal zoom}
WorkBitmap:=TPseudoBitmap.Create(NewWidth,SourceHeight,3);
try

if SourceWidth = 1 then XScale:=NewWidth / SourceWidth
else XScale:=(NewWidth - 1) / (SourceWidth - 1);

if SourceHeight = 1 then YScale:=NewHeight / SourceHeight
else YScale:=(NewHeight - 1) / (SourceHeight - 1);


{ Build horizontal weight array }
GetMem(ContribList, NewWidth* sizeof(TContribItem));
BuildWeightArray(NewWidth,SourceWidth,SampleRadius,XScale);

DoHorizontalSlices;

{ Free the memory allocated for horizontal Filter weights}
for I:=0 to NewWidth-1 do FreeMem(ContribList^[I].ListPointer);
FreeMem(ContribList);

{ Build Vertical Weight Array }
GetMem(ContribList, NewHeight* sizeof(TContribItem));
BuildWeightArray(NewHeight,SourceHeight,SampleRadius,YScale);

DoVerticalSlice;

{ Free the memory allocated for vertical FilterProc weights}
for I:=0 to NewHeight-1 do FreeMem(ContribList^[I].ListPointer);
FreeMem(ContribList);

finally WorkBitmap.Free; end;
end;



procedure StretchBitmap(SourceBitmap: TBitmap; NewWidth,NewHeight: integer; Quality: TImageQuality; ProgressProc:TProgressProc);
var FilterIndex: integer;
begin
case Quality of
 iqSoft: FilterIndex:=4;
 iqMedium: FilterIndex:=5;
 iqSharp: FilterIndex:=6;
 end;
StretchBitmap(SourceBitmap,NewWidth,NewHeight, FilterIndex, ProgressProc);
end;



{------------------ Support routines --------------------------------}


procedure GetFilterNames(List: TStrings);
var I: integer;
begin
List.Clear;
for I:=0 to MaxFilters-1 do
	List.Add(ResampleFilters[I].Name);
end;


function GetFunctionValue(Value: single; Inx: integer): single;
var FilterProc: TFilterProc;
begin
FilterProc:=ResampleFilters[Inx].Filter;
Result:=FilterProc(Value);
end;


end.
