unit SelectionBox;

interface

uses
  Windows, Messages, Forms, Graphics, ExtCtrls, SysUtils, Classes;

type TBoxCorners=(bcNone,bcTopLeft,bcTopRight,bcBottomLeft,bcBottomRight,bcCenter);

type
  TSelectionBox = class(TObject)
  private
   BackBitmap: TBitmap;
   FImageBox: TRect;
   FCanvas: TCanvas;
   FScrollBox: TScrollBox;
   BoxDragMode: TBoxCorners;
   StartX,StartY: integer;
   SelectBox,OldBox: TRect;
   LeftDown: boolean;
   BoxDrawn: boolean;
   function CornerHitTest(P: TPoint): TBoxCorners;
   procedure DrawRectangle(Canvas: TCanvas; R: TRect);
   procedure DrawHandle(Canvas: TCanvas; P: TPoint; Corner: TBoxCorners);
   procedure DrawBox(Canvas: TCanvas; R: TRect);
   procedure CursorScroll(X,Y: integer);
   procedure UpdateBox(X1,Y1,X2,Y2: integer);
   procedure HideBox;
   procedure RestoreBox;
   procedure RestoreBackground;
   procedure SaveBackground;
  protected
  public
   constructor Create(Canvas: TCanvas; Box: TRect; ScrollBox: TScrollBox);
   destructor Destroy; override;
   procedure BoxMouseDown(X,Y: integer);
   procedure BoxMouseMove(X,Y: integer);
   procedure BoxMouseUp;
  published
  end;


implementation



constructor TSelectionBox.Create(Canvas: TCanvas; Box: TRect; ScrollBox: TScrollBox);
begin
FImageBox:=Box;
FCanvas:=Canvas;
FScrollBox:=ScrollBox;
BackBitmap:=TBitmap.Create;
BackBitmap.Width:=FImageBox.Right-FImageBox.Left;
BackBitmap.Height:=FImageBox.Bottom-FImageBox.Top;
BoxDrawn:=False;
end;


destructor TSelectionBox.Destroy;
begin
BackBitmap.Free;
inherited Destroy;
end;



{-------------------Handle Drawing Selection Box --------------------}


function TSelectionBox.CornerHitTest(P: TPoint): TBoxCorners;
{Test if point falls in one of the corners}
const Size=5;
begin
Result:=bcNone;
if PtInRect(Rect(SelectBox.Left-Size,SelectBox.Top-Size,SelectBox.Left+Size,SelectBox.Top+Size),P) then Result:=bcTopLeft
else if PtInRect(Rect(SelectBox.Right-Size,SelectBox.Top-Size,SelectBox.Right+Size,SelectBox.Top+Size),P) then Result:=bcTopRight
else if PtInRect(Rect(SelectBox.Left-Size,SelectBox.Bottom-Size,SelectBox.Left+Size,SelectBox.Bottom+Size),P) then Result:=bcBottomLeft
else if PtInRect(Rect(SelectBox.Right-Size,SelectBox.Bottom-Size,SelectBox.Right+Size,SelectBox.Bottom+Size),P) then Result:=bcBottomRight
else if PtInRect(SelectBox,P) then Result:=bcCenter;
end;



procedure TSelectionBox.SaveBackground;
var R1,R2: TRect;
begin
R1:=Rect(0,0,FImageBox.Right-FImageBox.Left,FImageBox.Bottom-FImageBox.Top);
R2:=Rect(0,0,BackBitmap.Width,BackBitmap.Height);
BackBitmap.Canvas.CopyRect(R2,FCanvas,R1);
end;


procedure TSelectionBox.RestoreBackground;
var R1,R2: TRect;
begin
R1:=Rect(0,0,FImageBox.Right-FImageBox.Left,FImageBox.Bottom-FImageBox.Top);
R2:=Rect(0,0,BackBitmap.Width,BackBitmap.Height);
FCanvas.CopyRect(R1,BackBitmap.Canvas,R2);
BoxDrawn:=False;
end;




procedure TSelectionBox.DrawRectangle(Canvas: TCanvas; R: TRect);
{Draw a rectangle on specified canvas.}
begin
Canvas.MoveTo(R.Left,R.Top);
Canvas.LineTo(R.Right,R.Top);
Canvas.LineTo(R.Right,R.Bottom);
Canvas.LineTo(R.Left,R.Bottom);
Canvas.LineTo(R.Left,R.Top);
end;



procedure TSelectionBox.DrawHandle(Canvas: TCanvas; P: TPoint; Corner: TBoxCorners);
{Draw one of the corner handles}
var R: TRect;
const Size=3;
begin
R:=Rect(P.X-Size,P.Y-Size,P.X+Size,P.Y+Size);

Canvas.Brush.Style:=bsSolid;
Canvas.Brush.Color:=clAqua;
Canvas.Pen.Style:=psSolid;
Canvas.Pen.Color:=Canvas.Brush.Color;
Canvas.Pen.Mode:=pmCopy;
Canvas.Rectangle(R.Left,R.Top,R.Right,R.Bottom);
end;


procedure TSelectionBox.DrawBox(Canvas: TCanvas; R: TRect);
{Draw current clipping box}
var R2: TRect;
begin
SaveBackground;

Canvas.Pen.Style:=psSolid;
Canvas.Pen.Color:=clRed;
Canvas.Pen.Mode:=pmCopy;
DrawRectangle(Canvas,R);

R2:=R;
InflateRect(R2,1,1);
DrawRectangle(Canvas,R2);
if (R.Left=0) and (R.Top=0) and (R.Right=0) and (R.Bottom=0) then exit;

DrawHandle(Canvas,R.TopLeft,bcTopLeft);
DrawHandle(Canvas,R.BottomRight,bcBottomRight);
DrawHandle(Canvas,Point(R.Right,R.Top),bcTopRight);
DrawHandle(Canvas,Point(R.Left,R.Bottom),bcBottomLeft);
BoxDrawn:=True;
end;


procedure TSelectionBox.CursorScroll(X,Y: integer);
{Scroll box based on position X,Y}
var RX,RY: integer;
var XLow,XHigh: integer;
var MovX,MovY: integer;
begin
if FScrollBox=nil then exit;
RX:=X+FImageBox.Left;
RY:=Y+FImageBox.Top;
if RX<0 then MovX:=-RX
else if RX>FScrollBox.ClientWidth then MovX:=RX-FScrollBox.ClientWidth
else MovX:=0;

if RY<0 then MovY:=-RY
else if RY>FScrollBox.ClientHeight then MovY:=RY-FScrollBox.ClientHeight
else MovY:=0;

if RX<0 then FScrollBox.HorzScrollBar.Position:=FScrollBox.HorzScrollBar.Position-MovX;
if RX>FScrollBox.ClientWidth then FScrollBox.HorzScrollBar.Position:=FScrollBox.HorzScrollBar.Position+MovX;
if RY<0 then FScrollBox.VertScrollBar.Position:=FScrollBox.VertScrollBar.Position-MovY;
if RY>FScrollBox.ClientHeight then FScrollBox.VertScrollBar.Position:=FScrollBox.VertScrollBar.Position+MovY;
end;


procedure TSelectionBox.UpdateBox(X1,Y1,X2,Y2: integer);
{Erase old box and draw new box}
var T: integer;
begin
if BoxDrawn then RestoreBackground;
SelectBox:=Rect(X1,Y1,X2,Y2);
DrawBox(FCanvas,SelectBox);
end;


procedure TSelectionBox.HideBox;
{Hide current box}
begin
RestoreBackground;
end;


procedure TSelectionBox.RestoreBox;
{Restore hidden box}
begin
UpdateBox(OldBox.Left,OldBox.Top,OldBox.Right,OldBox.Bottom);
end;



{---------------------Selection Box Event Handlers -------------------}


procedure TSelectionBox.BoxMouseDown(X,Y: integer);
begin
LeftDown:=True;
BoxDragMode:=CornerHitTest(Point(X,Y));
if BoxDragMode=bcNone then UpdateBox(X,Y,X,Y);
StartX:=X; StartY:=Y;
end;


procedure TSelectionBox.BoxMouseMove(X,Y: integer);
var Rect: TRect;
begin
if LeftDown then
	begin
	CursorScroll(X,Y);
	case BoxDragMode of
	 bcNone,bcBottomRight: UpdateBox(SelectBox.Left,SelectBox.Top,X,Y);
	 bcTopLeft: UpdateBox(X,Y,SelectBox.Right,SelectBox.Bottom);
	 bcTopRight: UpdateBox(SelectBox.Left,Y,X,SelectBox.Bottom);
	 bcBottomLeft: UpdateBox(X,SelectBox.Top,SelectBox.Right,Y);
	 bcCenter:
		begin
		Rect:=SelectBox;
		OffsetRect(Rect,X-StartX,Y-StartY);
		UpdateBox(Rect.Left,Rect.Top,Rect.Right,Rect.Bottom);
		end;
	 end;
	StartX:=X; StartY:=Y;
	end;
end;

procedure TSelectionBox.BoxMouseUp;
begin
LeftDown:=False;
end;


end.
