unit StatsLib;

interface

uses HugeArry,Math;


type TRunningAverage = class(THugeDouble)
 private
  FRunMax: integer;
  function GetAverage: double;
  procedure SetRunMax(V: integer);
 public
  constructor Create;
  property RunMax: integer read FRunMax write SetRunMax;
  procedure AddItem(V: double);
  property Average: double read GetAverage;
 end;


function GaussianDistribution(X, STD: double): double;
function NormalProbability(X,STD,Mean: double): double;
function ScaledProbability(X,STD,Mean: double): double;
function Correlation(X,Y: THugeDouble): double;

function GetArraySum(AD: THugeDouble): double; overload;
function GetArraySum(AD: array of double): double; overload;

function GetArrayMean(AD: THugeDouble): double; overload;
function GetArrayMean(AD: array of double): double; overload;

function GetArrayVariance(AD: THugeDouble): double; overload;
function GetArrayVariance(AD: array of double): double; overload;

function GetArraySTD(AD: THugeDouble): double; overload;
function GetArraySTD(AD: array of double): double; overload;

procedure GetArrayMinMax(AD: THugeDouble; var Min,Max: double);

function GetPartialArraySum(AD: THugeDouble; First,Last: integer): double;
function GetPartialArrayMean(AD: THugeDouble; First,Last: integer): double;

function GetArrayMedian(AD: THugeDouble): double;


implementation

{ Pearson's Correlation Coefficient

                   Exy - 1/n Ex Ey
R =  ------------------------------------------------
     Sqrt(Ex^2 - 1/n (Ex)^2)  Sqrt(EY^2 - 1/n (EY)^2)
}



function Correlation(X,Y: THugeDouble): double;
var I,N: integer;
var S1,S2,S3,S4,S5,S6,S7: double;
var A,B,C,T: double;
begin
N:=X.ItemIndex;
S1:=0; for I:=0 to N-1 do S1:=S1+ (X[I]*Y[I]);
S2:=0; for I:=0 to  N-1 do S2:=S2 + X[I];
S3:=0; for I:=0 to  N-1 do S3:=S3 + Y[I];
S4:=0; for I:=0 to  N-1 do S4:=S4 + sqr(X[I]);

S5:=sqr(S2);

S6:=0; for I:=0 to  N-1 do S6:=S6 + sqr(Y[I]);
S7:=sqr(S3);
A:=S1 -  ((S2 * S3)/N);
B:=sqrt(S4 - (S5/N));
C:=sqrt(S6 - (S7/N));
T:=(B * C);
if T=0 then T:=1e-200;
Result:=A / T;
end;



function GaussianDistribution(X, STD: double): double;
{ Gaussian distribution Return Y value for specified STD  }
{Peak value at mean is always one}
begin
Result:=Exp(- sqr(X) / (2 * sqr(STD)));
end;



function NormalProbability(X,STD,Mean: double): double;
{Return Y value for specified STD}
{Peak value at mean is always one}
var T,E: double;
begin
if STD=0 then STD:=1e-200;
T:=1/(STD * Sqrt(2 * Pi));
E:=Exp(-Sqr(X - Mean)/(2 * sqr(STD)));
Result:=E * T;
end;





function ScaledProbability(X,STD,Mean: double): double;
{Return distribution scaled so mean value is 1}
var K: double;
begin
K:=NormalProbability(0,STD,Mean);
Result:=NormalProbability(X,STD,Mean) * (1/K);
end;


{---------------- Calculate Standard Statistical Functions -------------------}



function GetArraySum(AD: THugeDouble): double;
{Calculate the sum of array values}
var I: integer;
begin
Result:=0;
for I:=0 to AD.Count-1 do
 Result:=Result + AD[I];
end;


function GetArraySum(AD: array of double): double; overload;
var I: integer;
var HD: THugeDouble;
begin
HD:=THugeDouble.Create(5);
try
HD.Clear;
for I:=0 to High(AD) do HD.Add(AD[I]);
Result:=GetArraySum(HD);
finally HD.Free; end;
end;



function GetPartialArraySum(AD: THugeDouble; First,Last: integer): double;
{Calculate the sum of partial array values}
var I: integer;
begin
Result:=0;
if First<0 then First:=0;
if Last>=AD.Count then Last:=AD.Count-1;
for I:=First to Last do
 Result:=Result + AD[I];
end;



function GetPartialArrayMean(AD: THugeDouble; First,Last: integer): double;
{Calculate the mean of part of array values}
var Count: integer;
begin
if First<0 then First:=0;
if Last>=AD.Count then Last:=AD.Count-1;
Count:=Last - First;
if Count<1 then Result:=0
else Result:=GetPartialArraySum(AD,First,Last)/ Count;
end;



function GetArrayMean(AD: THugeDouble): double;
{Calcualte the mean of array values}
begin
if AD.Count<1 then Result:=0
else Result:=GetArraySum(AD)/ AD.Count;
end;



function GetArrayMean(AD: array of double): double;
var I: integer;
var HD: THugeDouble;
begin
HD:=THugeDouble.Create(5);
try
HD.Clear;
for I:=0 to High(AD) do HD.Add(AD[I]);
Result:=GetArrayMean(HD);
finally HD.Free; end;
end;



function GetArrayVariance(AD: THugeDouble): double;
{Calculate the variance of specified array}
var I: integer;
var Mean: double;
begin
Result:=0;
Mean:=GetArrayMean(AD);
for I:=0 to AD.Count-1 do Result:=Result + Sqr(AD[I] - Mean);
if AD.Count<1 then Result:=0
else Result:=Result / AD.Count;
end;



function GetArrayVariance(AD: array of double): double;
var I: integer;
var HD: THugeDouble;
begin
HD:=THugeDouble.Create(5);
try
HD.Clear;
for I:=0 to High(AD) do HD.Add(AD[I]);
Result:=GetArrayVariance(HD);
finally HD.Free; end;
end;



function GetArraySTD(AD: THugeDouble): double;
{Calculate the Standard Deviation of specified array}
begin
Result:=sqrt(GetArrayVariance(AD));
end;



function GetArraySTD(AD: array of double): double;
var I: integer;
var HD: THugeDouble;
begin
HD:=THugeDouble.Create(5);
try
HD.Clear;
for I:=0 to High(AD) do HD.Add(AD[I]);
Result:=GetArraySTD(HD);
finally HD.Free; end;
end;



function RealCompare(P1,P2: Pointer): integer;
type PDouble = ^double;
begin
Result:=round(PDouble(P1)^ - PDouble(P2)^);
end;



function GetArrayMedian(AD: THugeDouble): double;
var I: integer;
var HD: THugeDouble;
begin
HD:=THugeDouble.Create(5);
try
HD.Clear;
HD.Assign(AD);
HD.Sort(RealCompare);
Result:=HD[HD.Count div 2];
finally HD.Free; end;
end;





procedure GetArrayMinMax(AD: THugeDouble; var Min,Max: double);
{Calculate the sum of array values}
var I: integer;
begin
if AD.Count<1 then exit;
Min:=AD[0]; Max:=AD[0];
for I:=0 to AD.Count-1 do
	begin
	if AD[I]>Max then Max:=AD[I];
	if AD[I]<Min then Min:=AD[I];
	end;
end;

{-------------- Routines for running average object ---------------------------}


procedure TRunningAverage.SetRunMax(V: integer);
begin
if V<>FRunMax then
	begin
	FRunMax:=V;
	if V<Count then Count:=V;
	end;
end;


function TRunningAverage.GetAverage: double;
begin
Result:=GetArrayMean(Self);
end;



procedure TRunningAverage.AddItem(V: double);
begin
while Count>=FRunMax do Self.Delete(0);
Self.Add(V);
end;



constructor TRunningAverage.Create;
begin
inherited Create(10);
FRunMax:=10;
end;




end.
