unit TEACypherLib;

{Pascal/Delphi implementation of TEA - Tiny Encryption Algorithm}
// Algorithm: David Wheeler & Roger Needham, Cambridge University Computer Lab
// TEA:       http://www.cl.cam.ac.uk/ftp/papers/djw-rmn/djw-rmn-tea.html (1994)

interface

uses Classes,StreamLib;

const
  Delta: longword = $9e3779b9;

type

  TLong2 = array[0.. 1] of Longword;  // 64-bit
  TTeaKey = array[0..3] of longword;  // 128-bit
  TByte16 = array[0..15] of byte;     // 128-bit
  TByte4 = array[0..3] of byte;       // 32-bit
  TTeaData = array of longword;       // n*32-bit
  TByteArrData = array of byte;       // n*8-bit

procedure TeaEncrypt(var data: TLong2; const key: TTeaKey);
procedure TeaDecrypt(var data: TLong2; const key: TTeaKey);


procedure BlockTeaEncrypt(data: TTeaData; const key: TTeaKey);
procedure BlockTeaDecrypt(data: TTeaData; const key: TTeaKey);


function SameKey(const key1, key2: TTeaKey): boolean;

procedure StrToKey(const s: string; var key: TTeaKey);
function KeyToStr(const key: TTeaKey): string;
procedure StrToData(const s: string; var data: TTeaData);
procedure DataToStr(var s: string; const data: TTeaData);

procedure ReadData(const FileName: string; var data: TTeaData);
procedure WriteData(const FileName: string; var data: TTeaData);

procedure EncryptFile(Filename: string; const key: TTeaKey);
procedure DecryptFile(Filename: string; const key: TTeaKey);
procedure DecryptToString(Filename: string; const key: TTeaKey; var S: string);

procedure EncryptStream(Stream: TStream; const key: TTeaKey);
procedure DecryptStream(Stream: TStream; const key: TTeaKey);
function StringToKey(S: string): TTeaKey;

function DecryptString(S, Key: string): string;
function EncryptString(S, Key: string): string;


implementation

uses  math, SysUtils;



procedure TeaEncrypt(var data: TLong2; const key: TTeaKey);
var y,z,sum: Longword;
var a:byte;
begin
y:=data[0];
z:=data[1];
sum:=0;
for a:=0 to 31 do
	begin
	inc(sum,Delta);
	inc(y,((z shl 4)+key[0]) xor (z+sum) xor ((z shr 5)+key[1]));
	inc(z,((y shl 4)+key[2]) xor (y+sum) xor ((y shr 5)+key[3]));
	end;
data[0]:=y;
data[1]:=z
end;

procedure TeaDecrypt(var data: TLong2; const key: TTeaKey);
var y,z,sum: Longword;
var a:byte;
begin
y:=data[0];
z:=data[1];
sum:=delta shl 5;
for a:=0 to 31 do
	begin
	dec(z,((y shl 4)+key[2]) xor (y+sum) xor ((y shr 5)+key[3]));
	dec(y,((z shl 4)+key[0]) xor (z+sum) xor ((z shr 5)+key[1]));
	dec(sum,Delta);
	end;
data[0]:=y;
data[1]:=z
end;


procedure BlockTeaEncrypt(data: TTeaData; const key: TTeaKey);
var z, y, sum, e, p: longword;
var q, n: integer;

	function mx: longword;
	begin
	result := (((z shl 4) xor (z shr 5)) + z) xor (key[(p and 3) xor e] + sum);
	end;

begin
n := Length(data);
q := 6 + 52 div n;
z := data[n-1];
sum := 0;
repeat
	begin
	inc(sum,Delta);
	e := (sum shr 2) and 3;
	for p := 0 to n-1 do
		begin
		y := data[p];
		inc(y,mx);
		data[p] := y;
		z := y;
		end;
	dec(q);
	end
until q = 0;
end;

procedure BlockTeaDecrypt(data: TTeaData; const key: TTeaKey);
var z, y, sum, e, p, q: longword;
var n: integer;

	function mx: longword;
	begin
	result := (((z shl 4) xor (z shr 5)) + z) xor (key[(p and 3) xor e] + sum);
	end;

begin
n := Length(data);
q := 6 + 52 div n;
sum := q*Delta;
while sum <> 0 do
	begin
	e := (sum shr 2) and 3;
	for p := n-1 downto 1 do
		begin
		z := data[p-1];
		y := data[p];
		dec(y,mx);
		data[p] := y;
		end;
	z := data[n-1];
	y := data[0];
	dec(y,mx);
	data[0] := y;
	dec(sum,Delta);
	  end;
end;



function SameKey(const key1, key2: TTeaKey): boolean;
// comparison of TTeaKey type variables
var i: integer;
begin
result := false;
for i := 0 to 3 do
 if key1[i] <> key2[i] then exit;
result := true;
end;



procedure StrToKey(const s: string; var key: TTeaKey);
var sa, sb: AnsiString;
var i, n: integer;
begin
sa := AnsiString(s);
sb := StringOfChar(' ',16);
n := min(Length(sa),16);
for i := 1 to n do sb[i] := sa[i];
for i := 1 to 16 do TByte16(key)[i-1] := ord(sb[i]);
sa := '';
sb := '';
end;


function KeyToStr(const key: TTeaKey): string;
var s: AnsiString;
var i: integer;
begin
SetLength(s,16);
for i := 1 to 16 do
	begin
	s[i] := Chr(TByte16(key)[i-1]);
	end;
result := s;
end;


procedure StrToData(const s: string; var data: TTeaData);
var sa: AnsiString;
var i, n, m: integer;
begin
sa := AnsiString(s);
n := Length(sa) div 4;
m := Length(sa) mod 4;
if m <> 0 then
	begin
	inc(n);
	sa := sa + StringOfChar(' ',m);
	end;
if n < 2 then  // n = 1
	begin
	n := 2;
	sa := sa + StringOfChar(' ',4);
	end;

SetLength(data,n);
for i := 0 to n-1 do
 for m := 0 to 3 do TByte4(data[i])[m] := ord(sa[i*4+m+1]);
sa := '';
end;


procedure DataToStr(var s: string; const data: TTeaData);
var sa: AnsiString;
var i, n, m: integer;
var b: byte;
begin
n := Length(data);
SetLength(sa,n*4);
for i := 0 to n-1 do
 for m := 0 to 3 do
	begin
	b := TByte4(data[i])[m];
	sa[i*4+m+1] := Chr(b);
	end;
s := Trim(sa);
sa := '';
end;


procedure ReadData(const FileName: string; var data: TTeaData);
var i, n: integer;
var ww: longword;
var wwf: file of longword;
begin
try
AssignFile(wwf,FileName);
Reset(wwf);
n := FileSize(wwf);
SetLength(data,n);
for i := 0 to n-1 do
	begin
	read(wwf,ww);
	data[i] := ww;
	end;
finally CloseFile(wwf); end;
end;

procedure WriteData(const FileName: string; var data: TTeaData);
var i, n: integer;
var ww: longword;
var wwf: file of longword;
begin
try
AssignFile(wwf,FileName);
Rewrite(wwf);
n := Length(data);
for i := 0 to n-1 do
	begin
	ww := data[i];
	write(wwf,ww);
	end;
finally CloseFile(wwf); end;
end;


procedure EncryptFile(Filename: string; const key: TTeaKey);
var Data: TTeaData;
begin
ReadData(Filename,Data);
BlockTeaEncrypt(Data,Key);
WriteData(Filename,Data);
end;


procedure DecryptFile(Filename: string; const key: TTeaKey);
var Data: TTeaData;
begin
ReadData(Filename,Data);
BlockTeaDecrypt(Data,Key);
WriteData(Filename,Data);
end;


procedure DecryptToString(Filename: string; const key: TTeaKey; var S: string);
var Data: TTeaData;
begin
ReadData(Filename,Data);
BlockTeaDecrypt(Data,Key);
DataToStr(S,Data);
end;

{------------------------------------------------------------------------------}



procedure StreamToTeaData(Stream: TStream; var data: TTeaData);
var I, N: integer;
var WW: longword;
begin
Stream.Position:=0;
N:=Stream.Size;
N:=N div 4;
if (Stream.Size mod 4)>0 then Inc(N);
SetLength(Data,N);
for I := 0 to N-1 do
	begin
	WW:=GetStreamDWord(Stream);
	Data[I]:=WW;
	end;
end;




procedure TeaDataToStream(Stream: TStream; var data: TTeaData);
var I, N: integer;
var WW: longword;
begin
Stream.Position:=0;
Stream.Size:=0;
N:=Length(data);
for I := 0 to N-1 do
	begin
	WW:=Data[I];
	WriteStreamDWord(Stream,WW);
	end;
end;





procedure EncryptStream(Stream: TStream; const key: TTeaKey);
var Data: TTeaData;
begin
StreamToTeaData(Stream,Data);
BlockTeaEncrypt(Data,Key);
TeaDataToStream(Stream,Data);
end;


procedure DecryptStream(Stream: TStream; const key: TTeaKey);
var Data: TTeaData;
begin
StreamToTeaData(Stream,Data);
BlockTeaDecrypt(Data,Key);
TeaDataToStream(Stream,Data);
end;


function StringToKey(S: string): TTeaKey;
const RandomKey: TTeaKey = ($50F1D53E,$C9053B77,$43AC0EAE,$084ADCF9);
var Len,I: integer;
begin
Result:=RandomKey;
Len:=Length(S);
if Len>16 then Len:=16;
for I:=0 to Len-1 do PByteArray(@Result)^[I]:=byte(S[I+1]);
end;



function DecryptString(S, Key: string): string;
var Stream: TMemoryStream;
begin
Stream:=TMemoryStream.Create;
try
Stream.Write(PChar(S)^,Length(S));
DecryptStream(Stream,StringToKey(Key));
SetLength(Result, Stream.Size);
Stream.Position:=0;
Stream.Read(PChar(Result)^, Stream.Size);
finally Stream.Free; end;
end;



function EncryptString(S, Key: string): string;
var Stream: TMemoryStream;
begin
Stream:=TMemoryStream.Create;
try
Stream.Write(PChar(S)^,Length(S));
EncryptStream(Stream,StringToKey(Key));

SetLength(Result, Stream.Size);
Stream.Position:=0;
Stream.Read(PChar(Result)^, Stream.Size);
finally Stream.Free; end;
end;



end.
