unit ThreadLib;

interface

uses Windows,Messages,Forms,Contnrs,Classes,SysUtils,SyncObjs,MMSystem,DSupport;


{Object to help create and encapsulate threaded tasks. 1. Create a
 new object for each task. 2. The subroutine that you want to
 multitask is passed in the "Create" method. 3. The underlying
 ThreadCapsule is passed to the multitasked subroutine so you can
 access its methods. 4. Use the "SafeExecute" method to handle any
 subroutines that accesses Delphi VCL routines. 5. You must terminate
 the tread using the "Terminate" method before freeing the object }


type
 TThreadCapsule = class;

 TThreadTask = procedure(Thread: TThreadCapsule);
 TObjectTask = procedure(Thread: TThreadCapsule) of Object;
 TThreadCapsule = class(TThread)
  private
    FTask: TThreadTask;
    FObjectTask: TObjectTask;
    FProcedureTask: TProcedure;
    FSafeProc: TProcedure;
    FException: Exception;
    procedure DoSafeExecute;
    procedure ShowException;
    procedure HandleException;
  protected
    procedure Execute; override;
  public
    constructor Create(Task: TThreadTask); overload;
    constructor Create(Task: TObjectTask); overload;
    constructor Create(Task: TProcedure); overload;
    destructor Destroy; override;
    procedure SafeExecute(Proc: TProcedure); overload;
    procedure SafeExecute(Method: TThreadMethod); overload;
  property Terminated;
 end;


{Create a one time delay}

type TThreadDelay = class(TThread)
  private
    FDelay: integer;
    FProcedure: TProcedure;
    procedure HandleTerminate(Sender: TObject);
  protected
    procedure Execute; override;
  public
    constructor Create(Delay: integer; Proc: TProcedure);
 end;


{Thread-safe timer - this timer has a resolution of about 40 usecs}
{There appears to be a bug in this code that causes so sort of memory}
{access problems if the code runs in periodic mode long enough}

type TTimerMode = (tmOneShot,tmPeriodic);


type TThreadTimer = class(TObject)
 private
  FTimerID: Cardinal;
  FInterval: Cardinal;
  FOnTimer: TNotifyEvent;
  FEnabled: Boolean;
  FMode: TTimerMode;
  procedure UpdateTimer;
  procedure SetEnabled(Value: Boolean);
  procedure SetInterval(Value: Cardinal);
  procedure SetOnTimer(Value: TNotifyEvent);
  procedure SetMode(const Value: TTimerMode);
 protected
  procedure Timer; dynamic;
 public
  constructor Create;
  destructor Destroy; override;
  property Enabled: Boolean read FEnabled write SetEnabled default True;
  property Interval: Cardinal read FInterval write SetInterval default 1000;
  property OnTimer: TNotifyEvent read FOnTimer write SetOnTimer;
  property Mode: TTimerMode read FMode write SetMode;
 end;


const WM_HiresTimeout =  WM_USER + 1000;

{This high resolution timer is capable of about 20 microsecond resolution,}
{ however it also causes the program to use 100% of the CPU cycles for one core.}
{ If you can live with 40 usecond resolution, use the TThreadTimer instead}

type THiResTimer = class(TComponent)
 private
  FHWnd: THandle;
  EndTime: int64;
  FInterval: int64;
  FOnTimer: TNotifyEvent;
  FEnabled: Boolean;
  Thread: TThreadCapsule;
  procedure SetEnabled(Value: Boolean);
  procedure SetInterval(Value: int64);
  procedure SetOnTimer(Value: TNotifyEvent);
  procedure ThreadHandler(Thread: TThreadCapsule);
  procedure StartTimer;
  procedure MessageHandler(var Msg: TMessage);
 protected
  procedure Timer; dynamic;
 public
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
 published
  property Enabled: Boolean read FEnabled write SetEnabled default True;
  property Interval: int64 read FInterval write SetInterval default 1000000;
  property OnTimer: TNotifyEvent read FOnTimer write SetOnTimer;
 end;



{-----------------------------------------------------------------------------}

type TThreadManager = class(TThreadList)
 private
  FMaxThreads: integer;
  FAbortWait: boolean;
  function GetThreadCount: integer;
  function IsEmpty: boolean;
  procedure HandleTerminate(Sender: TObject);
 protected
 public
  constructor Create;
  function WaitOpenThread: boolean;
  procedure Add(Item: TThread);
  property Count: integer read GetThreadCount;
  property Empty: boolean read IsEmpty;
  property MaxThreads: integer read FMaxThreads write FMaxThreads;
 end;


{Thread object that makes it easy to execute any procedure in a thread
as a one shot event. The thread is automatically terminated and freed
so no instance variable is needed. The object is used this way:

TThreadOneShot.Create(MyProcedure);

"MyProcedure" is executed in its own thread one time.

}

type TThreadOneShot = class(TThreadCapsule)
 private
 protected
  procedure Execute; override;
 public
 published
 end;


{Thread safe version of the TStringList }

const WM_AddStringEvent =    WM_USER + $1000;

type TThreadStringList = class
 private
  FHWnd: THandle;
  FStringList: TStringList;
  FLock: TRTLCriticalSection;
  FOnAdd: TNotifyEvent;
  procedure MessageHandler(var Msg:TMessage);
  function GetDuplicates: TDuplicates;
  procedure SetDuplicates(dup: TDuplicates);
  function GetCapacity: Integer;
  procedure SetCapacity(capa: Integer);
  function GetCommaText: string;
  procedure SetCommaText(const S: string);
  function GetCount: Integer;
  function GetDelimiter: Char;
  procedure SetDelimiter(delim: Char);
  function GetDelimitedText: string;
  procedure SetDelimitedText(const S: string);
  function GetNames(Index: Integer): string;
  function GetValues(const Name: string): string;
  procedure SetValues(const Name: string; S: string);
  function GetStrings(Index: Integer): string;
  procedure SetStrings(Index: Integer; S: string);
  function GetAsText: string;
  procedure SetAsText(S: string);
 public
  constructor Create;
  destructor Destroy; override;
  function LockList: TStringList;
  procedure UnlockList;
  function Add(const S: string): Integer;
  procedure AddStrings(Strings: TStrings);
  procedure Delete(Index: Integer);
  procedure Clear;
  procedure Exchange(Index1, Index2: Integer);
  function Find(const S: string; var Index: Integer): Boolean;
  procedure Insert(Index: Integer; const S: string);
  function IndexOf(const S: string): Integer;
  function IndexOfName(const Name: string): Integer;
  procedure Sort;
  function GetText: PChar;
  procedure LoadFromFile(const FileName: string);
  procedure LoadFromStream(Stream: TStream);
  procedure SaveToFile(const FileName: string);
  procedure SaveToStream(Stream: TStream);
  procedure AddString;
  property Duplicates: TDuplicates read GetDuplicates write SetDuplicates;
  property Capacity: Integer read GetCapacity write SetCapacity;
  property CommaText: string read GetCommaText write SetCommaText;
  property Count: Integer read GetCount;
  property Delimiter: Char read GetDelimiter write SetDelimiter;
  property DelimitedText: string read GetDelimitedText write SetDelimitedText;
  property Names[Index: Integer]: string read GetNames;
  property Values[const Name: string]: string read GetValues write SetValues;
  property Strings[Index: Integer]: string read GetStrings write SetStrings; default;
  property Text: string read GetAsText write SetAsText;
  property OnAdd: TNotifyEvent read FOnAdd write FOnAdd;
 end;

{Thread-safe string - Access through the value property}


type TThreadString = class(TObject)
 private
  Semaphore: TCriticalSection;
  FString: string;
  function ReadString: string;
  procedure WriteString(const Value: string);
 protected
 public
  constructor Create;
  destructor Destroy; override;
  property Value: string read ReadString write WriteString;
 end;


function GetThreadPriorityStr(P: TThreadPriority): string;


implementation


{-------------------- Support Routines ----------------------------------------}


function GetThreadPriorityStr(P: TThreadPriority): string;
{Get string describing specified thread priorty}
begin
case P of
 tpIdle: Result:='Idle';
 tpLowest: Result:='Lowest';
 tpLower: Result:='Lower';
 tpNormal: Result:='Normal';
 tpHigher: Result:='Higher';
 tpHighest: Result:='Highest';
 tpTimeCritical: Result:='TimeCritical';
 else Result:='Unknown';
 end;
end;




{--------------- TThreadCapsule Methods -------------------------------}


procedure TThreadCapsule.DoSafeExecute;
begin
FSafeProc;
end;


procedure TThreadCapsule.SafeExecute(Proc: TProcedure);
begin
FSafeProc:=Proc;
Synchronize(DoSafeExecute);
end;


procedure TThreadCapsule.SafeExecute(Method: TThreadMethod);
begin
Synchronize(Method);
end;



constructor TThreadCapsule.Create(Task: TThreadTask);
begin
FTask:=Task;
FreeOnTerminate:=False;
inherited Create(False);
end;


constructor TThreadCapsule.Create(Task: TObjectTask);
begin
FObjectTask:=Task;
FreeOnTerminate:=False;
inherited Create(False);
end;


constructor TThreadCapsule.Create(Task: TProcedure);
begin
FProcedureTask:=Task;
FreeOnTerminate:=False;
inherited Create(False);
end;


destructor TThreadCapsule.Destroy;
begin
Terminate;
inherited Destroy;
end;


procedure TThreadCapsule.ShowException;
begin
if FException is Exception then Application.ShowException(FException)
else SysUtils.ShowException(FException, nil);
end;


procedure TThreadCapsule.HandleException;
begin
FException := Exception(ExceptObject);
if not (FException is EAbort) then Synchronize(ShowException);
FException := nil;
end;


procedure TThreadCapsule.Execute;
begin
try try
while True do
	begin
	if Assigned(FTask) then FTask(Self)
	else if Assigned(FObjectTask) then FObjectTask(Self)
	else if Assigned(FProcedureTask) then FProcedureTask;
	if Terminated then Exit;
	end;
finally
 Terminate;
 end;
except
 HandleException;
 end;
end;




{------------------------------------------------------------------------------}


procedure TThreadOneShot.Execute;
begin
{Make sure object is freed when it terminates so}
{it can be created without an instance variable}
FreeOnTerminate:=True;
{Set terminate flag true so it only executes once}
Terminate;
{Do the actual code execution}
inherited;
end;


{------------------------------------------------------------------------------}



{The callback arguments are defined differently for version XE3 and above}

{$IF CompilerVersion >= 17}
type TTimerArg = NativeUInt;
{$ELSE}
type TTimerArg = Cardinal;
{$IFEND}



procedure HandleTimer(TimerId, Msg: Cardinal; User, dw1, dw2: TTimerArg); stdcall;
begin
TThreadTimer(User).Timer;
end;


constructor TThreadTimer.Create;
begin
inherited Create;
FTimerID:=0;
FEnabled := False;
FInterval := 1000;
FMode:=tmOneShot;
end;


destructor TThreadTimer.Destroy;
begin
FEnabled := False;
UpdateTimer;
inherited Destroy;
end;


procedure TThreadTimer.UpdateTimer;
var Flags: cardinal;
begin
if FTimerID<>0 then timeKillEvent(FTimerID);
if (FInterval <> 0) and FEnabled and Assigned(FOnTimer) then
	begin
	if FMode=tmPeriodic then Flags:=TIME_PERIODIC
	else Flags:=TIME_ONESHOT;
	FTimerId:=timeSetEvent(FInterval, 0, HandleTimer, DWORD(Self),Flags);
	if FTimerId = 0 then raise Exception.Create('Unable to Create Timer');
	end;
end;


procedure TThreadTimer.SetEnabled(Value: Boolean);
begin
if Value <> FEnabled then
	begin
	FEnabled := Value;
	UpdateTimer;
	end;
end;



procedure TThreadTimer.SetInterval(Value: Cardinal);
begin
if Value <> FInterval then
	begin
	FInterval := Value;
	UpdateTimer;
	end;
end;



procedure TThreadTimer.SetOnTimer(Value: TNotifyEvent);
begin
FOnTimer := Value;
UpdateTimer;
end;



procedure TThreadTimer.Timer;
begin
if Assigned(FOnTimer) then FOnTimer(Self);
end;



procedure TThreadTimer.SetMode(const Value: TTimerMode);
begin
if FMode<>Value then
	begin
	FMode:=Value;
	UpdateTimer;
	end;
end;

{------------------------------------------------------------------------------}

procedure TThreadDelay.HandleTerminate(Sender: TObject);
begin
if Assigned(FProcedure) then FProcedure;
end;


constructor TThreadDelay.Create(Delay: integer; Proc: TProcedure);
begin
FDelay:=Delay;
FProcedure:=Proc;
Self.OnTerminate:=HandleTerminate;
inherited Create(False);
end;


procedure TThreadDelay.Execute;
begin
Sleep(FDelay);
end;

{------------------------------------------------------------------------------}



constructor TThreadManager.Create;
begin
inherited;
FMaxThreads:=10;
end;

function TThreadManager.GetThreadCount: integer;
begin
Result:=LockList.Count;
UnlockList;
end;



function TThreadManager.IsEmpty: boolean;
begin
Result:=Count<1;
end;



function TThreadManager.WaitOpenThread: boolean;
begin
Result:=False;
FAbortWait:=False;
repeat
	begin
	if FAbortWait then exit;
	Application.ProcessMessages;
	Sleep(0);
	end
until Count< FMaxThreads;
Result:=True;
end;



procedure TThreadManager.HandleTerminate(Sender: TObject);
begin
Self.Remove(Sender);
end;



procedure TThreadManager.Add(Item: TThread);
begin
Inherited Add(Item);
Item.OnTerminate:=HandleTerminate;
end;





{ -------- TThreadStringList --------------------------------------------------}

constructor TThreadStringList.Create;
begin
inherited Create;
InitializeCriticalSection(FLock);
FStringList := TStringList.Create;
FStringList.Duplicates := dupIgnore;
FHWnd:=Classes.AllocateHWnd(MessageHandler);
end;



destructor TThreadStringList.Destroy;
begin
LockList;
try
FStringList.Free;
inherited Destroy;
finally
 UnlockList;
 DeleteCriticalSection(FLock);
 Classes.DeallocateHWnd(FHWnd);
 end;
end;



function TThreadStringList.LockList: TStringList;
begin
EnterCriticalSection(FLock);
Result := FStringList;
end;



procedure TThreadStringList.UnlockList;
begin
LeaveCriticalSection(FLock);
end;

function TThreadStringList.Add(const S: string): Integer;
begin
Result := -1;
LockList;
try
Result := FStringList.Add(S);
 finally
 UnlockList;
 AddString;
 end;
end;



procedure TThreadStringList.AddStrings(Strings: TStrings);
begin
LockList;
try
FStringList.AddStrings(Strings);
finally
 UnlockList;
 AddString;
 end;
end;



procedure TThreadStringList.Delete(Index: Integer);
begin
LockList;
try
FStringList.Delete(Index);
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.Clear;
begin
LockList;
try
FStringList.Clear;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.Exchange(Index1, Index2: Integer);
begin
LockList;
try
FStringList.Exchange(Index1, Index2);
finally
 UnlockList;
 end;
end;



function TThreadStringList.Find(const S: string; var Index: Integer): Boolean;
begin
LockList;
try
Result := FStringList.Find(S, Index);
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.Insert(Index: Integer; const S: string);
begin
LockList;
try
FStringList.Insert(Index, S);
finally
 UnlockList;
 end;
end;



function TThreadStringList.IndexOf(const S: string): Integer;
begin
Result := -1;
LockList;
try
Result := FStringList.IndexOf(S);
finally
 UnlockList;
 end;
end;



function TThreadStringList.IndexOfName(const Name: string): Integer;
begin
Result := -1;
LockList;
try
Result := FStringList.IndexOfName(Name);
finally UnlockList; end;
end;



procedure TThreadStringList.Sort;
begin
LockList;
try
FStringList.Sort;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetText: PChar;
begin
LockList;
try
Result := FStringList.GetText;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.LoadFromFile(const FileName: string);
begin
LockList;
try
FStringList.LoadFromFile(FileName);
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.LoadFromStream(Stream: TStream);
begin
LockList;
try
FStringList.LoadFromStream(Stream);
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SaveToFile(const FileName: string);
begin
LockList;
try
FStringList.SaveToFile(FileName);
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SaveToStream(Stream: TStream);
begin
LockList;
try
FStringList.SaveToStream(Stream);
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetDuplicates: TDuplicates;
begin
LockList;
try
Result := FStringList.Duplicates;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetDuplicates(dup: TDuplicates);
begin
LockList;
try
FStringList.Duplicates := dup;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetCapacity: Integer;
begin
LockList;
try
Result := FStringList.Capacity;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetCapacity(capa: Integer);
begin
LockList;
try
FStringList.Capacity := capa;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetCommaText: string;
begin
LockList;
try
Result := FStringList.CommaText;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetCommaText(const S: string);
begin
LockList;
try
FStringList.CommaText := S;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetCount: Integer;
begin
LockList;
try
Result := FStringList.Count;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetDelimiter: Char;
begin
LockList;
try
Result := FStringList.Delimiter;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetDelimiter(Delim: Char);
begin
LockList;
try
FStringList.Delimiter:=Delim;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetDelimitedText: string;
begin
LockList;
try
Result := FStringList.DelimitedText;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetDelimitedText(const S: string);
begin
LockList;
try
FStringList.DelimitedText := S;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetNames(Index: Integer): string;
begin
LockList;
try
Result := FStringList.Names[Index];
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetValues(const Name: string): string;
begin
LockList;
try
Result := FStringList.Values[Name];
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetValues(const Name: string; S: string);
begin
LockList;
try
FStringList.Values[Name] := S;
finally
 UnlockList;
 end;
end;



function TThreadStringList.GetStrings(Index: Integer): string;
begin
LockList;
try
Result := FStringList.Strings[Index];
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetStrings(Index: Integer; S: string);
begin
LockList;
try
FStringList.Strings[Index] := S;
finally
 UnlockList;
 AddString;
 end;
end;



function TThreadStringList.GetAsText: string;
begin
LockList;
try
Result := FStringList.Text;
finally
 UnlockList;
 end;
end;



procedure TThreadStringList.SetAsText(S: string);
begin
LockList;
try
FStringList.Text := S;
finally
 UnlockList;
 AddString;
 end;
end;




procedure TThreadStringList.MessageHandler(var Msg: TMessage);
begin
if Assigned(OnAdd) then OnAdd(Self);
end;



procedure TThreadStringList.AddString;
begin
PostMessage(FHwnd,WM_AddStringEvent,0,0);
end;


{-------------THiResTimer ----------------------------------------------------}


procedure THiResTimer.StartTimer;
begin
EndTime:=GetHiResTick64+Interval;
end;



procedure THiResTimer.SetEnabled(Value: Boolean);
begin
if Value <> FEnabled then
	begin
        if Value then StartTimer;
	FEnabled := Value;
	end;
end;




procedure THiResTimer.Timer;
begin
if Assigned(FOnTimer) then FOnTimer(Self);
end;




procedure THiResTimer.MessageHandler(var Msg:TMessage);
begin
case Msg.Msg of
 WM_HiresTimeout: Timer;
 end;
end;



procedure THiResTimer.ThreadHandler(Thread: TThreadCapsule);
var CurTime: Int64;
var I: integer;
begin
while FEnabled and not Thread.Terminated do
	begin
	CurTime:=GetHiResTick64;
	Application.ProcessMessages;
	if CurTime>=EndTime then
		begin
		PostMessage(FHwnd,WM_HiresTimeout,0,0);
		//Thread.SafeExecute(Timer);
		StartTimer;
		end;
	end;
end;




constructor THiResTimer.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
Thread:=TThreadCapsule.Create(ThreadHandler);
FHWnd := Classes.AllocateHWnd(MessageHandler);
FEnabled := True;
FInterval := 1000000;
end;



destructor THiResTimer.Destroy;
begin
Thread.Terminate;
FEnabled := False;
Thread.Free;
Classes.DeallocateHWnd(FHWnd);
inherited Destroy;
end;



procedure THiResTimer.SetInterval(Value: int64);
begin
if Value <> FInterval then
	begin
	FInterval := Value;
	end;
end;



procedure THiResTimer.SetOnTimer(Value: TNotifyEvent);
begin
FOnTimer := Value;
end;

{-------------------------------------------------------------------------------}



{ TThreadString }

constructor TThreadString.Create;
begin
Semaphore:=TCriticalSection.Create;
end;



destructor TThreadString.Destroy;
begin
Semaphore.Free;
inherited Destroy;
end;


function TThreadString.ReadString: string;
begin
Semaphore.Acquire;
try
Result:=FString;
finally Semaphore.Release; end;
end;



procedure TThreadString.WriteString(const Value: string);
begin
Semaphore.Acquire;
try
FString:=Value;
finally Semaphore.Release; end;
end;


end.
