unit VRMLWriter;

{VRML Writer - Routines to write VRLM files. To use:

 1. Pass groups of points to "StoreVRMLPoints"

 2. If "VRMLPolygonMode" is true each group of points
     represents a polygon. Otherwise, the each group
     represents a continuous line segment.

 3. If "VRMLColorMode" is true, you can send a set
    of colors to "StoreVRMLColor" and then store indexes
    to those colors with "StoreVRMLColorIndex"

 4. If "VRMLVertexColors" is true, you should send
    a color index for each vertex and an index of -1
    at the end of each polygon or line segment.

 5. When all options have been set, points stored and
    colors set, call WriteVRMLStream  or WriteVRMLFile.


 Each group of points represents }
{Point represent Polygons or Lines depending on "StoreVRMLPoints"}
{When all points have been stored, call "StoreVRMLPoints"}

interface

uses Windows,Classes,Graphics,SysUtils,DSupport,HugeArry,Vector,VectorHash,StreamLib;

type TVRMLMaterial = record
  Diffuse,Emissive,Specular: TColor;
  Transparency,Ambient,Shininess: double;
  DoTransparency: boolean;
  end;


{Parameters to control VRML output}

var VRMLMaterial: TVRMLMaterial;	{Material attributes of current object}
var VRMLBackgroundColor: TColor;	{Background color for VRML display}
var VRMLPolygonMode: boolean = True;	{Flags that we are drawing polygons}
var VRMLColorMode: boolean = False;	{Flags output color data}
var VRMLVertexColors: boolean;		{Flags one color per vertex}
var VRMLApplyTexture: boolean = False;	{Flags texture will be applied}
var VRMLDoYawRotation: boolean = False;	{Flags yaw rotation animations}
var VRMLSplineSurface: boolean = False;	{Flags spline surface (Cortona only feature?)}
var VRMLTextureFilename: string;	{Filename for texture file}

{Data arrays }

var VRMLVertexArray: TVectorHash;	{Holds accumlated vectors for object}
var VRMLIndexArray: THugeInteger;	{Holds accumlated indexes for object}
var VRMLColorList: THuge3DVector;	{Holds list of colors for the object}
var VRMLColorIndexes: THugeInteger;	{Holds list of face color indexes}

{Points array that is passed to "StoreVRMLPoints" }

const VRMLMaxPoints = 200;		{Maxium number of points that can be passed}

type TVRMLPointArray = record
 Points: array [0..VRMLMaxPoints-1] of T3DVector;
 Count: integer;
 end;

var MeshMin,MeshMax: T3DVector;		{Holds accumulated extents of mesh}

{Prototype for user defined "Progress" call back routine }

type TVRMLProgress = procedure(Percent: integer);

procedure StoreVRMLPoints(P: TVRMLPointArray);
procedure ClearVRMLPoints;
procedure StoreVRMLColor(C: TColor);
procedure StoreVRMLColorIndex(I: integer);

function OpenVRMLFile(Filename: string): boolean;
procedure WriteVRMLFile(Progress: TVRMLProgress);
procedure CloseVRMLFile;


implementation


var VRMLStream: TFileStream;
var HasHeader: boolean;		{Flags that current file has header}

{Indent Parameters}

const TabIncrement = 4;		{Indent for file}
var CurrentIndent: integer=0;	{Current number of spaces file is indented}

{Defines quaternion for four item argumenets }

type TQuaternion = record
 case integer of
  0: (X,Y,Z,W: double);
  1: (V: T3DVector);
 end;


{-------------------Support Routines -----------------------------------------}

function MakeQuaternion(X,Y,Z,W: double): TQuaternion;
{Create a quaternion from individual values}
begin
Result.V:=MakeVector3D(X,Y,Z);
Result.W:=W;
end;



function GetColorVector(Color: TColor): T3DVector;
{Convert Delphi color to color vector}
var C: longint;
begin
C:=ColorToRGB(Color);
Result.X:=(C and $FF)/255;
Result.Y:=((C shr 8) and $FF)/255;
Result.Z:=((C shr 16) and $FF)/255;
end;


{--------------- Low level I/o Routines ---------------------------------------}


procedure VRMLNewLine(Stream: TStream);
{Write CRLF to VRML stream}
begin
WriteStreamString(Stream,CRLF_Char);
end;


procedure VRMLWriteLine(Stream: TStream; S: string);
{Write string and EOL to VRML stream}
begin
WriteStreamString(Stream,S);
VRMLNewLine(Stream);
end;


procedure VRMLIndent(Stream: TStream);
{Write indented to VRML stream}
var I: integer;
var S: string;
begin
S:='';
for I:=1 to CurrentIndent do S:=' '+S;
WriteStreamString(Stream,S);
end;


procedure VRMLWriteIndented(Stream: TStream; S: string);
{Write indented string to VRML stream}
begin
VRMLIndent(Stream);
WriteStreamString(Stream,S);
end;


procedure VRMLWriteIndentEOL(Stream: TStream; S: string);
{Write indented string and EOL to VRML stream}
begin
VRMLWriteIndented(Stream,S);
VRMLNewLine(Stream);
end;

{------------------ Number I/O Routines --------------------------------------}


procedure VRMLWriteBoolean(Stream: TStream; B: boolean);
{Write boolean value}
var T: string;
begin
if B then T:='TRUE' else T:='FALSE';
WriteStreamString(Stream,T);
end;


procedure VRMLWriteInteger(Stream: TStream; I: integer);
{Write integer to VRML stream}
begin
WriteStreamString(Stream,IntToStr(I));
end;


procedure VRMLWriteDouble(Stream: TStream; D: double);
{Write double to VRML stream}
begin
WriteStreamString(Stream,GetXPLRealStr(D));
end;


procedure VRMLWriteVector3D(Stream: TStream; V: T3DVector);
{Write vector to VRML stream}
begin
VRMLWriteDouble(Stream,V.X); WriteStreamString(Stream,' ');
VRMLWriteDouble(Stream,V.Y); WriteStreamString(Stream,' ');
VRMLWriteDouble(Stream,V.Z);
end;


procedure VRMLWriteQuaternion(Stream: TStream; Q: TQuaternion);
{Write quaternion to VRML stream}
begin
VRMLWriteVector3D(Stream,Q.V);
WriteStreamString(Stream,' ');
VRMLWriteDouble(Stream,Q.W);
end;



procedure VRMLWriteColor(Stream: TStream; C: TColor);
{Write color as a vector}
begin
VRMLWriteVector3D(Stream,GetColorVector(C));
end;


{------------------ Write Labeled Values --------------------------------------}


procedure VRMLWriteLabeledVector3D(Stream: TStream; S: String; V: T3DVector);
{Write a labeled Vector}
begin
VRMLWriteIndented(Stream,S+' ');
VRMLWriteVector3D(Stream,V);
VRMLNewLine(Stream);
end;


procedure VRMLWriteLabeledQuaternion(Stream: TStream; S: String; Q: TQuaternion);
{Write a labeled Auaternion}
begin
VRMLWriteIndented(Stream,S);
VRMLWriteQuaternion(Stream,Q);
VRMLNewLine(Stream);
end;


procedure VRMLWriteLabeledDouble(Stream: TStream; S: String; D: double);
{Write a labeled double}
begin
VRMLWriteIndented(Stream,S+' ');
VRMLWriteDouble(Stream,D);
VRMLNewLine(Stream);
end;



procedure VRMLWriteLabeledColor(Stream: TStream; S: String; C: TColor);
{Write a labeled color as vector}
begin
VRMLWriteLabeledVector3D(Stream,S,GetColorVector(C));
end;



procedure VRMLWriteLabeledBool(Stream: TStream; S: string; B: boolean);
{Write labeled boolean value}
begin
VRMLWriteIndented(Stream,S+' ');
VRMLWriteBoolean(Stream,B);
VRMLNewLine(Stream);
end;



{------ Open/Close Various Block types (each requires explicit closing ) -------}


procedure VRMLWriteBegin(Stream: TStream);
{Write beginning bracket and increment indent}
begin
CurrentIndent:=CurrentIndent + TabIncrement;
VRMLWriteIndented(Stream,'{');
VRMLNewLine(Stream);
end;



procedure VRMLWriteEnd(Stream: TStream);
{Write ending bracket and decrement indent}
begin
VRMLWriteIndented(Stream,'}');
VRMLNewLine(Stream);
CurrentIndent:=CurrentIndent - TabIncrement;
if CurrentIndent<0 then CurrentIndent:=0;
end;


procedure VRMLOpenItem(Stream: TStream; Name,Delimiter: string);
{Open named block and with specified delimiter}
begin
VRMLWriteIndentEOL(Stream,Name);
CurrentIndent:=CurrentIndent + TabIncrement;
VRMLWriteIndentEOL(Stream,Delimiter);
end;


procedure VRMLOpenChildren(Stream: TStream);
{Open a child block}
begin
VRMLOpenItem(stream,'children','[');
end;



procedure VRMLCloseChildren(Stream: TStream);
{Close a child block}
begin
VRMLWriteIndented(Stream,']');
VRMLNewLine(Stream);
CurrentIndent:=CurrentIndent - TabIncrement;
if CurrentIndent<0 then CurrentIndent:=0;
end;



procedure VRMLOpenTransform(Stream: TStream; Translation,Scale: T3DVector);
{Open a transform block}
begin
VRMLOpenItem(Stream,'Transform','{');
VRMLWriteLabeledVector3D(Stream,'translation',Translation);
VRMLWriteLabeledVector3D(Stream,'scale',Scale);
end;


procedure VRMLOpenShape(Stream: TStream);
{Open a shape block}
begin
VRMLOpenItem(Stream,'Shape','{');
end;


procedure VRMLOpenAppearance(Stream: TStream);
{Open Appearance Block}
begin
VRMLOpenItem(Stream,'appearance Appearance','{');
end;


{-----------------Write misc objects ------------------------------------------}



procedure VRMLWriteSignature(Stream: TStream);
{Write standard VRML Signature }
begin
VRMLWriteIndentEOL(Stream,'#VRML V2.0 utf8');
VRMLWriteIndentEOL(Stream,'# Compass');
VRMLWriteIndentEOL(Stream,'# Fountain Computer Products');
end;


procedure VRMLWriteBackgroundColor(Stream: TStream);
begin
VRMLOpenItem(Stream,'Background','{');
VRMLWriteLabeledColor(Stream,'skyColor',VRMLBackgroundColor);
VRMLWriteEnd(Stream);
end;


procedure VRMLWriteNameDefinition(Stream: TStream; Name: string);
{Write a specific name definition}
begin
VRMLWriteIndentEOL(Stream,'DEF '+Name);
end;


{------------------ Write complete blocks (don't require closing) ------------}


procedure VRMLWriteViewpoint(Stream: TStream);
var Delta,Position: T3DVector;
begin
Delta:=VectorSubtract3D(MeshMax,MeshMin);
Position.X:=0;
Position.Y:=0;
Position.Z:=Delta.X * 1.5;
VRMLOpenItem(Stream,'Viewpoint','{');
VRMLWriteLabeledVector3D(Stream,'position',Position);
VRMLWriteLabeledQuaternion(Stream,'orientation ',MakeQuaternion(0,0,0,0));

VRMLWriteEnd(Stream);
end;


procedure VRMLWriteMaterial(Stream: TStream);
begin
with VRMLMaterial do
 begin
 VRMLOpenItem(Stream,'material Material','{');
 VRMLWriteLabeledColor(Stream,'diffuseColor',Diffuse);
 VRMLWriteLabeledColor(Stream,'emissiveColor',Emissive);
 VRMLWriteLabeledColor(Stream,'specularColor',Specular);

 VRMLWriteLabeledDouble(Stream,'ambientIntensity',Ambient);
 VRMLWriteLabeledDouble(Stream,'shininess',Shininess);
 if DoTransparency then VRMLWriteLabeledDouble(Stream,'transparency ',Transparency);
 end;
VRMLWriteEnd(Stream);
end;


{--------- Routines for Writing User Specified Vertex/Face colors -------------}


procedure VRMLWriteColorList(Stream: TStream);
{Write user specified base colors}
var I: integer;
begin
VRMLOpenItem(Stream,'color Color','{');
VRMLOpenItem(Stream,'color','[');
for I:=0 to VRMLColorList.ItemIndex-1 do
	begin
	VRMLIndent(Stream);
	VRMLWriteVector3D(Stream,VRMLColorList.Value[I]);
	VRMLNewLine(Stream);
	end;
VRMLCloseChildren(Stream);
VRMLWriteEnd(Stream);
end;


procedure VRMLWriteColorIndexes(Stream: TStream);
{Write the user specified index that correspond to the base colors}
var I: integer;
begin
VRMLOpenItem(Stream,'colorIndex','[');
for I:=0 to VRMLColorIndexes.ItemIndex-1 do
	begin
	if (I mod 50)=0 then VRMLNewLine(Stream);
	if (I mod 50)=0 then VRMLIndent(Stream);
	VRMLWriteInteger(Stream,VRMLColorIndexes.Value[I]);
	WriteStreamString(Stream,' ');
	end;
VRMLNewLine(Stream);
VRMLCloseChildren(Stream);
end;


procedure VRMLWriteColorData(Stream: TStream);
{Write base colors and color indices for each face or vertex}
begin
VRMLWriteLabeledBool(Stream,'colorPerVertex',VRMLVertexColors);
VRMLWriteColorList(Stream);
VRMLWriteColorIndexes(Stream);
end;


{------------------- Write Animation Data -------------------------------------}

procedure VRMLWriteSensor(Stream: TStream);
begin
VRMLWriteNameDefinition(Stream,'Sensor1');
VRMLOpenItem(Stream,'TouchSensor','{');
VRMLWriteEnd(Stream);
end;


procedure VRMLWriteTimer(Stream: TStream);
begin
VRMLWriteNameDefinition(Stream,'Timer1');
VRMLOpenItem(Stream,'TimeSensor','{');
VRMLWriteLabeledDouble(Stream,'cycleInterval',18);
VRMLWriteLabeledBool(Stream,'loop',False);
VRMLWriteLabeledDouble(Stream,'startTime',0);
VRMLWriteLabeledDouble(Stream,'stopTime',-1);
VRMLWriteEnd(Stream);
end;


procedure VRMLWriteRotator(Stream: TStream);
begin
VRMLWriteSensor(Stream);
VRMLWriteNameDefinition(Stream,'Rotator1');
VRMLOpenItem(Stream,'OrientationInterpolator','{');
VRMLWriteIndented(Stream,'key [');
VRMLWriteVector3D(Stream,MakeVector3D(0.0,0.5,1.0));
WriteStreamString(Stream,']'+CRLF_Char);

VRMLWriteIndented(Stream,'keyValue [');
VRMLWriteQuaternion(Stream,MakeQuaternion(0,1,0,0));
WriteStreamString(Stream,', ');
VRMLWriteQuaternion(Stream,MakeQuaternion(0,1,0,Pi));
WriteStreamString(Stream,', ');
VRMLWriteQuaternion(Stream,MakeQuaternion(0,1,0,Pi*2));
WriteStreamString(Stream,']'+CRLF_Char);
VRMLWriteEnd(Stream);
end;


procedure VRMLWriteRoute(Stream: TStream; FromItem,ToItem: string);
begin
WriteStreamString(Stream,'ROUTE '+FromItem+' TO '+ToItem);
VRMLNewLine(Stream);
end;



procedure VRMLWriteAnimation(Stream: TStream);
begin
VRMLWriteTimer(Stream);
VRMLWriteRotator(Stream);
VRMLWriteRoute(Stream,'Sensor1.isActive','Timer1.loop');
VRMLWriteRoute(Stream,'Timer1.fraction_changed','Rotator1.set_fraction');
VRMLWriteRoute(Stream,'Rotator1.value_changed','Frame1.set_rotation');
end;


{------------------ Routines For Writing Actual VRML File ---------------------}

procedure VRMLWritePoints(Stream: TStream; Progress: TVRMLProgress);
{Write all specified points the have been stored}
var I: integer;
const CreaseAngle = 1.571;
begin
RealFormat(1,2);
if VRMLPolygonMode then
	begin
	if VRMLSplineSurface then VRMLOpenItem(Stream,'geometry SplineFaceSet','{')
	else VRMLOpenItem(Stream,'geometry IndexedFaceSet','{');
	end
else VRMLOpenItem(Stream,'geometry IndexedLineSet','{');
if VRMLPolygonMode then VRMLWriteIndentEOL(Stream,'creaseAngle '+FloatToStrF(CreaseAngle,ffFixed,18,3));

VRMLOpenItem(Stream,'coord Coordinate','{');

VRMLOpenItem(Stream,'point','[');
for I:=0 to VRMLVertexArray.ItemIndex-1 do
	begin
	if (I mod 10)=0 then VRMLNewLine(Stream);
	if (I mod 10)=0 then VRMLIndent(Stream);
	VRMLWriteVector3D(Stream,VRMLVertexArray.Value[I]);
	WriteStreamString(Stream,', ');
	if Assigned(Progress) and ((I mod 100)=0) then
		Progress(MulDiv(50,I,VRMLVertexArray.ItemIndex));
	end;
VRMLNewLine(Stream);
VRMLCloseChildren(Stream);	{Close Point block}
VRMLWriteEnd(Stream);		{close Coordinate block}

VRMLOpenItem(Stream,'coordIndex','[');
for I:=0 to VRMLIndexArray.ItemIndex-1 do
	begin
	if (I mod 50)=0 then VRMLNewLine(Stream);
	if (I mod 50)=0 then VRMLIndent(Stream);
	VRMLWriteInteger(Stream,VRMLIndexArray.Value[I]);
	WriteStreamString(Stream,', ');
	if Assigned(Progress) and ((I mod 100)=0) then
		Progress(50+MulDiv(50,I,VRMLIndexArray.ItemIndex));
	end;
VRMLNewLine(Stream);
VRMLCloseChildren(Stream);	{close coordIndex block}

if (VRMLColorList.ItemIndex<>0) and VRMLColorMode then VRMLWriteColorData(Stream);

VRMLWriteEnd(Stream);		{close IndexedFaceSet block}
if Assigned(Progress) then Progress(100);
end;


procedure VRMLWriteTexture(Stream: TStream);
begin
VRMLOpenItem(Stream,'texture ImageTexture','{');
VRMLWriteIndentEOL(Stream,'url "'+VRMLTextureFilename+'"');
VRMLWriteEnd(Stream);
end;


procedure VRMLWriteAppearance(Stream: TStream);
begin
VRMLOpenAppearance(Stream);
VRMLWriteMaterial(Stream);
if VRMLApplyTexture and (VRMLTextureFilename<>'') then VRMLWriteTexture(Stream);
VRMLWriteEnd(Stream);
end;


procedure WriteVRMLHeader(Stream: TStream);
{Write header info to VRML stream}
{Note: Inorder for the transform ot be properly set}
{Lines, points or a mesh must be written to set dimenison}
begin
RealFormat(1,5);
VRMLWriteSignature(Stream);
VRMLWriteBackgroundColor(Stream);
VRMLWriteNameDefinition(Stream,'ViewPoint1');
VRMLWriteViewPoint(Stream);
VRMLWriteNameDefinition(Stream,'Frame1');
VRMLOpenTransform(Stream,MakeVector3D(-5,-5,0),MakeVector3D(1,1,1));
VRMLOpenChildren(Stream);
HasHeader:=true;
end;



procedure WriteVRMLShape(Stream: TStream; Progress: TVRMLProgress);
{Write currently stored point set to VRML stream}
begin
RealFormat(1,5);
VRMLOpenShape(Stream);
VRMLWriteAppearance(Stream);
VRMLWritePoints(Stream,Progress);
VRMLWriteEnd(Stream);		{For shape}
end;



procedure WriteVRMLTrailer(Stream: TStream);
{Write data to close out VRML file}
begin
VRMLCloseChildren(Stream);
VRMLWriteEnd(Stream);		{For transform}
if VRMLDoYawRotation then
	begin
	VRMLNewLine(Stream);
	VRMLWriteAnimation(Stream);
	end;
end;


function OpenVRMLFile(Filename: string): boolean;
{Open file for VRML Writing}
begin
VRMLStream := TFileStream.Create(Filename,fmCreate);
HasHeader:=False;
end;



procedure WriteVRMLFile(Progress: TVRMLProgress);
{Write accumulated VRML data to disk file}
begin
if not HasHeader then WriteVRMLHeader(VRMLStream);
WriteVRMLShape(VRMLStream, Progress);
end;


procedure CloseVRMLFile;
{Clese VRML file}
begin
WriteVRMLTrailer(VRMLStream);
VRMLStream.Free;
end;


{-----------------Routines to allow user to store points. ---------------}



procedure AddVRMLPoint(P: T3DVector);
{Store individual point}
begin
VRMLVertexArray.Insert(P);
VRMLIndexArray.Add(VRMLVertexArray.ItemIndex-1);
MeshMin:=GetMinimumVector(MeshMin,P);
MeshMax:=GetMaximumVector(MeshMax,P);
end;




procedure StoreVRMLPoint(P: T3DVector);
{Store individual point, dealing with duplicates}
var Inx: integer;
begin
{Have we already stored this point?}
Inx:=VRMLVertexArray.Search(P);
{if not just add it, otherwise just reference the index}
if Inx<0 then AddVRMLPoint(P)
else VRMLIndexArray.Add(Inx);
end;



procedure StoreVRMLPoints(P: TVRMLPointArray);
{Store points into VRML arrays}
var I: integer;
begin
for I:=0 to P.Count-1 do StoreVRMLPoint(P.Points[I]);
VRMLIndexArray.Add(-1);
end;


{------------------------------------------------------------------------------}

procedure StoreVRMLColor(C: TColor);
{Store single VRML Color}
begin
VRMLColorList.Add(GetColorVector(C));
end;

procedure StoreVRMLColorIndex(I: integer);
{Store a face index}
begin
VRMLColorIndexes.Add(I);
end;

{------------------------------------------------------------------------------}


procedure ClearVRMLPoints;
{Clear all stored points}
begin
VRMLVertexArray.ItemIndex:=0;
VRMLVertexArray.Rehash;
VRMLIndexArray.ItemIndex:=0;
VRMLColorList.ItemIndex:=0;
VRMLColorIndexes.ItemIndex:=0;
MeshMin:=MakeVector3D(1e99,1e99,1e99);
MeshMax:=MakeVector3D(-1e99,-1e99,-1e99);
end;



initialization
VRMLVertexArray:=TVectorHash.Create(5);
VRMLIndexArray:=THugeInteger.Create(5);
VRMLColorList:=THuge3DVector.Create(5);
VRMLColorIndexes:=THugeInteger.Create(5);
finalization
VRMLVertexArray.Free;
VRMLIndexArray.Free;
VRMLColorList.Free;
VRMLColorIndexes.Free;
end.
