unit WaveLib;

interface

uses Forms,Windows,SysUtils,MMSystem,ShellAPI;

var WaveOutHandle: hWaveOut;
var PendingOutBuffers: integer = 0;



procedure WaveBufferOut(Data: Pointer; Size: integer; Wait: boolean); overload;
procedure WaveBufferOut(Data: Pointer; Size: integer); overload;
procedure SetupOutputFormat(OutFormat: TWaveFormatEx);
procedure WaveOutShutdown;
procedure DisplayAudioPropertiesDialog;
procedure WaitForSoundEnd;
procedure SynthesizeSound8(Freq,Sec: double; Wait: boolean);
procedure PlayWaveArray(OutFormat: TWaveFormatEX; Data: Pointer; Size: integer; Wait: boolean);

procedure PlaybackSoundFile(Filename: string);
procedure StopSoundFile;
procedure CloseSoundFile;



implementation




procedure DisplayAudioPropertiesDialog;
begin
WinExec('rundll32.exe shell32.dll,Control_RunDLL mmsys.cpl,,0',sw_normal);
end;


{----------------- Output Routines --------------------------------------------}


function GetErrorStr(R: MMRESULT): string;
begin
Result:=IntToStr(R);
case R of
 MMSYSERR_INVALHANDLE: Result:=Result+': Specified device handle is invalid.';
 MMSYSERR_NODRIVER: Result:=Result+': No device driver is present.';
 MMSYSERR_NOMEM: Result:=Result+' Unable to allocate or lock memory.';
 WAVERR_UNPREPARED: Result:=Result+' The data block pointed to by the pwh parameter hasn''''t been prepared.';
 end;
end;


{Variables to handle abort - waveOutProc cannot call certain }
{ system routines during waveOutRest. They are disposed of outside}
{the waveOutProc in the WaveOutShutdown}

var Aborted: boolean;		{Flags that the operation is aborted}
var AbortHeader: PWaveHdr;	{Holds the header for the aborted sound}
var AbortHandle:  HWaveOut;	{Holds the handle for the anorted sound}


procedure waveOutProc(WaveHandle: HWaveOut; Msg: UINT; Instance, Param1, Param2: dword); stdcall;
{Call back procedure to handle destroying the header }
begin
case Msg of
 WOM_CLOSE: ;
 WOM_DONE:
	begin
	if not Aborted then
		begin
		{Unprepare cannot be called here unless the sound has finished playing}
		waveOutUnprepareHeader(WaveHandle,PWaveHdr(Param1),SizeOf(TWaveHdr));
		Dispose(PWaveHdr(Param1));
		PWaveHdr(Param1):=nil;
		end
	else
		begin
		{Save the data so we can use it later}
		AbortHeader:=PWaveHdr(Param1);
		AbortHandle:=WaveHandle;
		end;
	Dec(PendingOutBuffers);
	end;
 WOM_OPEN: ;
 end;
end;


procedure SetupOutputFormat(OutFormat: TWaveFormatEX);
{Create outoutput wave handle based on specified handle}
begin
if WaveOutHandle<>0 then
	begin
	{Required or the memory will continue to grow}
	waveOutClose(WaveOutHandle);
	WaveOutHandle:=0;
	end;
if waveOutOpen(@WaveOutHandle, WAVE_MAPPER,@OutFormat, 0, 0, WAVE_FORMAT_QUERY) <> MMSYSERR_NOERROR then raise Exception.Create('Can''t play format');
if waveOutOpen(@WaveOutHandle, WAVE_MAPPER,@OutFormat, dword(@waveOutProc), 0, CALLBACK_FUNCTION) <> MMSYSERR_NOERROR then raise Exception.Create('Failed to open output device');
end;



procedure WaveBufferOut(Data: Pointer; Size: integer; Wait: boolean);
{Play specified wave data}
var OutBufferHeader: PWaveHdr;
var R: MMRESULT;
begin
New(OutBufferHeader);
FillMemory(OutBufferHeader,SizeOf(TWaveHdr),0);
OutBufferHeader.lpData:= Data;
OutBufferHeader.dwBufferLength:= Size;
Aborted:=False;
AbortHeader:=nil;
AbortHandle:=0;

if waveOutPrepareHeader(WaveOutHandle,OutBufferHeader,  SizeOf(TWaveHdr)) <> MMSYSERR_NOERROR then raise Exception.Create('Unable to prepare header');
R:=waveOutWrite(WaveOutHandle,OutBufferHeader,SizeOf(TWaveHdr));
if R <>  MMSYSERR_NOERROR then  raise Exception.Create(GetErrorStr(R));
Inc(PendingOutBuffers);
if Wait then
	begin
	repeat until (OutBufferHeader.dwFlags and WHDR_DONE)<> 0;
	end;
end;


procedure WaveBufferOut(Data: Pointer; Size: integer);
begin
WaveBufferOut(Data,Size,False);
end;



procedure WaitForSoundEnd;
begin
while (PendingOutBuffers>0) and
      not Application.Terminated do
	Application.ProcessMessages;
end;



procedure PlayWaveArray(OutFormat: TWaveFormatEX; Data: Pointer; Size: integer; Wait: boolean);
begin
SetupOutputFormat(OutFormat);
WaveBufferOut(Data, Size,Wait);
end;


procedure WaveOutShutdown;
begin
Aborted:=True;
waveOutReset(waveOutHandle);
WaitForSoundEnd;
waveOutUnprepareHeader(AbortHandle,AbortHeader,SizeOf(TWaveHdr));
if AbortHeader<> nil then
	begin
	Dispose(AbortHeader);
	AbortHeader:=nil;
	end;
waveOutClose(WaveOutHandle);
PendingOutBuffers:=0;
end;



{------------------------------------------------------------------------------}


type TSoundData = array of byte;

const SampPerSec = 11025;

procedure BuildWaveTable(var Data: TSoundData; Freq,Sec,Amp: double);
var I: integer;
var W,V: double;
begin
SetLength(Data,round(SampPerSec*Sec));
for I:=0 to High(Data) do
	begin
	W:=(I * Freq * 2 * Pi)/SampPerSec;
	V:=Sin(W) * 128 * Amp + 128;
	Data[I]:=trunc(V);
	end;
end;


procedure SynthesizeSound8(Freq,Sec: double; Wait: boolean);
var OutFormat: TWaveFormatEx;
var Data: TSoundData;
begin
OutFormat.nChannels:=1;
OutFormat.nSamplesPerSec:=SampPerSec;
OutFormat.nAvgBytesPerSec:=SampPerSec;
OutFormat.nBlockAlign:=1;
OutFormat.wBitsPerSample:=8;
OutFormat.cbSize:=0;
OutFormat.wFormatTag := WAVE_FORMAT_PCM;
SetupOutputFormat(OutFormat);
BuildWaveTable(Data,Freq,Sec,0.2);
WaveBufferOut(Data,Length(Data));
if Wait then WaitForSoundEnd;
end;



{------------------MCI Library Routines ---------------------------------------}



procedure PlaybackSoundFile(Filename: string);
begin
mciSendString(PChar('close SoundFile'), nil, 0, 0);
mciSendString(PChar('open "' + Filename + '" alias SoundFile'), nil, 0, 0);
mciSendString('play SoundFile', nil, 0, 0);
end;



procedure StopSoundFile;
begin
mciSendString('stop SoundFile', 0,0,0);
end;



procedure CloseSoundFile;
begin
mciSendString('close SoundFile', 0,0,0);
end;



end.
