unit avilib;

interface

uses VFW,Windows,classes,graphics,sysutils;

type
 TVideoWriter = class(TObject)
 private
  AVIFile: IAVIFile;
  AVIStream,CompressStream: IAVISTREAM;
  StreamHeader: TAVISTREAMINFO;
  CompressOptions: TAVICOMPRESSOPTIONS;
  BMHeader: PBitmapInfoHeader;
  MBuffer: TMemoryStream;
  hr: longint;
  CurrentFrame: integer;
  procedure GetBitmapHeader(BM: TBitmap;var BH: PBitmapInfoHeader);
 public
  FramesPerSecond: integer;
  FrameInc: integer;
  Quality: integer;
  constructor Create;
  destructor Destroy; override;
  procedure CloseFile;
  procedure OpenFile(BaseBitmap: TBitmap; Filename: string);
  procedure WriteFrame(Bitmap: TBitmap);
 end;


type
 TVideoReader = class(TObject)
 private
  ReturnCode: Integer;
  AVIFile: IAVIFile;
  AVIStream: IAVIStream;
  GetFrame: IGETFRAME;
  function GetHeaderSize(BI: TBitmapInfoHeader): integer;
 public
  FileInfo: TAVIFileInfo;
  StreamInfo: TAVIStreamInfo;
  FrameCount, FirstFrame: integer;
  procedure CloseFile;
  function OpenFile(Filename: string): boolean;
  function ReadFrame(Bitmap: TBitmap; FrameNumber: integer): boolean;
 end;

type
 TVideoCapture = class(TObject)
 private
  CaptureCanvas: TCanvas;
  AVIBitmap: TBitmap;
  AVIWriter: TVideoWriter;
  SourceRect,DestRect: TRect;
 public
  Capturing: boolean;
  constructor Create;
  destructor Destroy; override;
  procedure StartCapture(Filename: string; Canvas: TCanvas; FrameRect: TRect; FPS,Quality: integer);
  procedure CaptureWindow;
  procedure CaptureBitmap(BM: TBitmap);
  procedure EndCapture;
 end;


implementation




function TestAVIReturnValue(V: integer; BaseError: string): boolean;
begin
Result:=False;
case V of
 AVIERR_OK: Result:=True;
 AVIERR_BADFORMAT: raise Exception.Create(BaseError+': Bad/corrupt file format.');
 AVIERR_MEMORY: raise Exception.Create(BaseError+': Insufficient memory.');
 AVIERR_FILEREAD: raise Exception.Create(BaseError+': Disk read error.');
 AVIERR_FILEOPEN: raise Exception.Create(BaseError+': Disk open error.');
 REGDB_E_CLASSNOTREG: raise Exception.Create(BaseError+': Unregistered File Type');

 AVIERR_UNSUPPORTED:  raise Exception.Create(BaseError+': Compression Not Supported');
 AVIERR_NOCOMPRESSOR:  raise Exception.Create(BaseError+': Suitable Compressor Not Found');
 AVIERR_READONLY:  raise Exception.Create(BaseError+': File opened Read only');
 AVIERR_NODATA:  raise Exception.Create(BaseError+': Files does not contains specified data type');
 AVIERR_BUFFERTOOSMALL:  raise Exception.Create(BaseError+': Buffer Smaller Than Sample');

 AVIERR_INTERNAL:  raise Exception.Create(BaseError+': AVI Internal Error');
 AVIERR_BADFLAGS:  raise Exception.Create(BaseError+': Bad Flags');
 AVIERR_BADPARAM:  raise Exception.Create(BaseError+': Bad Parameters');
 AVIERR_BADSIZE:  raise Exception.Create(BaseError+': Bad size');
 AVIERR_BADHANDLE:  raise Exception.Create(BaseError+': Bad Handle');
 AVIERR_FILEWRITE:  raise Exception.Create(BaseError+': File write error');
 AVIERR_COMPRESSOR:  raise Exception.Create(BaseError+': Compressor Error');
 AVIERR_CANTCOMPRESS:  raise Exception.Create(BaseError+': Unable to compress');
 AVIERR_USERABORT:  raise Exception.Create(BaseError+': User abort');
 AVIERR_ERROR:  raise Exception.Create(BaseError+': AVI Error');

 else  raise Exception.Create(BaseError+' Unknown Error: '+IntToHex(V,8));
 end;
end;


{---------------- Video Writer Routines ------------------------------}


constructor TVideoWriter.Create;
begin
inherited Create;
MBuffer:=TMemoryStream.Create;
Quality:=6500;
end;



destructor TVideoWriter.Destroy;
begin
MBuffer.Free;
inherited Destroy;
end;



procedure TVideoWriter.CloseFile;
begin
if AVIStream<>Nil then AVIStreamRelease(AVIStream);
if CompressStream<>Nil then AVIStreamRelease(CompressStream);
if AVIFile<>Nil then AVIFileRelease(AVIFile);
AVIFileExit;
end;



procedure TVideoWriter.GetBitmapHeader(BM: TBitmap;var BH: PBitmapInfoHeader);
begin
MBuffer.Clear;
BM.SaveToStream(MBuffer);
BH:=PBitmapInfoHeader(integer(MBuffer.Memory)+14);
end;



procedure TVideoWriter.WriteFrame(Bitmap: TBitmap);
begin
if CompressStream=Nil then raise Exception.Create('File not open');
GetBitmapHeader(Bitmap,BMHeader);
hr:=AVIStreamWrite(CompressStream,CurrentFrame,1,
                PBYTE(cardinal(BMHeader)+BMHeader^.biSize+BMHeader^.biClrUsed*sizeof(TRGBQuad)),
                  BMHeader^.biSizeImage,AVIIF_KEYFRAME,Nil,Nil);
TestAVIReturnValue(hr,'Unable to write AVI Frame');
Inc(CurrentFrame,FrameInc);
end;




procedure TVideoWriter.OpenFile(BaseBitmap: TBitmap; Filename: string);
var Version: integer;
begin
try
Version := HIWORD(VideoForWindowsVersion);
if Version<$010A then raise Exception.Create('Invalid AVI version');
AVIFile:=nil; AVIStream:=Nil; CompressStream:=Nil;
CurrentFrame:=0;
GetBitmapHeader(BaseBitmap,BMHeader);

AVIFileInit;

hr:=AVIFileOpen(AVIFile,PChar(Filename),OF_WRITE  or  OF_CREATE,nil);
TestAVIReturnValue(hr,'Unable to create AVI file ("'+ExtractFileExt(Filename)+'")');


AVIFile._AddRef;

{Create AVI stream }
FillChar(StreamHeader,sizeof(StreamHeader), 0);
StreamHeader.fccType := streamtypeVIDEO;
StreamHeader.fccHandler := 0;
StreamHeader.dwScale := 1;
StreamHeader.dwRate := FramesPerSecond;
StreamHeader.dwSuggestedBufferSize := BMHeader^.biSizeImage;

SetRect(StreamHeader.rcFrame, 0, 0,BMHeader^.biWidth,BMHeader^.biHeight);
hr:=AVIFileCreateStream(AVIFile,AVIStream,StreamHeader);
TestAVIReturnValue(hr,'Unable to create AVI stream');
AVIStream._AddRef;

{Create compression stream} 
FillChar(CompressOptions, sizeof(CompressOptions),0);
CompressOptions.dwFlags:=AVICOMPRESSF_VALID;
CompressOptions.fccType:=streamtypeVIDEO;
CompressOptions.fccHandler:=mmioFOURCC('M','S','V','C');
CompressOptions.dwQuality:=Quality;
CompressOptions.dwKeyFrameEvery:=7;

hr:=AVIMakeCompressedStream(CompressStream,AVIStream,@CompressOptions,Nil);
TestAVIReturnValue(hr,'Unable to create Compression Stream');
CompressStream._AddRef;

hr:=AVIStreamSetFormat(CompressStream,0,BMHeader,BMHeader^.biSize +BMHeader^.biClrUsed*sizeof(TRGBQUAD));
TestAVIReturnValue(hr,'Unable to set Compression Format');

except
 AVIStream:=Nil;
 CompressStream:=Nil;
 AVIFile:=Nil;
 Raise;
 end
end;


{-----------------AVI Reader Routines ---------------------------------}


function TVideoReader.ReadFrame(Bitmap: TBitmap; FrameNumber: integer): boolean;
var BitmapInfo: PBITMAPINFO;
var BMPBits: Pointer;
var BMPHandle: HBITMAP;
var IP: Pointer;
begin
Result:=True;
try
BitmapInfo:=PBitmapInfo(AVIStreamGetFrame(GetFrame, FrameNumber));
if BitmapInfo = nil then raise Exception.Create('Unable to read frame');

BMPHandle:=CreateDIBSection(0, BitmapInfo^, DIB_RGB_COLORS, BMPBits, 0, 0);
if (BMPHandle = 0) or (BMPBits = nil) then raise Exception.Create('Insuficient Memory');

IP:=Pointer(Integer(BitmapInfo)+GetHeaderSize(BitmapInfo.bmiHeader));
Move(IP^,BMPBits^,BitmapInfo.bmiHeader.biSizeImage);
Bitmap.Handle:=BMPHandle;
except
 Result:=False;
 end;
end;

procedure TVideoReader.CloseFile;
begin
if GetFrame<>Nil then AVIStreamGetFrameClose(GetFrame);
if AVIStream<>Nil then AVIStreamRelease(AVIStream);
if AVIFile<>Nil then AVIFileRelease(AVIFile);
AVIFileExit;
end;


function TVideoReader.GetHeaderSize(BI: TBitmapInfoHeader): integer;
begin
if BI.biBitCount > 8 then
	begin
	Result := SizeOf(TBitmapInfoHeader);
	if (BI.biCompression and BI_BITFIELDS) <> 0 then Inc(Result, 12);
	end
else Result := SizeOf(TBitmapInfoHeader) + SizeOf(TRGBQuad) * (1 shl BI.biBitCount);
end;


function TVideoReader.OpenFile(Filename: string): boolean;
begin
Result:=True;
AVIFile:=Nil; AVIStream:=Nil; GetFrame:=Nil;
try
AVIFileInit;

ReturnCode:=AVIFileOpen(AVIFile, PChar(Filename), 0, nil);
TestAVIReturnValue(ReturnCode,'Unable to open file');
AVIFile._AddRef;


ReturnCode:=AVIFileGetStream(AVIFile, AVIStream, 0, 0);
TestAVIReturnValue(ReturnCode,'Unable to open stream');
AVIStream._AddRef;

GetFrame:=AVIStreamGetFrameOpen(AVIStream, nil);
if GetFrame = nil then  raise Exception.Create('Unable to setup frame-reader');
GetFrame._AddRef;

AVIFileInfo(AVIFile, FileInfo, Sizeof(TAVIFileInfo));
AVIStreamInfo(AVIStream, StreamInfo, SizeOf(TAVIStreamInfo));
FrameCount:=AVIStreamLength(AVIStream);
FirstFrame:=AVIStreamStart(AVIStream);

except
 Result:=False;
 GetFrame:=Nil;
 AVIStream:=Nil;
 AVIFile:=Nil;
 end
end;

{-------------------Video Capture Routines ---------------------------}


constructor TVideoCapture.Create;
begin
inherited Create;
AVIBitmap := TBitmap.Create;
AVIBitmap.Canvas.CopyMode := cmSrcCopy;
end;



destructor TVideoCapture.Destroy;
begin
AVIBitmap.Free;
if Capturing then AVIWriter.Free;
inherited Destroy;
end;



procedure TVideoCapture.StartCapture(Filename: string; Canvas: TCanvas;FrameRect: TRect;FPS,Quality: integer);
var W,H: integer;
var Bits: integer;
begin
AVIWriter:=TVideoWriter.Create;
Capturing:=True;
CaptureCanvas:=Canvas;
SourceRect:=FrameRect;
H:=SourceRect.Bottom-SourceRect.Top; W:=SourceRect.Right-SourceRect.Left;
DestRect:=Rect(0,0,W,H);
AVIWriter.FramesPerSecond:=FPS;
AVIWriter.FrameInc:=1;
AVIWriter.Quality:=Quality;
{Setup bitmap }
AVIBitmap.Width:=W; AVIBitmap.Height:=H;
Bits:=GetDeviceCaps(Canvas.Handle,BITSPIXEL)*GetDeviceCaps(Canvas.Handle,PLANES);
case Bits of
 1: AVIBitmap.PixelFormat:=pf4bit;
 4: AVIBitmap.PixelFormat:=pf4bit;
 8: AVIBitmap.PixelFormat:=pf8bit;
 15: AVIBitmap.PixelFormat:=pf15bit;
 16,24: AVIBitmap.PixelFormat:=pf24bit;
 32: AVIBitmap.PixelFormat:=pf32bit;
 end;
try
AVIWriter.OpenFile(AVIBitmap,Filename);
except EndCapture; end;
end;


procedure TVideoCapture.CaptureWindow;
begin
AVIBitmap.Canvas.CopyRect(DestRect,CaptureCanvas,SourceRect);
AVIWriter.WriteFrame(AVIBitmap);
end;


procedure TVideoCapture.CaptureBitmap(BM: TBitmap);
begin
BM.PixelFormat:=AVIBitmap.PixelFormat;
AVIWriter.WriteFrame(BM);
end;



procedure TVideoCapture.EndCapture;
begin
AVIWriter.CloseFile;
AVIWriter.Free;
Capturing:=False;
end;

end.