unit tifread;

{ Unit to read common versions of 6.0 TIFF file formats }

interface

uses Windows,SysUtils,Classes,Graphics, TifGlobal, GeoTiff, StreamLib,
     Math,BitmapLib, DSupport;


{Type to hold the parameters from current TIF file}

const PalSize=256;

Type
TTIFFInfo = record
 Width,Depth,
 BytesPerRow,Bits: Word;
 PhotoMetric: Word;
 ExtraSamples: Word;
 RowsPerStrip: Word;
 StripCount: Word;
 Samples: Word;
 Predictor: Word;
 PlanarConfig: Word;
 Compression: Word;
 BitsPerSample: Word;
 ByteCount: DWord;
 Offset: DWord;
 XRes, YRes: double;
 ResUnit: DWord;
 Palette: array [0..PalSize-1] of TRGBQuad;
 TagList: array of word;
 end;

var VerNumber: word;		{Identifies file as TIFF }
var BMPBytesPerLine: integer;	{Bytes/Line for Windows Bitmap (dword aligned)}

var TiffFilename: string;	{Current input filename }
var TIFInfo: TTIFFInfo;		{Contain all info about current file }

var StartPos: Int64;		{Used for streams whose 1st position in not zero}


{ Public procedures that can be called from external programs }

procedure ReadTIFFile(BM: TBitmap; FileName: string);
procedure ReadTIFStream(Stream: TStream; BM: TBitmap);
function TIFFInfoToStr(TI: TTIFFInfo): string;


implementation



function TIFFInfoToStr(TI: TTIFFInfo): string;
var I: integer;
begin
Result:='';
Result:=Result+'Width/Height: '+IntToStr(TI.Width)+','+IntToStr(TI.Depth)+CRLF_Char;
Result:=Result+'Bits: '+IntToStr(TI.Bits)+CRLF_Char;
Result:=Result+'Bytes/Row: '+IntToStr(TI.BytesPerRow)+CRLF_Char;
Result:=Result+'PhotoMetric: '+IntToStr(TI.PhotoMetric)+CRLF_Char;
Result:=Result+'Extra Samples: '+IntToStr(TI.ExtraSamples)+CRLF_Char;
Result:=Result+'Rows/Strip: '+IntToStr(TI.RowsPerStrip)+CRLF_Char;
Result:=Result+'Strip Count: '+IntToStr(TI.StripCount)+CRLF_Char;
Result:=Result+'Samples: '+IntToStr(TI.Samples)+CRLF_Char;
Result:=Result+'Predictor: '+IntToStr(TI.Predictor)+CRLF_Char;
Result:=Result+'Planar Config: '+IntToStr(TI.PlanarConfig)+CRLF_Char;
Result:=Result+'Compression: '+IntToStr(TI.Compression)+CRLF_Char;
Result:=Result+'Bits/Sample: '+IntToStr(TI.BitsPerSample)+CRLF_Char;

Result:=Result+'Byte Count: '+IntToStr(TI.ByteCount)+CRLF_Char;
Result:=Result+'Offset: '+IntToStr(TI.Offset)+CRLF_Char;
Result:=Result+'XRes: '+FloatToStrF(TI.XRes,ffFixed,18,1)+CRLF_Char;
Result:=Result+'YRes: '+FloatToStrF(TI.YRes,ffFixed,18,1)+CRLF_Char;
Result:=Result+'ResUnit: '+IntToStr(TI.ResUnit)+CRLF_Char;
Result:=Result+'Tags: '+CRLF_Char;
for I:=0 to High(TI.TagList) do
	begin
	Result:=Result+'        ';
	Result:=Result+IntToStr(TI.TagList[I]);
	Result:=Result+' - '+GetTagStr(TI.TagList[I])+CRLF_Char;
	end;
end;



procedure DoRelativeSeek(Stream: TStream; Offset: Int64);
{Some streams may not start from beginning of device}
{This seeks relative to the starting position of the stream}
begin
{Use Position rather than Seek, see TSectorStream.Seek}
Stream.Position:=Offset+StartPos;
end;



procedure UnhandledCompression(C: integer);
{Exception to deal with unhandled compression}
var S: string;
begin
S:='Unsupported Compression = '+IntToStr(C);
raise Exception.Create(S);
end;




{----------------- TIFF Routines ------------------------------------}


procedure SetDefaults(var FI: TTIFFInfo);
{Fill parameter structure with default values}
var I : Integer;
begin
SetLength(FI.TagList,0);
FI.Width:=0;
FI.Depth:=0;
FI.Bits:=0;
FI.PlanarConfig:=1;
FI.Samples:=1;
FI.PhotoMetric:=1;
FI.Compression:=1;
FI.Offset:=0;
FI.BytesPerRow:=0;
FI.RowsPerStrip:=0;
FI.StripCount:=0;
FI.BitsPerSample:=0;
FI.ByteCount:=0;
FI.ExtraSamples:=0;
FI.Predictor:=0;
SetLength(FI.TagList,0);
for I:=0 to PalSize-1 do
	begin
	FI.Palette[I].rgbRed:=I;
	FI.Palette[I].rgbGreen:=I;
	FI.Palette[I].rgbBlue:=I;
	FI.Palette[I].rgbReserved:=0;
	end;
end;


procedure SetBilevelPalette(OneIsBlack: boolean; var FI: TTIFFInfo);
{Set bilevel paletter so first two entries control black and white}
{if OneisBlack is true, a one in the data signifies a black pixel}
var B,W: integer;
begin
if OneIsBlack then begin B:=1; W:=0; end
else begin B:=0; W:=1; end;
FI.Palette[B].rgbRed:=0; FI.Palette[B].rgbGreen:=0; FI.Palette[B].rgbBlue:=0;
FI.Palette[W].rgbRed:=255; FI.Palette[W].rgbGreen:=255; FI.Palette[W].rgbBlue:=255;
end;


procedure GetColorMap(Stream: TStream;Offset: integer; var FI: TTIFFInfo);
{Read the TIF file color map into TIF parameters}
var I: integer;
var Pos: int64;
begin
Pos:=Stream.Position;
DoRelativeSeek(Stream,Offset);
for I:=0 to (1 shl FI.BitsPerSample)-1 do
        begin
        if I >= 256 then break;
        FI.Palette[I].rgbRed:= GetStreamWord(Stream) shr 8;
        end;
for I:=0 to (1 shl FI.BitsPerSample)-1 do
        begin
        if I >= 256 then break;
        FI.Palette[I].rgbGreen:= GetStreamWord(Stream) shr 8;
        end;
for I:=0 to (1 shl FI.BitsPerSample)-1 do
        begin
        if I >= 256 then break;
        FI.Palette[I].rgbBlue:=GetStreamWord(Stream) shr 8;
        end;
{Becuase Pos is absolute, no need to use DoRelativeSeek}
Stream.Position:=Pos;
end;


function GetRational(Stream: TStream; Offset: integer): double;
{Get rational number pointed to by Offset}
{And return the equivalent double value}
var N,D: integer;
var OldPos: int64;
begin
OldPos:=Stream.Position;
DoRelativeSeek(Stream,Offset);
N:=GetStreamDWord(Stream);
D:=GetStreamDWord(Stream);
{Becuase Pos is absolute, no need to use DoRelativeSeek}
Stream.Position:=OldPos;
Result:=N/D;
end;




procedure DecodeTag(Stream: TStream; var FI: TTIFFInfo);
{Read and decode a TIF tag }
var Len,Offset,Value : DWord;
var Pos: int64;
var Tag,FieldType : Integer;
var Rational: double;
begin
Tag:=GetStreamWord(Stream);
FieldType:=GetStreamWord(Stream);
Len:=GetStreamDWord(Stream);
Offset:=GetStreamDWord(Stream);
if (FieldType = tifFractional) then Rational:=GetRational(Stream,Offset)
else if (FieldType = tifLong) or (TifNumberType=TifLittleEnd) then  Value:=Offset
else Value:=Offset shr 16;
SetLength(FI.TagList,Length(FI.TagList)+1);
FI.TagList[High(FI.TagList)]:=Tag;
case Tag of
 SubfileType: ;
 NewSubFile: ;
 ImageWidth: FI.Width:=Value;
 ImageLength: FI.Depth:=Value;
 RowsPerStrip: FI.RowsPerStrip:=Value;
 StripOffsets:
	begin
	FI.Offset:=Value;
	FI.StripCount:=Len;
	end;
 StripByteCounts: FI.ByteCount:=Value;
 SamplesPerPixel: FI.Samples:=Value;
 Predictor: FI.Predictor:=Value;
 BitsPerSample:
	begin
	if Len > 1 then
		begin
		Pos:=Stream.Position;
		DoRelativeSeek(Stream,Offset);
		FI.BitsPerSample:=GetStreamWord(Stream);
		{Becuase Pos is absolute, no need to use DoRelativeSeek}
		Stream.Position:=Pos;
		end
	else FI.BitsPerSample:=Value;
	end;
 PlanarConfiguration: FI.PlanarConfig:=Value;
 Compression: FI.Compression:=Value;
 Group3Options: ;
 Group4Options: ;
 FillOrder: ;
 Threshholding: ;
 CellWidth: ;
 CellLength: ;
 MinSampleValue: ;
 MaxSampleValue: ;
 PhotometricInterp: FI.PhotoMetric:=Value;
 GrayResponseUnit: ;
 GrayResponseCurve: ;
 ColorResponseUnit: ;
 ColorResponseCurves: ;
 XResolution: FI.XRes:=Rational;
 YResolution: FI.YRes:=Rational;
 ResolutionUnit: FI.ResUnit:=Value;
 Orientation: ;
 DocumentName: ;
 PageName: ;
 XPosition: ;
 YPosition: ;
 PageNumber: ;
 ImageDescription: ;
 Make: ;
 Model: ;
 FreeOffsets: ;
 FreeByteCounts: ;
 ColorMap: GetColorMap(Stream,Offset,FI);
 ExtraSamples: FI.ExtraSamples:=Value;
 else DecodeGeoTiffTag(Stream,Tag,Len,Offset);
 end;
end;


{--------------- Handle Color Scheme Conversions ------------------------------}




type TCMYK = packed record
 Cyan,Magenta,Yellow,Black: byte;
 end;

type TCMYKArray = array [0..32768] of TCMYK;
type PCMYKArray = ^TCMYKArray;



function CMYKToRGB(CMYK: TCMYK): TRGBTriple;
{Convert CYMYK to RGB}
const MaxValue = 255;
begin
Result.rgbtRed:=MaxValue-EnsureRange(CMYK.Cyan+CMYK.Black,0,255);
Result.rgbtGreen:=MaxValue-EnsureRange(CMYK.Magenta+CMYK.Black,0,255);
Result.rgbtBlue:=MaxValue-EnsureRange(CMYK.Yellow+CMYK.Black,0,255);
end;


procedure ReorderRGB(Destination: PByteArray; Bytes: integer);
{Change BGR color order to RGB for Windows BMP}
var I: integer;
var T: byte;
begin
I:=0;
while I<=Bytes-3 do
	begin
	T:=Destination[I];
	Destination[I]:=Destination[I+2];
	Destination[I+2]:=T;
	Inc(I,3);
	end;
end;



procedure DoCMYKToRGBQuad(Destination: PCMYKArray; Bytes: integer);
{Change CMYK color order to RGB for Windows BMP}
var I: integer;
var RGBTriple: TRGBTriple;
begin
for I:=0 to (Bytes div 4) do
	begin
	RGBTriple:=CMYKToRGB(Destination[I]);
	PRGBTriple(@Destination[I])^:=RGBTriple;
	end;
end;




procedure ReorderRGBQuad(Destination: PByteArray; Bytes: integer);
{Change BGR color order to RGB for Windows BMP}
var I: integer;
var T: byte;
begin
I:=0;
while I<=Bytes-4 do
	begin
	T:=Destination[I];
	Destination[I]:=Destination[I+2];
	Destination[I+2]:=T;
	Inc(I,4);
	end;
end;


procedure InvertColorScheme(Destination: PByteArray; Bytes: integer);
var I: integer;
begin
for I:=0 to Bytes-1 do Destination[I]:=255 - Destination[I];
end;



procedure GrayScaleToRGB16(Destination: PWordArray; Bytes: integer);
var I: integer;
var T: Word;
begin
for I:=0 to (Bytes div 2)-1 do
	begin
	T:=Destination[I] shr 11;
	Destination[I]:=T or (T shl 5) or (T shl 10);
	end;
end;




procedure ConvertColorScheme(var Buffer: Pointer; FI: TTIFFInfo);
begin
case FI.Bits of
 8:	begin
	if FI.PhotoMetric=0 then InvertColorScheme(Buffer,FI.BytesPerRow);
	end;
 16: GrayScaleToRGB16(Buffer,FI.BytesPerRow);
 24: ReorderRGB(Buffer,FI.BytesPerRow);
 32:	begin
	if FI.PhotoMetric=Photometric_Separated then DoCMYKToRGBQuad(Buffer,FI.BytesPerRow)
	else ReorderRGBQuad(Buffer,FI.BytesPerRow);
	end;
 end;
end;

{------------------ LZW Decompression Routines ----------------------------}


var LZWWData,LZWBits,CodeWidth: cardinal;


const ClearCode = 256;
const EOICode = 257;
const NullCode = 4096;


function GetLZWCode(Stream:TStream; CodeSize: dword): integer;
{Read one variable-width code from the stream}
{CodeSize holds the number of bits wide the code is}
var CodeMask: dword;
var Atom: byte;
begin
Result:=EOICode;
CodeMask:=(1 shl CodeSize) - 1;
while LZWBits < CodeSize do
	begin
	if Stream.Read(Atom,1)<1 then exit;
	Inc(LZWWData, Cardinal(Atom) shl (24 - LZWBits));
	Inc(LZWBits, 8);
	end;

Result:=(LZWWData and ($FFFFFFFF - CodeMask)) shr (32 - CodeSize);
LZWWData:=LZWWData shl CodeSize;
Dec(LZWBits, CodeSize);
end;


function ReadLZWStrip(Stream: TStream; var Destination: Pointer; Lines: integer; FI: TTIFFInfo): boolean;
{Reads and decompresses a single 'Strip' from input file}
{Each 'Strip' can contain multiple scan lines of data }
{Each scan line must be stored in the Windows BMP dword aligned}
var L: integer;
var LZWIndex: cardinal;
var OldCode: cardinal;
var BackLink: array[0..4095] of DWord;
var DecodeValues: array [0..4095] of Byte;
var Stack: array [0..4095] of Byte;
var StackPointer: PByte;
var FirstChar: Byte;
var Strip: PByteArray;

	procedure InitializeLZWTable;
	{Initialize decode table}
	var I: integer;
	begin
	for I:=0 to ClearCode - 1 do
		begin
		BackLink[I]:=NullCode;
		DecodeValues[I]:=I;
		end;
	end;


	procedure ResetLZW;
	{Reset state of LZW decoder}
	begin
	CodeWidth:=9;
	LZWIndex:=ClearCode + 2;
	OldCode:=NullCode;
	end;



	function DecodeLZW(Code: DWord; var Buffer: PByte): Boolean;
	{Decode current LZW code }
	var InCode: DWord; { buffer for passed code}
	begin
	if Code > LZWIndex then raise Exception.Create('Invalid LZW Code: '+IntToStr(Code)+'>'+IntToStr(LZWIndex)+' @:'+IntToStr(Stream.Position));

	{ handling for the first LZW code: print and keep it}
	if OldCode = NullCode then
		begin
		FirstChar:=DecodeValues[Code];
		Buffer^:=FirstChar;
		Inc(Buffer);
		OldCode:=Code;
		Result:=True;
		Exit;
		end;

	{ keep the passed LZW code}
	InCode:=Code;

	{ the first LZW code is always smaller than FFirstCode}
	if Code = LZWIndex then
		begin
		StackPointer^:=FirstChar;
		Inc(StackPointer);
		Code:=OldCode;
		end;

	{ loop to put decoded bytes onto the stack}
	while Code > ClearCode do
		begin
		StackPointer^:=DecodeValues[Code];
		Inc(StackPointer);
		Code:=BackLink[Code];
		end;

	{ place new code into code table}
	FirstChar:=DecodeValues[Code];
	StackPointer^:=FirstChar;
	Inc(StackPointer);
	BackLink[LZWIndex]:=OldCode;
	DecodeValues[LZWIndex]:=FirstChar;
	if LZWIndex < 4096 then Inc(LZWIndex);

	{ put decoded bytes (from the stack) into the target buffer}
	OldCode:=InCode;
	repeat
		begin
		Dec(StackPointer);
		Buffer^:=StackPointer^;
		Inc(Buffer);
		end
	until DWord(StackPointer) <= DWord(@Stack);
	Result:=True;
	end;



	function ReadStrip(Stream: TStream; Buffer: Pointer): integer;
	{Read the strip into buffer }
	var Code: dword;
	var Target: Pointer;
	begin
	Target:=Buffer;
	while true do
		begin
		Code:=GetLZWCode(Stream,CodeWidth);
		case Code of
		 ClearCode: ResetLZW;
		 EOICode: break;
		 else DecodeLZW(Code,PByte(Target));
		 end;
		if LZWIndex>=(1 shl CodeWidth)-1 then Inc(CodeWidth);
		Result:=Cardinal(Target)-Cardinal(Buffer);
		if Result>=(TIFInfo.BytesPerRow*TIFInfo.RowsPerStrip) then break;
		end;
	end;

begin
GetMem(Strip,FI.BytesPerRow*Lines);
try
LZWWData:=0; LZWBits:=0;
InitializeLZWTable;
ResetLZW;
StackPointer:=@Stack;
ReadStrip(Stream,Strip);
for L:=0 to Lines-1 do
	begin
	AdvanceProgress(psRunning,0);
	Move(Strip[L*Fi.BytesPerRow],Destination^,FI.BytesPerRow);
	ConvertColorScheme(Destination,FI);
	{Copy backward to match BMP standard}
	Dec(DWord(Destination),BMPBytesPerLine);
	end;
finally FreeMem(Strip); end;
Result:=True;
end;


{---------------- Decode PackBits Compression -------------------------}

function ReadPackBitsLine(Stream: TStream; Destination: Pointer; BytesOut: Integer): Integer;
{Read single scanline of packbits data into Destination}
var B: Byte;
var RepCount: integer;
var P: Pointer;
begin
P:=Destination;
repeat
	begin
	RepCount:=SignedByteToInteger(GetStreamByte(Stream));
	if RepCount>=0 then
		begin
		{0-128 = Copy N bytes from input}
		Stream.Read(P^,RepCount+1);
		Inc(Cardinal(P),RepCount+1);
		end
	else if RepCount>-128 then
		begin
		{-128 to -1 = Repeat next byte N times}
		B:=GetStreamByte(Stream);
		FillChar(P^,-RepCount+1,B);
		Inc(Cardinal(P),-RepCount+1);
		end
	else ;	{-128 is a NOP }
	Result:=Cardinal(P)-Cardinal(Destination);
	end
until Result >= BytesOut;
end;


function ReadPackBitsStrip(Stream: TStream; var Buffer: Pointer; Lines: integer; FI: TTIFFInfo): boolean;
{Read all 'PackBits' scanlines in a 'strip' and copy into BMP DWord aligned}
var L: integer;
begin
Result:=False;
for L:=0 to Lines-1 do
	begin
	AdvanceProgress(psRunning,0);     
	if ReadPackBitsLine(Stream,Buffer,FI.BytesPerRow) <> FI.BytesPerRow then exit;
	ConvertColorScheme(Buffer,FI);
	{Copy backward to match BMP standard}
	Dec(DWord(Buffer),BMPBytesPerLine);
	end;
Result:=True;
end;

{------------------ Handle Uncompressed data ---------------------------------}



function ReadUncompressedStrip(Stream: TStream; var Buffer: Pointer; Lines: integer; FI: TTIFFInfo): boolean;
{Read uncompressed scanlines in a 'strip' and copy into BMP DWord aligned}
var L: integer;
begin
Result:=False;
for L:=0 to Lines-1 do
	begin
	AdvanceProgress(psRunning,0);
	if Stream.Read(Buffer^,FI.BytesPerRow) <> FI.BytesPerRow then exit;
	ConvertColorScheme(Buffer,FI);
	{Copy backward to match BMP standard}
	Dec(DWord(Buffer),BMPBytesPerLine);
	end;
Result:=True;
end;

{------------------Rountines To Copy Data into BMP ---------------------}




function ReadStrip(Stream: TStream; var Buffer: Pointer; Lines: integer; FI: TTIFFInfo): boolean;
{Read a 'Strip' using appropriate Decompression }
begin
Result:=True;
case FI.Compression of
 Compress_None: Result:=ReadUncompressedStrip(Stream,Buffer,Lines,FI);
 Compress_LZW: Result:=ReadLZWStrip(Stream,Buffer,Lines,FI);
 Compress_PackBits: Result:=ReadPackBitsStrip(Stream,Buffer,Lines,FI);
 else UnhandledCompression(FI.Compression);
 end;
end;



function CopyData(Stream: TStream; Buffer: Pointer; FI: TTIFFInfo): boolean;
{Copy all 'Strips' into buffer }
var I,C: cardinal;
begin
Result:=False;
if FI.StripCount=1 then
	begin
	{StripCount=1, the Offset points to actual data}
	AdvanceProgress(psStarting,FI.Depth);
	DoRelativeSeek(Stream,FI.Offset);
	if not ReadStrip(Stream,Buffer,FI.Depth,FI) then exit;
	end
else
	begin
	{Offset points to table of values }
	AdvanceProgress(psStarting,FI.Depth);
	for I:=0 to FI.StripCount-1 do
		begin
		DoRelativeSeek(Stream,FI.Offset+(I*sizeof(DWord)));
		DoRelativeSeek(Stream,GetStreamDWord(Stream));
		if (((I+1) * FI.RowsPerStrip)-1) >= FI.Depth then C:=FI.Depth-(I*FI.RowsPerStrip)
		else C:=FI.RowsPerStrip;
		if not ReadStrip(Stream,Buffer,C,FI) then exit;
		end;
	end;
Result:=True;
end;


{------------------------------------------------------------------------------}


procedure Predictor8Bits(Buffer: Pointer; FI: TTIFFInfo);
var X,Y,I,RowSize: integer;
var BA: PByteArray;
begin
RowSize:=((FI.Bits * FI.Width + 31) div 32) * 4;
BA:=PByteArray(Buffer);
for Y:=0 to FI.Depth-1 do
	begin
	for X:=1 to FI.Width-1 do
		begin
		I:=X + (RowSize * Y);
		BA[I]:=BA[I-1]+BA[I];
		end;
	end;
end;



procedure Predictor24Bits(Buffer: Pointer; FI: TTIFFInfo);
var X,Y,RowSize: integer;
var TA: PRGBTripleArray;
begin
RowSize:=((FI.Bits * FI.Width + 31) div 32) * 4;
for Y:=0 to FI.Depth-1 do
	begin
	TA:=PRGBTripleArray(Integer(Buffer)+RowSize*Y);
	for X:=1 to FI.Width-1 do
		begin
		TA[X].rgbtRed:=ShortInt(TA[X].rgbtRed)+ShortInt(TA[X-1].rgbtRed);
		TA[X].rgbtGreen:=ShortInt(TA[X].rgbtGreen)+ShortInt(TA[X-1].rgbtGreen);
		TA[X].rgbtBlue:=ShortInt(TA[X].rgbtBlue)+ShortInt(TA[X-1].rgbtBlue);
		end;
	end;
end;



procedure Predictor32Bits(Buffer: Pointer; FI: TTIFFInfo);
var X,Y,RowSize: integer;
var TA: PRGBQuadArray;
var Old: TRGBQuad;
begin
RowSize:=((FI.Bits * FI.Width + 31) div 32) * 4;
for Y:=0 to FI.Depth-1 do
	begin
	{The rows are stored in the BMP from the bottom up}
	TA:=PRGBQuadArray(Integer(Buffer)+RowSize*(FI.Depth - Y-1));
	{Restart the value at the beginning of each line }
	//Old Code: doesn't seem to work: if (Y mod FI.RowsPerStrip)=0 then 
	Old:=TA[0];
	for X:=1 to FI.Width-1 do
		begin
		TA[X].rgbRed:=ShortInt(TA[X].rgbRed)+ShortInt(Old.rgbRed);
		TA[X].rgbGreen:=ShortInt(TA[X].rgbGreen)+ShortInt(Old.rgbGreen);
		TA[X].rgbBlue:=ShortInt(TA[X].rgbBlue)+ShortInt(Old.rgbBlue);
		TA[X].rgbReserved:=ShortInt(TA[X].rgbReserved)+ShortInt(Old.rgbReserved);
		Old:=TA[X];
		end;
	end;
end;



procedure HandlePredictors(Buffer: Pointer; FI: TTIFFInfo);
begin
case FI.Bits of
  8: Predictor8Bits(Buffer,FI);
 24: Predictor24Bits(Buffer,FI);
 32: Predictor32Bits(Buffer,FI);
 end;
end;


{------------------------------------------------------------------------------}

procedure GetResolution(FI: TTIFFInfo; var X,Y: integer);
begin
case FI.ResUnit of
 ResUnit_None:
	begin
	X:=Trunc(FI.XRes);
	Y:=Trunc(FI.YRes);
	end;
 ResUnit_Inch:
	begin
	X:=Trunc(FI.XRes * Meters2Inches);
	Y:=Trunc(FI.YRes * Meters2Inches);
	end;
 ResUnit_Centimeter:
	begin
	X:=Trunc(FI.XRes * 100);
	Y:=Trunc(FI.YRes) * 100;
	end;
 end;
end;



procedure StoreInBitmap(Stream: TStream; BM: TBitmap; FI: TTIFFInfo);
{Copy Tiff data into Delphi TBitmap Object }
var BitsMem,P: Pointer;
var BitmapInfo: PBitmapInfo;
var BMHandle: HBITMAP;
var I,Size,BitsPerPixel: integer;
var XRes,YRes: integer;
begin
if (FI.Width<=0) or (FI.Depth<=0) then raise Exception.Create('Width or Height is Zero');
GetMem(BitmapInfo, Sizeof(TBitmapInfoHeader) + 256 * sizeof(TRGBQuad));
try
GetResolution(FI,XRes,YRes);

BitsPerPixel:=FI.Bits;

BMPBytesPerLine:=BytesPerScanline(FI.Width,BitsPerPixel, 32);
Size:=BMPBytesPerLine * FI.Depth;

BitmapInfo^.bmiHeader.biSize:=Sizeof(TBitmapInfoHeader);
BitmapInfo^.bmiHeader.biWidth:=FI.Width;
BitmapInfo^.bmiHeader.biHeight:=FI.Depth;
BitmapInfo^.bmiHeader.biPlanes:=1;
BitmapInfo^.bmiHeader.biBitCount:=BitsPerPixel;
BitmapInfo^.bmiHeader.biCompression:=BI_RGB;
BitmapInfo^.bmiHeader.biSizeImage:=Size;
BitmapInfo^.bmiHeader.biXPelsPerMeter:=XRes;
BitmapInfo^.bmiHeader.biYPelsPerMeter:=YRes;
BitmapInfo^.bmiHeader.biClrUsed:=0;
BitmapInfo^.bmiHeader.biClrImportant:=0;

for I:=0 to 255 do BitmapInfo^.bmiColors[I]:=FI.Palette[I];

BMHandle:=CreateDIBSection(0, BitmapInfo^, DIB_RGB_COLORS, BitsMem, 0, 0);
if (BMHandle = 0) or (BitsMem = nil) then raise Exception.Create('Unsupported Format: Could not create Bitmap.');

{Copy data from end to begining, to make it compatable with old BMPs}
P:=Pointer(integer(BitsMem)+Size-BMPBytesPerLine);
CopyData(Stream,P,FI);
if FI.Predictor=2 then HandlePredictors(BitsMem,FI);


BM.ReleasePalette;
BM.Handle:=BMHandle;
{We have to set the values again because}
{creating a TBitmap loses the values}
SetBmpPPM(BM,XRes,YRes);
finally FreeMem(BitmapInfo); end;
end;



procedure ReadTIFStream(Stream: TStream; BM: TBitmap);
{Read TIF file from Stream and copy it into the BM bitmap}
{Currently configured to read single images, Add a loop}
{between comments "*Loop*" to read multiple images}
var Loc : int64;
var I,N : Integer;
begin
{Used for streams whose 1st position in not zero}
StartPos:=Stream.Position;
TifNumberType:=GetStreamWord(Stream);
if TifNumberType = TifBigEnd then StreamEndian:=BigEnd
else StreamEndian:=LittleEnd;

VerNumber:=GetStreamWord(Stream);
if (TifNumberType<>TifLittleEnd) and (TifNumberType<>TifBigEnd) and (VerNumber<>42) then
	raise Exception.Create('Invalid TIF File');

{ Only read the first image }
{ *LOOP* to read multiple images}
Loc:=GetStreamDWord(Stream);
if Loc=0 then exit;
DoRelativeSeek(Stream,Loc);

{ Read TIF information tags }
SetDefaults(TIFInfo);
N:=GetStreamWord(Stream);
for I:=0 to N-1 do DecodeTag(Stream,TIFInfo);

{ save the file position }
Loc:=Stream.Position;

{ figure the bits per pixel}
TIFInfo.Bits:=TIFInfo.Samples * TIFInfo.BitsPerSample;
if TIFInfo.Bits=1 then SetBilevelPalette(TIFInfo.PhotoMetric=0,TIFInfo);

{ figure the bytes }
if TIFInfo.Bits =1 then TIFInfo.BytesPerRow:=MulDiv(TIFInfo.Width,TIFInfo.Bits,8)+1
else if TIFInfo.Bits < 8 then TIFInfo.BytesPerRow:=MulDiv(TIFInfo.Width,TIFInfo.Bits,8)
else if TIFInfo.Bits=8 then TIFInfo.BytesPerRow:=TIFInfo.Width
else if TIFInfo.Bits=16 then TIFInfo.BytesPerRow:=TIFInfo.Width * 2
else if TIFInfo.Bits=24 then TIFInfo.BytesPerRow:=TIFInfo.Width * 3
else if TIFInfo.Bits=32 then TIFInfo.BytesPerRow:=TIFInfo.Width * 4
else raise Exception.Create('Unsupported Bits/Pixel: '+IntToStr(TIFInfo.Bits));;

StoreInBitmap(Stream,BM,TIFInfo);
AdvanceProgress(psEnding,0);
{ Restore pointer to next entry }
{Becuase Pos is absolute, no need to use DoRelativeSeek}

Stream.Position:=Loc;
{ *LOOP* to read multiple images}

OnTIFProgress:=nil;
end;





procedure ReadTIFFile(BM: TBitmap; FileName: string);
{Read TIF from file }
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.LoadFromFile(Filename);
try
ReadTIFStream(Stream,BM);
finally Stream.Free; end;
end;



end.
