unit ImageUndoer;

interface

uses Classes,SysUtils,ExtCtrls,Graphics, BitmapLib;

type TBitmapUndoer = class(TComponent)
 private
  FMaxUndo: integer;
 protected
  FUndoList: TBitmapList;
  FRedoList: TBitmapList;
 public
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure StoreUndo(BM: TBitmap); dynamic;
  procedure Undo(BM: TBitmap); dynamic;
  procedure Redo(BM: TBitmap); dynamic;
  procedure Clear; dynamic;
  procedure DeleteLastUndo;
 published
  property MaxUndo: integer read FMaxUndo write FMaxUndo;
 end;


type TStoreQueryEvent = procedure(Sender: TObject; var CanStore: Boolean) of object;



type TImageUndoer = class(TBitmapUndoer)
 private
  FImage: TImage;
  FOnChange: TNotifyEvent;
  FOnStoreQuery: TStoreQueryEvent;
  FAutoUpdate: boolean;
    procedure SetAutoUpdate(const Value: boolean);
 protected
  procedure SetImage(V: TImage); virtual;
  function GetCount: integer; virtual;
  procedure HandleChangeEvent(Sender: TObject); virtual;
  procedure Changed; virtual;
  function  StoreQuery: boolean; virtual;
 public
  Enabled: boolean;
  constructor Create(AOwner: TComponent); override;
  procedure StoreUndo(Bitmap: TBitmap); overload;
  procedure StoreUndo; overload;
  procedure Undo;
  procedure Redo;
 published
  property Image: TImage read FImage write SetImage;
  property OnStoreQuery: TStoreQueryEvent read FOnStoreQuery write FOnStoreQuery;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property Count: integer read GetCount;
  property AutoUpdate: boolean read FAutoUpdate write SetAutoUpdate;
 end;

procedure Register;

implementation


procedure Register;
begin
RegisterComponents('Fountain', [TImageUndoer]);
end;


{ TBitmapUndoer }

constructor TBitmapUndoer.Create(AOwner: TComponent);
begin
inherited;
FUndoList:=TBitmapList.Create;
FRedoList:=TBitmapList.Create;
FMaxUndo:=100;
end;



destructor TBitmapUndoer.Destroy;
begin
FUndoList.Free;
FRedoList.Free;
inherited;
end;


procedure TBitmapUndoer.Clear;
begin
FUndoList.Clear;
FRedoList.Clear;
end;


procedure TBitmapUndoer.StoreUndo(BM: TBitmap);
begin
FUndoList.AddBitmap(BM);
if FUndoList.Count>FMaxUndo then FUndoList.Delete(0);
FRedoList.Clear;
end;


procedure TBitmapUndoer.DeleteLastUndo;
begin
if FUndoList.Count>0 then FUndoList.Delete(FUndoList.Count-1);
end;

procedure TBitmapUndoer.Undo(BM: TBitmap);
begin
if FUndoList.Count=0 then exit;
FRedoList.AddBitmap(BM);
BM.Assign(FUndoList.Bitmaps[FUndoList.Count-1]);
FUndoList.Delete(FUndoList.Count-1);
end;





procedure TBitmapUndoer.Redo(BM: TBitmap);
begin
if FRedoList.Count=0 then exit;
FUndoList.AddBitmap(BM);
if FUndoList.Count>FMaxUndo then FUndoList.Delete(0);
BM.Assign(FRedoList.Bitmaps[FRedoList.Count-1]);
FRedoList.Delete(FRedoList.Count-1);
end;



{ TImageUndoer }


procedure TImageUndoer.Changed;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;



constructor TImageUndoer.Create(AOwner: TComponent);
begin
inherited;
FImage:=nil;
Enabled:=True;
end;


function TImageUndoer.GetCount: integer;
begin
Result:=FUndoList.Count;
end;

procedure TImageUndoer.HandleChangeEvent(Sender: TObject);
{Note the change event only capture drawing to canvas}
{Does not capture bitmap changing size - could maybe}
{make this work by creating a bitmap decendent and }
{overriding the SetWidth and SetHeight procedures}
begin
if not FAutoUpdate then exit;
StoreUndo;
end;



procedure TImageUndoer.SetImage(V: TImage);
begin
if FImage<>V then
	begin
	FImage:=V;
	end;
end;



procedure TImageUndoer.StoreUndo;
begin
if not Enabled then exit;
if FImage=nil then Raise Exception.Create('No Image Set');
if (FImage.Picture.Bitmap.Width=0) or (FImage.Picture.Bitmap.Height=0) then exit;
if not StoreQuery then exit;
inherited StoreUndo(FImage.Picture.Bitmap);
Changed;
end;



procedure TImageUndoer.StoreUndo(Bitmap: TBitmap);
begin
inherited StoreUndo(Bitmap);
Changed;
end;



function TImageUndoer.StoreQuery: boolean;
begin
Result:=True;
if Assigned(FOnStoreQuery) then FOnStoreQuery(Self,Result);
end;



procedure TImageUndoer.Undo;
begin
if FImage=nil then Raise Exception.Create('No Image Set');
inherited Undo(FImage.Picture.Bitmap);
Changed;
end;




procedure TImageUndoer.Redo;
begin
if FImage=nil then Raise Exception.Create('No Image Set');
inherited Redo(FImage.Picture.Bitmap);
Changed;
end;



procedure TImageUndoer.SetAutoUpdate(const Value: boolean);
begin
if FAutoUpdate<>Value then
	begin
	FAutoUpdate:= Value;
	if FAutoUpdate then
		begin
		{Assign event to bitmap canvas vs the image canvas, otherwise )
		{Image will create a bitmap of Width/Height size at startup}
		{Also there is a problem working with ImageEx where the undo images overwrite the backbitmap}
		if not (csDesigning in ComponentState) then
			Fimage.Picture.Bitmap.Canvas.OnChanging:=HandleChangeEvent;
		end
	else
		begin
		if not (csDesigning in ComponentState) then
			Fimage.Picture.Bitmap.Canvas.OnChanging:=nil;
		end;
	end;
end;



end.
