unit DottedBox;

interface

uses Windows,Classes,Controls,Types,SysUtils,Graphics,ExtCtrls,DSupport,IntVector,
     GraphicsLib, GLPolygons,Vector;


type TDottedBase = class(TComponent)
 private
  FLooper: integer;
  procedure HandleTimer(Sender: TObject);
  function AdjustPolygon(var Points: TPolygonArray): TPolygonArray;
  procedure DrawPolygon(var Poly: TPolygonArray);
 protected
  CounterStart: integer;
  Timer: TTimer;
  FImage: TImage;
  FAnimated: boolean;
  FXORMode: boolean;
  FDotWidth: integer;
  procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  procedure SetAnimated(const Value: boolean);
  procedure SetImage(const Value: TImage);
  procedure MovingDots(X, Y: Integer; P: Pointer);
  procedure SetXORMode(const Value: boolean);
  procedure SetDotWidth(const Value: integer);
  procedure RemoveCursorRestrictions;
  procedure RestrictCursorToDrawingArea(const Image: TImage);
  procedure ShowLines; virtual;
  procedure HideLines; virtual;
  procedure SetObjectPosition(X,Y: integer); virtual;
 public
  TargetCanvas:TControlCanvas;
  Scale: double;
  Visible: boolean;
  Polygon: TPolygonArray;
 published
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  property Animated: boolean read FAnimated write SetAnimated;
  property Image: TImage read FImage write SetImage;
  property XORMode: boolean read FXORMode write SetXORMode;
  property DotWidth: integer read FDotWidth write SetDotWidth default 5;
end;


type TDottedBox = class(TDottedBase)
  private
  procedure DrawRectangle;
 protected
  procedure ShowLines; override;
  procedure HideLines; override;
 public
  Box: TRect;
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure ShowRectangle;
  procedure HideRectangle;
 published
 end;


type TDottedPolygon = class(TDottedBase)
  private
 protected
  procedure ShowLines; override;
  procedure HideLines; override;
 public
  Box: TRect;
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure ShowPolygon;
  procedure HidePolygon;
 published
 end;


procedure Register;



implementation





procedure Register;
begin
RegisterComponents('Fountain', [TDottedBox,TDottedPolygon]);
end;



{ TDottedBase }



procedure TDottedBase.MovingDots(X,Y: Integer; P: Pointer);
var C: TColor;
begin
Inc(FLooper);
if (X>FImage.Width) or (Y>Image.Height) then exit;
if XORMode then TargetCanvas.Pen.Mode:=pmXOR
else TargetCanvas.Pen.Mode:=pmCopy;

{Remember, we drawing to the canvas under the image, not the image itself}
if ((FLooper div 5) and 1)=1 then C:=clWhite else C:=clBlack;
{This is twice as fast as using Pixels[X,Y] }
Windows.SetPixel(TargetCanvas.Handle, X, Y, ColorToRGB(C));
end;




function TDottedBase.AdjustPolygon(var Points: TPolygonArray): TPolygonArray;
{Offset and scale polygon - The offset is necessary
because we use The underlying parent control as }
{the canvas and the image will be offset by the scrollbars}
var Off,P: T2DVector;
var I,J: integer;
begin
SetLength(Result, Length(Points));
for I:=0 to High(Points) do SetLength(Result[I],Length(Points[I]));

Off:=ScalarDivide2D(MakeVector2D(Image.Left,Image.Top),Scale);
for I:=0 to High(Points) do
 for J:=0 to High(Points[I]) do
	begin
	P:=PointToVector2D(Points[I,J]);
	P:=VectorAdd2D(P,Off);
	P:=ScalarProduct2D(P,Scale);
	Result[I,J]:=Vector2DToPoint(p);
	end;
end;




procedure TDottedBase.DrawPolygon(var Poly: TPolygonArray);
var P1,P2: TPoint;
var I,J: integer;
var P: TPolygonArray;
begin
if FImage=nil then raise Exception.Create('No Image Selected');
P:=AdjustPolygon(Poly);
FLooper:=CounterStart;
for I:=0 to High(Poly) do
 for J:=0 to High(Poly[I]) do
	begin
	P1:=P[I,J];
	if J=High(P[i]) then P2:=P[I,0]
	else P2:=P[I,J+1];
	DrawLine(P1.X,P1.Y ,P2.X,P2.Y, MovingDots, nil);
	end;
end;



procedure TDottedBase.SetAnimated(const Value: boolean);
begin
if FAnimated<>Value then
	begin
	FAnimated := Value;
	Timer.Enabled:=Value;
	end;
end;



procedure TDottedBase.Notification(AComponent: TComponent;
  Operation: TOperation);
{Handle the live deletion of an image controlled by the object at designtime }
begin
inherited Notification(AComponent, Operation);
if (Operation = opRemove) and (AComponent = FImage) then FImage := nil;
end;



constructor TDottedBase.Create(AOwner: TComponent);
begin
inherited;
TargetCanvas:=TControlCanvas.Create;
Timer:=TTimer.Create(nil);
Timer.Interval:=100;
Timer.Enabled:=False;
Timer.OnTimer:=HandleTimer;
Scale:=1;
SetLength(Polygon,0);
FLooper:=CounterStart;
DotWidth:=5;
XORMode:=True;
Visible:=False;
end;



destructor TDottedBase.Destroy;
begin
Timer.Free;
inherited;
end;



procedure TDottedBase.SetImage(const Value: TImage);
begin
if FImage<>Value then
	begin
	FImage:=Value;
	if csDesigning in ComponentState then exit;
	{We use the parent control, not the image, because the XOR works better}
	TargetCanvas.Control:=FImage.Parent;
	end;
end;




procedure TDottedBase.SetXORMode(const Value: boolean);
begin
FXORMode := Value;
end;




procedure TDottedBase.SetDotWidth(const Value: integer);
begin
if FDotWidth<>Value then
	begin
	if Value<1 then FDotWidth:=1
	else FDotWidth := Value;
	end;
end;



procedure TDottedBase.RestrictCursorToDrawingArea(const Image: TImage);
var CursorClipArea: TRect;
begin
CursorClipArea := Bounds(Image.ClientOrigin.X, Image.ClientOrigin.Y, Image.Width, Image.Height);
Windows.ClipCursor(@CursorClipArea)
end;



procedure TDottedBase.RemoveCursorRestrictions;
begin
Windows.ClipCursor(nil)
end;


procedure TDottedBase.HideLines;
begin
//To be filled in by descendent classes
end;

procedure TDottedBase.ShowLines;
begin
//To be filled in by descendent classes
end;


procedure TDottedBase.SetObjectPosition(X,Y: integer);
begin
//To be filled in by descendent classes
end;



procedure TDottedBase.HandleTimer(Sender: TObject);
begin
if csDesigning in ComponentState then exit;
HideLines;
Inc(CounterStart);
ShowLines;
end;




{ TDottedBox }





procedure TDottedBox.DrawRectangle;
var R: TRect;
var Poly: TPolygonArray;
begin
SetLength(Poly,1);
SetLength(Poly[0],4);
Poly[0,0]:=Box.TopLeft;
Poly[0,1]:=Point(Box.Right,Box.Top);
Poly[0,2]:=Box.BottomRight;
Poly[0,3]:=Point(Box.Left,Box.Bottom);
DrawPolygon(Poly);
end;



procedure TDottedBox.ShowRectangle;
begin
if Visible then exit;
Timer.Enabled:=FAnimated;
if not Visible then DrawRectangle;
Visible:=True;
end;



procedure TDottedBox.HideRectangle;
begin
Timer.Enabled:=False;
{Tis works because the selection b0x is drawn to the canvas of the }
{control under the image, so refreshing reloads the bitmap in the image}
if Visible then
 if XORMode then DrawRectangle
 else Image.Refresh;
Visible:=False;
end;




constructor TDottedBox.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
Box:=Rect(0,0,0,0);
end;



destructor TDottedBox.Destroy;
begin
inherited;
end;



procedure TDottedBox.HideLines;
begin
inherited;
HideRectangle;
end;

procedure TDottedBox.ShowLines;
begin
inherited;
ShowRectangle;
end;


{ TDottedPolygon }


constructor TDottedPolygon.Create(AOwner: TComponent);
begin
inherited;
end;



destructor TDottedPolygon.Destroy;
begin
inherited;
end;



procedure TDottedPolygon.HidePolygon;
begin
Timer.Enabled:=False;
if Visible then ShowPolygon;
Visible:=False;
end;




procedure TDottedPolygon.ShowPolygon;
begin
if Visible then exit;
Timer.Enabled:=FAnimated;
DrawPolygon(Polygon);
Visible:=True;
end;



procedure TDottedPolygon.HideLines;
begin
inherited;
HidePolygon;
end;



procedure TDottedPolygon.ShowLines;
begin
inherited;
ShowPolygon;
end;







end.
