unit GLViewer;

interface

uses Windows,SysUtils,Forms,Graphics,Messages,Controls,Classes,ClipBrd,OpenGL15,
     GLLibrary,GLRenderer,GLMesh,GLLines,GL3DFigures,GLFonts, GLXMLGlobal,GLLights,
     GLTexture,GLXFileGlobal,GLFrame,GLOctTree,GLFog,SyncObjs,paintpan,DSupport,
     HugeArry,Vector,BitmapLib,StdCtrls;


{ Define clip plane distances. Note, all these value are
relative to the eye or camera. When the camera is positioned,
the actual, real world coordinates of the frustum will be
defined. }


const PixelsPerUnit= 100;	{Scaling factor for screen}
const WorkDistance = 10;	{Distance at which above scaling applies}

{Values based on ideas here: https://www.sjbaker.org/steve/omniv/love_your_z_buffer.html}

const FrustNear = 1;		{Position of the near frustrum}
const FrustFar = 1000000;	{Position of the far frustrum}



var Cnt: integer;

type TGLWindow = class(TObject)
  private
    WinControl: TWinControl;
    FRealWidth,FRealHeight: double;
    FCurrentWinHandle: HWND;
    FAntiAlias: boolean;
    function GetAntialiasPixelFormat(NumberSamples: integer;
      var AntialiasFormat: Integer): boolean;
  protected
    procedure BuildFormatDescriptor(var PFD: TPixelFormatDescriptor; DesignTime: boolean);
    procedure SetupPixelFormat(DC: HDC; DesignTime: boolean);
  public
    FrustWidth,FrustHeight,PortWidth,PortHeight: double;
    GLWindowDC: HDC;
    OpenGLRC: HGLRC;
    constructor Create(Control: TWinControl; Width,Height: integer; DesignTime: boolean);
    destructor Destroy; override;
    procedure SetupGLWindow(Handle: HWND; DesignTime: boolean);
    procedure SetViewPort(Width,Height: integer);
    function PixelToWorld(X, Y: integer): T3DVector;
    property RealWidth: double read FRealWidth;
    property RealHeight: double read FRealHeight;
    property AntiAlias: boolean read FAntiAlias write FAntiAlias;
  end;



{Camera object - note, the clipping planes and frustum are relative to camera position}

const DefaultCameraPos: T3DVector = (X:0; Y: 0; Z:WorkDistance);
const DefaultCameraTarget: T3DVector = (X:0; Y:0; Z:0);
const DefaultCameraOrientation: T3DVector = (X:0; Y:1; Z:0);



type TGLCamera = class(TObject)
 private
 protected
 public
 CameraPos,CameraTarget,CameraOrientation: T3DVector;
 procedure Reset;
 procedure SetPosition;
 end;


{GL Viewer object}

{Array and variables for hit testing}

type TGLHitRecord = packed record
 StackCount: GLuint;
 DepthMin,DepthMax: GLuint;
 ObjectIndex: GLuint;
 end;

const HitArraySize = 512;
const HitIntCount = HitArraySize * Sizeof(TGLHitRecord) div SizeOf(GLuint);

type TGLHitArray = array [0..HitArraySize-1] of TGLHitRecord;

var GLHitArray: TGLHitArray;
var GLHitCount: integer;

type TSceneType = (stSoccerBall,stTorus);


type TGLViewer = class(TPaintPanel)
 private
  FRenderSemaphore: TCriticalSection;
  FDefaultScene: boolean;
  FFrameStart,FFrameEnd: cardinal;
  FOnChange: TNotifyEvent;
  FScreenColor: TColor;
  FOnDisplayChanging: TNotifyEvent;
  FOnRender: TNotifyEvent;
  FOnCreateWindow: TNotifyEvent;
  FOnDisplayChange: TNotifyEvent;
  FOnObjectChange: TNotifyEvent;
  FEnabled: boolean;
  FBackCulling: boolean;
  FWireLighting: TGLArrayf4;
  FWireFrameLight: boolean;
  FSceneType: TSceneType;
  FScaleFactor: double;
  FAntialias: boolean;
  procedure ObjectChange;
  procedure DisplayChange;
  procedure DisplayChanging;
  procedure Change;
  procedure SetScreenColor(const Value: TColor);
  function GetExtentMax: T3DVector;
  function GetExtentMin: T3DVector;
  procedure HandleFrameChange(Sender: TObject);
  procedure SetBackCulling(const Value: boolean);
  procedure SetupAmbientLight;
  procedure SetWireFrameLight(const Value: boolean);
  procedure SetSceneType(const Value: TSceneType);
  procedure Rendering;
  procedure SetScaleFactor(const Value: double);
  procedure SetAntialias(const Value: boolean);
 protected
  procedure Loaded; override;
  procedure Paint; override;
  procedure CreateWnd; override;
  procedure DestroyWnd; override;
  procedure Resize; override;
  procedure SetDefaultScene(V: boolean);
  function GetGLObject(I: integer): TGLObject;
  function GetObjectCount: integer;
 public
  GLWindow: TGLWindow;
  BaseFrame: TGLBaseFrame;
  HitArray: THugePoint;
  FrameTime,FramesPerSec: double;
  Lights: TGLLightArray;
  GlobalAmbientLight: TGLArrayf4;
  BackTexture: TGLBackTexture;
  Camera: TGLCamera;
  Fog: TGLFog;
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  procedure AddAllObjectPos(X, Y, Z: double);
  procedure BuildDefaultScene;
  procedure RenderAll(DoEvents: boolean);
  procedure ResetLightParams;
  procedure CenterScene;
  procedure FitSceneByScaling;
  procedure FitSceneByDistance;
  procedure FitSceneByCamera;
  procedure FitObjectByDistance(Obj: TGLObject); overload;
  procedure FitObjectByScaling(Obj: TGLObject); overload;
  procedure FitObjectByDistance(Inx: integer); overload;
  procedure FitObjectByScaling(Inx: integer); overload;
  function FindNearestObject(X,Y: integer; var O: TGLObject): boolean; overload;
  function FindNearestObject(X,Y: integer): integer; overload;
  function FindNearestObject(X,Y: integer; var ObjectType: TRenderType; var Index: integer; var GLObject: TGLObject): boolean; overload;
  function FindNearestObject(X,Y: integer; var Index: integer): boolean; overload;
  procedure GetScreenImage(BM: TBitmap; Scale: double =1);
  procedure SaveToClipboard(Scale: double =1);
  procedure SaveToFile(Filename: string);
  procedure LoadXFile(Filename: string; ProgressProc: TProgressProcedure; NewMeshes: boolean);
  procedure LoadXStream(Stream: TStream; ProgressProc: TProgressProcedure; NewMeshes: boolean);
  procedure SaveXFile(Filename: string; ProgressProc: TProgressProcedure);
  procedure SaveXStream(Stream: TStream; ProgressProc: TProgressProcedure);

  procedure LoadXMLFile(Filename: string; ProgressProc: TXMLProgressProc; NewMeshes: boolean);
  procedure SaveXMLFile(Filename: string; ProgressProc: TXMLProgressProc);

  property GLObjects[Index: Integer]: TGLObject read GetGLObject;
  property ObjectCount: integer read GetObjectCount;
  property BackCulling: boolean read FBackCulling write SetBackCulling default true;
  procedure DeleteObject(I: integer);
  procedure Reset;
  property Antialias: boolean read FAntialias write SetAntialias default False;
 published
  property ScaleFactor: double read FScaleFactor write SetScaleFactor;
  property Enabled: boolean read FEnabled write FEnabled;
  property ShowDefaultScene: boolean read FDefaultScene write SetDefaultScene;
  property SceneType: TSceneType read FSceneType write SetSceneType default stSoccerBall;
  property OnObjectChange: TNotifyEvent read FOnObjectChange write FOnObjectChange;
  property OnDisplayChanging: TNotifyEvent read FOnDisplayChanging write FOnDisplayChanging;
  property OnDisplayChange: TNotifyEvent read FOnDisplayChange write FOnDisplayChange;
  property OnRender: TNotifyEvent read FOnRender write FOnRender;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property OnCreateWindow: TNotifyEvent read FOnCreateWindow write FOnCreateWindow;
  property ScreenColor: TColor read FScreenColor write SetScreenColor;
  property ExtentMin: T3DVector read GetExtentMin;
  property ExtentMax: T3DVector read GetExtentMax;
  property WireFrameLight: boolean read FWireFrameLight write SetWireFrameLight;
 end;


procedure Register;


implementation

uses GLXFileReader,GLXFileWriter, GLXMLReader, GLXMLWriter;

procedure Register;
begin
RegisterComponents('Fountain', [TGLViewer]);
end;



{---------------- Routines to setup the OpenGL window ------------------------}

const PFD_SUPPORT_COMPOSITION = $00008000;

{===============================================================================}



function TGLWindow.GetAntialiasPixelFormat(NumberSamples: integer; var AntialiasFormat: Integer): boolean;
{Test for multibuffer formats and returns one if available}
{NumberSamples controls the number Antialiasing samples requested}
var fAttributes: array [0..1] of Single;
var iAttributes: array [0..21] of Integer;
var pfd        : PIXELFORMATDESCRIPTOR;
var hwnd       : Cardinal;
var wnd        : TWndClassEx;
var numFormats : Cardinal;
var Format     : Integer;
var DC: Hdc;
begin
Result:=False;
ZeroMemory(@wnd, SizeOf(wnd));
wnd.cbSize        := SizeOf(wnd);
wnd.lpfnWndProc   := @DefWindowProc;
wnd.hCursor       := LoadCursor(0, IDC_ARROW);
wnd.lpszClassName := 'GetPixelFormat';
RegisterClassEx(wnd);
hwnd := CreateWindow('GetPixelFormat', nil, WS_POPUP, 0, 0, 0, 0, 0, 0, HInstance,nil);
DC := GetDC(hwnd);
FillChar(pfd, SizeOf(pfd), 0);
pfd.nSize        := SizeOf(TPIXELFORMATDESCRIPTOR);
pfd.nVersion     := 1;
pfd.dwFlags      := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER;
pfd.iPixelType   := PFD_TYPE_RGBA;
pfd.cColorBits   := 32;
pfd.cDepthBits   := 24;
pfd.cStencilBits := 8;
pfd.iLayerType   := PFD_MAIN_PLANE;
SetPixelFormat(DC, ChoosePixelFormat(DC, @pfd), @pfd);
wglMakeCurrent(DC, wglCreateContext(DC));

fAttributes[0] := 0;
fAttributes[1] := 0;
iAttributes[0] := WGL_DRAW_TO_WINDOW_ARB;
iAttributes[1] := 1;
iAttributes[2] := WGL_SUPPORT_OPENGL_ARB;
iAttributes[3] := 1;
iAttributes[4] := WGL_ACCELERATION_ARB;
iAttributes[5] := WGL_FULL_ACCELERATION_ARB;
iAttributes[6] := WGL_COLOR_BITS_ARB;
iAttributes[7] := 24;
iAttributes[8] := WGL_ALPHA_BITS_ARB;
iAttributes[9] := 8;
iAttributes[10] := WGL_DEPTH_BITS_ARB;
iAttributes[11] := 16;
iAttributes[12] := WGL_STENCIL_BITS_ARB;
iAttributes[13] := 0;
iAttributes[14] := WGL_DOUBLE_BUFFER_ARB;
iAttributes[15] := 1;
iAttributes[16] := WGL_SAMPLE_BUFFERS_ARB;
iAttributes[17] := 1;
iAttributes[18] := WGL_SAMPLES_ARB;
iAttributes[19] := NumberSamples;
iAttributes[20] := 0;
iAttributes[21] := 0;
if not wglChoosePixelFormatARB(GetDC(hWnd), @iattributes, @fattributes, 1, @Format, @numFormats) then exit;
if numFormats <1 then exit;
AntialiasFormat := Format;
ReleaseDC(hwnd, DC);
DestroyWindow(hwnd);
wglMakeCurrent(0, 0);
wglDeleteContext(DC);
Result:=True;
end;



procedure TGLWindow.BuildFormatDescriptor(var PFD: TPixelFormatDescriptor; DesignTime: boolean);
{Set the pixel format descriptor for our window }
var Flag: cardinal;
begin
FillMemory(@PFD,Sizeof(TPixelFormatDescriptor),0);
PFD.nSize:=sizeof(TPixelFormatDescriptor);
PFD.nVersion:=1;
{With areo enabled, selecting the viewer at design time causes the
{ viewer not get repainted and remain grayed out. Using the PFD_SUPPORT_GDI}
{ flag fixes the problem but would be too slow for runtime operation}
if DesignTime then Flag:=PFD_SUPPORT_GDI else Flag:=PFD_DOUBLEBUFFER;
PFD.dwFlags:=PFD_SUPPORT_OPENGL or PFD_DRAW_TO_WINDOW or Flag;
PFD.iPixelType:=PFD_TYPE_RGBA;
PFD.cColorBits:=24;
PFD.cDepthBits:=16;
PFD.iLayerType:=PFD_MAIN_PLANE;
end;



procedure TGLWindow.SetupPixelFormat(DC: HDC; DesignTime: boolean);
{Setup multisampled pixel format for antialiasing}
var PixelFormat: integer;
var PFD: TPixelFormatDescriptor;
begin
BuildFormatDescriptor(PFD,DesignTime);

{for some reason? Antialiasing doesn't work at design time}
if DesignTime then PixelFormat:=ChoosePixelFormat(DC, @PFD)
{Not design-times then try four points per pixel}
else if not GetAntialiasPixelFormat(4,PixelFormat) then
 {If that fails, try two points per pixel}
if not GetAntialiasPixelFormat(2,PixelFormat) then
  {If that fails revert to non-antialias}
  PixelFormat:=ChoosePixelFormat(DC, @PFD);

if PixelFormat = 0 then
	begin
	MessageBox(WindowFromDC(DC), 'ChoosePixelFormat failed.', 'Error', MB_ICONERROR or MB_OK);
	exit;
	end;

SetPixelFormat(DC, PixelFormat, @pfd);
if not SetPixelFormat(DC, pixelFormat, @pfd) then
	begin
	MessageBox(WindowFromDC(DC), 'SetPixelFormat failed.', 'Error', MB_ICONERROR or MB_OK);
	exit;
	end;
end;



procedure TGLWindow.SetupGLWindow(Handle: HWND; DesignTime: boolean);
{Create the GL handle for our window}
begin
FCurrentWinHandle:=Handle;
GLWindowDC:=GetDC(FCurrentWinHandle);
SetupPixelFormat({True,}GLWindowDC,DesignTime);
OpenGLRC:=wglCreateContext(GLWindowDC);
wglMakeCurrent(GLWindowDC, OpenGLRC);
end;



{===============================================================================}



procedure TGLWindow.SetViewPort(Width,Height: integer);
{Set the view port parameter for our window}
begin
PortWidth:=Width;
PortHeight:=Height;

FrustWidth:= Width / PixelsPerUnit / WorkDistance;
FrustHeight:= Height / PixelsPerUnit / WorkDistance;
FRealWidth:=FrustWidth * WorkDistance;
FRealHeight:=FrustHeight * WorkDistance;
glViewport(0, 0, Width, Height);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glFrustum(-(FrustWidth/2), FrustWidth/2, -(FrustHeight/2), FrustHeight/2, FrustNear,FrustFar);
glMatrixMode(GL_MODELVIEW);
end;



function TGLWindow.PixelToWorld(X,Y: integer): T3DVector;
begin
Result:=MakeVector3D(((X/PortWidth) * FRealWidth)-(FRealWidth/2), (FRealHeight/2)-((Y/PortHeight) * FRealHeight), 0);
end;



constructor TGLWindow.Create(Control: TWinControl; Width,Height: integer; DesignTime: boolean);
begin
inherited Create;
WinControl:=Control;
SetupGLWindow(Control.Handle, DesignTime);
SetViewPort(Width,Height);
end;


destructor TGLWindow.Destroy;
begin
wglMakeCurrent(0, 0);
wglDeleteContext(OpenGLRC);
ReleaseDC(FCurrentWinHandle, GLWindowDC);
inherited Destroy;
end;


{------------------ GLCamera --------------------------------------------------}

procedure TGLCamera.Reset;
begin
CameraPos:=DefaultCameraPos;
CameraTarget:=DefaultCameraTarget;
CameraOrientation:=DefaultCameraOrientation;
end;


procedure TGLCamera.SetPosition;
begin
gluLookAt(CameraPos.X,CameraPos.Y,CameraPos.Z,
	  CameraTarget.X,CameraTarget.Y,CameraTarget.Z,
	  CameraOrientation.X,CameraOrientation.Y,CameraOrientation.Z);
end;

{GLViewer}

procedure TGLViewer.ResetLightParams;
begin
Lights.Reset;
GlobalAmbientLight:= MakeGLArrayf4(0.1,0.1,0.1,0.1);
end;

procedure TGLViewer.SetupAmbientLight;
begin
if FWireFrameLight then glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @FWireLighting)
else glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @GlobalAmbientLight);
end;



procedure TGLViewer.BuildDefaultScene;
{Build the default scene}
begin
{avoid problems when reading defaults for the first time}
if BaseFrame=nil then exit;
if BaseFrame.MeshList.Count=0 then exit;
GLWindow.SetViewPort(Width,Height);
glMatrixMode(GL_MODELVIEW);
glEnable(GL_DEPTH_TEST);
if FDefaultScene then
	begin
	BaseFrame.MeshList[0].Clear;
	BaseFrame.StringList3D[0].Clear;
	case FSceneType of
	 stSoccerBall: 	DrawTruncateIcosahedron(BaseFrame.MeshList[0],MakeVector3D(1,1,1),clBlack,clWhite);
	 stTorus: BuildTorus(BaseFrame.MeshList[0],11,11,nil);
	 end;
	end
else
	begin
	BaseFrame.MeshList[0].Clear;
	end;
BaseFrame.MeshList[0].ResetTransforms;
end;



procedure TGLViewer.RenderAll(DoEvents: boolean);
{Display current mesh object}
const Res=1000000;
var R,G,B: single;
begin
if Application.Terminated then exit;
FRenderSemaphore.Acquire;
try
FFrameStart:=GetHiResTick;

{Setup screen}
wglMakeCurrent(GLWindow.GLWindowDC, GLWindow.OpenGLRC);
glMatrixMode(GL_MODELVIEW);

{Clear screen}
GetRGB(FScreenColor,R,G,B);
glClearColor(R,G,B,1);


{Set rendering modes}
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
glEnable(GL_NORMALIZE);
glCullFace(GL_BACK);
if FBackCulling then glEnable(GL_CULL_FACE)
else glDisable(GL_CULL_FACE);
glEnable(GL_COLOR_MATERIAL);

glEnable(GL_LIGHTING);
SetupAmbientLight;
Lights.RenderLights;

{Render and display}
DisplayChanging;
if BackTexture.Enabled then BackTexture.Render;

{Setup hit testing parameters}
HitArray.Clear;
glInitNames;		{ Initializes the name stack }
glPushName(0);		{ Push 0 (at least one entry) onto the stack }

glLoadIdentity;
Camera.SetPosition;

BaseFrame.Render(HitArray);
Rendering;
SwapBuffers(wglGetCurrentDC);
FFrameEnd:=GetHiResTick;
FrameTime:=(FFrameEnd-FFrameStart)/Res;
FramesPerSec:=1/FrameTime;
{There is a problem if these last two event handlers access the matrix}
{The mouse handler's scale changes - has to do with 'LastMatrix'}
if DoEvents then
	begin
	Change;
	DisplayChange;
	end;
finally FRenderSemaphore.Release; end;
end;




function TGLViewer.FindNearestObject(X,Y: integer): integer;
var ViewPort: TVector4i;
var Choose,I: integer;
var Depth: GLuint;
var A,UnitHeight: GLDouble;
begin
{ The size of the viewport: [0] = <x>, [1] = <y>, [2] = <length>, [3] = <width> }
{ This sets the array <viewport> to the size and location of the screen relative to the window }

glGetIntegerv(GL_VIEWPORT, @viewport);
glSelectBuffer(HitIntCount, @GLHitArray);		{ Tell Opengl to use our array for selection }

{ Puts opengl in selection mode. Nothing will be drawn.  Object id's and extents are stored in the buffer. }
glRenderMode(GL_SELECT);

glMatrixMode(GL_PROJECTION);	{ Selects the projection matrix }
glPushMatrix;			{ Push the projection matrix }
glLoadIdentity;			{ Resets the matrix }

{ This creates a matrix that will zoom up to a small portion of the screen, where the mouse is. }
gluPickMatrix(X, viewport[3]-Y, 1, 1, viewport);

{ Apply the perspective matrix }
UnitHeight:=Height / PixelsPerUnit/WorkDistance;
A:=ArcTan(UnitHeight/FrustNear)*Radian2Degree;
gluPerspective(A, (viewport[2]-viewport[0])/(viewport[3]-viewport[1]), FrustNear,FrustFar);

glMatrixMode(GL_MODELVIEW);	{ Select the modelview matrix }
glLoadIdentity;			{ Reset the modelview matrix }

glPushMatrix;
RenderAll(False);
glPopMatrix; 

glMatrixMode(GL_PROJECTION);	{ Select the projection matrix }
glPopMatrix;			{ Pop the projection matrix }
glMatrixMode(GL_MODELVIEW);	{ Select the modelview matrix }
GLHitCount:=glRenderMode(GL_RENDER);	{ Switch to render mode, find out how many }

{ Objects were drawn where the mouse was }

if GLHitCount > 0 then
	begin			{ If there were more than 0 hits }
	choose:=GLHitArray[0].ObjectIndex;	{ Make our selection the first object }
	depth:=GLHitArray[0].DepthMin;	{ Store how far away it is }
	for I:=1 to GLHitCount-1 do		{ Loop through all the detected hits }
		begin
		{ If this object is closer to us than the one we have selected }
		if GLHitArray[I].DepthMin < depth then
			begin
			choose:=GLHitArray[I].ObjectIndex;	{ Select the closer object }
			depth:=GLHitArray[I].DepthMin;		{ Store how far away it is }
			end;
		end; 
	Result:=choose;
	end
else  Result:=-1;
end;



function TGLViewer.FindNearestObject(X,Y: integer; var ObjectType: TRenderType; var Index: integer; var GLObject: TGLObject): boolean;
var Inx: integer;
begin
GLObject:=nil;
Inx:=FindNearestObject(X,Y);
Result:=Inx>=0;
if Result then
	begin
	ObjectType:=TRenderType(HitArray[Inx].X);
	GLObject:=TGLObject(HitArray[Inx].Y);
	Index:=BaseFrame.MasterList.IndexOf(GLObject);
	end;
end;




function TGLViewer.FindNearestObject(X,Y: integer; var Index: integer): boolean;
{Find nearest object using returning object list index}
var RT: TRenderType;
var O: TGLObject;
begin
Result:=FindNearestObject(X,Y,RT,Index,O);
end;




function TGLViewer.FindNearestObject(X,Y: integer; var O: TGLObject): boolean;
{Find nearest object using returning object list index}
var RT: TRenderType;
var Index: integer;
begin
Result:=FindNearestObject(X,Y,RT,Index,O);
end;


{---------------------------------------------------------------------------=-}


procedure TGLViewer.Resize;
begin
inherited Resize;
if GLWindow<>nil then GLWindow.SetViewPort(Width,Height);
Invalidate;
end;




procedure TGLViewer.CreateWnd;
var DesignTime: boolean;
begin
inherited CreateWnd;
DesignTime:=csDesigning in ComponentState;
{Create everything here because the control }
{needs a handle before the GLWindow can be created}
GLWindow:=TGLWindow.Create(Self,Self.Width,Self.Height,DesignTime);
Fog:=TGLFog.Create;
{You can only add 3DStrings after window is created}
BaseFrame.StringList3D.Add;
BaseFrame.MeshList.Add;
BaseFrame.FastMeshList.Add;
BaseFrame.LineList.Add;

BaseFrame.StringList3D[0].Name:='BaseString';
BaseFrame.MeshList[0].Name:='BaseMesh';
BaseFrame.FastMeshList[0].Name:='BaseFastMesh';
BaseFrame.LineList[0].Name:='BaseLines';

BaseFrame.OnChange:=HandleFrameChange;

ResetLightParams;
BuildDefaultScene;
glDisable(GL_MULTISAMPLE);
FAntiAlias:=False;
{Don't let the parent turn double buffering on}
{$IF CompilerVersion >= 17}
ParentDoubleBuffered:=False;
{$ifend}
{The the control must never be double buffered}
DoubleBuffered:=False;
if Assigned(FOnCreateWindow) then FOnCreateWindow(Self);
end;




procedure TGLViewer.DestroyWnd;
{ Destroy the window here so there is a handle to destroy}
begin
inherited DestroyWnd;
end;



procedure TGLViewer.Loaded;
var H: HWND;
begin
inherited;
{Force the handle into existance so it will work in a}
H:=Handle;
end;


constructor TGLViewer.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
ControlStyle:=ControlStyle + [csOpaque];
FRenderSemaphore:=TCriticalSection.Create;

Lights:=TGLLightArray.Create;
Lights[0].Enabled:=True;
Camera:=TGLCamera.Create;
BaseFrame:=TGLBaseFrame.Create;
BaseFrame.Name:='BaseFrame';
BackTexture:=TGLBackTexture.Create;
HitArray:=THugePoint.Create(5);
FScreenColor:=clBlack;
Camera.Reset;
FEnabled:=True;
FBackCulling:=True;
FWireLighting:=MakeGLArrayf4(1,1,1,1);
FSceneType:=stSoccerBall;
FScaleFactor:=1;
end;



destructor TGLViewer.Destroy;
begin
{So no events are triggered on shut down}
BaseFrame.OnChange:=nil;
GLWindow.Free;
Camera.Free;
BaseFrame.Free;
BackTexture.Free;
HitArray.Free;
Lights.Free;
FRenderSemaphore.Free;
Fog.Free;
inherited;
end;



procedure TGLViewer.Paint;
begin
{Don't run the inherited method here or it will flicker}
//inherited;
if not FEnabled then exit;
if Assigned(OnPaint) then OnPaint(Self);
if GLWindow.GLWindowDC<>0 then
	begin
	glRenderMode(GL_RENDER);
       	RenderAll(True);
	end;
end;




procedure TGLViewer.SetDefaultScene(V: boolean);
begin
if V=FDefaultScene then exit;
FDefaultScene:=V;
BuildDefaultScene;
Invalidate;
end;


procedure TGLViewer.SetSceneType(const Value: TSceneType);
begin
if FSceneType<>Value then
	begin
	FSceneType := Value;
	BuildDefaultScene;
	Invalidate;
	end;
end;

procedure TGLViewer.Change;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;



procedure TGLViewer.ObjectChange;
begin
if Assigned(FOnObjectChange) then FOnObjectChange(Self);
end;


procedure TGLViewer.DisplayChanging;
begin
if Assigned(FOnDisplayChanging) then FOnDisplayChanging(Self);
end;



procedure TGLViewer.Rendering;
begin
if Assigned(FOnRender) then FOnRender(Self);
end;



procedure TGLViewer.DisplayChange;
begin
if Assigned(FOnDisplayChange) then FOnDisplayChange(Self);
end;



procedure TGLViewer.HandleFrameChange(Sender: TObject);
begin
ObjectChange;
Change;
end;



function TGLViewer.GetGLObject(I: integer): TGLObject;
begin
Result:=TGLObject(BaseFrame.MasterList[I]);
end;



procedure TGLViewer.DeleteObject(I: integer);
var O: TGLObject;
var Inx: integer;
begin
O:=GLObjects[I];
if (O=nil) or (O is TGLBaseFrame) then exit;
Inx:=O.OwnerList.IndexOf(O);
if Inx<0 then exit;
O.OwnerList.Delete(Inx);
end;



function TGLViewer.GetObjectCount: integer;
begin
Result:=BaseFrame.MasterList.Count;
end;



procedure TGLViewer.SetScreenColor(const Value: TColor);
begin
if FScreenColor<>Value then
	begin
	FScreenColor:=Value;
	{These are necessary to prevent erasing background}
	{to wrong color in certain situations}
	Color:=Value;
	Canvas.Brush.Color:=Value;
	Invalidate;
	end;
end;



procedure TGLViewer.SetBackCulling(const Value: boolean);
begin
if FBackCulling<>Value then
	begin
	FBackCulling:=Value;
	Invalidate;
	end;
end;




procedure TGLViewer.SetWireFrameLight(const Value: boolean);
begin
if FWireFrameLight<>Value then
	begin
	FWireFrameLight:=Value;
	Invalidate;
	end;
end;



function TGLViewer.GetExtentMax: T3DVector;
var I: integer;
var First: boolean;
begin
Result:=MakeVector3D(0,0,0);
First:=False;
for I:=0 to ObjectCount-1 do
 if Self.GLObjects[I].Enabled then
	begin
	if First then Result:=GLObjects[I].ExtentMax
	else Result:=GetMaximumVector(GLObjects[I].ExtentMax,Result);
	end;
end;



function TGLViewer.GetExtentMin: T3DVector;
var I: integer;
var First: boolean;
begin
Result:=MakeVector3D(0,0,0);
First:=False;
for I:=0 to ObjectCount-1 do
 if Self.GLObjects[I].Enabled then
	begin
	if First then Result:=GLObjects[I].ExtentMin
	else Result:=GetMinimumVector(GLObjects[I].ExtentMin,Result);
	end;
end;


procedure TGLViewer.AddAllObjectPos(X,Y,Z: double);
var I: integer;
begin
for I:=0 to ObjectCount-1 do
	begin
	GLObjects[I].AddPosition(X,Y,Z);
	end;
end;



{------------------------------------------------------------------------------}


procedure TGLViewer.CenterScene;
var C1: T3DVector;
begin
C1:=ScalarDivide3D(VectorAdd3D(ExtentMax,ExtentMin),2);
BaseFrame.RotationCenter:=C1;
BaseFrame.Position:=VectorMultiply3D(C1,MakeVector3D(-1,-1,-1));
end;



procedure TGLViewer.FitObjectByDistance(Obj: TGLObject);
{Fit object to screen using distance from View Port}
var Size,Scale: double;
begin
Size:=VectorABS3D(VectorSubtract3D(Obj.ExtentMax,Obj.ExtentMin));
Scale:=1/(GLWindow.RealWidth/Size);
Obj.Position.Z:=-(WorkDistance * Scale);
Obj.ResetPos:=Obj.Position;
end;



procedure TGLViewer.FitObjectByScaling(Obj: TGLObject);
{Fit Object to screen using Scaling}
var Size,Scale: double;
begin
Size:=VectorABS3D(VectorSubtract3D(Obj.ExtentMax,Obj.ExtentMin));
Scale:=1.414*(GLWindow.RealWidth/Size);
Obj.Scale:=Scale;
end;



procedure TGLViewer.FitObjectByDistance(Inx: integer);
{Fit object to screen using distance from View Port}
begin
FitObjectByDistance(GLObjects[Inx]);
end;



procedure TGLViewer.FitObjectByScaling(Inx: integer);
{Fit Object to screen using Scaling}
begin
FitObjectByScaling(GLObjects[Inx]);
end;



procedure TGLViewer.FitSceneByCamera;
{Fit scene to screen using distance of camera}
var Size,Scale: double;
begin
Size:=VectorABS3D(VectorSubtract3D(ExtentMax,ExtentMin));
Scale:=1/(GLWindow.RealWidth/Size);
Camera.Reset;
Camera.CameraPos.Z:=(WorkDistance * Scale);
end;





procedure TGLViewer.FitSceneByDistance;
{Fit scene to screen using distance from View Port}
var Size,Scale: double;
var I: integer;
begin
Size:=VectorABS3D(VectorSubtract3D(ExtentMax,ExtentMin));
Scale:=1/(GLWindow.RealWidth/Size);
for I:=0 to ObjectCount-1 do
 if GLObjects[I]<>Self.BaseFrame then
	begin
	GLObjects[I].ResetTransforms;
	GLObjects[I].Position.Z:=-((WorkDistance * Scale)-Camera.CameraPos.Z);
	GLObjects[I].ResetPos:=GLObjects[I].Position;
	end;
end;



procedure TGLViewer.FitSceneByScaling;
{Fit scene to screen using Scaling}
var Size,Scale: double;
var I: integer;
const Sqrt2 = 1.41421356237;
begin
Size:=VectorABS3D(VectorSubtract3D(ExtentMax,ExtentMin));
Scale:=Sqrt2 * (GLWindow.RealWidth/Size);
for I:=0 to ObjectCount-1 do
 if GLObjects[I].Enabled then
	begin
	GLObjects[I].ResetTransforms;
	GLObjects[I].Scale:=Scale;
	end;
end;


{------------------------------------------------------------------------------}



procedure TGLViewer.GetScreenImage(BM: TBitmap; Scale: double =1);
{Return copy of the screen as Bitmap}
var Buffer,Pixel: PRGBQuad;
var Scan: PRGBQuadArray;
var X,Y: integer;
var ImageWidth,ImageHeight: integer;
var framebuffer,colorbuffer,depthbuffer: GLuint;

	procedure CreateFBO(Width,Height: integer);
	{create a Frame Buffer Object so drawing isn't constrained by the screen}
	var status: GLuint;
	begin
	glGenFramebuffersEXT(1, @framebuffer);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);

	glGenRenderbuffersEXT(1,@depthbuffer);
	glGenRenderbuffersEXT(1,@colorbuffer);
	glBindRenderbufferEXT (GL_RENDERBUFFER_EXT,depthbuffer);
	glRenderbufferStorageEXT (GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, Width, Height);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT, depthbuffer);
	glBindRenderbufferEXT (GL_RENDERBUFFER_EXT,colorbuffer);
	glRenderbufferStorageEXT (GL_RENDERBUFFER_EXT, GL_RGBA, Width, Height);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_RENDERBUFFER_EXT, colorbuffer);
	Status:= glCheckFramebufferStatusEXT (GL_FRAMEBUFFER_EXT);
	if Status <> GL_FRAMEBUFFER_COMPLETE_EXT then Raise Exception.Create('Error in framebuffer activation: '+IntToStr(Status));
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);
	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
	end;

	procedure ScaleViewPort(Width,Height: integer; Scale: double);
	{Scale view port by the specified value}
	var FW,FH: double;
	begin
	FW:= Width / PixelsPerUnit / WorkDistance;
	FH:= Height / PixelsPerUnit / WorkDistance;
	glMatrixMode(GL_PROJECTION);
	glViewport(0, 0, Round(Width * Scale), Round(Height * Scale));
	glLoadIdentity;
	glFrustum(-(FW/2), FW/2, -(FH/2), FH/2, FrustNear,FrustFar);
	glMatrixMode(GL_MODELVIEW);
	end;


begin
ImageWidth:=Round(ClientWidth * Scale);
ImageHeight:=Round(ClientHeight * Scale);
GetMem(Buffer, ImageWidth * ImageHeight * 4);
try
glPixelStorei(GL_PACK_ALIGNMENT, 4);
glPixelStorei(GL_PACK_ROW_LENGTH, 0);
glPixelStorei(GL_PACK_SKIP_ROWS, 0);
glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
CreateFBO(ImageWidth, ImageHeight);
ScaleViewPort(ClientWidth, ClientHeight, Scale);
{Do not trigger events here, infinite loop}
RenderAll(False);
{User font buffer because the drawing has been swapped}
glReadBuffer(GL_FRONT);
glReadPixels(0, 0, ImageWidth, ImageHeight, GL_BGRA, GL_UNSIGNED_BYTE, Buffer);
if not Assigned(BM) then BM:=TBitmap.Create;
BM.PixelFormat:=pf32Bit;
BM.Width:=ImageWidth;
BM.Height:=ImageHeight;
Pixel:=Buffer;
for Y:=BM.Height - 1 downto 0 do
	begin
	Scan:=BM.ScanLine[Y];
	for X:=0 to BM.Width - 1 do
		begin
		Scan[X]:=Pixel^;
		inc(Pixel);
		end;
	end;
glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
GLWindow.SetViewPort(Width,Height);
RenderAll(False);
finally FreeMem(Buffer); end;
end;



procedure TGLViewer.SaveToClipboard(Scale: double =1);
var BM: TBitmap;
var MyFormat : Word;
var AData: THandle;
var APalette: HPALETTE;
begin
BM:=TBitmap.Create;
try
GetScreenImage(BM,Scale);
BM.SaveToClipboardFormat(MyFormat,AData,APalette);
ClipBoard.SetAsHandle(MyFormat,AData);
finally BM.Free; end;
end;



procedure TGLViewer.SaveToFile(Filename: string);
var BM: TBitmap;
begin
BM:=TBitmap.Create;
try
GetScreenImage(BM);
BM.SaveToFile(Filename);
finally BM.Free; end;
end;


{------------------------------------------------------------------------------}


procedure TGLViewer.LoadXFile(Filename: string; ProgressProc: TProgressProcedure; NewMeshes: boolean);
var OldEnabled: boolean;
begin
OldEnabled:=FEnabled;
FEnabled:=False;
try
ReadXFile(FileName,Self,ProgressProc,NewMeshes);
finally FEnabled:=OldEnabled; end;
end;



procedure TGLViewer.LoadXStream(Stream: TStream; ProgressProc: TProgressProcedure; NewMeshes: boolean);
var OldEnabled: boolean;
begin
OldEnabled:=FEnabled;
FEnabled:=False;
try
ReadXStream(Stream,Self,ProgressProc,NewMeshes);
finally FEnabled:=OldEnabled; end;
end;



procedure TGLViewer.SaveXFile(Filename: string; ProgressProc: TProgressProcedure);
begin
WriteXFile(FileName,Self, ProgressProc);
end;



procedure TGLViewer.SaveXStream(Stream: TStream; ProgressProc: TProgressProcedure);
begin
WriteXStream(Stream,Self, ProgressProc);
end;



procedure TGLViewer.LoadXMLFile(Filename: string;  ProgressProc: TXMLProgressProc; NewMeshes: boolean);
var OldEnabled: boolean;
begin
OldEnabled:=FEnabled;
FEnabled:=False;
try
if NewMeshes then BaseFrame.Clear;
LoadGLXMLFile(BaseFrame,FileName,ProgressProc);
BaseFrame.Reset;
finally FEnabled:=OldEnabled; end;
end;




procedure TGLViewer.SaveXMLFile(Filename: string; ProgressProc: TXMLProgressProc);
begin
SaveGLXMLFile(BaseFrame,FileName,ProgressProc);
end;


procedure TGLViewer.Reset;
var I: integer;
begin
Camera.Reset;
BaseFrame.Clear;
ResetLightParams;
for I:=0 to MaxLights-1 do Lights[I].Enabled:=False;
Lights[0].Enabled:=True;
BaseFrame.ResetTransforms;
Color:=clBlack;
end;

var BaseLeft,BaseTop,BaseWidth,BaseHeight: integer;

procedure TGLViewer.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
inherited;
BaseLeft:=ALeft;
BaseTop:=ATop;
BaseWidth:=AWidth;
BaseHeight:=AHeight;
end;

procedure TGLViewer.SetScaleFactor(const Value: double);
begin
if FScaleFactor = Value then exit;
FScaleFactor := Value;
inherited SetBounds(Round(BaseLeft * FScaleFactor), Round(BaseTop * FScaleFactor),
Round(BaseWidth * FScaleFactor), Round(BaseHeight * FScaleFactor));
end;

procedure TGLViewer.SetAntialias(const Value: boolean);
begin
FAntialias := Value;
if Value then glEnable(GL_MULTISAMPLE)
else glDisable(GL_MULTISAMPLE);
Invalidate;
end;

end.
