unit NTPClient;

interface

{Uses declarations needed to work with Delph 2007}

{$IF CompilerVersion<22}
uses Classes, IdGlobal, IdUDPClient, DSupport;
{$ELSE}
uses Classes, IdGlobal,IdGlobalProtocols, IdAssignedNumbers, IdUDPClient, DSupport;
{$IFEND}


const NTPMaxInt = 4294967297.0;

{ Raw NTP Datagram format }

type TNTPGram = packed record
 Head1, Head2, Head3, Head4: byte;
 RootDelay: integer;
 RootDispersion: integer;
 RefID: integer;
 Ref1, Ref2: integer;
 Org1, Org2: integer;
 Rcv1, Rcv2: integer;
 Xmit1, Xmit2: integer;
 end;

{structure for disassembling integers into bytes}

type TByteCracker = packed record
  B1,B2,B3,B4: byte;
  end;


{Holds processed NTP information}

type TNTPInfo = record
 LeapInd: byte;
 Version: byte;
 Mode: byte;
 Stratum: byte;
 PollInterval: double;
 Precision: double;

 RootDelay: double;
 RootDispersion: double;
 ReferenceID: string;

 DestinationTimestamp: TDateTime;
 LocalClockOffset: TDateTime;
 OriginateTimestamp: TDateTime;
 ReceiveTimestamp: TDateTime;
 RoundTripDelay: TDateTime;
 TransmitTimestamp: TDateTime;

 LocalTime: TDateTime;
 UTCTime: TDateTime;
 Valid: boolean;
 end;




type TNTPClient = class(TIdUDPClient)
 protected
  function GetAdjustmentTime: TDateTime;
  function RequestTimeData(var DataGram: TNTPGram): boolean;
 public
  TimeInfo: TNTPInfo;
  constructor Create(AOwner: TComponent);
  function SyncTime: Boolean;
  property AdjustmentTime: TDateTime read GetAdjustmentTime;
  function GetDateTime: TDateTime;
  property DateTime: TDateTime read GetDateTime;
  property RoundTripDelay: TDateTime read TimeInfo.RoundTripDelay;
  property Port default IdPORT_SNTP;
 end;


procedure Register;

implementation


uses SysUtils;


procedure Register;
begin
RegisterComponents('Fountain', [TNTPClient]);
end;

function Flip(var Number: integer): integer;
{Reverse byte order to convert between big and little endian}
var Number1, Number2: TByteCracker;
begin
Number1:= TByteCracker(Number);
Number2.B1:= Number1.B4;
Number2.B2:= Number1.B3;
Number2.B3:= Number1.B2;
Number2.B4:= Number1.B1;
Result:= integer(Number2);
end;



procedure DateTimeToNTP(ADateTime: TDateTime; var Second, Fraction: integer);
{Convert DateTime to NTP format values}
var Value1, Value2: Double;
begin
Value1:= ADateTime + TimeZoneBias - 2;
Value1:= Value1 * 86400;
Value2:= Value1;
if Value2 > NTPMaxInt then Value2:= Value2 - NTPMaxInt;
Second:= Trunc(Value2);
Value2:= ((Frac(Value1) * 1000) / 1000) * NTPMaxInt;
if Value2 > NTPMaxInt then Value2:= Value2 - NTPMaxInt;
Fraction:= Trunc(Value2);
end;



function NTPToDateTime(Second, Fraction: integer): TDateTime;
{Convert NTP time information to Delphi TDateTime}
var Value1,Value2: Double;
begin
Value1:= Second;
if Value1 < 0 then
	begin
	Value1:= NTPMaxInt + Value1 - 1;
	end;
Value2:= Fraction;
if Value2 < 0 then
	begin
	Value2:= NTPMaxInt + Value2 - 1;
	end;
Value2:= Value2 / NTPMaxInt;
Value2:= Trunc(Value2 * 1000) / 1000;
Result:= ((Value1 + Value2) / 86400);
Result:= Result - TimeZoneBias + 2;
end;



constructor TNTPClient.Create(AOwner: TComponent);
begin
inherited;
Port:= IdPORT_SNTP;
ReceiveTimeout:=500;
end;




function TNTPClient.GetAdjustmentTime: TDateTime;
begin
Result:= TimeInfo.LocalClockOffset;
end;



function BuildTransmitDatagramStr: string;
{Build NTP transmit Data gram as string}
var DataGram: TNTPGram;
begin
FillChar(DataGram, SizeOf(TNTPGram), 0);
DataGram.Head1:= $1B;
DateTimeToNTP(Now, DataGram.Xmit1, DataGram.Xmit2);
DataGram.Xmit1:= Flip(DataGram.Xmit1);
DataGram.Xmit2:= Flip(DataGram.Xmit2);
SetLength(Result, SizeOf(TNTPGram));
Move(DataGram, Result[1], SizeOf(TNTPGram));
end;



function TNTPClient.RequestTimeData(var DataGram: TNTPGram): boolean;
{Request NTP data and return in DataGram}
var S: string;
begin
S:=BuildTransmitDatagramStr;
BufferSize:= SizeOf(TNTPGram);
Send(S);
S:= ReceiveString;
Result:=Length(S) >= SizeOf(TNTPGram);
if Result then Move(S[1],DataGram,SizeOf(TNTPGram));
end;



function GetNTPShortReal(I: integer): double;
{convert integer with fraction point between 15-16 bits}
begin
Result:=Flip(I)/32768;
end;



function GetSignBytePower2(B: byte): double;
begin
Result:=Power(2,ExtendSign(B));
end;




function ExtractNTPInfo(DataGram: TNTPGram): TNTPInfo;
{Extract info from NTP DataGram and convert to meaning full units}
var S: string;
begin
Result.DestinationTimeStamp:= Now;
{Extract Header}
Result.LeapInd:=(DataGram.Head1 and $C0) shr 6;
Result.Version:=(DataGram.Head1 and $38) shr 3;
Result.Mode:=DataGram.Head1 and $03;
Result.Stratum:=DataGram.Head2;
Result.PollInterval:=GetSignBytePower2(DataGram.Head3);
Result.Precision:=GetSignBytePower2(DataGram.Head4);

Result.RootDelay:=GetNTPShortReal(DataGram.RootDelay);
Result.RootDispersion:=GetNTPShortReal(DataGram.RootDispersion);

if Result.Stratum<2 then
	begin
	S:=Char(PByteArray(@DataGram.RefID)[0]);
	S:=S+Char(PByteArray(@DataGram.RefID)[1]);
	S:=S+Char(PByteArray(@DataGram.RefID)[2]);
	S:=S+Char(PByteArray(@DataGram.RefID)[3]);
        end
else
	begin
	S:=IntToStr(PByteArray(@DataGram.RefID)[0])+'.';
	S:=S+IntToStr(PByteArray(@DataGram.RefID)[1])+'.';
	S:=S+IntToStr(PByteArray(@DataGram.RefID)[2])+'.';
	S:=S+IntToStr(PByteArray(@DataGram.RefID)[3]);
	end;
Result.ReferenceID:=S;
Result.LocalTime:= NTPToDateTime(Flip(DataGram.Xmit1), Flip(DataGram.Xmit2));
Result.UTCTime:= Result.LocalTime + TimeZoneBias - 2;
Result.OriginateTimeStamp:= NTPToDateTime(Flip(DataGram.Org1), Flip(DataGram.Org2));
Result.ReceiveTimestamp:= NTPToDateTime(Flip(DataGram.Rcv1),  Flip(DataGram.Rcv2));
Result.TransmitTimestamp:= NTPToDateTime(Flip(DataGram.Xmit1), Flip(DataGram.Xmit2));
Result.RoundTripDelay:= (Result.DestinationTimestamp - Result.OriginateTimestamp) - (Result.ReceiveTimestamp - Result.TransmitTimestamp);
Result.LocalClockOffset:= ((Result.ReceiveTimestamp - Result.OriginateTimestamp) + (Result.TransmitTimestamp - Result.DestinationTimestamp)) / 2;
Result.Valid:= (Result.LeapInd <> 3) and (Result.Stratum <= 15) and (Result.Stratum <> 0) and
	(((Int(Result.TransmitTimestamp)) <> 0.0) or (Frac(Result.TransmitTimestamp) <> 0.0));
end;



function TNTPClient.GetDateTime: TDateTime;
{Load and return NTP Local Date/Time}
var NTPDataGram: TNTPGram;
begin
Result:=0;
TimeInfo.Valid:=False;
if RequestTimeData(NTPDataGram) then TimeInfo:=ExtractNTPInfo(NTPDataGram);
if TimeInfo.Valid then Result:=TimeInfo.LocalTime;
end;



function TNTPClient.SyncTime: Boolean;
{Synchronize local time with NTP time}
begin
Result:= DateTime <> 0;
if Result then
	begin
	{$IF CompilerVersion<22}
	result:= SetLocalTime(TimeInfo.OriginateTimestamp + TimeInfo.LocalClockOffset + TimeInfo.RoundTripDelay);
	{$ELSE}
	result:= IndySetLocalTime(TimeInfo.OriginateTimestamp + TimeInfo.LocalClockOffset + TimeInfo.RoundTripDelay);
	{$IFEND}
	end;
end;


end.
