unit PrintPreview;

interface

uses Windows,Classes,Forms,SysUtils,Types,Graphics,Controls,ExtCtrls,Contnrs,Printers,
     Vector,IntVector,GLMesh,GLFrame,BitmapLib,Math,PrintLIb;


type TProgressProc = procedure(Percent: integer);


type TPaperSize = record
 PhysicalSize: TPoint;
 PortSize: TPoint;
 Offset: TPoint;
 PPI: TPoint;
 PaperAspect: double;
 end;


type TPrinterPage = class(TObject)
 private
  FPaperRect,PortRect: TRect;
  procedure DrawPaper(Canvas: TCanvas; WinRect: TRect);
  procedure GetDrawingSizes(WinRect: TRect);
 protected
 public
  PaperSize: TPaperSize;
  Bitmap: TBitmap;
  ShadowColor,BackgroundColor: TColor;
  Title: string;
  constructor Create;
  destructor Destroy; override;
  function IsBlankPage: boolean;
  procedure Update;
  procedure DrawPreview(Canvas: TCanvas; WinRect: TRect);
  procedure PrintPage;
  procedure DrawScaledBitmap(BM: TBitmap);
  property PaperRect: TRect read FPaperRect;
 end;



type TPageList = class(TObjectList)
 private
  FOnChange: TNotifyEvent;
  procedure Change;
 protected
  function GetItem(Index: Integer): TPrinterPage;
  procedure SetItem(Index: Integer; AObject: TPrinterPage);
 public
  function Add: integer; overload;
  function Add(var Page: TPrinterPage): integer; overload;
  property Items[Index: Integer]: TPrinterPage read GetItem write SetItem; default;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
 end;



type TPrintPreview = class(TScrollBox)
 private
  PageList: TPageList;
  FColumns: integer;
  FAutoArrange: boolean;
  FUpdateCount: Integer;
    FOnImageMouseMove: TMouseMoveEvent;
  function GetPage(Index: integer): TPrinterPage;
  function GetCount: integer;
  procedure RebuildPages;
  procedure DisplayPages;
  procedure RemoveAllControls;
  procedure SetColumns(const Value: integer);
  function GetShadowColor: TColor;
  procedure SetAutoArrange(const Value: boolean);
  procedure SetCount(const Value: integer);
  procedure BeginUpdate;
  procedure EndUpdate;
  procedure ImageMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
 protected
  procedure Resize; override;
  procedure PaintWindow(DC: HDC); override;
 public
  FitScale: double;
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  property Pages[Index: integer]: TPrinterPage read GetPage;
  function Add: integer; overload;
  function Add(var Page: TPrinterPage): integer; overload;
  procedure Delete(Index: integer);
  procedure Clear;
  procedure Update;
  procedure TileBitmap(BM: TBitmap; Offset,Limit: integer);
  procedure FitToPage(BM: TBitmap);
  procedure PrintScaled(BM: TBitmap; Scale: double);
  procedure PrintAllPages(IgnoreBlanks: boolean; Progress: TProgressProc);
 published
  property Columns: integer read FColumns write SetColumns default 3;
  property AutoArrange: boolean read FAutoArrange write SetAutoArrange;
  property Count: integer read GetCount write SetCount;
  property OnImageMouseMove: TMouseMoveEvent read FOnImageMouseMove write FOnImageMouseMove;
 end;

function GetPaperSize: TPaperSize;
function GetScreenToPaperScale: T2DVector;


procedure Register;

implementation


procedure Register;
begin
RegisterComponents('Fountain', [TPrintPreview]);
end;




function GetPaperSize: TPaperSize;
{Get Paper Size parameter based on current printer settings}
begin
Result.PPI.X:=GetDeviceCaps(Printer.Handle,LOGPIXELSX);
Result.PPI.Y:=GetDeviceCaps(Printer.Handle,LOGPIXELSY);
Result.PhysicalSize.X:=GetDeviceCaps(Printer.Handle,PHYSICALWIDTH);
Result.PhysicalSize.Y:=GetDeviceCaps(Printer.Handle,PHYSICALHEIGHT);

Result.PortSize.X:=GetDeviceCaps(Printer.Handle,HORZRES);
Result.PortSize.Y:=GetDeviceCaps(Printer.Handle,VERTRES);

Result.Offset.X:=GetDeviceCaps(Printer.Handle,PHYSICALOFFSETX);
Result.Offset.Y:=GetDeviceCaps(Printer.Handle,PHYSICALOFFSETY);
Result.PaperAspect:=Result.PhysicalSize.Y/Result.PhysicalSize.X;
end;


function GetScreenToPaperScale: T2DVector;
{Get the scale between Screen and Paper}
var PrinterDPI,ScreenDPI: T2DVector;
var PaperSize: TPaperSize;
var DC: HDC;
begin
PaperSize:=GetPaperSize;
PrinterDPI:=PointToVector2D(PaperSize.PPI);
DC:=GetDC(0);
ScreenDPI.X:=GetDeviceCaps(DC,LOGPIXELSX);
ScreenDPI.Y:=GetDeviceCaps(DC,LOGPIXELSY);
Result:=VectorDivide2D(PrinterDPI,ScreenDPI);
end;


procedure GetScaledRectangle(var R: TRect; Width,Height: integer);
{Get retangle scaled from screen DPI to Printer DPI}
var Scale: T2DVector;
begin
Scale:=GetScreenToPaperScale;
R:=Rect(0,0,round(Scale.X*Width),round(Scale.Y*Height));
end;



procedure GetUnScaledRectangle(var R: TRect; Width,Height: integer);
{Get retangle scaled from Printer DPI to screen DPI}
var Scale,PrinterDPI,ScreenDPI: T2DVector;
var PaperSize: TPaperSize;
var DC: HDC;
begin
PaperSize:=GetPaperSize;
PrinterDPI:=PointToVector2D(PaperSize.PPI);
DC:=GetDC(0);
ScreenDPI.X:=GetDeviceCaps(DC,LOGPIXELSX);
ScreenDPI.Y:=GetDeviceCaps(DC,LOGPIXELSY);
Scale:=VectorDivide2D(PrinterDPI,ScreenDPI);
R:=Rect(0,0,round(Width/Scale.X),round(Height/Scale.Y));
end;



constructor TPrinterPage.Create;
begin
Bitmap:=TBitmap.Create;
Bitmap.HandleType:=bmDIB;
BackgroundColor:=clBtnFace;
ShadowColor:=clBtnShadow;
Update;
end;



destructor TPrinterPage.Destroy;
begin
Bitmap.Free;
inherited;
end;



procedure TPrinterPage.GetDrawingSizes(WinRect: TRect);
var AC: double;
var WinSize,Offset: TPoint;
var Scale: T2DVector;
begin
WinSize:=IntVectorSubtract(WinRect.BottomRight,WinRect.TopLeft);
AC:=WinSize.Y/WinSize.X;
FPaperRect.TopLeft:=Point(0,0);
if AC<PaperSize.PaperAspect then
	begin
	{ Paper constrained in Y axis }
	FPaperRect.Bottom:=MulDiv(WinSize.Y,93,100);
	FPaperRect.Right:=round(FPaperRect.Bottom/PaperSize.PaperAspect);
	end
else
	begin
	{ Paper constrained by X Axis }
	FPaperRect.Right:=MulDiv(WinSize.X,93,100);
	FPaperRect.Bottom:=round(FPaperRect.Right*PaperSize.PaperAspect);
	end;
Scale:=VectorDivide2D(PointToVector2D(FPaperRect.BottomRight),PointToVector2D(PaperSize.PhysicalSize));
{Center Rectangle in client area}
Offset:=IntVectorSubtract(WinSize,FPaperRect.BottomRight);
Offset:=IntScalarDivide(Offset,2);
OffsetRect(FPaperRect,Offset.X,Offset.Y);

PortRect.TopLeft:=IntVectorAdd(FPaperRect.TopLeft, Point(Round(PaperSize.Offset.X * Scale.X),Round(PaperSize.Offset.Y * Scale.Y)));
PortRect.BottomRight:=IntVectorAdd(PortRect.TopLeft,Point(Round(PaperSize.PortSize.X * Scale.X),Round(PaperSize.PortSize.Y * Scale.Y)));
end;




procedure TPrinterPage.DrawPaper(Canvas: TCanvas; WinRect: TRect);
const ShadowWidth=5;
begin
GetDrawingSizes(WinRect);
{Fill background}
Canvas.Brush.Color:=BackgroundColor;
Canvas.Pen.Color:=BackgroundColor;
Canvas.Rectangle(WinRect);

{ Draw Paper box }
Canvas.Brush.Color:=clWhite;
Canvas.Pen.Color:=clBlack;
Canvas.Rectangle(FPaperRect);
{Draw interior port}
Canvas.Pen.Color:=RGB(220,220,220);
Canvas.Rectangle(PortRect);
{ Draw Shadows }
Canvas.Brush.Color:=ShadowColor;
Canvas.Pen.Color:=ShadowColor;
Canvas.Rectangle(FPaperRect.Right,FPaperRect.Top+ShadowWidth,FPaperRect.Right+ShadowWidth,FPaperRect.Bottom+ShadowWidth);
Canvas.Rectangle(FPaperRect.Left+ShadowWidth,FPaperRect.Bottom,FPaperRect.Right+1,FPaperRect.Bottom+ShadowWidth);
end;



procedure TPrinterPage.DrawPreview(Canvas: TCanvas; WinRect: TRect);
begin
DrawPaper(Canvas,WinRect);
Canvas.StretchDraw(PortRect,Bitmap);
end;



procedure TPrinterPage.Update;
begin
PaperSize:=GetPaperSize;
Bitmap.Width:=PaperSize.PortSize.X;
Bitmap.Height:=PaperSize.PortSize.Y;
end;



procedure TPrinterPage.PrintPage;
begin
Printer.Title:=Title;
Printer.BeginDoc;
Printer.Canvas.CopyMode:=cmSrcCopy;
Printer.Canvas.Draw(0,0,Bitmap);
Printer.EndDoc;
end;




function TPrinterPage.IsBlankPage: boolean;
begin
Result:=IsBitmapBlank(Bitmap);
end;




procedure TPrinterPage.DrawScaledBitmap(BM: TBitmap);
var R: TRect;
begin
GetScaledRectangle(R,BM.Width,BM.Height);
Bitmap.Canvas.StretchDraw(R,BM)
end;


{ TPageList }


procedure TPageList.Change;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;


function TPageList.Add: integer;
var Page: TPrinterPage;
begin
Result:=Add(Page);
end;


function TPageList.Add(var Page: TPrinterPage): integer;
begin
Page:=TPrinterPage.Create;
Result:=inherited Add(Page);
Change;
end;


function TPageList.GetItem(Index: Integer): TPrinterPage;
begin
Result := TPrinterPage(inherited GetItem(Index));
end;



procedure TPageList.SetItem(Index: Integer; AObject: TPrinterPage);
begin
inherited SetItem(Index, AObject);
end;




{ TPrintPreview }



function TPrintPreview.GetShadowColor: TColor;
var R,G,B: byte;
begin
ColorToRGBBytes(Color,R,G,B);
R:=MulDiv(R,50,100);
G:=MulDiv(G,50,100);
B:=MulDiv(B,50,100);
Result:=RGB(R,G,B);
end;



procedure TPrintPreview.DisplayPages;
var I,Inx: integer;
var Image: TImage;
begin
for I:=0 to PageList.Count-1 do
	begin
	Image:=TImage(Controls[I]);
	Image.Canvas.Brush.Color:=Color;
	Image.Canvas.Pen.Color:=Color;
	Image.Canvas.Rectangle(Image.ClientRect);
	Inx:=Image.Tag;
	Pages[Inx].ShadowColor:=GetShadowColor;
	Pages[Inx].BackgroundColor:=Color;
	Pages[Inx].DrawPreview(Image.Canvas,Image.ClientRect);
	end;
end;



procedure TPrintPreview.RemoveAllControls;
var I: integer;
var C: TControl;
begin
for I:=ControlCount-1 downto 0 do
	begin
	C:=Controls[I];
	RemoveControl(C);
	C.Free;
	end;
end;


procedure TPrintPreview.ImageMouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
begin
if Assigned(FOnImageMouseMove) then FOnImageMouseMove(Sender,Shift,X,Y);
end;



procedure TPrintPreview.RebuildPages;
var I,X,Y: integer;
var Image: TImage;
var CellWidth,CellHeight: integer;
var PaperSize: TPaperSize;
var OldScroll: boolean;
begin
{Turn off autoscroll so dashed frame doesn't show up in design mode}
OldScroll:=AutoScroll;
AutoScroll:=False;
PaperSize:=GetPaperSize;
CellWidth:=Width div Columns;
CellHeight:=round(CellWidth * PaperSize.PaperAspect);
RemoveAllControls;
X:=0; Y:=0;
for I:=0 to PageList.Count-1 do
	begin
	Pages[I].Update;
	Pages[I].Title:='Page: '+IntToStr(I+1);
	Image:=TImage.Create(Self);
	Image.Parent:=Self;
	Image.Left:=X * CellWidth ;
	Image.Top:=Y*(CellHeight);
	Image.Width:=CellWidth;
	Image.Height:=CellHeight;
	Image.Tag:=I;
	Image.OnMouseMove:=ImageMouseMove;
	Inc(X);
	if X>=Columns then
		begin
		X:=0;
		Inc(Y);
		end;
	end;
AutoScroll:=OldScroll;
end;



function TPrintPreview.Add(var Page: TPrinterPage): integer;
begin
Result:=PageList.Add(Page);
Update;
end;


function TPrintPreview.Add: integer;
begin
Result:=PageList.Add;
Update;
end;


procedure TPrintPreview.Clear;
begin
PageList.Clear;
Update;
end;


constructor TPrintPreview.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
PageList:=TPageList.Create;
Columns:=3;
doubleBuffered:=True;
Add;Add;Add;
end;



procedure TPrintPreview.Delete(Index: integer);
begin
PageList.Delete(Index);
Update;
end;



destructor TPrintPreview.Destroy;
begin
PageList.Free;
inherited;
end;



function TPrintPreview.GetCount: integer;
begin
Result:=PageList.Count;
end;



procedure TPrintPreview.SetCount(const Value: integer);
var I,D: integer;
begin
if Value<>PageList.Count then
	begin
	BeginUpdate;
	try
	D:=Value-PageList.Count;
	if Value>PageList.Count then
		begin
		for I:=0 to abs(D)-1 do Add;
		end
	else
		begin
		for I:=0 to abs(D)-1 do Delete(PageList.Count-1);
		end;
	finally EndUpdate; end;
	end;
end;


function TPrintPreview.GetPage(Index: integer): TPrinterPage;
begin
Result:=PageList[Index];
end;



procedure TPrintPreview.BeginUpdate;
begin
Inc(FUpdateCount);
end;



procedure TPrintPreview.EndUpdate;
begin
Dec(FUpdateCount);
if FUpdateCount <= 0 then
	begin
	FUpdateCount := 0;
	Update;
	end;
end;



procedure TPrintPreview.Update;
begin
if FUpdateCount<>0 then exit;
RebuildPages;
Invalidate;
end;



procedure TPrintPreview.SetColumns(const Value: integer);
begin
if FColumns<>Value then
	begin
	FColumns:=Value;
	Update;
	end;
end;



procedure TPrintPreview.Resize;
begin
inherited;
if FAutoArrange then Update;
end;



procedure TPrintPreview.SetAutoArrange(const Value: boolean);
begin
if FAutoArrange<>Value then
	begin
	FAutoArrange:=Value;
	if FAutoArrange then Update;
	end;
end;



procedure TPrintPreview.PaintWindow(DC: HDC);
begin
DisplayPages;
inherited PaintWindow(DC);
end;




procedure GetPageParameters(BM: TBitmap; var XPages,YPages,Total: integer);
var PaperSize: TPaperSize;
var BR: TRect;
begin
PaperSize:=GetPaperSize;
GetScaledRectangle(BR,BM.Width,BM.Height);
XPages:=ceil(BR.Right/PaperSize.PortSize.X);
YPages:=ceil(BR.Bottom/PaperSize.PortSize.Y);
Total:=XPages*YPages;
end;



procedure ScaleBitmap(var BM: TBitmap);
var R: TRect;
begin
GetScaledRectangle(R,BM.Width,BM.Height);
BM.Width:=R.Right;
BM.Height:=R.Bottom;
BM.Canvas.StretchDraw(R,BM)
end;




function GetSourceRect(BM: TBitmap; var R: TRect; X,Y: integer): boolean;
var PR,BR: TRect;
var Paper: TPaperSize;
begin
Paper:=GetPaperSize;
GetUnScaledRectangle(PR,Paper.PortSize.X,Paper.PortSize.Y);
OffsetRect(PR,PR.Right*X,PR.Bottom*Y);
BR:=Rect(0,0,BM.Width,BM.Height);
Result:=IntersectRect(R,PR,BR);
end;




function GetDestRect(BM: TBitmap; var R: TRect; X,Y: integer): boolean;
var PR,BR: TRect;
var Paper: TPaperSize;
begin
Paper:=GetPaperSize;
PR:=Rect(0,0,Paper.PortSize.X,Paper.PortSize.Y);
GetScaledRectangle(BR,BM.Width,BM.Height);
OffsetRect(PR,Paper.PortSize.X*X,Paper.PortSize.Y*Y);
Result:=IntersectRect(R,PR,BR);
Result:=OffsetRect(R,-R.Left,-R.Top);
end;



procedure TPrintPreview.TileBitmap(BM: TBitmap; Offset,Limit: integer);
var XPages,YPages,Total: integer;
var X,Y,Inx: integer;
var Source,Dest: TRect;
begin
Clear;
GetPageParameters(BM,XPages,YPages,Total);
if Total<1 then exit;
Count:=Total;
Columns:=XPages;

for Y:=0 to YPages-1 do
 for X:=0 to XPages-1 do
	begin
	GetSourceRect(BM,Source,X,Y);
	GetDestRect(BM,Dest,X,Y);
	Inx:=(Y * Columns + X)+Offset;
	Pages[Inx].Bitmap.Canvas.CopyRect(Dest,BM.Canvas,Source);
	if Inx>=Limit then exit;
	end;
end;





procedure TPrintPreview.FitToPage(BM: TBitmap);
var Dest: TRect;
begin
Clear;
Count:=1;
Columns:=1;
Dest:=Rect(0,0,BM.Width,BM.Height);
if (Pages[0].Bitmap.Width-BM.Width)<(Pages[0].Bitmap.Height-BM.Height) then
	FitScale:=Pages[0].Bitmap.Width/BM.Width
else FitScale:=Pages[0].Bitmap.Height/BM.Height;
Dest.BottomRight:=IntVector.IntScalarProduct(Dest.BottomRight,FitScale);
Pages[0].Bitmap.Canvas.StretchDraw(Dest,BM);
end;



procedure TPrintPreview.PrintScaled(BM: TBitmap; Scale: double);
var BM2: TBitmap;
begin
BM2:=TBitmap.Create;
try
ResizeBitmap(BM,BM2,Scale,True);
TileBitmap(BM2,0,30);
finally BM2.Free; end;
end;



procedure TPrintPreview.PrintAllPages(IgnoreBlanks: boolean; Progress: TProgressProc);
var I: integer;
begin
for I:=0 to Count-1 do
	begin
	if Assigned(Progress) then Progress(MulDiv(100,I,Count));
	if IgnoreBlanks and Pages[I].IsBlankPage then continue;
	Pages[I].PrintPage;
	end;
end;



end.
