unit Reopen;

interface

uses Classes,Dialogs,Menus,SysUtils,IniFiles,DSupport;

type TMenuClickEvent = procedure(Sender: TObject; Path: String) of object;

type
TReopenMenu=class(TComponent)
 private
  FINIFile,FINISection: string;
  FMenuItem: TMenuItem;
  FOnMenuClick: TMenuClickEvent;
  FLimit: integer;
  FNumberItems: boolean;
  FNoDuplicates: boolean;
  FShowPath: boolean;
  FShowErrors: boolean;
  PathList: TStringList;
  FUpdateOnClick: boolean;
    FMustExist: boolean;
  procedure GetReOpenPath(M: TMenuItem;var P: string);
  procedure DeleteSubMenu(M: TMenuItem);
  procedure PutMenuItem(M: TMenuItem;Path: string;Top: boolean);
  procedure MenuClick(Sender: TObject);
  procedure TrimSubMenu(M: TMenuItem);
  procedure Renumber(M: TMenuItem);
  function FindItem(M: TMenuItem;Target: string): integer;
  procedure DeleteItem(M: TMenuItem;Inx: integer);
  procedure IniFileError;
protected
 public
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure LoadSubMenu;
  procedure SaveSubMenu;
  procedure AddMenuItem(Path: string);
  procedure Clear;
  procedure Notification(AComponent: TComponent; Operation: TOperation); override;
 published
  property MustExist: boolean read FMustExist write FMustExist default True;
  property Limit: integer read FLimit write FLimit default 5;
  property INIFile: string read FINIFile write FINIFile;
  property INISection: string read FINISection write FINISection;
  property MenuItem: TMenuItem read FMenuItem write FMenuItem;
  property OnMenuClick: TMenuClickEvent read FOnMenuClick write FOnMenuClick;
  property NumberItems: boolean read FNumberItems write FNumberItems default True;
  property NoDuplicates: boolean read FNoDuplicates write FNoDuplicates default True;
  property ShowPath: boolean read FShowPath write FShowPath default False;
  property ShowErrors: boolean read FShowErrors write FShowErrors default false;
  property UpdateOnClick: boolean read FUpdateOnClick write FUpdateOnClick default false;
 end;

{Note: the word Register must have upper case R or it won't Register}

procedure Register;

implementation



{--------------- Routines to manipulate menu items ----------------}

procedure Register;
begin
  RegisterComponents('Fountain', [TReopenMenu]);
end;


procedure TReopenMenu.IniFileError;
var S: string;
begin
if not FShowErrors then exit;
S:='Unable to access: '+FINIFile+CRLF_Char;
S:=S+'You may not permission to access the Windows folder';
MessageDlg(S,mtWarning,[mbOK],0);
end;




constructor TReopenMenu.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
PathList:=TStringList.Create;
FLimit:=5;
FMustExist:=True;
FNumberItems:=True;
FNoDuplicates:=True;
FShowPath:=False;
end;



destructor TReopenMenu.Destroy;
begin
{Don't clear because the menu
items may already be gone}
PathList.Free;
inherited Destroy;
end;



procedure TReopenMenu.GetReOpenPath(M: TMenuItem; var P: string);
{Get Reopen Path from menu item}
begin
P:=PathList[M.MenuIndex];
end;





procedure TReopenMenu.DeleteItem(M: TMenuItem; Inx: integer);
{Delete all submenus items}
begin
PathList.Delete(Inx);
M.Delete(Inx);
end;



procedure TReopenMenu.DeleteSubMenu(M: TMenuItem);
{Delete all submenus items}
var P: longint;
begin
M.Clear;
PathList.Clear;
exit;
for P:=M.Count-1 downto 0 do
	begin
	PathList.Delete(P);
	M.Delete(P);
	end;
end;



procedure TReopenMenu.Clear;
begin
if FMenuItem=nil then exit;
DeleteSubMenu(FMenuItem);
end;



procedure TReopenMenu.TrimSubMenu(M: TMenuItem);
{Trim submenus to limit}
var P: longint;
begin
for P:=M.Count-1 downto FLimit do
	begin
	PathList.Delete(P);
	M.Delete(P);
	end;
end;



procedure TReopenMenu.MenuClick(Sender: TObject);
{Execute an external program based on menu item}
var Path: string;
begin
GetReOpenPath(TMenuItem(Sender),Path);
if Assigned(FOnMenuClick) then FOnMenuClick(Self,Path);
if Self.FUpdateOnClick then AddMenuItem(Path);
end;




procedure TReopenMenu.Renumber(M: TMenuItem);
{Renumber menu items }
var P: integer;
var Path: string;
begin
for P:=0 to FMenuItem.Count-1 do
	begin
	GetReOpenPath(M.Items[P],Path);
	if not FShowPath then Path:=ExtractFilename(Path);
	M.Items[P].Caption:='&'+IntToStr(P+1)+' '+Path;
	end;
end;




function TReopenMenu.FindItem(M: TMenuItem;Target: string): integer;
{Return index of target in list}
begin
Result:=PathList.IndexOf(Target);
end;



procedure TReopenMenu.PutMenuItem(M: TMenuItem; Path: string; Top: boolean);
{Place a new submenu item}
var NewItem: TMenuItem;
var L: integer;
var MenuPath: string;
begin
if FNoDuplicates then
	begin
	L:=FindItem(M,Path);
	if L>-1 then DeleteItem(M,L);
	end;
NewItem:=TMenuItem.Create(M.Owner);
if not FShowPath then MenuPath:=ExtractFilename(Path) else MenuPath:=Path;
NewItem.Caption:= MenuPath;
NewItem.OnClick:=MenuClick;
if Top then
	begin
	M.Insert(0,NewItem);
	PathList.Insert(0,Path);
	end
else
	begin
	M.Add(NewItem);
	PathList.Add(Path);
	end;
if FNumberItems then Renumber(M);
end;



procedure TReopenMenu.AddMenuItem(Path: string);
begin
if FMenuItem=nil then exit;
if MustExist and not FileExists(Path) then exit;
PutMenuItem(FMenuItem,Path,True);
TrimSubMenu(FMenuItem);
end;




procedure TReopenMenu.LoadSubMenu;
{Read INI file and build submenu}
var WorkINI: TIniFile;
var C,P: integer;
var Path: string;
begin
if FMenuItem=nil then exit;
DeleteSubMenu(FMenuItem);
WorkINI := TIniFile.Create(FINIFile);
try try
C:=WorkINI.ReadInteger(FINISection,'Count',0);
for P:=0 to C-1 do
	begin
	Path:=WorkINI.ReadString(FINISection,'Path'+IntToStr(P),'');
	PutMenuItem(FMenuItem,Path,False);
	end;
finally WorkINI.Free; end;
except IniFileError; end;
end;



procedure TReopenMenu.SaveSubMenu;
{Save submenu to INI file}
var WorkINI: TIniFile;
var P: integer;
var Path: string;
begin
if (FMenuItem=nil) or (FINIFile='') then exit;
WorkINI := TIniFile.Create(FINIFile);
try try
WorkINI.EraseSection(FINISection);
WorkINI.WriteInteger(FINISection,'Count',FMenuItem.Count);

for P:=0 to FMenuItem.Count-1 do
	begin
	GetReOpenPath(FMenuItem.Items[P],Path);
	WorkINI.WriteString(FINISection,'Path'+IntToStr(P),Path);
	end;

finally WorkINI.Free; end;
except  IniFileError; end;
end;




procedure TReopenMenu.Notification(AComponent: TComponent;
  Operation: TOperation);
{Handle the live deletion of a menu item controlled by the object at designtime }
begin
inherited Notification(AComponent, Operation);
if (Operation = opRemove) and (AComponent = FMenuItem) then FMenuItem := nil;
end;


end.
