unit SizLimit;

interface

uses
  Dialogs,
  SysUtils, WinTypes, WinProcs, Messages, 
  Classes;

type
  TMinSizeError = class(Exception);
  TNotSizeableError = class(Exception);
  TNotOwnerError = class(Exception);
  TDuplicationError = class(Exception);

  TFormSizeLimit = class(TComponent)
  private
    { Private declarations }
    FActive: boolean;
    FAutoTrack: boolean;
    FAbsMinHeight: word;
    FAbsMinWidth: word;                 
    FMaxWidth: word;                    
    FMinWidth: word;                    
    FMaxHeight: word;                   
    FMinHeight: word;                   
    FFormHandle: Hwnd;                  
    FWndHandlerPtr: Pointer;            
    FOldWndHandler: Pointer;
    FDesignPixPerIn: integer;
    FRunPixPerIn: integer;

    procedure AdjustMinMaxMessage(MMPointer: Pointer);
    procedure ChkMaxHeight(Setting: word);
    procedure ChkMaxWidth(Setting: word);
    procedure ChkMinHeight(Setting: word);      
    procedure ChkMinWidth(Setting: word);       
    procedure NWndProc(var Messg: TMessage);
    procedure SetNewSize(NewClWidth, NewClHeight: word );

  protected
    { Protected declarations }

  public
    { Public declarations }
    constructor 
      Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published
    { Published declarations }
    property Active: boolean read FActive write FActive;
    property AutoTrack: boolean read FAutoTrack write FAutoTrack;
    property MaxHeight: word read FMaxHeight write ChkMaxHeight;
    property MaxWidth: word read FMaxWidth write ChkMaxWidth;
    property MinHeight: word read FMinHeight write ChkMinHeight;
    property MinWidth: word read FMinWidth write ChkMinWidth;
    property DesignPixPerIn: integer read FDesignPixPerIn write FDesignPixPerIn;

  end;

procedure Register;

implementation

uses  Forms, Controls;

constructor TFormSizeLimit.Create(AOwner: TComponent);
var i: word;
var FormSizeLimitCount: byte;
begin
FormSizeLimitCount := 0;
inherited Create(AOwner);
if (AOwner is TForm) then
	begin
	with (AOwner as TForm) do
	 begin
	 if csDesigning in ComponentState then
	 	begin
	 	for i := 0 to ComponentCount - 1 do
		 	if Components[i] is TFormSizeLimit then inc(FormSizeLimitCount);
	 	if FormSizeLimitCount > 1 then raise TDuplicationError.Create(
			'Already a FormSizeLimit component' + ' on selected Form');
		if BorderStyle <> bsSizeable then raise TNotSizeableError.Create(
			'Form does not have BorderStyle' + ' of bsSizeable');
		FMinHeight := GetSystemMetrics(SM_CYMIN);
		FAbsMinHeight := FMinHeight;
		FMinWidth := GetSystemMetrics(SM_CXMIN);
		FAbsMinWidth := FMinWidth;
		MaxHeight := Height;
		MaxWidth := Width;
		FDesignPixPerIn:=PixelsPerInch;
		end;
	 FRunPixPerIn:=Screen.PixelsPerInch;
	 FFormHandle:=Handle;
	 FOldWndHandler:=Pointer(GetWindowLong(FFormHandle,GWL_WNDPROC));
	 FWndHandlerPtr:=MakeObjectInstance(NWndProc);
	 if FWndHandlerPtr = nil then raise EOutOfResources.Create(
		          'Windows Resources exhausted');
	 SetWindowLong(FFormHandle, GWL_WNDPROC, Longint(FWndHandlerPtr));
	 end;
	end
else raise TNotOwnerError.Create('A Form MUST be the owner');
end;




destructor TFormSizeLimit.Destroy;
begin
FAutoTrack := False;
SetWindowLong(FFormHandle, gwl_WndProc, Longint(FOldWndHandler));
if FWndHandlerPtr <> nil then FreeObjectInstance(FWndHandlerPtr);
inherited Destroy;
end;


procedure TFormSizeLimit.ChkMaxHeight(Setting: word);
begin
if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then
 if (Setting < MinHeight) then raise TMinSizeError.Create(
        'The Maximum Height must not be'#13 + 'less than the Minimum height');
FMaxHeight := Setting;
end;

procedure TFormSizeLimit.ChkMaxWidth(Setting: word);
begin
if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then
 if (Setting < MinWidth) then raise TMinSizeError.Create(
        'The Maximum Width must not be'#13 +
        'less than the Minimum width');
FMaxWidth := Setting;
end;

procedure TFormSizeLimit.ChkMinHeight(Setting: word);
begin
if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then
	begin
	if (Setting < FAbsMinHeight) then raise TMinSizeError.Create(
		'The Minimum Height must not'#13 +
		'be less than the caption height');
	if (Setting >= FMaxHeight) then raise TMinSizeError.Create(
		'The Minimum Height must not be'#13 +
		'greater than the Maximum height');
	end;
FMinHeight := Setting;
end;

procedure  TFormSizeLimit.ChkMinWidth(Setting: word);
begin
if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then
	begin
	if (Setting < FAbsMinWidth) then raise TMinSizeError.Create(
		'The Minimum Width must not be less'#13 +
		'than the caption controls width');
	if (Setting >= FMaxWidth) then raise TMinSizeError.Create(
		'The Minimum width must not be'#13 +
		'greater than the Maximum width');
	end;
FMinWidth := Setting;
end;

procedure TFormSizeLimit.SetNewSize(NewClWidth, NewClHeight: word);
var NewWidth: word;
var NewHeight: word;
var SavActive: boolean;
begin
NewWidth := NewClWidth + (GetSystemMetrics(SM_CXFRAME) * 2 );
NewHeight := NewClHeight + (GetSystemMetrics(SM_CYFRAME) * 2 ) +
    (GetSystemMetrics(SM_CYCAPTION) - GetSystemMetrics(SM_CYBORDER));
SavActive := FActive;
FActive := False;
if (NewWidth < MinWidth) then FMinWidth := NewWidth;
if (NewHeight < MinHeight) then FMinHeight := NewHeight;
MaxWidth := NewWidth;
MaxHeight := NewHeight;
FActive := SavActive;
end;

procedure  TFormSizeLimit.AdjustMinMaxMessage(MMPointer: Pointer);
var MinMaxPtr: ^TMinMaxInfo;
begin
MinMaxPtr := MMPointer;
MinMaxPtr^.ptMinTrackSize.X := MulDiv(FMinWidth,FRunPixPerIn,FDesignPixPerIn);
MinMaxPtr^.ptMinTrackSize.Y := MulDiv(FMinHeight,FRunPixPerIn,FDesignPixPerIn);
MinMaxPtr^.ptMaxTrackSize.X := MulDiv(FMaxWidth,FRunPixPerIn,FDesignPixPerIn);
MinMaxPtr^.ptMaxTrackSize.Y := MulDiv(FMaxHeight,FRunPixPerIn,FDesignPixPerIn);
end;

procedure TFormSizeLimit.NWndProc(var Messg: TMessage);
begin
with Messg do
 begin
 case Msg of
  WM_GETMINMAXINFO: if (FActive and not (csDesigning in ComponentState)) then
            AdjustMinMaxMessage(Pointer(lParam));
  WM_SIZE: if ((csDesigning in ComponentState) and
		(FAutoTrack) and
		(wParam = SIZE_RESTORED)) then SetNewSize(LoWord(lParam), HiWord(lParam));
  end;
 Result := CallWindowProc(FOldWndHandler, FFormHandle, Msg, wParam, lParam);
 end;
end;



procedure Register;
begin
RegisterComponents('Fountain', [TFormSizeLimit]);
end;



end.
