unit ScrollBarEx;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Consts, Math;

type TScrollEventEx = procedure(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: int64) of object;

type
  TScrollBarEx = class(TWinControl)
  private
    FKind: TScrollBarKind;
    FPosition: int64;
    FMin: int64;
    FMax: int64;
    FPageSize: int64;
    FRTLFactor: Integer;
    FSmallChange: int64;
    FLargeChange: int64;
    FOnChange: TNotifyEvent;
    FOnScroll: TScrollEventEx;
    procedure DoScroll(var Msg: TWMScroll);
    function NotRightToLeft: Boolean;
    procedure SetKind(Value: TScrollBarKind);
    procedure SetMax(Value: int64);
    procedure SetMin(Value: int64);
    procedure SetPosition(Value: int64);
    procedure SetPageSize(Value: int64);
    procedure DoSetRange(Min,Max: int64; Redraw: boolean);
    procedure DoSetPosition(Pos: int64; Redraw: boolean);
    procedure CNHScroll(var Message: TWMHScroll); message CN_HSCROLL;
    procedure CNVScroll(var Message: TWMVScroll); message CN_VSCROLL;
    procedure CNCtlColorScrollBar(var Message: TMessage); message CN_CTLCOLORSCROLLBAR;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure Change; dynamic;
    procedure Scroll(ScrollCode: TScrollCode; var ScrollPos: int64); dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetParams(APosition, AMin, AMax: int64);
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Kind: TScrollBarKind read FKind write SetKind default sbHorizontal;
    property LargeChange: int64 read FLargeChange write FLargeChange default 1;
    property Max: int64 read FMax write SetMax default 100;
    property Min: int64 read FMin write SetMin default 0;
    property PageSize: int64 read FPageSize write SetPageSize;
    property ParentBiDiMode;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
    property Position: int64 read FPosition write SetPosition default 0;
    property ShowHint;
    property SmallChange: int64 read FSmallChange write FSmallChange default 1;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnContextPopup;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnScroll: TScrollEventEx read FOnScroll write FOnScroll;
    property OnStartDock;
    property OnStartDrag;
  end;



procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Fountain', [TScrollBarEx]);
end;


{ TScrollBar }




procedure TScrollBarEx.DoSetRange(Min,Max: int64; Redraw: boolean);
{Scale Min/Max to 0 - 2^31 range}
var SMin,SMax: cardinal;
begin
SMax:=High(Integer);
SMin:=Trunc((Min/Max) * SMax);
SetScrollRange(Handle, SB_CTL, SMin, SMax, Redraw);
end;



procedure TScrollBarEx.DoSetPosition(Pos: int64; Redraw: boolean);
{Scale Pos to 0 - 2^31 range}
var SPos: cardinal;
begin
SPos:=Trunc((Pos/FMax) * High(Integer));
SetScrollPos(Handle, SB_CTL, SPos, Redraw);
end;


constructor TScrollBarEx.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
Width := 121;
Height := GetSystemMetrics(SM_CYHSCROLL);
TabStop := True;
ControlStyle := [csFramed, csDoubleClicks, csOpaque];
FKind := sbHorizontal;
FPosition := 0;
FMin := 0;
FMax := 100;
FSmallChange := 1;
FLargeChange := 1;
if SysLocale.FarEast and (Win32Platform = VER_PLATFORM_WIN32_NT) then ImeMode := imDisable;
end;



procedure TScrollBarEx.CreateParams(var Params: TCreateParams);
const Kinds: array[TScrollBarKind] of DWORD = (SBS_HORZ, SBS_VERT);
begin
inherited CreateParams(Params);
CreateSubClass(Params, 'SCROLLBAR');
Params.Style := Params.Style or Kinds[FKind];
if FKind = sbVertical then
 if not UseRightToLeftAlignment then Params.Style := Params.Style or SBS_RIGHTALIGN
 else Params.Style := Params.Style or SBS_LEFTALIGN;
if NotRightToLeft then FRTLFactor := 1
else FRTLFactor := -1;
end;



procedure TScrollBarEx.CreateWnd;
var ScrollInfo: TScrollInfo;
var LBounds: TRect;
begin
// Windows' does not always create the window size we ask for, so we have
//  insist sometimes.  Setting BoundsRect will only adjust the size if needed.
LBounds := BoundsRect;
inherited CreateWnd;
BoundsRect := LBounds;

DoSetRange(FMin, FMax, False);
ScrollInfo.cbSize := SizeOf(ScrollInfo);
ScrollInfo.nPage := FPageSize;
ScrollInfo.fMask := SIF_PAGE;
SetScrollInfo(Handle, SB_CTL, ScrollInfo, False);
if NotRightToLeft then DoSetPosition(FPosition, True)
else DoSetPosition(FMax - FPosition, True);
end;



function TScrollBarEx.NotRightToLeft: Boolean;
begin
Result := (not IsRightToLeft) or (FKind = sbVertical);
end;



procedure TScrollBarEx.SetKind(Value: TScrollBarKind);
begin
if FKind <> Value then
	begin
	FKind := Value;
	if not (csLoading in ComponentState) then SetBounds(Left, Top, Height, Width);
	RecreateWnd;
	end;
end;




procedure TScrollBarEx.SetParams(APosition, AMin, AMax: int64);
begin
if (AMax < AMin) or (AMax < FPageSize) then raise EInvalidOperation.Create(SScrollBarRange);
if APosition < AMin then APosition := AMin;
if APosition > AMax then APosition := AMax;
if (FMin <> AMin) or (FMax <> AMax) then
	begin
	FMin := AMin;
	FMax := AMax;
	if HandleAllocated then DoSetRange(AMin, AMax, FPosition = APosition);
	end;
if FPosition <> APosition then
	begin
	FPosition := APosition;
	if HandleAllocated then
	 if NotRightToLeft then DoSetPosition( FPosition, True)
	 else DoSetPosition(FMax - FPosition, True);
	 Enabled := True;
	 Change;
	 end;
end;



procedure TScrollBarEx.SetPosition(Value: int64);
begin
SetParams(Value, FMin, FMax);
end;



procedure TScrollBarEx.SetPageSize(Value: int64);
var ScrollInfo: TScrollInfo;
begin
if (FPageSize = Value) or (Value > FMax) then exit; FPageSize := Value;
ScrollInfo.cbSize := SizeOf(ScrollInfo);
ScrollInfo.nPage := Value;
ScrollInfo.fMask := SIF_PAGE;
if HandleAllocated then SetScrollInfo(Handle, SB_CTL, ScrollInfo, True);
end;



procedure TScrollBarEx.SetMin(Value: int64);
begin
SetParams(FPosition, Value, FMax);
end;



procedure TScrollBarEx.SetMax(Value: int64);
begin
SetParams(FPosition, FMin, Value);
end;

procedure TScrollBarEx.Change;
begin
inherited Changed;
if Assigned(FOnChange) then FOnChange(Self);
end;


procedure TScrollBarEx.Scroll(ScrollCode: TScrollCode; var ScrollPos: int64);
begin
if Assigned(FOnScroll) then FOnScroll(Self, ScrollCode, ScrollPos);
end;




procedure TScrollBarEx.DoScroll(var Msg: TWMScroll);
var ScrollPos: int64;
var NewPos: int64;
var ScrollInfo: TScrollInfo;
begin
NewPos := FPosition;
case TScrollCode(Msg.ScrollCode) of
  scLineUp: Dec(NewPos, FSmallChange * FRTLFactor);
  scLineDown: Inc(NewPos, FSmallChange * FRTLFactor);
  scPageUp: Dec(NewPos, FLargeChange * FRTLFactor);
  scPageDown: Inc(NewPos, FLargeChange * FRTLFactor);
  scPosition, scTrack:
   with ScrollInfo do
	begin
	cbSize := SizeOf(ScrollInfo);
	fMask := SIF_ALL;
	GetScrollInfo(Handle, SB_CTL, ScrollInfo);
	{nTrackPos in a 32-bit integer, so scale up FMax range}
	NewPos := Trunc(nTrackPos * (FMax/High(Integer)));
	{ We need to reverse the positioning because SetPosition below
	calls SetParams that reverses the position. This acts as a
	double negative. }
	if not NotRightToLeft then NewPos := FMax - NewPos;
	end;
  scTop: NewPos := FMin;
  scBottom: NewPos := FMax;
  end;
 if NewPos < FMin then NewPos := FMin;
 if NewPos > FMax then NewPos := FMax;
 ScrollPos := NewPos;
 Scroll(TScrollCode(Msg.ScrollCode), ScrollPos);
 SetPosition(ScrollPos);
end;



procedure TScrollBarEx.CNHScroll(var Message: TWMHScroll);
begin
DoScroll(Message);
end;


procedure TScrollBarEx.CNVScroll(var Message: TWMVScroll);
begin
DoScroll(Message);
end;


procedure TScrollBarEx.CNCtlColorScrollBar(var Message: TMessage);
begin
with Message do
 CallWindowProc(DefWndProc, Handle, Msg, WParam, LParam);
end;



procedure TScrollBarEx.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
DefaultHandler(Message);
end;

end.
